#ifndef MC_BSDF_H
#define MC_BSDF_H

#include <memory>

#include "vec.h"

#include "renderer/bsdf/bsdfutils.h"
#include "renderer/texture/texture.h"
#include "renderer/bsdf/microfacet.h"
#include "renderer/bsdf/ior.h"

#include "renderer/sampler.h"
#include "renderer/interpolation.h"
#include "renderer/world.h"

#include <omp.h>

///////////// LAMBERT MODEL /////////////
class BsdfLambert;

///////////// NOT PHYSICAL /////////////
class BsdfReflection;
class BsdfTransmission;

///////////// SMOOTH MODELS /////////////
class BsdfConductor;
class BsdfDielectric;
class BsdfThinDielectric;
class BsdfPlastic;
class BsdfSmoothCoat; // need further test 

///////////// ROUGH MODELS /////////////
class BsdfOrenNayar;
class BsdfRoughDiffuse;
class BsdfRoughConductor;
class BsdfRoughDielectric;
class BsdfRoughPlastic;
class BsdfRoughCoat; // need further test 

///////////// MIX MODEL /////////////
class BsdfMix;

///////////// PHASE MODEL /////////////
class BsdfHenyeyGreenstein;
class BsdfMie;

///////////// BSDF LOBES /////////////
enum BsdfType
{
    BSDF_NULL                 = 0,
    BSDF_GLOSSYREFLECTION     = (1 << 0),
    BSDF_GLOSSYTRANSMISSION   = (1 << 1),
    BSDF_DIFFUSEREFLECTION    = (1 << 2),
    BSDF_DIFFUSETRANSMISSION  = (1 << 3),
    BSDF_SPECULARREFLECTION   = (1 << 4),
    BSDF_SPECULARTRANSMISSION = (1 << 5),
    BSDF_ANISOTROPIC          = (1 << 6),
    BSDF_PHASE                = (1 << 7),
    BSDF_FORWARD              = (1 << 8),
    BSDF_BACKWARD             = (1 << 9),
    BSDF_GLOSSY               = BSDF_GLOSSYREFLECTION | BSDF_GLOSSYTRANSMISSION,
    BSDF_DIFFUSE              = BSDF_DIFFUSEREFLECTION | BSDF_DIFFUSETRANSMISSION,
    BSDF_SPECULAR             = BSDF_SPECULARREFLECTION | BSDF_SPECULARTRANSMISSION,
    BSDF_TRANSMISSIVE         = BSDF_GLOSSYTRANSMISSION | BSDF_DIFFUSETRANSMISSION | BSDF_SPECULARTRANSMISSION,
    BSDF_REFLECTIVE           = BSDF_GLOSSYREFLECTION | BSDF_DIFFUSEREFLECTION | BSDF_SPECULARREFLECTION,
};

///////////// BSDF SAMPLE /////////////
struct BsdfSample
{
    Vector wi = Vector(0);
    float pdf = 0;
    Color weight = Black();
    // uint32_t lobe;
};

///////////// BSDF INTERFACE /////////////
class Bsdf
{
public:
    Bsdf(BsdfType type = BSDF_NULL) : m_type(type) {}
    virtual ~Bsdf() {}
    
    virtual Color eval( const Vector& wi, const Vector& wo, const vec2& uv = vec2(0,0)) const = 0;
    
    virtual float pdf( const Vector& wi, const Vector& wo, const vec2& uv = vec2(0,0)) const = 0;
    
    virtual BsdfSample sample( Sampler& sampler, const Vector& wo, const vec2& uv = vec2(0,0)) const = 0;

    // test components
    virtual bool hasGlossy() const { return (m_type & BSDF_GLOSSY) != 0; }
    virtual bool hasDiffuse() const { return (m_type & BSDF_DIFFUSE) != 0; }
    virtual bool hasSpecular() const { return (m_type & BSDF_SPECULAR) != 0; }
    virtual bool hasForward() const { return (m_type & BSDF_FORWARD) != 0; }
    virtual bool hasBackward() const { return (m_type & BSDF_BACKWARD) != 0; }
    virtual bool isPhase() const { return (m_type & BSDF_PHASE) != 0; }
    virtual bool isForward() const { return m_type != 0 && (m_type & ~BSDF_FORWARD) == 0; }
    virtual bool isBackward() const { return m_type != 0 && (m_type & ~BSDF_BACKWARD) == 0; }
    virtual bool isPureGlossy() const { return m_type != 0 && (m_type & ~BSDF_GLOSSY) == 0; }
    virtual bool isPureSpecular() const { return m_type != 0 && (m_type & ~BSDF_SPECULAR) == 0; }
    virtual bool isPureDiffuse() const { return m_type != 0 && (m_type & ~BSDF_DIFFUSE) == 0; }
    virtual bool isTransmissive() const { return (m_type & BSDF_TRANSMISSIVE) != 0; }
    virtual bool isAnisotropic() const { return (m_type & BSDF_ANISOTROPIC) != 0; }
    virtual bool isThinDielectric() const { return false; }
    virtual bool isDelta() const { return isPureSpecular(); }           
    virtual bool hasDelta() const { return hasSpecular(); }           
    virtual bool isObserverDependent() const { return hasGlossy() || hasSpecular() || hasForward() || hasBackward(); }

    // White furnace test for the given outgoing direction
    virtual Color whiteFurnaceTest(const Vector& wo, int n = 1e6) const
    {
        Sampler sampler;
        Color mean = Black();
        float count = 0;
        vec2 uv(0.5, 0.5);
        for(int i = 0; i < n; ++i)
        {
            BsdfSample bs = sample(sampler, wo, uv); 
            if(bs.pdf > 0)
            {
                ++count;
                mean = mean + (bs.weight - mean) / count;
            }
        }
        return Color(mean, 1);
    }

    // White furnace image, top view sphere, one direction per pixel, normal observer direction (0,0,1)
    virtual Image whiteFurnaceTest(int width = 100, int height = 100, int n = 1000) const
    {
        const Color bg(0.75f);
        Image image(width, height, bg);

        #pragma omp parallel for
        for(int py = 0; py < image.height(); ++py)
        for(int px = 0; px < image.width(); ++px)
        {
            const float x = -1.1f + 2.2f * px / (float) image.width();
            const float y = -1.1f + 2.2f * py / (float) image.height();
            if(x*x + y*y > 1.0f) continue;
            const float z = std::max(0.f, std::sqrt(1.0f - x*x - y*y));
            Vector wo = normalize(Vector(x,y,z));
            if( wo.z <= 0 ) continue;
            image( px, image.height() - 1 - py ) = bg * whiteFurnaceTest(wo, n);
        }

        return image;
    }

    virtual float TotalVariationApprox(const Vector& wo1, const Vector& wo2, const vec2& uv) const
    {
        // specail cas if distributions does not depend on the observing direction
        if( !isObserverDependent() )
            return 1;

        Vector wi1, wi2;
        // in media we take the forward / backward direction depending 
        // whether the phase function is scattering forward or backward
        // to determine the peak of the distribution
        if( isPhase() )
        {
            wi1 = hasBackward() ? wo1 : -wo1;
            wi2 = hasBackward() ? wo2 : -wo2;
            //wi1 = hasBackward() ? -wo1 : wo1;
            //wi2 = hasBackward() ? -wo2 : wo2;
        }
        // on surfaces we take the mirror direction
        // which is the peak of the bsdf distribution
        // in the Cook-Torrance model
        else
        {
            wi1 = Reflect(wo1);
            wi2 = Reflect(wo2);
        }

        // normalized discrete distrib P (wo1) on set {wi1, wi2}
        float p1 = pdf(wi1, wo1, uv);
        float p2 = pdf(wi2, wo1, uv);
        float sp = p1+p2;

        // normalized discrete distri Q (wo2) on set {wi1, wi2}
        float q1 = pdf(wi1, wo2, uv);
        float q2 = pdf(wi2, wo2, uv);
        float sq = q1+q2;

        // total variation approximation with two discrete samples 
        // 1 -> distributions mismatch
        // 0 -> distributions are equal
        float tv = 0.5f * (std::abs(p1/sp - q1/sq) + std::abs(p2/sp - q2/sq));
        // float tv = 0.5f * (std::pow(p1/sp - q1/sq, 1000.f) + std::pow(p2/sp - q2/sq, 1000.f));
        // float tv = 0.5f * (std::sqrt(p1/sp - q1/sq) + std::sqrt(p2/sp - q2/sq));

        // similarity is one minus TV
        // 1 -> distributions are equal
        // 0 -> distributions mismatch
        return std::max(0.f, 1.f - tv);
    }

    virtual BsdfType type() const { return m_type; }
    
    BsdfType m_type;
};



////////////// LAMBERT BSDF //////////////

class BsdfLambert : public Bsdf
{
public:
    BsdfLambert(std::shared_ptr<Texture> reflectance) : 
        Bsdf(BSDF_DIFFUSEREFLECTION), 
        m_reflectance(reflectance) 
    {}
    
    Color eval( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        return (*m_reflectance)(uv) / M_PI;
    }
    
    float pdf( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        return PdfCosineWeightedHemisphere(wi);
        // return wi.z < 0 ? 0 : PdfCosineWeightedHemisphere(wi);
    }

    BsdfSample sample( Sampler& sampler, const Vector& wo, const vec2& uv ) const
    {
        BsdfSample bs;
        bs.wi = SampleCosineWeightedHemisphere(sampler.sample(), sampler.sample()); 
        bs.pdf = PdfCosineWeightedHemisphere(bs.wi); 
        bs.weight = (*m_reflectance)(uv); 
        return bs;
    }

    std::shared_ptr<Texture> m_reflectance;
};



////////////// 100% REFLECTIVE BSDF //////////////

class BsdfReflection : public Bsdf
{
public:
    BsdfReflection(std::shared_ptr<Texture> reflectance) : 
        Bsdf(BSDF_SPECULARREFLECTION), 
        m_reflectance(reflectance) 
    {}
    
    Color eval( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        if(wi.z <= 0 || wo.z <= 0) return Black();
        return (*m_reflectance)(uv) / wi.z;
    }
    
    float pdf( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        return CheckReflectionConstraint(wi, wo);
    }

    BsdfSample sample( Sampler& sampler, const Vector& wo, const vec2& uv ) const
    {
        BsdfSample bs;
        bs.wi = Reflect(wo); 
        bs.pdf = 1; 
        bs.weight = (*m_reflectance)(uv); 
        return bs; 
    }

    std::shared_ptr<Texture> m_reflectance;
};



////////////// 100% TRANSMISSIVE BSDF //////////////

class BsdfTransmission : public Bsdf
{
public:
    BsdfTransmission(std::shared_ptr<Texture> reflectance, float intIOR = 1.5f, float extIOR = 1.f) : 
        Bsdf(BSDF_SPECULARTRANSMISSION),
        m_reflectance(reflectance), 
        m_intIOR(intIOR), 
        m_extIOR(extIOR), 
        m_eta(m_intIOR / m_extIOR), 
        m_invEta(1.f / m_eta)
    {}
    
    Color eval( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        if(wi.z * wo.z >= 0) return Black();
        return (*m_reflectance)(uv) / std::abs(wi.z);
    }
    
    float pdf( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        if(wi.z * wo.z >= 0) return 0;
        float eta = wo.z < 0.0f ? m_eta : m_invEta;
        float cosThetaT = ComputeThetaTransmitted(eta, wo);
        return CheckRefractionConstraint(wi, wo, eta, cosThetaT);
    }

    BsdfSample sample( Sampler& sampler, const Vector& wo, const vec2& uv ) const
    {
        float eta = wo.z < 0.0f ? m_eta : m_invEta;
        BsdfSample bs;
        bs.wi = Refract(eta, wo);
        bs.pdf = 1; 
        bs.weight = (*m_reflectance)(uv); 
        return bs; 
    }

    std::shared_ptr<Texture> m_reflectance;
    float m_intIOR, m_extIOR, m_eta, m_invEta;
};



////////////// DIFFUSE TRANSMISSION BSDF //////////////

class BsdfDiffuseTransmission : public Bsdf
{
public:
    BsdfDiffuseTransmission(
        std::shared_ptr<Texture> reflectance, 
        std::shared_ptr<Texture> transmittance = std::make_shared<TextureConstant>(0.5f)
    ) : 
        Bsdf((BsdfType)(BSDF_DIFFUSE)),
        m_reflectance(reflectance),
        m_transmittance(transmittance)
    {}

    std::shared_ptr<Texture> m_reflectance;
    std::shared_ptr<Texture> m_transmittance;
    
    Color eval( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        float transmittance = (*m_transmittance)(uv).r;
        float prob = wi.z * wo.z < 0.0f ? transmittance : 1.0f - transmittance;
        return (*m_reflectance)(uv) * prob / M_PI;
    }
    
    float pdf( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        if( wo.z * wi.z == 0 ) return 0;
        float transmittance = (*m_transmittance)(uv).r;
        float prob = wi.z * wo.z < 0.0f ? transmittance : 1.0f - transmittance;
        return prob * PdfCosineWeightedHemisphere(wi);
    }

    BsdfSample sample( Sampler& sampler, const Vector& wo, const vec2& uv ) const
    {
        Vector wi = SampleCosineWeightedHemisphere(sampler.sample(), sampler.sample());
        float prob = (*m_transmittance)(uv).r;

        BsdfSample bs;
        if( sampler.sample() < prob )
        {
            bs.wi = -std::copysign(1.f, wo.z) * wi;
            bs.pdf = prob * PdfCosineWeightedHemisphere(wi);
        }
        else
        {
            bs.wi = std::copysign(1.f, wo.z) * wi;
            bs.pdf = (1.f - prob) * PdfCosineWeightedHemisphere(wi);
        }
        bs.weight = (*m_reflectance)(uv); 

        return bs; 
    }
};



////////////// CONDUCTOR BSDF //////////////

class BsdfConductor : public Bsdf
{
public:
    // constructor with complex IOR
    // eval using full Fresnel equations
    BsdfConductor(ComplexIor ior = CopperIor) :
        Bsdf(BSDF_SPECULARREFLECTION),
        m_complexIor(ior) {}

    // constructor with tint (R0)
    // eval using Schlick Fresnel approximation
    BsdfConductor(Color tint = CopperTint) :
        Bsdf(BSDF_SPECULARREFLECTION),
        m_useTint(true),
        m_tint(tint) {}

    // constructor with tint (R0) and edge tint (R82)
    // eval using Lazanyi - Schlick Fresnel approximation
    BsdfConductor(Color tint, Color edgeTint) :
        Bsdf(BSDF_SPECULARREFLECTION),
        m_useEdgeTint(true),
        m_tint(tint),
        m_edgeTint(edgeTint) {}

    // constructor with tint (R0)
    // eval using Schlick Fresnel approximation
    BsdfConductor(std::shared_ptr<Texture> tint = std::make_shared<TextureConstant>(CopperTint)) :
        Bsdf(BSDF_SPECULARREFLECTION),
        m_useTint(true),
        m_reflectance(tint) {}

    // constructor with tint (R0) and edge tint (R82)
    // eval using Lazanyi - Schlick Fresnel approximation
    BsdfConductor(
        std::shared_ptr<Texture> tint, 
        std::shared_ptr<Texture> edgeTint
    ) :
        Bsdf(BSDF_SPECULARREFLECTION),
        m_useEdgeTint(true),
        m_reflectance(tint),
        m_edge_reflectance(edgeTint) {}

    ComplexIor m_complexIor;
    bool m_useTint = false, m_useEdgeTint = false;
    Color m_tint, m_edgeTint;
    std::shared_ptr<Texture> m_reflectance, m_edge_reflectance;

    Color eval( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        if(CheckReflectionConstraint(wi, wo))
        {
            if( m_useEdgeTint )
            {
                const Color tint = m_reflectance ? (*m_reflectance)(uv) : m_tint;
                const Color edgeTint = m_edge_reflectance ? (*m_edge_reflectance)(uv) : m_edgeTint;
                return LazanyiSchlickConductorReflectance(tint, edgeTint, wo.z) / std::abs(wi.z);
            } 
            if( m_useTint )
            {
                const Color tint = m_reflectance ? (*m_reflectance)(uv) : m_tint;
                return SchlickTintedConductorReflectance(tint, wo.z) / std::abs(wi.z);
            } 
            return ConductorReflectance(m_complexIor.eta, m_complexIor.k, wo.z) / std::abs(wi.z);
        }
        return Black();
    }
    
    float pdf( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        if(wi.z * wo.z <= 0) return 0;
        return CheckReflectionConstraint(wi, wo);
    }
    
    BsdfSample sample( Sampler& sampler, const Vector& wo, const vec2& uv ) const
    {
        BsdfSample bs;
        bs.wi = Reflect(wo);
        bs.pdf = 1;

        if( m_useEdgeTint )
        {
            const Color tint = m_reflectance ? (*m_reflectance)(uv) : m_tint;
            const Color edgeTint = m_edge_reflectance ? (*m_edge_reflectance)(uv) : m_edgeTint;
            bs.weight = LazanyiSchlickConductorReflectance(tint, edgeTint, wo.z);
        } 
        else if( m_useTint )
        {
            const Color tint = m_reflectance ? (*m_reflectance)(uv) : m_tint;
            bs.weight = SchlickTintedConductorReflectance(tint, wo.z);
        } 
        else 
            bs.weight = ConductorReflectance(m_complexIor.eta, m_complexIor.k, wo.z);

        return bs;
    }
};



////////////// DIELECTRIC BSDF //////////////

class BsdfDielectric : public Bsdf
{
public:
    BsdfDielectric(
        float intIOR = 1.5f, 
        float extIOR = 1.f, 
        std::shared_ptr<Texture> reflectance = nullptr, 
        std::shared_ptr<Texture> transmittance = nullptr 
    ) : 
        Bsdf(BSDF_SPECULAR),
        m_intIOR(intIOR), 
        m_extIOR(extIOR), 
        m_eta(m_intIOR / m_extIOR), 
        m_invEta(1.f / m_eta),
        m_reflectance(reflectance), 
        m_transmittance(transmittance) 
    {}

    float m_intIOR, m_extIOR, m_eta, m_invEta;
    std::shared_ptr<Texture> m_reflectance, m_transmittance;

    Color eval( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        if(wi.z == 0 || wo.z == 0) return Black();

        float eta = wo.z < 0.0f ? m_eta : m_invEta;
        float cosThetaT = 0.0f;
        float F = DielectricReflectance(eta, std::abs(wo.z), cosThetaT);

        if(wo.z * wi.z >= 0.0f) 
        {
            if(CheckReflectionConstraint(wi, wo))
                // add a tint outside the object
                return (m_reflectance ? (*m_reflectance)(uv) : White()) * F / std::abs(wi.z);
            else
                return Black();
        } 
        else 
        {
            if(CheckRefractionConstraint(wi, wo, eta, cosThetaT))
                // add a tint inside the object
                // better do that with transmittance color at distance and Beer-Lambert later ...
                return (m_transmittance ? (*m_transmittance)(uv) : White()) * (1.0f - F) / std::abs(wi.z);
            else
                return Black();
        }
    }
    
    float pdf( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        if(wi.z == 0 || wo.z == 0) return 0;

        float eta = wo.z < 0.0f ? m_eta : m_invEta;
        float cosThetaT = 0.0f;
        float F = DielectricReflectance(eta, std::abs(wo.z), cosThetaT);

        if (wo.z * wi.z >= 0.0f) 
        {
            if(CheckReflectionConstraint(wi, wo))
                return F;
            else
                return 0.0f;
        } 
        else 
        {
            if(CheckRefractionConstraint(wi, wo, eta, cosThetaT))
                return 1.0f - F;
            else
                return 0.0f;
        }
    }

    BsdfSample sample( Sampler& sampler, const Vector& wo, const vec2& uv ) const
    {
        BsdfSample bs;

        float eta = wo.z < 0.0f ? m_eta : m_invEta;
        float cosThetaT = 0.0f;
        float F = DielectricReflectance(eta, std::abs(wo.z), cosThetaT);

        if(sampler.sample() <= F) 
        {
            bs.wi = Reflect(wo);
            bs.pdf = F;
            bs.weight = m_reflectance ? (*m_reflectance)(uv) : White();
        } 
        else 
        {
            if (F == 1.0f) 
                return bs;

            bs.wi = Vector(-wo.x * eta, -wo.y * eta, -std::copysign(cosThetaT, wo.z));
            bs.pdf = 1.f - F;
            bs.weight = m_transmittance ? (*m_transmittance)(uv) : White();
        }

        return bs;
    }
};



////////////// THIN DIELECTRIC BSDF //////////////

class BsdfThinDielectric : public Bsdf
{
public:
    BsdfThinDielectric(
        float intIOR = 1.5f, 
        float extIOR = 1.f, 
        std::shared_ptr<Texture> reflectance = nullptr, 
        std::shared_ptr<Texture> transmittance = nullptr 
    ) : 
        Bsdf((BsdfType)(BSDF_SPECULAR | BSDF_FORWARD) ),
        m_intIOR(intIOR), 
        m_extIOR(extIOR), 
        m_eta(m_intIOR / m_extIOR), 
        m_invEta(1.f / m_eta),
        m_reflectance(reflectance), 
        m_transmittance(transmittance) 
    {}

    bool isThinDielectric() const override { return true; }

    float m_intIOR, m_extIOR, m_eta, m_invEta;
    std::shared_ptr<Texture> m_reflectance, m_transmittance;

    Color eval( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        float eta = wo.z < 0.0f ? m_eta : m_invEta;
        float cosThetaT = 0.0f;
        float F = DielectricReflectance(eta, std::abs(wo.z), cosThetaT);
        // Account for internal reflections: r' = r + trt + tr^3t + ..
        if (F < 1)
            F += (1-F)*(1-F) * F / (1-F*F);
        
        // if (F < 1)
            // F *= 2.f / (1.f + F);

        if(wo.z * wi.z >= 0.0f) 
        {
            if(CheckReflectionConstraint(wi, wo))
                // add a tint outside the object
                return (m_reflectance ? (*m_reflectance)(uv) : White()) * F / std::abs(wi.z);
            else
                return Black();
        } 
        else 
        {
            if(wi.x == -wo.x && wi.y == -wo.y && wi.z == -wo.z)
                // add a tint inside the object
                // better do that with transmittance color at distance and Beer-Lambert later ...
                return (m_transmittance ? (*m_transmittance)(uv) : White()) * (1.0f - F) / std::abs(wi.z);
            else
                return Black();
        }
        
        return Black();
    }
    
    float pdf( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        if(wi.z == 0 || wo.z == 0) return 0;

        float eta = wo.z < 0.0f ? m_eta : m_invEta;
        float cosThetaT = 0.0f;
        float F = DielectricReflectance(eta, std::abs(wo.z), cosThetaT);
        // Account for internal reflections: r' = r + trt + tr^3t + ..
        if (F < 1)
            F += (1-F)*(1-F) * F / (1-F*F);
        
        // if (F < 1)
            // F *= 2.f / (1.f + F);

        if (wo.z * wi.z >= 0.0f) 
        {
            if(CheckReflectionConstraint(wi, wo))
                return F;
            else
                return 0.0f;
        } 
        else 
        {
            if(wi.x == -wo.x && wi.y == -wo.y && wi.z == -wo.z)
                return 1.0f - F;
            else
                return 0.0f;
        }

        return 0;
    }

    BsdfSample sample( Sampler& sampler, const Vector& wo, const vec2& uv ) const
    {
        BsdfSample bs;

        float eta = wo.z < 0.0f ? m_eta : m_invEta;
        float F = DielectricReflectance(eta, std::abs(wo.z));
        // Account for internal reflections: r' = r + trt + tr^3t + ..
        if (F < 1)
            F += (1-F)*(1-F) * F / (1-F*F);
        
        // if (F < 1)
            // F *= 2.f / (1.f + F);

        if(sampler.sample() <= F) 
        {
            bs.wi = Reflect(wo);
            bs.pdf = F;
            bs.weight = m_reflectance ? (*m_reflectance)(uv) : White();
            // printf("thin dielectric:: reflect cosi %f coso %f dot(wi, wo) %f\n", bs.wi.z, wo.z, dot(bs.wi, wo));
        } 
        else 
        {
            // if( F >= 1.f )
            //     return bs;

            bs.wi = -wo;
            bs.pdf = 1.f - F;
            bs.weight = (m_transmittance ? (*m_transmittance)(uv) : White());
            // printf("thin dielectric:: refract cosi %f coso %f dot(wi, wo) %f\n", bs.wi.z, wo.z, dot(bs.wi, wo));
        }

        return bs;
    }
};



///////////// PLASTIC BSDF /////////////

class BsdfPlastic : public Bsdf
{
public:
    BsdfPlastic(
        std::shared_ptr<Texture> diffuse_reflectance, 
        std::shared_ptr<Texture> specular_reflectance = nullptr, 
        float intIOR = 1.5f, 
        float extIOR = 1.f,
        bool nonlinear = false,
        float thickness = 1.f,
        Color transmittanceColor = White()
    ) : 
        Bsdf((BsdfType)(BSDF_SPECULARREFLECTION | BSDF_DIFFUSEREFLECTION)),
        m_diffuse_reflectance(diffuse_reflectance), 
        m_specular_reflectance(specular_reflectance), 
        m_intIOR(intIOR), 
        m_extIOR(extIOR), 
        m_eta(m_intIOR / m_extIOR), 
        m_invEta(1.f / m_eta),
        m_thickness(thickness),
        m_nonlinear(nonlinear)

    {
        m_sigmaA = -log(transmittanceColor) / m_thickness;
        
        m_scaledSigmaA = m_thickness * m_sigmaA;
        m_avgTransmittance = std::exp(-2.0f * m_scaledSigmaA.power());
        m_diffuseFresnel = FresnelDiffuseReflectanceApprox(m_invEta);
        m_substrateWeight = m_diffuse_reflectance->average().power();
    }

    std::shared_ptr<Texture> m_diffuse_reflectance, m_specular_reflectance;
    float m_intIOR, m_extIOR, m_eta, m_invEta;
    float m_thickness = 1.f;
    Color m_sigmaA = Black();
    float m_diffuseFresnel;
    float m_avgTransmittance;
    Color m_scaledSigmaA;
    bool m_nonlinear = false;
    float m_substrateWeight;
    
    Color eval( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        if (wi.z <= 0.0f || wo.z <= 0.0f)
            return Black();

        float Fo = DielectricReflectance(m_invEta, wo.z);
        float Fi = DielectricReflectance(m_invEta, wi.z);

        if(CheckReflectionConstraint(wi, wo))
        {
            return (m_specular_reflectance ? (*m_specular_reflectance)(uv) : White()) * Fo / std::abs(wi.z);
        } 
        else
        {
            Color diffuseAlbedo = (*m_diffuse_reflectance)(uv);
            Color weight = m_nonlinear ? 
                    (diffuseAlbedo / (White() - diffuseAlbedo * m_diffuseFresnel)):
                    (diffuseAlbedo / (1.f - m_diffuseFresnel));
            // weight.r = std::max(0.f, weight.r);
            // weight.g = std::max(0.f, weight.g);
            // weight.b = std::max(0.f, weight.b);

            weight = clamp(weight);
            
            Color brdf = ((1.0f - Fi)*(1.0f - Fo) * m_invEta * m_invEta / M_PI) * weight;

            if(m_scaledSigmaA.max() > 0.0f)
            {
                brdf = brdf * exp(m_scaledSigmaA*(-1.0f / wi.z - 1.0f / wo.z));
            }
            
            return brdf;
        } 
        
        return Black();
    }
    
    float pdf( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        if (wi.z <= 0.0f || wo.z <= 0.0f)
            return 0.0f;

        float Fo = DielectricReflectance(m_invEta, wo.z);
        float substrateWeight = m_substrateWeight * m_avgTransmittance * (1.0f - Fo);
        float specularWeight = Fo;
        float specularProbability = specularWeight / (specularWeight + substrateWeight);
        
        if(CheckReflectionConstraint(wi, wo))
        {
            return specularProbability;
        }
        else
        {
            return PdfCosineWeightedHemisphere(wi) * (1.f - specularProbability);
        }
    }

    BsdfSample sample( Sampler& sampler, const Vector& wo, const vec2& uv ) const
    {
        BsdfSample bs;
        
        float Fo = DielectricReflectance(m_invEta, wo.z);
        float substrateWeight = m_substrateWeight * m_avgTransmittance * (1.0f - Fo);
        float specularWeight = Fo;
        float specularProbability = specularWeight / (specularWeight + substrateWeight);

        if(sampler.sample() <= specularProbability) 
        {
            bs.wi = Reflect(wo);
            bs.pdf = specularProbability;
            bs.weight = (specularWeight + substrateWeight) * (m_specular_reflectance ? (*m_specular_reflectance)(uv) : White());
            // bs.weight = (specularWeight + substrateWeight) * Red(); //debug
        } 
        else 
        {
            bs.wi = SampleCosineWeightedHemisphere(sampler.sample(), sampler.sample());
            bs.pdf = (1.f - specularProbability) * PdfCosineWeightedHemisphere(bs.wi);

            Color diffuseAlbedo = (*m_diffuse_reflectance)(uv);
            Color weight = m_nonlinear ? 
                    clamp(diffuseAlbedo / (White() - diffuseAlbedo * m_diffuseFresnel)):
                    clamp(diffuseAlbedo / (1.f - m_diffuseFresnel));
            
            float Fi = DielectricReflectance(m_invEta, bs.wi.z);
            bs.weight = ((1.0f - Fi) * m_invEta * m_invEta * (specularWeight + substrateWeight) / m_avgTransmittance) * weight;

            if(m_scaledSigmaA.max() > 0.0f)
            {
                bs.weight = bs.weight * exp(m_scaledSigmaA*(-1.0f / bs.wi.z - 1.0f / wo.z));
            }
        }

        // pdf
        bs.pdf = specularProbability + (1.f - specularProbability) * PdfCosineWeightedHemisphere(bs.wi); 

        // final value
        bs.weight = bs.weight / bs.pdf;

        // bs.weight = eval(bs.wi, wo, uv)/bs.pdf*std::abs(bs.wi.z);

        return bs;
    }
};

// Linear roughness remapping
enum AnisotropicRemapping { Burley = 0, Kulla, Neubelt };

static float minAlpha = 1e-3f;

static void ComputeAnisotropicAlpha(
    const float linearRoughness, 
    const float linearAnisotropy, 
    float& alpha_x, 
    float& alpha_y, 
    AnisotropicRemapping remapping = AnisotropicRemapping::Kulla)
{
    // squared remapping from linear roughness to alpha
    float alpha = linearRoughness * linearRoughness;

    // isotropic -> no remapping
    if( linearAnisotropy <= 0 )
    {
        alpha_x = alpha;
        alpha_y = alpha;
    }
    // anisotropic -> remapping using the linear anisotropy parameter
    else
    {
        switch(remapping)
        {
            case AnisotropicRemapping::Burley:
            {
                float remap = std::sqrt(1.f - 0.9f * linearAnisotropy);
                alpha_x = alpha / remap;
                alpha_y = alpha * remap;
                break;
            }
            case AnisotropicRemapping::Kulla:
            {
                alpha_x = alpha * (1.f + linearAnisotropy);
                alpha_y = alpha * (1.f - linearAnisotropy);
                break;
            }
            case AnisotropicRemapping::Neubelt:
            {
                alpha_x = alpha;
                alpha_y = lerp(0.f, alpha, 1.f - linearAnisotropy);
                break;
            }
            default:
            {
                alpha_x = alpha;
                alpha_y = alpha;
                break;
            }
        }
    }

    // return alpha clamped to the minimum value
    alpha_x = std::max(minAlpha, alpha_x);
    alpha_y = std::max(minAlpha, alpha_y);
}

class BsdfRoughDielectric : public Bsdf
{
public:

    BsdfRoughDielectric(
        float intIOR = 1.5f, 
        float extIOR = 1.f,
        std::shared_ptr<Texture> reflectance = nullptr, 
        std::shared_ptr<Texture> transmittance = nullptr, 
        std::shared_ptr<Texture> roughness = std::make_shared<TextureConstant>(0.1f), 
        std::shared_ptr<Texture> anisotropy = nullptr,
        std::shared_ptr<MicroFacet> mf = std::make_shared<MicroFacetGGX>(),
        std::shared_ptr<Texture3D> compensationLowEta = nullptr,
        std::shared_ptr<Texture3D> compensationHighEta = nullptr
    ) : 
        Bsdf((anisotropy ? (BsdfType)(BSDF_GLOSSY | BSDF_ANISOTROPIC) : BSDF_GLOSSY)),
        m_intIOR(intIOR), 
        m_extIOR(extIOR), 
        m_eta(m_intIOR / m_extIOR), 
        m_invEta(1.f / m_eta),
        m_reflectance(reflectance),
        m_transmittance(transmittance),
        m_roughness(roughness), 
        m_anisotropy(anisotropy), 
        m_distribution(mf),  
        m_compensationLowEta(compensationLowEta),  
        m_compensationHighEta(compensationHighEta)  
    {}
    
    float m_intIOR, m_extIOR, m_eta, m_invEta;
    std::shared_ptr<Texture> m_reflectance;
    std::shared_ptr<Texture> m_transmittance;
    std::shared_ptr<Texture> m_roughness;
    std::shared_ptr<Texture> m_anisotropy;
    std::shared_ptr<MicroFacet> m_distribution;
    std::shared_ptr<Texture3D> m_compensationLowEta;
    std::shared_ptr<Texture3D> m_compensationHighEta;
    bool m_reflective = true;
    bool m_transmissive = true;

    float compensation( float cosTheta, float roughness, float eta ) const
    {
        float w = 1;
        if( m_compensationLowEta && m_compensationHighEta )
        {
            float x = cosTheta;
            float y = roughness;
            float Ess = 1;
            if( eta > 1 )
            {
                // remap z to [0, 1]
                float z = std::min(1.f, std::max(0.f, eta * 0.5f - 0.5f));
                Ess = (*m_compensationHighEta)(x, y, z).r;
            }
            else
            {
                // remap z to [0, 1]
                float z = std::min(1.f, std::max(0.f, eta * 1.5f - 0.5f));
                Ess = (*m_compensationLowEta)(x, y, z).r;
            }
            w = Ess > 0 ? 1.f / Ess : 1;
        }
        return w;
    }

    Color eval( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        //
        float wiDotN = wi.z;
        float woDotN = wo.z;

        if( wiDotN * woDotN == 0 )
            return Black();

        // get roughness value
        float roughness = (*m_roughness)(uv).r;
        float anisotropy = m_anisotropy ? (*m_anisotropy)(uv).r : 0;
        float alpha_x, alpha_y;
        ComputeAnisotropicAlpha(roughness, anisotropy, alpha_x, alpha_y);

        bool reflect = wiDotN * woDotN >= 0.0f;
        if ((reflect && !m_reflective) || (!reflect && !m_transmissive))
            return Black();

        float eta = woDotN < 0.0f ? m_eta : m_invEta;
        Vector wh;
        if (reflect)
            wh = std::copysign(1.f, woDotN) * normalize(wi + wo); // reflection ok
        else
            wh = -normalize(wo * eta + wi); // transmission ok

        float wiDotWh = dot(wi, wh);
        float woDotWh = dot(wo, wh);

        float F = DielectricReflectance(m_invEta, woDotWh);
        float G = m_distribution->G2(alpha_x, alpha_y, wh, wi, wo);
        float D = m_distribution->D(alpha_x, alpha_y, wh, wo);

        float w = compensation(std::abs(wo.z), roughness, eta);

        if (reflect) 
        {   
            float fr = (F * G * D * 0.25f) / std::abs(woDotN * wiDotN);
            return (m_reflectance ? (*m_reflectance)(uv) : White()) * fr * w;
        } 
        else 
        {
            if( F == 1.0f ) 
                return Black();

            float rr = eta * woDotWh + wiDotWh;
            float ft = std::abs(wiDotWh * woDotWh) / std::abs(wiDotN * woDotN) * (1.0f - F) * G * D / (rr * rr);
            return (m_transmittance ? (*m_transmittance)(uv) : White()) * ft * w;
        }

        return Black();
    }

    float pdf( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        // cosines 
        float wiDotN = wi.z;
        float woDotN = wo.z;

        // if one of cosine terme is zero return pdf of 0
        if( wiDotN * woDotN == 0 )
            return 0;

        // get roughness value
        float roughness = (*m_roughness)(uv).r;
        float anisotropy = m_anisotropy ? (*m_anisotropy)(uv).r : 0;
        float alpha_x, alpha_y;
        ComputeAnisotropicAlpha(roughness, anisotropy, alpha_x, alpha_y);

        // sampled event
        bool reflect = wiDotN * woDotN > 0.0f;
        if ((reflect && !m_reflective) || (!reflect && !m_transmissive))
            return 0.0f;

        float eta = woDotN < 0.0f ? m_eta : m_invEta;
        Vector wh;
        if (reflect)
            wh = std::copysign(1.f, woDotN) * normalize(wi + wo);
        else
            wh = -normalize(wo*eta + wi);

        float wiDotWh = dot(wi, wh);
        float woDotWh = dot(wo, wh);

        float F = DielectricReflectance(m_invEta, woDotWh);
        float pdf_wh = m_distribution->pdf(alpha_x, alpha_y, wh, wo);

        float pdf_wi = 0;
        if (reflect)
        {
            pdf_wi = pdf_wh * 0.25f / std::abs(woDotWh);
        }
        else
        {
            if( F == 1.0f ) 
                return 0;

            float rr = eta * woDotWh + wiDotWh;
            float jacobian = 1.f / (rr * rr);
            pdf_wi = pdf_wh * std::abs(wiDotWh) * jacobian;
        }

        if (m_reflective && m_transmissive) 
        {
            if (reflect)
                pdf_wi *= F;
            else
                pdf_wi *= 1.0f - F;
        }

        return pdf_wi;
    }
    
    BsdfSample sample( Sampler& sampler, const Vector& wo, const vec2& uv ) const
    {
        BsdfSample bs;

        // sampled event
        bool sampleR = m_reflective;
        bool sampleT = m_transmissive;

        // eta
        float woDotN = wo.z;
        float eta = woDotN < 0.0f ? m_eta : m_invEta;

        // get roughness value
        float roughness = (*m_roughness)(uv).r;
        float anisotropy = m_anisotropy ? (*m_anisotropy)(uv).r : 0;
        float alpha_x, alpha_y;
        ComputeAnisotropicAlpha(roughness, anisotropy, alpha_x, alpha_y);

        // sample microfacet normal
        Vector wh = m_distribution->sample(alpha_x, alpha_y, sampler.sample(), sampler.sample(), wo);
        float pdfWh = m_distribution->pdf(alpha_x, alpha_y, wh, wo);

        float woDotWh = dot(wo, wh);
        float cosThetaT = 0.0f;
        float F = DielectricReflectance(m_invEta, woDotWh, cosThetaT);
        float etaWh = woDotWh < 0.0f ? m_eta : m_invEta;

        bool reflect;
        if( sampleR && sampleT ) 
        {
            reflect = (sampler.sample() <= F);
        } 
        else if( sampleT ) 
        {
            if( F == 1.0f ) 
                return BsdfSample();
            reflect = false;
        } 
        else if( sampleR ) 
        {
            reflect = true;
        } 
        else 
        {
            return BsdfSample();
        }

        float w = compensation(std::abs(wo.z), roughness, eta);

        if( reflect )
        {
            // direction
            bs.wi = Reflect(wo, wh);
            
            // pdf
            float jacobian = 0.25f / std::abs(woDotWh);
            bs.pdf = pdfWh * jacobian * F;
            
            // value
            float G2 = m_distribution->G2(alpha_x, alpha_y, wh, bs.wi, wo);
            float G1 = m_distribution->G1(alpha_x, alpha_y, wh, wo);
            float fr = G1 > 0 ? w * G2 / G1 : 0;
            bs.weight = (m_reflectance ? (*m_reflectance)(uv) : White()) * fr;
        }
        else
        {
            // direction
            bs.wi = (etaWh * woDotWh - std::copysign(1.f, woDotWh) * cosThetaT) * wh - etaWh * wo;
            
            // pdf
            float wiDotWh = dot(bs.wi, wh);
            float r = eta * woDotWh + wiDotWh;
            float jacobian = std::abs(wiDotWh) / (r * r);
            bs.pdf = pdfWh * jacobian * (1.f - F);

            // value
            float G2 = m_distribution->G2(alpha_x, alpha_y, wh, bs.wi, wo);
            float G1 = m_distribution->G1(alpha_x, alpha_y, wh, wo);
            float ft = G1 > 0 ? w * G2 / G1 : 0;
            bs.weight = (m_transmittance ? (*m_transmittance)(uv) : White()) * ft;
        }

        // if(bs.pdf < 1e-8f)
            // return BsdfSample();

        float wiDotN = bs.wi.z;

        bool reflected = wiDotN * woDotN > 0.0f;
        if (reflected != reflect)
            return BsdfSample();

        return bs;
    }
};

class BsdfRoughDiffuse : public Bsdf
{
public:
    // Helper function: creates 1D uniform random numbers
    // needed in the bsdf evaluation
    // to change 
    static float rand1D() 
    {   
        using G = std::default_random_engine;    
        using D = std::uniform_real_distribution<float>;    
        
        static std::vector<G> gen = std::vector<G>(omp_get_max_threads());
        static std::vector<D> dist = std::vector<D>(omp_get_max_threads(), D(0.f, 1.f));

        return dist[omp_get_thread_num()](gen[omp_get_thread_num()]);
    }

    std::shared_ptr<Texture> m_reflectance;
    std::shared_ptr<Texture> m_roughness;
    std::shared_ptr<Texture> m_anisotropy;
    std::shared_ptr<MicroFacet> m_distribution;
    std::shared_ptr<Texture> m_compensation;

    BsdfRoughDiffuse(
            std::shared_ptr<Texture> reflectance, 
            std::shared_ptr<Texture> roughness = std::make_shared<TextureConstant>(0.1f), 
            std::shared_ptr<Texture> anisotropy = nullptr, 
            std::shared_ptr<MicroFacet> mf = std::make_shared<MicroFacetGGX>(),
            std::shared_ptr<Texture> compensation = nullptr 
        ) : 
        Bsdf((anisotropy ? (BsdfType)(BSDF_DIFFUSEREFLECTION | BSDF_ANISOTROPIC) : BSDF_DIFFUSEREFLECTION)),
        m_reflectance(reflectance),
        m_roughness(roughness),
        m_anisotropy(anisotropy),
        m_distribution(mf),
        m_compensation(compensation)
    {}

    float compensation( float cosTheta, float roughness ) const
    {
        float w = 1;
        if( m_compensation )
        {
            float x = cosTheta;
            float y = roughness;
            float Ess = (*m_compensation)(x, y).r;
            w = 1.f / Ess;
        }
        return w;
    }

    Color eval( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        // cosines 
        float wiDotN = wi.z;
        float woDotN = wo.z;

        // if one of cosine terme is zero return 0
        if( wiDotN * woDotN == 0 ) return Black();

        // get roughness value
        float roughness = (*m_roughness)(uv).r;
        float anisotropy = m_anisotropy ? (*m_anisotropy)(uv).r : 0;
        float alpha_x, alpha_y;
        ComputeAnisotropicAlpha(roughness, anisotropy, alpha_x, alpha_y);

        // sample microfacet normal
        Vector wh = m_distribution->sample(alpha_x, alpha_y, rand1D(), rand1D(), wo);

        // compute bsdf value
        float wiDotWh = std::max(0.f, dot(wi, wh));
        float G2 = m_distribution->G2(alpha_x, alpha_y, wh, wi, wo);
        float G1 = m_distribution->G1(alpha_x, alpha_y, wh, wo);
        float w = compensation( std::max(0.f, wo.z), roughness );

        return (*m_reflectance)(uv) * wiDotWh * G2 / G1 / M_PI * w;
    }
    
    float pdf( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        if( wi.z <= 0 || wo.z <= 0 ) return 0;
        return PdfCosineWeightedHemisphere(wi);
    }

    BsdfSample sample( Sampler& sampler, const Vector& wo, const vec2& uv ) const
    {
        BsdfSample bs;

        // get roughness value
        float roughness = (*m_roughness)(uv).r;
        float anisotropy = m_anisotropy ? (*m_anisotropy)(uv).r : 0;
        float alpha_x, alpha_y;
        ComputeAnisotropicAlpha(roughness, anisotropy, alpha_x, alpha_y);

        // sample microfacet normal
        Vector wh = m_distribution->sample(alpha_x, alpha_y, sampler.sample(), sampler.sample(), wo);

        // sample diffuse direction in the hemisphere centered on the microfacet normal
        Frame frame(wh);
        int test = 0;

        // try several times to get a sample in the upper hemisphere 
        while( bs.wi.z <= 0 && test < 10 )
        {
            bs.wi = frame.ToWorld(SampleCosineWeightedHemisphere(sampler.sample(), sampler.sample()));
            ++test;
        }

        // if the sampling failed
        if(bs.wi.z <= 0) 
            return BsdfSample();

        // pdf
        bs.pdf = PdfCosineWeightedHemisphere(bs.wi);

        // value
        float wiDotWh = std::max(0.f, dot(bs.wi, wh));
        float G2 = m_distribution->G2(alpha_x, alpha_y, wh, bs.wi, wo);
        float G1 = m_distribution->G1(alpha_x, alpha_y, wh, wo);
        float w = compensation( std::max(0.f, wo.z), roughness );
        bs.weight = (*m_reflectance)(uv) * wiDotWh * G2 / G1 * w;

        return bs;
    }
};

class BsdfOrenNayar : public Bsdf
{
public:
    BsdfOrenNayar(
        std::shared_ptr<Texture> reflectance, 
        std::shared_ptr<Texture> roughness = std::make_shared<TextureConstant>(0.1f),
        std::shared_ptr<Texture> compensation = nullptr
    ) : 
        Bsdf(BSDF_DIFFUSEREFLECTION),
        m_reflectance(reflectance),
        m_roughness(roughness),
        m_compensation(compensation) 
    {}

    std::shared_ptr<Texture> m_reflectance;
    std::shared_ptr<Texture> m_roughness;
    std::shared_ptr<Texture> m_compensation;
    
    static float LinearRoughnessToSigma(const float linearRoughness)
    {   
        // remap to [0, pi/2], the std dev of the gaussian surface
        return linearRoughness * M_PI * 0.5f;
    }

    float compensation( float cosTheta, float roughness ) const
    {
        float w = 1;
        if( m_compensation )
        {
            float x = cosTheta;
            float y = roughness;
            float Ess = (*m_compensation)(x, y).r;
            w = 1.f / Ess;
        }
        return w;
    }

    Color eval( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {   
        // get roughness value
        float roughness = (*m_roughness)(uv).r;
        
        // roughness parameter is considered linear
        float sigma = LinearRoughnessToSigma(roughness);

        //
        float sinThetaIsq = std::max(0.f, 1.f - wi.z * wi.z);
        float sinThetaI = std::sqrt(sinThetaIsq);
        float sinThetaOsq = std::max(0.f, 1.f - wo.z * wo.z);
        float sinThetaO = std::sqrt(sinThetaOsq);

        // Compute cosine term of Oren-Nayar model
        float maxCos = 0.f;
        if (sinThetaI > 1e-4f && sinThetaO > 1e-4f) 
        {
            float phiI = std::atan2(wi.y, wi.x);
            float cosPhiI = std::cos(phiI);
            float sinPhiI = std::sin(phiI);

            float phiO = std::atan2(wo.y, wo.x);
            float cosPhiO = std::cos(phiO);
            float sinPhiO = std::sin(phiO);

            float dCos = cosPhiI * cosPhiO + sinPhiI * sinPhiO;
            maxCos = std::max(0.f, dCos);
        }

        // Compute sine and tangent terms of Oren-Nayar model
        float sinAlpha, tanBeta;
        if(std::abs(wi.z) > std::abs(wo.z)) 
        {
            sinAlpha = sinThetaO;
            tanBeta = sinThetaI / std::abs(wi.z);
        } 
        else 
        {
            sinAlpha = sinThetaI;
            tanBeta = sinThetaO / std::abs(wo.z);
        }

        float sigma2 = sigma * sigma;
        float A = 1.f - (sigma2 / (2.f * (sigma2 + 0.33f)));
        float B = 0.45f * sigma2 / (sigma2 + 0.09f);

        float w = compensation( std::max(0.f, wo.z), roughness );
        return (*m_reflectance)(uv) / M_PI * (A + B * maxCos * sinAlpha * tanBeta) * w;
    }
    
    float pdf( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        return wi.z <= 0 ? 0 : PdfCosineWeightedHemisphere(wi);
    }

    BsdfSample sample( Sampler& sampler, const Vector& wo, const vec2& uv ) const
    {   
        BsdfSample bs;
        bs.wi = SampleCosineWeightedHemisphere(sampler.sample(), sampler.sample());
        bs.pdf = bs.wi.z <= 0 ? 0 : PdfCosineWeightedHemisphere(bs.wi);
        bs.weight = bs.pdf > 0 ? eval(bs.wi, wo, uv) * std::abs(bs.wi.z) / bs.pdf : Black();
        return bs;
    }
};

class BsdfRoughConductor : public Bsdf
{
public:
    // constructor with complex IOR
    // eval using fulll Fresnel equations
    BsdfRoughConductor(
        ComplexIor ior = CopperIor,
        std::shared_ptr<Texture> roughness = std::make_shared<TextureConstant>(0.1f), 
        std::shared_ptr<Texture> anisotropy = nullptr, 
        std::shared_ptr<MicroFacet> mf = std::make_shared<MicroFacetGGX>(),
        std::shared_ptr<Texture> compensation = nullptr 
    ) :
        Bsdf(),
        m_roughness(roughness),
        m_distribution(mf),
        m_anisotropy(anisotropy),
        m_complexIor(ior),
        m_compensation(compensation)
    {
        if( m_anisotropy )
            m_type = (BsdfType)(BSDF_GLOSSYREFLECTION | BSDF_ANISOTROPIC);
        else
            m_type = BSDF_GLOSSYREFLECTION;
    }

    // constructor with tint (R0)
    // eval using Schlick Fresnel approximation
    BsdfRoughConductor(
        // Color tint,
        std::shared_ptr<Texture> tint, 
        std::shared_ptr<Texture> roughness = std::make_shared<TextureConstant>(0.1f), 
        std::shared_ptr<MicroFacet> mf = std::make_shared<MicroFacetGGX>(),
        std::shared_ptr<Texture> anisotropy = nullptr, 
        std::shared_ptr<Texture> compensation = nullptr 
    ) :
        Bsdf(),
        m_roughness(roughness),
        m_distribution(mf),
        m_anisotropy(anisotropy),
        m_useTint(true),
        // m_tint(tint),
        m_reflectance(tint),
        m_compensation(compensation)
    {
        if( m_anisotropy )
            m_type = (BsdfType)(BSDF_GLOSSYREFLECTION | BSDF_ANISOTROPIC);
        else
            m_type = BSDF_GLOSSYREFLECTION;
    }

    // constructor with tint (R0) and edge tint (R82)
    // eval using Lazanyi - Schlick Fresnel approximation
    BsdfRoughConductor(
        // Color tint,
        // Color edgeTint,
        std::shared_ptr<Texture> tint, 
        std::shared_ptr<Texture> edgeTint, 
        std::shared_ptr<Texture> roughness, 
        std::shared_ptr<MicroFacet> mf = std::make_shared<MicroFacetGGX>(),
        std::shared_ptr<Texture> anisotropy = nullptr, 
        std::shared_ptr<Texture> compensation = nullptr 
    ) :
        Bsdf(),
        m_roughness(roughness),
        m_distribution(mf),
        m_anisotropy(anisotropy),
        m_useEdgeTint(true),
        // m_tint(tint),
        // m_edgeTint(edgeTint),
        m_reflectance(tint),
        m_edge_reflectance(edgeTint),
        m_compensation(compensation)
    {
        if( m_anisotropy )
            m_type = (BsdfType)(BSDF_GLOSSYREFLECTION | BSDF_ANISOTROPIC);
        else
            m_type = BSDF_GLOSSYREFLECTION;
    }

    std::shared_ptr<Texture> m_roughness;
    std::shared_ptr<MicroFacet> m_distribution;
    std::shared_ptr<Texture> m_anisotropy;
    ComplexIor m_complexIor;
    bool m_useTint = false, m_useEdgeTint = false;
    Color m_tint, m_edgeTint;
    std::shared_ptr<Texture> m_reflectance;
    std::shared_ptr<Texture> m_edge_reflectance;
    bool m_useComp;
    std::shared_ptr<Texture> m_compensation;

    Color eval( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        // cosines 
        float wiDotN = wi.z;
        float woDotN = wo.z;

        // if one of cosine terme is zero return pdf of 0
        if( wiDotN * woDotN == 0 )
            return Black();

        // get roughness value
        float roughness = (*m_roughness)(uv).r;
        float anisotropy = m_anisotropy ? (*m_anisotropy)(uv).r : 0;
        float alpha_x, alpha_y;
        ComputeAnisotropicAlpha(roughness, anisotropy, alpha_x, alpha_y);

        // compute microfacet normal
        Vector wh = normalize(wi + wo);
        float cosTheta = dot(wo, wh);

        Color F;
        // if( m_useEdgeTint )
        //     F = LazanyiSchlickConductorReflectance(m_tint, m_edgeTint, cosTheta);
        // else if( m_useTint ) 
        //     F = SchlickTintedConductorReflectance(m_tint, cosTheta);
        // else
        //     F = ConductorReflectance(m_complexIor.eta, m_complexIor.k, cosTheta);

        if( m_useEdgeTint )
        {
            const Color tint = m_reflectance ? (*m_reflectance)(uv) : m_tint;
            const Color edgeTint = m_edge_reflectance ? (*m_edge_reflectance)(uv) : m_edgeTint;
            F = LazanyiSchlickConductorReflectance(tint, edgeTint, cosTheta);
        } 
        else if( m_useTint )
        {
            const Color tint = m_reflectance ? (*m_reflectance)(uv) : m_tint;
            F = SchlickTintedConductorReflectance(tint, cosTheta);
        } 
        else
            F = ConductorReflectance(m_complexIor.eta, m_complexIor.k, cosTheta);

        float G = m_distribution->G2(alpha_x, alpha_y, wh, wi, wo);
        float D = m_distribution->D(alpha_x, alpha_y, wh, wo);
        float fr = G * D / (4.f * std::abs(woDotN * wiDotN));

        Color rho; 
        if( m_compensation )
        {
            // emmanuel turquin's compensation term
            // cf: https://blog.selfshadow.com/publications/turquin/ms_comp_final.pdf
            float x = std::max(0.f, wo.z);
            float y = roughness;
            float Ess = (*m_compensation)(x, y).r;
            rho = Color(F * fr) + (1.f - Ess) / Ess * F * F * fr;
        }
        else
        {
            rho = F * fr;
        }

        return Color(rho, 1);
    }
    
    float pdf( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        // cosines 
        float wiDotN = wi.z;
        float woDotN = wo.z;

        // if one of cosine terme is zero return pdf of 0
        if( wiDotN * woDotN == 0 )
            return 0;

        // get roughness value
        float roughness = (*m_roughness)(uv).r;
        float anisotropy = m_anisotropy ? (*m_anisotropy)(uv).r : 0;
        float alpha_x, alpha_y;
        ComputeAnisotropicAlpha(roughness, anisotropy, alpha_x, alpha_y);

        // compute microfacet normal
        Vector wh = normalize(wi + wo);
        float pdf_wh = m_distribution->pdf(alpha_x, alpha_y, wh, wo);
        return pdf_wh * 0.25f / std::abs(dot(wo, wh));
    }

    BsdfSample sample( Sampler& sampler, const Vector& wo, const vec2& uv ) const
    {   
        BsdfSample bs;

        // get roughness value
        float roughness = (*m_roughness)(uv).r;
        float anisotropy = m_anisotropy ? (*m_anisotropy)(uv).r : 0;
        float alpha_x, alpha_y;
        ComputeAnisotropicAlpha(roughness, anisotropy, alpha_x, alpha_y);

        // sample microfacet normal
        Vector wh = m_distribution->sample(alpha_x, alpha_y, sampler.sample(), sampler.sample(), wo);

        // direction
        bs.wi = Reflect(wo, wh);
        
        // pdf
        float pdfWh = m_distribution->pdf(alpha_x, alpha_y, wh, wo);
        float woDotWh = dot(wo, wh);
        bs.pdf = pdfWh * 0.25f / std::abs(woDotWh);

        // value
        float G2 = m_distribution->G2(alpha_x, alpha_y, wh, bs.wi, wo);
        float G1 = m_distribution->G1(alpha_x, alpha_y, wh, wo);
        float fr = G2 / G1;

        Color F;
        // if( m_useEdgeTint )
            // F = LazanyiSchlickConductorReflectance(m_tint, m_edgeTint, woDotWh);
        // else if( m_useTint ) 
            // F = SchlickTintedConductorReflectance(m_tint, woDotWh);
        // else
            // F = ConductorReflectance(m_complexIor.eta, m_complexIor.k, woDotWh);

        if( m_useEdgeTint )
        {
            const Color tint = m_reflectance ? (*m_reflectance)(uv) : m_tint;
            const Color edgeTint = m_edge_reflectance ? (*m_edge_reflectance)(uv) : m_edgeTint;
            F = LazanyiSchlickConductorReflectance(tint, edgeTint, woDotWh);
        } 
        else if( m_useTint )
        {
            const Color tint = m_reflectance ? (*m_reflectance)(uv) : m_tint;
            F = SchlickTintedConductorReflectance(tint, woDotWh);
        } 
        else
            F = ConductorReflectance(m_complexIor.eta, m_complexIor.k, woDotWh);

        if( m_compensation )
        {
            // emmanuel turquin's compensation term
            // cf: https://blog.selfshadow.com/publications/turquin/ms_comp_final.pdf
            float x = std::max(0.f, wo.z);
            float y = roughness;
            float Ess = (*m_compensation)(x, y).r;
            bs.weight = fr * F + (1.f - Ess) / Ess * fr * F * F;
        }
        else
        {
            bs.weight = fr * F;
        }

        return bs; 
    }
};

class BsdfRoughPlastic : public Bsdf
{
public:
    BsdfRoughPlastic(
        std::shared_ptr<Texture> diffuse_reflectance, 
        std::shared_ptr<Texture> specular_reflectance = nullptr, 
        std::shared_ptr<Texture> roughness = std::make_shared<TextureConstant>(0.1f), 
        std::shared_ptr<Texture> anisotropy = nullptr, 
        std::shared_ptr<MicroFacet> mf = std::make_shared<MicroFacetGGX>(),
        float intIOR = 1.5f, 
        float extIOR = 1.f, 
        bool nonlinear = false,
        float thickness = 1.f,
        Color transmittanceColor = White()) :

        Bsdf(),
        m_diffuse_reflectance(diffuse_reflectance), 
        m_specular_reflectance(specular_reflectance), 
        m_roughness(roughness),
        m_anisotropy(anisotropy),
        m_distribution(mf),
        m_intIOR(intIOR), 
        m_extIOR(extIOR), 
        m_eta(m_intIOR / m_extIOR), 
        m_invEta(1.f / m_eta),
        m_nonlinear(nonlinear),
        m_thickness(thickness)
    {
        if( m_anisotropy )
            m_type = (BsdfType)(BSDF_GLOSSYREFLECTION | BSDF_DIFFUSEREFLECTION | BSDF_ANISOTROPIC);
        else
            m_type = (BsdfType)(BSDF_GLOSSYREFLECTION | BSDF_DIFFUSEREFLECTION);

        m_sigmaA = -log(transmittanceColor) / m_thickness;

        m_scaledSigmaA = m_thickness * m_sigmaA;
        m_avgTransmittance = std::exp(-2.0f * m_scaledSigmaA.power());
        m_diffuseFresnel = FresnelDiffuseReflectanceApprox(m_invEta);
        m_substrateWeight = m_diffuse_reflectance->average().power();
    }

    std::shared_ptr<Texture> m_diffuse_reflectance;
    std::shared_ptr<Texture> m_specular_reflectance;
    std::shared_ptr<Texture> m_roughness;
    std::shared_ptr<Texture> m_anisotropy;
    std::shared_ptr<MicroFacet> m_distribution;
    float m_intIOR, m_extIOR, m_eta, m_invEta;
    bool m_nonlinear = false;
    float m_thickness;
    Color m_sigmaA = Black();
    float m_diffuseFresnel;
    float m_avgTransmittance;
    Color m_scaledSigmaA;
    float m_substrateWeight;
    
    Color eval( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        if (wi.z <= 0.0f || wo.z <= 0.0f)
            return Black();

        // get roughness value
        float roughness = (*m_roughness)(uv).r;
        float anisotropy = m_anisotropy ? (*m_anisotropy)(uv).r : 0;
        float alpha_x, alpha_y;
        ComputeAnisotropicAlpha(roughness, anisotropy, alpha_x, alpha_y);

        // compute microfacet normal
        Vector wh = normalize(wi + wo);

        // compute glossy bsdf value
        float F = DielectricReflectance(m_invEta, dot(wo, wh));
        float G = m_distribution->G2(alpha_x, alpha_y, wh, wi, wo);
        float D = m_distribution->D(alpha_x, alpha_y, wh, wo);
        // the cosine term wi.z is added because we multiply by cosine in the main loop...
        float fr = (F * G * D * 0.25f) / std::abs(wo.z * wi.z);
        Color glossyR = (m_specular_reflectance ? (*m_specular_reflectance)(uv) : White()) * fr;

        // compute diffuse bsdf value
        float Fo = DielectricReflectance(m_invEta, wo.z);
        float Fi = DielectricReflectance(m_invEta, wi.z);

        Color diffuseAlbedo = (*m_diffuse_reflectance)(uv);
        Color weight = m_nonlinear ? 
                (diffuseAlbedo / (White() - diffuseAlbedo * m_diffuseFresnel)):
                (diffuseAlbedo / (1.f - m_diffuseFresnel));

        weight = clamp(weight);

        // the cosine term wi.z should be removed here because we already multiply by cosine in the main loop...
        // Color diffuseR = ((1.0f - Fi) * (1.0f - Fo) * m_invEta * m_invEta * wi.z / M_PI) * weight;
        
        // enregy conservation / reciprocity term * coating compression * substrate bsdf
        Color diffuseR = ((1.f - Fi) * (1.f - Fo) * m_invEta * m_invEta / M_PI) * weight;

        if(m_scaledSigmaA.max() > 0)
        {
            diffuseR = diffuseR * exp(m_scaledSigmaA*(-1.f / wi.z - 1.f / wo.z));
        }

        return (glossyR + diffuseR);
    }
    
    float pdf( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        if (wi.z <= 0.0f || wo.z <= 0.0f)
            return 0.0f;
        
        // get roughness value
        float roughness = (*m_roughness)(uv).r;
        float anisotropy = m_anisotropy ? (*m_anisotropy)(uv).r : 0;
        float alpha_x, alpha_y;
        ComputeAnisotropicAlpha(roughness, anisotropy, alpha_x, alpha_y);

        // compute microfacet normal
        Vector wh = normalize(wi + wo);

        // compute microfacet pdf
        float pdf_wh = m_distribution->pdf(alpha_x, alpha_y, wh, wo);

        // compute reflected direction pdf for glossy reflection
        float glossyPdf = pdf_wh * 0.25f / std::abs(dot(wo, wh));

        // compute reflected direction pdf for diffuse reflection
        float diffusePdf = PdfCosineWeightedHemisphere(wi);

        // weight by specular prob
        float Fo = DielectricReflectance(m_invEta, wo.z);
        float substrateWeight = m_substrateWeight * m_avgTransmittance * (1.0f - Fo);
        float specularWeight = Fo;
        float specularProbability = specularWeight / (specularWeight + substrateWeight);

        glossyPdf *= specularProbability;
        diffusePdf *= 1 - specularProbability;

        return glossyPdf + diffusePdf;
    }
    
    BsdfSample sample( Sampler& sampler, const Vector& wo, const vec2& uv ) const
    {
        BsdfSample bs;
        if(wo.z <= 0.0f) 
            return bs;
        
        // get roughness value
        float roughness = (*m_roughness)(uv).r;
        float anisotropy = m_anisotropy ? (*m_anisotropy)(uv).r : 0;
        float alpha_x, alpha_y;
        ComputeAnisotropicAlpha(roughness, anisotropy, alpha_x, alpha_y);

        // compute glossy probability
        float Fo = DielectricReflectance(m_invEta, wo.z);
        float substrateWeight = m_substrateWeight * m_avgTransmittance * (1.0f - Fo);
        float specularWeight = Fo;
        float specularProbability = specularWeight / (specularWeight + substrateWeight);

        // sample rough dielectric coating
        Vector wh;
        if(sampler.sample() <= specularProbability) 
        {

            // sample microfacet normal
            wh = m_distribution->sample(alpha_x, alpha_y, sampler.sample(), sampler.sample(), wo);
            
            // direction
            bs.wi = Reflect(wo, wh);
        } 
        // sample diffuse substrate
        else 
        {
            bs.wi = SampleCosineWeightedHemisphere(sampler.sample(), sampler.sample());
            wh = normalize(bs.wi + wo);
        }

        // pdf
        float pdf_wh = m_distribution->pdf(alpha_x, alpha_y, wh, wo);
        bs.pdf = specularProbability * pdf_wh * 0.25f / std::abs(dot(wo, wh)) + 
                (1.f - specularProbability) * PdfCosineWeightedHemisphere(bs.wi); 

        // compute glossy bsdf value
        float F = DielectricReflectance(m_invEta, dot(wo, wh));
        float G = m_distribution->G2(alpha_x, alpha_y, wh, bs.wi, wo);
        float D = m_distribution->D(alpha_x, alpha_y, wh, wo);
        float fr = (F * G * D * 0.25f) / std::abs(wo.z);
        Color glossyR = (m_specular_reflectance ? (*m_specular_reflectance)(uv) : White()) * fr;

        // compute diffuse bsdf value
        float Fi = DielectricReflectance(m_invEta, bs.wi.z);

        Color diffuseAlbedo = (*m_diffuse_reflectance)(uv);
        Color weight = m_nonlinear ? 
                (diffuseAlbedo / (White() - diffuseAlbedo * m_diffuseFresnel)):
                (diffuseAlbedo / (1.f - m_diffuseFresnel));
        weight = clamp(weight);

        // enregy conservation / reciprocity term * coating compression * substrate bsdf
        Color diffuseR = ((1.f - Fi) * (1.f - Fo) * m_invEta * m_invEta * std::abs(bs.wi.z) / M_PI) * weight;

        if(m_scaledSigmaA.max() > 0.0f)
        {
            diffuseR = diffuseR * exp(m_scaledSigmaA*(-1.0f / bs.wi.z - 1.0f / wo.z));
        }

        // final value
        bs.weight = (glossyR + diffuseR) / bs.pdf;

        return bs;
    }
};



///////////// MIX BSDF /////////////

class BsdfMix : public Bsdf
{
public:
    BsdfMix(
        std::shared_ptr<Bsdf> bsdf1,
        std::shared_ptr<Bsdf> bsdf2,
        std::shared_ptr<Texture> mix = std::make_shared<TextureConstant>(Color(0.5f))
    ):
        m_bsdf1(bsdf1),
        m_bsdf2(bsdf2),
        m_mix(mix)
    {}

    std::shared_ptr<Bsdf> m_bsdf1;
    std::shared_ptr<Bsdf> m_bsdf2;
    std::shared_ptr<Texture> m_mix;

    Color eval( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        float mix = (*m_mix)(uv).r;
        // replace with lerp
        return mix * m_bsdf1->eval(wi, wo, uv) + std::max(0.f, (1.f - mix)) * m_bsdf2->eval(wi, wo, uv);
    }
    
    float pdf( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        float mix = (*m_mix)(uv).r;
        // replace with lerp
        return mix * m_bsdf1->pdf(wi, wo, uv) + std::max(0.f, (1.f - mix)) * m_bsdf2->pdf(wi, wo, uv);
    }
    
    BsdfSample sample( Sampler& sampler, const Vector& wo, const vec2& uv ) const
    {
        BsdfSample bs;
        float mix = (*m_mix)(uv).r;

        if(sampler.sample() < mix)
        {
            BsdfSample bs1 = m_bsdf1->sample(sampler, wo, uv);
            bs.wi = bs1.wi;
        }   
        else
        {
            BsdfSample bs2 = m_bsdf2->sample(sampler, wo, uv);
            bs.wi = bs2.wi;
        }
        
        bs.pdf = pdf(bs.wi, wo, uv);
        bs.weight = eval(bs.wi, wo, uv) * std::abs(bs.wi.z) / bs.pdf;
        return bs;
    }
};



///////////// SMOOTH COAT BSDF /////////////

class BsdfSmoothCoat : public Bsdf
{
public:
    BsdfSmoothCoat(
        std::shared_ptr<Bsdf> substrate = std::make_shared<BsdfConductor>(CopperIor), 
        float intIOR = 1.5f, 
        float extIOR = 1.f,
        std::shared_ptr<Texture> specular_reflectance = nullptr,
        float thickness = 1.f,
        Color transmittanceColor = White()
    ) : 
        Bsdf((BsdfType)(BSDF_SPECULARREFLECTION | substrate->type())),
        m_substrate(substrate), 
        m_intIOR(intIOR), 
        m_extIOR(extIOR), 
        m_eta(m_intIOR / m_extIOR), 
        m_invEta(1.f / m_eta),
        m_specular_reflectance(specular_reflectance),
        m_thickness(thickness)
    {
        m_sigmaA = -log(transmittanceColor) / m_thickness;
        m_scaledSigmaA = m_thickness * m_sigmaA;
        m_avgTransmittance = std::exp(-2.0f * m_scaledSigmaA.power());
    }

    std::shared_ptr<Bsdf> m_substrate;
    float m_intIOR, m_extIOR, m_eta, m_invEta;
    std::shared_ptr<Texture> m_specular_reflectance;
    float m_thickness = 1.f;
    Color m_sigmaA = Black();
    float m_avgTransmittance;
    Color m_scaledSigmaA;
    
    Color eval( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        if (wi.z <= 0.0f || wo.z <= 0.0f)
            return Black();

        float cosThetaTi, cosThetaTo;
        float Fo = DielectricReflectance(m_invEta, wo.z, cosThetaTo);
        float Fi = DielectricReflectance(m_invEta, wi.z, cosThetaTi);

        if(CheckReflectionConstraint(wi, wo))
        {
            return (m_specular_reflectance ? (*m_specular_reflectance)(uv) : White()) * Fo / std::abs(wi.z);
        } 
        else
        {
            Vector wiSubstrate(wi.x * m_invEta , wi.y * m_invEta , std::copysign(cosThetaTi, wi.z));
            Vector woSubstrate(wo.x * m_invEta , wo.y * m_invEta , std::copysign(cosThetaTo, wo.z));

            // float laplacian =  m_invEta * m_invEta * wi.z / cosThetaTi;
            float laplacian =  m_invEta * m_invEta / cosThetaTi;

            Color substrateF = m_substrate->eval(wiSubstrate, woSubstrate, uv);

            // remove the extra cosine term of perfectly specular materials
            if( m_substrate->isDelta() )
                substrateF = substrateF * std::abs(wiSubstrate.z);

            if(m_scaledSigmaA.max() > 0.0f)
            {
                substrateF = substrateF * exp(m_scaledSigmaA*(-1.0f / wi.z - 1.0f / wo.z));
            }

            return laplacian * (1.0f - Fi) * (1.0f - Fo) * substrateF;
        }
        
        return Black();
    }
    
    float pdf( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        if (wi.z <= 0.0f || wo.z <= 0.0f)
            return 0.0f;

        float cosThetaTi, cosThetaTo;
        float Fo = DielectricReflectance(m_invEta, wo.z, cosThetaTo);
        float Fi = DielectricReflectance(m_invEta, wi.z, cosThetaTi);
        
        Vector wiSubstrate(wi.x * m_invEta , wi.y * m_invEta , std::copysign(cosThetaTi, wi.z));
        Vector woSubstrate(wo.x * m_invEta , wo.y * m_invEta , std::copysign(cosThetaTo, wo.z));

        float substrateWeight = m_avgTransmittance * (1.0f - Fo);
        float specularWeight = Fo;
        float specularProbability = specularWeight / (specularWeight + substrateWeight);

        if(CheckReflectionConstraint(wi, wo))
            return specularProbability;
        else
            return m_substrate->pdf(wiSubstrate, woSubstrate, uv) *
                   (1.0f - specularProbability) *
                   m_invEta * m_invEta *
                   std::abs(wi.z / cosThetaTi);
    }
    
    BsdfSample sample( Sampler& sampler, const Vector& wo, const vec2& uv ) const
    {
        BsdfSample bs;

        if(wo.z <= 0.0f) 
            return BsdfSample();

        float cosThetaTo;
        float Fo = DielectricReflectance(m_invEta, wo.z, cosThetaTo);
        
        Vector woSubstrate(wo.x * m_invEta , wo.y * m_invEta , std::copysign(cosThetaTo, wo.z));

        float substrateWeight = m_avgTransmittance * (1.0f - Fo);
        float specularWeight = Fo;
        float specularProbability = specularWeight / (specularWeight + substrateWeight);

        if(sampler.sample() <= specularProbability) 
        {
            bs.wi = Reflect(wo);
            bs.pdf = specularProbability;
            bs.weight = (m_specular_reflectance ? (*m_specular_reflectance)(uv) : White()) * (specularWeight + substrateWeight);
        } 
        else 
        {
            BsdfSample substrateSample = m_substrate->sample(sampler, woSubstrate, uv);
            if( substrateSample.wi.z == 0 )
                return BsdfSample();

            float cosThetaTi;
            float Fi = DielectricReflectance(m_eta, substrateSample.wi.z, cosThetaTi);

            if (Fi >= 1.0f)
                return BsdfSample();

            // direction
            bs.wi = Vector(substrateSample.wi.x * m_eta, substrateSample.wi.y * m_eta, cosThetaTi);

            // pdf
            bs.pdf = substrateSample.pdf * 
                    (1.0f - specularProbability) *
                    m_invEta * m_invEta *
                    std::abs(bs.wi.z / cosThetaTi);

            // value
            float laplacian =  m_invEta * m_invEta * bs.wi.z / cosThetaTi;
            Color substrateF = m_substrate->eval(substrateSample.wi, woSubstrate, uv);

            // remove the extra cosine term of perfectly specular materials
            if( m_substrate->isDelta() )
                substrateF = substrateF * std::abs(substrateSample.wi.z);

            // transmittance effect
            if(m_scaledSigmaA.max() > 0.0f)
            {
                substrateF = substrateF * exp(m_scaledSigmaA*(-1.0f / bs.wi.z - 1.0f / wo.z));
            }

            bs.weight = laplacian * (1.0f - Fi) * (1.0f - Fo) * substrateF / bs.pdf;
        }

        return bs;
    }
};


///////////// ROUGH COAT BSDF /////////////

class BsdfRoughCoat : public Bsdf
{
public:
    BsdfRoughCoat(
        std::shared_ptr<Bsdf> substrate = std::make_shared<BsdfConductor>(CopperIor), 
        std::shared_ptr<Texture> roughness = std::make_shared<TextureConstant>(0.1f), 
        std::shared_ptr<Texture> anisotropy = nullptr, 
        std::shared_ptr<MicroFacet> mf = std::make_shared<MicroFacetGGX>(),
        float intIOR = 1.5f, 
        float extIOR = 1.f,
        std::shared_ptr<Texture> specular_reflectance = nullptr,
        float thickness = 1.f,
        Color transmittanceColor = White()
    ) : 
        Bsdf((BsdfType)(BSDF_GLOSSYREFLECTION | substrate->type())),
        m_substrate(substrate), 
        m_roughness(roughness),
        m_anisotropy(anisotropy),
        m_distribution(mf),
        m_intIOR(intIOR), 
        m_extIOR(extIOR), 
        m_eta(m_intIOR / m_extIOR), 
        m_invEta(1.f / m_eta),
        m_specular_reflectance(specular_reflectance),
        m_thickness(thickness)
    {
        m_sigmaA = -log(transmittanceColor) / m_thickness;
        m_scaledSigmaA = m_thickness * m_sigmaA;
        m_avgTransmittance = std::exp(-2.0f * m_scaledSigmaA.power());
    }

    std::shared_ptr<Bsdf> m_substrate;
    std::shared_ptr<Texture> m_roughness;
    std::shared_ptr<Texture> m_anisotropy;
    std::shared_ptr<MicroFacet> m_distribution;
    float m_intIOR, m_extIOR, m_eta, m_invEta;
    std::shared_ptr<Texture> m_specular_reflectance;
    float m_thickness = 1.f;
    Color m_sigmaA = Black();
    float m_avgTransmittance;
    Color m_scaledSigmaA;

    Color eval( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        if (wi.z <= 0.0f || wo.z <= 0.0f)
            return Black();

        // get roughness value
        float roughness = (*m_roughness)(uv).r;
        float anisotropy = m_anisotropy ? (*m_anisotropy)(uv).r : 0;
        float alpha_x, alpha_y;
        ComputeAnisotropicAlpha(roughness, anisotropy, alpha_x, alpha_y);

        // compute microfacet normal
        Vector wh = normalize(wi + wo);

        // compute glossy bsdf value
        float F = DielectricReflectance(m_invEta, dot(wo, wh));
        float G = m_distribution->G2(alpha_x, alpha_y, wh, wi, wo);
        float D = m_distribution->D(alpha_x, alpha_y, wh, wo);
        float fr = (F * G * D * 0.25f) / std::abs(wo.z * wi.z);
        Color glossyR = (m_specular_reflectance ? (*m_specular_reflectance)(uv) : White()) * fr;
        
        // compute substrate bsdf value
        float cosThetaTi, cosThetaTo;
        float Fo = DielectricReflectance(m_invEta, wo.z, cosThetaTo);
        float Fi = DielectricReflectance(m_invEta, wi.z, cosThetaTi);

        Vector wiSubstrate(wi.x * m_invEta , wi.y * m_invEta , std::copysign(cosThetaTi, wi.z));
        Vector woSubstrate(wo.x * m_invEta , wo.y * m_invEta , std::copysign(cosThetaTo, wo.z));

        float laplacian =  m_invEta * m_invEta / cosThetaTi;
        Color substrateF = m_substrate->eval(wiSubstrate, woSubstrate, uv);

        // remove the extra cosine term of perfectly specular materials
        if( m_substrate->isDelta() )
            substrateF = substrateF * std::abs(wiSubstrate.z);

        if(m_scaledSigmaA.max() > 0.0f)
        {
            substrateF = substrateF * exp(m_scaledSigmaA*(-1.0f / wi.z - 1.0f / wo.z));
        }

        Color substrateR = laplacian * (1.0f - Fi) * (1.0f - Fo) * substrateF;

        return glossyR + substrateR;
    }
    
    float pdf( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        if (wi.z <= 0.0f || wo.z <= 0.0f)
            return 0.0f;
        
        // get roughness value
        float roughness = (*m_roughness)(uv).r;
        float anisotropy = m_anisotropy ? (*m_anisotropy)(uv).r : 0;
        float alpha_x, alpha_y;
        ComputeAnisotropicAlpha(roughness, anisotropy, alpha_x, alpha_y);

        // compute microfacet normal
        Vector wh = normalize(wi + wo);

        // compute microfacet pdf
        float pdf_wh = m_distribution->pdf(alpha_x, alpha_y, wh, wo);

        // compute reflected direction pdf for glossy reflection
        float glossyPdf = pdf_wh * 0.25f / std::abs(dot(wo, wh));

        // compute substrate pdf
        float cosThetaTi, cosThetaTo;
        float Fo = DielectricReflectance(m_invEta, wo.z, cosThetaTo);
        float Fi = DielectricReflectance(m_invEta, wi.z, cosThetaTi);
        
        Vector wiSubstrate(wi.x * m_invEta , wi.y * m_invEta , std::copysign(cosThetaTi, wi.z));
        Vector woSubstrate(wo.x * m_invEta , wo.y * m_invEta , std::copysign(cosThetaTo, wo.z));

        float substratePdf = m_substrate->pdf(wiSubstrate, woSubstrate, uv) * 
                            m_invEta * m_invEta * 
                            std::abs(wi.z) / cosThetaTi;

        // weight by specular prob
        float substrateWeight = m_avgTransmittance * (1.0f - Fo);
        float specularWeight = Fo;
        float specularProbability = specularWeight / (specularWeight + substrateWeight);

        glossyPdf *= specularProbability;
        substratePdf *= 1 - specularProbability;

        return glossyPdf + substratePdf;
    }
    
    BsdfSample sample( Sampler& sampler, const Vector& wo, const vec2& uv ) const
    {
        BsdfSample bs;
        if(wo.z <= 0.0f) 
            return bs;
        
        // get roughness value
        float roughness = (*m_roughness)(uv).r;
        float anisotropy = m_anisotropy ? (*m_anisotropy)(uv).r : 0;
        float alpha_x, alpha_y;
        ComputeAnisotropicAlpha(roughness, anisotropy, alpha_x, alpha_y);

        // compute glossy probability
        float cosThetaTo;
        float Fo = DielectricReflectance(m_invEta, wo.z, cosThetaTo);
        float substrateWeight = m_avgTransmittance * (1.0f - Fo);
        float specularWeight = Fo;
        float specularProbability = specularWeight / (specularWeight + substrateWeight);

        // sample rough dielectric coating
        Vector wh;
        bool reflect = true;
        float glossyPdf = 0, substratePdf = 0;
        Color glossyR = Black(), substrateR = Black();

        if(sampler.sample() <= specularProbability) 
        {
            // sample microfacet normal
            wh = m_distribution->sample(alpha_x, alpha_y, sampler.sample(), sampler.sample(), wo);
            
            // direction
            bs.wi = Reflect(wo, wh);

            // glossy pdf
            glossyPdf = m_distribution->pdf(alpha_x, alpha_y, wh, wo) * 
                        specularProbability *
                        0.25f / 
                        std::abs(dot(wo, wh));

            // glossy bsdf value
            float F = DielectricReflectance(m_invEta, dot(wo, wh));
            float G = m_distribution->G2(alpha_x, alpha_y, wh, bs.wi, wo);
            float D = m_distribution->D(alpha_x, alpha_y, wh, wo);
            // remove cosine term
            // float fr = (F * G * D * 0.25f) / std::abs(wo.z * bs.wi.z);
            float fr = (F * G * D * 0.25f) / std::abs(wo.z);
            glossyR = (m_specular_reflectance ? (*m_specular_reflectance)(uv) : White()) * fr;
        } 
        // sample substrate
        else 
        {
            Vector woSubstrate(wo.x * m_invEta , wo.y * m_invEta , std::copysign(cosThetaTo, wo.z));

            BsdfSample substrateSample = m_substrate->sample(sampler, woSubstrate, uv);
            if( substrateSample.wi.z == 0 )
                return BsdfSample();

            float cosThetaTi;
            float Fi = DielectricReflectance(m_eta, substrateSample.wi.z, cosThetaTi);

            // direction
            bs.wi = Vector(substrateSample.wi.x * m_eta, substrateSample.wi.y * m_eta, cosThetaTi);

            // substrate pdf
            float laplacian = m_invEta * m_invEta * std::abs(bs.wi.z) / cosThetaTi;
            substratePdf =  substrateSample.pdf * 
                            laplacian * 
                            (1 - specularProbability);

            // compute substrate bsdf value
            Color substrateF = m_substrate->eval(substrateSample.wi, woSubstrate, uv);

            // remove the extra cosine term of perfectly specular materials
            if( m_substrate->isDelta() )
                substrateF = substrateF * std::abs(substrateSample.wi.z);

            if(m_scaledSigmaA.max() > 0.0f)
            {
                substrateF = substrateF * exp(m_scaledSigmaA*(-1.0f / bs.wi.z - 1.0f / wo.z));
            }

            // substrateR = laplacian * (1.0f - Fi) * (1.0f - Fo) * substrateF;
            substrateR = laplacian * (1 - Fi) * (1 - Fo) * substrateF;

            // eval microfacet normal
            wh = normalize(bs.wi + wo);

            reflect = false;

            if (Fi >= 1.0f || Fo >= 1.f){
                return BsdfSample();
            }
        }

        if( reflect )
        {
            // compute substrate pdf
            float cosThetaTi;
            float Fi = DielectricReflectance(m_invEta, bs.wi.z, cosThetaTi);
            
            Vector wiSubstrate(bs.wi.x * m_invEta , bs.wi.y * m_invEta , std::copysign(cosThetaTi, bs.wi.z));
            Vector woSubstrate(wo.x * m_invEta , wo.y * m_invEta , std::copysign(cosThetaTo, wo.z));

            float laplacian = m_invEta * m_invEta * std::abs(bs.wi.z) / cosThetaTi;
            substratePdf =  m_substrate->pdf(wiSubstrate, woSubstrate, uv) * 
                            laplacian * 
                            (1 - specularProbability);

            // compute substrate bsdf value
            Color substrateF = m_substrate->eval(wiSubstrate, woSubstrate, uv);

            // remove the extra cosine term of perfectly specular materials
            if( m_substrate->isDelta() )
                substrateF = substrateF * std::abs(wiSubstrate.z);

            if(m_scaledSigmaA.max() > 0.0f)
            {
                substrateF = substrateF * exp(m_scaledSigmaA*(-1.0f / bs.wi.z - 1.0f / wo.z));
            }

            substrateR = laplacian * (1.0f - Fi) * (1.0f - Fo) * substrateF;

            if (Fi >= 1.0f || Fo >= 1.f){
                return BsdfSample();
            }
        }
        else
        {
            // compute microfacet pdf
            float pdf_wh = m_distribution->pdf(alpha_x, alpha_y, wh, wo);

            // compute reflected direction pdf for glossy reflection
            glossyPdf = pdf_wh * 0.25f / std::abs(dot(wo, wh));

            // compute glossy bsdf value
            float F = DielectricReflectance(m_invEta, dot(wo, wh));
            float G = m_distribution->G2(alpha_x, alpha_y, wh, bs.wi, wo);
            float D = m_distribution->D(alpha_x, alpha_y, wh, wo);
            // remove cosine term
            // float fr = (F * G * D * 0.25f) / std::abs(wo.z * bs.wi.z);
            float fr = (F * G * D * 0.25f) / std::abs(wo.z);
            glossyR = (m_specular_reflectance ? (*m_specular_reflectance)(uv) : White()) * fr;
        }

        bs.pdf = glossyPdf + substratePdf;

        // final bsdf value
        bs.weight = (glossyR + substrateR) / bs.pdf;

        return bs;
    }
};



///////////// HENYEY-GREENSTEIN BSDF /////////////

static inline float EvalHenyeyGreenstein( float cosTheta, float g ) 
{
    static const float Inv4Pi = 0.25f / (float)M_PI;
    float denom = 1.f + g * g - 2.f * g * cosTheta;
    return std::abs(g) < 1e-4 ? Inv4Pi : Inv4Pi * (1 - g * g) / (denom * std::sqrt(denom));
}

static inline Vector SampleHenyeyGreenstein( float g, float u1, float u2 ) 
{
    // compute cosine of Henyey-Greenstein sample
    float cosTheta;
    if(std::abs(g) < 1e-4)
    {
        cosTheta = 1 - 2 * u1;
    }
    else 
    {
        float sqrTerm = (1 - g * g) / (1 - g + 2 * g * u1);
        cosTheta = (1 + g * g - sqrTerm * sqrTerm) / (2 * g);
    }

    // compute direction of Henyey-Greenstein sample
    float sinTheta = std::sqrt(std::max(0.f, 1.f - cosTheta * cosTheta));
    float phi = 2.f * (float)M_PI * u2;
    return Vector(std::cos(phi) * sinTheta, std::sin(phi) * sinTheta, cosTheta);
}

class BsdfHenyeyGreenstein : public Bsdf
{
public:
    BsdfHenyeyGreenstein(float g = 0) : 
        Bsdf(BSDF_PHASE),
        m_g(std::min(1.f, std::max(-1.f, g))) 
    {
        if( g > 0 )
            m_type = (BsdfType)(m_type | BSDF_FORWARD);
        if( g < 0 )
            m_type = (BsdfType)(m_type | BSDF_BACKWARD);
    }
    
    Color eval( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        // Warning! In integrators and other BSDFs, wo points to the cameras, but wi to the lights, so their orientations are opposite.
        return Color(EvalHenyeyGreenstein(-dot(wi, wo), m_g));
    }
    
    float pdf( const Vector& wi, const Vector& wo, const vec2& uv ) const
    {
        // Warning! In integrators and other BSDFs, wo points to the cameras, but wi to the lights, so their orientations are opposite.
        return EvalHenyeyGreenstein(-dot(wi, wo), m_g);
    }
    
    BsdfSample sample( Sampler& sampler, const Vector& wo, const vec2& uv ) const
    {
        BsdfSample bs;
        const float u = sampler.sample();
        const float v = sampler.sample();
        // Warning! The sampling function returns directions in the local frame of wo = (0, 0, 1).
        //assert(wo == Vector(0, 0, 1));
        //bs.wi = SampleHenyeyGreenstein(m_g, u, v);
        // Without matching frames when calling this function, use:
        bs.wi = Frame(wo).ToWorld(SampleHenyeyGreenstein(m_g, u, v));
        // Warning! In integrators and other BSDFs, wo points to the cameras, but wi to the lights, so their orientations are opposite.
        bs.wi = -bs.wi;

        bs.pdf = pdf(bs.wi, wo, uv);
        bs.weight = White(); //perfect importance sampling
        return bs;
    }

    float m_g;
};


///////////// MIE BSDF /////////////

// static inline float EvalMie( float cosTheta ) 
// {
    
// }

// static inline Vector SampleMie( float u1, float u2 ) 
// {
//     // compute angle of Mie sample


//     // compute cosine of Mie sample
//     float cosTheta;

//     // compute direction of Mie sample
//     float sinTheta = std::sqrt(std::max(0.f, 1.f - cosTheta * cosTheta));
//     float phi = 2.f * (float)M_PI * u2;
//     return Vector(std::cos(phi) * sinTheta, std::sin(phi) * sinTheta, cosTheta);
// }

// class BsdfMie : public Bsdf
// {
// public:
//     static constexpr int bins = 512;
//     static float mie[bins]; 
//     static float total = 0.f; 

// public:
//     BsdfMie() : Bsdf(BsdfType(BSDF_PHASE | BSDF_FORWARD))
//     {
//         if( total == 0.f )
//         {
//             // Cloud Mie Phase Function : using Kokhanovsky 2004 ' analytical form...
//             // gaussian mixture fit
//             const float b[5] =     { 1744.f, 0.17f, 0.3f, 0.2f, 0.15f };
//             const float beta[5] =  { 1200.f, 75.f, 4826.f, 50.f, 1.f };
//             const float t[5] =     { 0.f, 2.5f, (float)M_PI, (float)M_PI, (float)M_PI };

//             for(int i = 0; i < bins; ++i)
//             {
//                 float theta = (float)M_PI * (float)i / (bins - 1.f);
//                 float p = 17.7f * std::exp(-3.9f*theta);
//                 for(int j = 0; j < 5; ++j)
//                 {
//                     float a = theta - t[j];
//                     p += b[j] * std::exp(-beta[j] * a*a);
//                 }
//                 mie[i] = p;
//                 total += p;
//             }
//         }
//     }
    
//     Color eval( const Vector& wi, const Vector& wo, const vec2& uv ) const
//     {
//         float cosTheta = dot(wi, wo);
//         return Color(EvalHenyeyGreenstein(dot(wi, wo), m_g));
//     }
    
//     float pdf( const Vector& wi, const Vector& wo, const vec2& uv ) const
//     {
//         return EvalHenyeyGreenstein(dot(wi, wo), m_g);
//     }
    
//     BsdfSample sample( Sampler& sampler, const Vector& wo, const vec2& uv ) const
//     {
//         // compute cosine of Henyey-Greenstein sample
//         float cosTheta;
//         float theta =
//         float u = sampler.get1D();

//         float v = sampler.get1D();

//         // compute direction of Henyey-Greenstein sample
//         float sinTheta = std::sqrt(std::max(0.f, 1.f - cosTheta * cosTheta));
//         float phi = 2.f * (float)M_PI * u2;
//         return Vector(std::cos(phi) * sinTheta, std::sin(phi) * sinTheta, cosTheta);

//         BsdfSample bs;
//         bs.wi = SampleHenyeyGreenstein(m_g, sampler.sample(), sampler.sample());
//         bs.pdf = EvalHenyeyGreenstein(dot(bs.wi, wo), m_g);
//         bs.weight = White();
//         return bs;
//     }

//     float m_g;
// };


#endif
