#ifndef MC_BSDF_UTILS_H
#define MC_BSDF_UTILS_H

///////////// REFLECTION UTILS /////////////

static inline bool CheckReflectionConstraint(const Vector& wi, const Vector& wo)
{
    return std::abs(wi.z * wo.z - wi.x * wo.x - wi.y * wo.y - 1.0f) < 1e-4;
}

static inline Vector Reflect(const Vector& wo, const Vector& wh)
{
    return normalize(2.f * dot(wo, wh) * wh - wo);
}

static inline Vector Reflect(const Vector& wo)
{
    return Vector(-wo.x, -wo.y, wo.z);
}



///////////// REFRACTION UTILS /////////////

static inline bool CheckRefractionConstraint(const Vector& wi, const Vector& wo, float eta, float cosThetaT)
{
    float dotP = -wi.x * wo.x * eta - wi.y * wo.y * eta - std::copysign(cosThetaT, wo.z) * wi.z;
    return std::abs(dotP - 1.0f) < 1e-4;
}

// warning this function does not return the right sign of cosThetaT
static inline float ComputeThetaTransmitted(const float eta, const Vector &wo)
{
    float sinThetaTSq = (1.0f - wo.z * wo.z) * eta * eta;                                
    if (sinThetaTSq >= 1.0f) return 0; // total internal reflection
    return std::sqrt(std::max(1.0f - sinThetaTSq, 0.0f));
}

static inline Vector Refract(const float eta, const Vector &wo, const Vector &wh)
{
    float cosThetaT = ComputeThetaTransmitted(eta, wo);
    return (eta * dot(wo, wh) - std::copysign(cosThetaT, wo.z)) * wh - eta * wo;
}   

static inline Vector Refract(const float eta, const Vector &wo)
{
    float cosThetaT = ComputeThetaTransmitted(eta, wo);
    return Vector(-wo.x * eta, -wo.y * eta, -std::copysign(cosThetaT, wo.z));
}   



///////////// DIELECTRIC FRESNEL /////////////

// cf: http://www.pbr-book.org/3ed-2018/Reflection_Models/Specular_Reflection_and_Transmission.html
static inline float DielectricReflectance( float eta, float cosThetaI, float &cosThetaT )
{
    if (cosThetaI < 0.0f) 
    {
        eta = 1.0f/eta;
        cosThetaI = -cosThetaI;
    }

    float sinThetaTSq = eta*eta*(1.0f - cosThetaI*cosThetaI);
    if (sinThetaTSq >= 1.0f) 
    {
        cosThetaT = 0.0f;
        return 1.0f;
    }
    cosThetaT = std::sqrt(std::max(1.0f - sinThetaTSq, 0.0f));

    float Rs = (eta*cosThetaI - cosThetaT)/(eta*cosThetaI + cosThetaT);
    float Rp = (eta*cosThetaT - cosThetaI)/(eta*cosThetaT + cosThetaI);

    return (Rs*Rs + Rp*Rp)*0.5f;
}

static inline float DielectricReflectance( float eta, float cosThetaI )
{
    float cosThetaT;
    return DielectricReflectance(eta, cosThetaI, cosThetaT);
}

static inline float SchlickDielectricReflectance( float eta, float cosThetaI )
{
    if (cosThetaI < 0.0f) 
    {
        eta = 1.0f/eta;
        cosThetaI = -cosThetaI;
    }

    float r0 = (eta - 1.0f) * (eta - 1.0f) / ((eta + 1.0f) * (eta + 1.0f));

    float u = 1.f - cosThetaI;
    float uSq = u*u;
    return r0 + (1.f - r0) * uSq * uSq * u;
}



///////////// CONDUCTOR FRESNEL /////////////

// From "PHYSICALLY BASED LIGHTING CALCULATIONS FOR COMPUTER GRAPHICS" by Peter Shirley
// http://www.cs.virginia.edu/~jdl/bib/globillum/shirley_thesis.pdf
static inline float ConductorReflectance( float eta, float k, float cosThetaI )
{
    float cosThetaISq = cosThetaI*cosThetaI;
    float sinThetaISq = std::max(1.0f - cosThetaISq, 0.0f);
    float sinThetaIQu = sinThetaISq*sinThetaISq;

    float innerTerm = eta*eta - k*k - sinThetaISq;
    float aSqPlusBSq = std::sqrt(std::max(innerTerm*innerTerm + 4.0f*eta*eta*k*k, 0.0f));
    float a = std::sqrt(std::max((aSqPlusBSq + innerTerm)*0.5f, 0.0f));

    float Rs = ((aSqPlusBSq + cosThetaISq) - (2.0f*a*cosThetaI))/
               ((aSqPlusBSq + cosThetaISq) + (2.0f*a*cosThetaI));
    float Rp = ((cosThetaISq*aSqPlusBSq + sinThetaIQu) - (2.0f*a*cosThetaI*sinThetaISq))/
               ((cosThetaISq*aSqPlusBSq + sinThetaIQu) + (2.0f*a*cosThetaI*sinThetaISq));

    return 0.5f*(Rs + Rs*Rp);
}

static inline float ConductorReflectanceApprox( float eta, float k, float cosThetaI) 
{
    float cosThetaISq = cosThetaI*cosThetaI;
    float ekSq = eta*eta* + k*k;
    float cosThetaEta2 = cosThetaI*2.0f*eta;

    float Rp = (ekSq*cosThetaISq - cosThetaEta2 + 1.0f)/(ekSq*cosThetaISq + cosThetaEta2 + 1.0f);
    float Rs = (ekSq - cosThetaEta2 + cosThetaISq)/(ekSq + cosThetaEta2 + cosThetaISq);
    return (Rs + Rp)*0.5f;
}

static inline Color ConductorReflectance( const Color& eta, const Color& k, float cosThetaI )
{
    return Color(
        ConductorReflectance(eta.r, k.r, cosThetaI),
        ConductorReflectance(eta.g, k.g, cosThetaI),
        ConductorReflectance(eta.b, k.b, cosThetaI)
    );
}

static inline float SchlickConductorReflectance( float eta, float k, float cosThetaI )
{
    float a = (eta - 1.0f) * (eta - 1.0f) + k * k;
    float b = (eta + 1.0f) * (eta + 1.0f) + k * k;
    float r0 = a / b;

    float u = 1.f - cosThetaI;
    float uSq = u*u;
    return r0 + (1.f - r0) * uSq * uSq * u;
}

static Color SchlickConductorReflectance( const Color& eta, const Color& k, float cosThetaI )
{
    return Color(
        SchlickConductorReflectance(eta.r, k.r, cosThetaI),
        SchlickConductorReflectance(eta.g, k.g, cosThetaI),
        SchlickConductorReflectance(eta.b, k.b, cosThetaI)
    );
}

static inline float SchlickTintedConductorReflectance( float r0, float cosThetaI )
{
    float u = 1.f - cosThetaI;
    float uSq = u*u;
    return r0 + (1.f - r0) * uSq * uSq * u;
}

static Color SchlickTintedConductorReflectance( const Color& tint, float cosThetaI )
{
    return Color(
        SchlickTintedConductorReflectance(tint.r, cosThetaI),
        SchlickTintedConductorReflectance(tint.g, cosThetaI),
        SchlickTintedConductorReflectance(tint.b, cosThetaI)
    );
}

// Lazanyi - Shlick model
// cf http://renderwonk.com/publications/mam2019/
static inline float LazanyiSchlickConductorReflectance( float r0, float r82, float cosThetaI )
{
    float u = 1.f - cosThetaI;
    float u2 = u*u;
    float u5 = u * u2 * u2;
    float u6 = u5 * u;

    static const float thetaMax = std::acos(1.f / 7.f);
    static const float cosThetaMax = std::cos(thetaMax);
    static const float v = 1.f - cosThetaMax;
    static const float v2 = v * v;
    static const float v5 = v * v2 * v2;
    static const float v6 = v5 * v;

    float a = (r0 + (1.f - r0) * v5 - r82) / (cosThetaMax * v6);

    return std::max(0.f, r0 + (1.f - r0) * u5 - a * cosThetaI * u6);
}

static Color LazanyiSchlickConductorReflectance( const Color& tint, const Color& edgetint, float cosThetaI )
{
    return Color(
        LazanyiSchlickConductorReflectance(tint.r, edgetint.r, cosThetaI),
        LazanyiSchlickConductorReflectance(tint.g, edgetint.g, cosThetaI),
        LazanyiSchlickConductorReflectance(tint.b, edgetint.b, cosThetaI)
    );
}



///////////// DIFFUSE FRESNEL /////////////

// Computes the average diffuse Fresnel reflectance 
// The hemispherical integral of DielectricReflectance(eta, cos(theta))*cos(theta)
// cf: Jensen et al, A Practical Model for Subsurface Light Transport, 2001
// https://graphics.stanford.edu/papers/bssrdf/bssrdf.pdf
static inline float FresnelDiffuseReflectance(float eta, const int sampleCount)
{
    double diffuseFresnel = 0.0;
    float fb = DielectricReflectance(eta, 0.0f);
    for (int i = 1; i <= sampleCount; ++i) 
    {
        float cosThetaSq = float(i) / sampleCount;
        float fa = DielectricReflectance(eta, std::min(std::sqrt(cosThetaSq), 1.0f));
        diffuseFresnel += double(fa + fb)*(0.5/sampleCount);
        fb = fa;
    }
    return float(diffuseFresnel);
}

// Computes an approximation of the average diffuse Fresnel reflectance
static inline float FresnelDiffuseReflectanceApprox(float eta)
{
    const float inv_eta   = 1.f / eta,
        inv_eta_2 = inv_eta   * inv_eta,
        inv_eta_3 = inv_eta_2 * inv_eta,
        inv_eta_4 = inv_eta_3 * inv_eta,
        inv_eta_5 = inv_eta_4 * inv_eta;
          
    return eta < 1 ?
        // Fit by Egan and Hilgeman (1973). Works reasonably well for "normal" IOR values (<2).
        // Max rel. error in 1.0 - 1.5 : 0.1%
        // Max rel. error in 1.5 - 2   : 0.6%
        // Max rel. error in 2.0 - 5   : 9.5%
        -1.4399f * (eta * eta) + 0.7099f * eta + 0.6681f + 0.0636f / eta :
        // Fit by d'Eon and Irving (2011)
        // Maintains a good accuracy even for unrealistic IOR values.
        // Max rel. error in 1.0 - 2.0   : 0.1%
        // Max rel. error in 2.0 - 10.0  : 0.2% 
        0.919317f - 3.4793f * inv_eta + 6.75335f * inv_eta_2 - 7.80989f * inv_eta_3 + 4.98554f * inv_eta_4 - 1.36881f * inv_eta_5;
}


// ///////////// REMAPPING PHONG -> ALPHA /////////////

static inline float ShininessToRoughness( const float aShininess )
{
    // cf: Brian Karis blog 
    // http://graphicrants.blogspot.com/2013/08/specular-brdf-reference.html
    return std::pow(2.f/(2.f+aShininess), 0.25f);
    // return std::pow(2.f/(2.f+aShininess), 0.5f);
}

// static inline float RoughnessToShininess( const float aRoughness )
// {
//     // cf: Brian Karis blog 
//     // http://graphicrants.blogspot.com/2013/08/specular-brdf-reference.html
//     float sqrRoughness = aRoughness * aRoughness;
//     // return 2.f / (sqrRoughness * sqrRoughness) - 2.f;
//     return 2.f / (sqrRoughness) - 2.f;
// }

// // cf: Brent Burley Disney shading model
// static inline float RoughnessToSmoothness( const float aRoughness )
// {
//     return 1.f - std::sqrt(aRoughness);
// }

// static inline float SmoothnessToRoughness( const float aSmoothness )
// {
//     float k = 1 - aSmoothness;  
//     return k * k;
// }


/////////////// HEMISPHERE COSINE WEIGTED SAMPLING /////////////

static inline Vector SampleCosineWeightedHemisphere( const float u1, const float u2 )
{
    float phi = 2.f * (float) M_PI * u1;
    float cos_theta = std::sqrt(u2);
    float sin_theta = std::sqrt(1.f - std::min(1.f, cos_theta*cos_theta));
    return Vector(std::cos(phi) * sin_theta, std::sin(phi) * sin_theta, cos_theta);
}

static inline float PdfCosineWeightedHemisphere( const Vector& w )
{
    return std::abs(w.z) / (float)(M_PI); 
}


#endif