#pragma once

#include "bsdf.h"
#include "image.h"
#include "image_io.h"
#include "../texture/texture.h"

#include <string.h>
#include <omp.h>

// TODO - new approximations at:
// https://github.com/materialx/MaterialX/blob/main/libraries/pbrlib/genglsl/lib/mx_microfacet_specular.glsl#L82

///////////// COMPENSATION /////////////

static void ComputeRoughConductorCompensationLUTBeckmann();
static void ComputeRoughConductorCompensationLUTGGX();

static void ComputeRoughDielectricCompensationLUTBeckmann();
static void ComputeRoughDielectricCompensationLUTGGX();

static void ComputeRoughDiffuseCompensationLUTBeckmann();
static void ComputeRoughDiffuseCompensationLUTGGX();

static void ComputeOrenNayarCompensationLUT();


///////////// ROUGH DIELECTRIC COMPENSATION /////////////

static void ComputeRoughDielectricCompensationLUT( 
	std::string filename, 
	std::shared_ptr<MicroFacet> distribution, 
	int samples = 10000 )
{
	int width = 32, height = 32, depth = 32;

	// store energy term for eta in [1/3,1]
    Image compensationLowEta(width, height * depth);
	
	// store energy term for eta in [1,3]
    Image compensationHighEta(width, height * depth);

    #pragma omp parallel
    {
    	Sampler sampler;

    	#pragma omp for
        for(int z = 0; z < depth; ++z)
        for(int y = 0; y < height; ++y)
        {
        	// interior ior from 1 to 3 <=> eta between 1/3 and 1 <=> 1/eta between 1 and 3
        	float intIOR = 1.f + 2.f * float(z) / float(depth-1);
        	float extIOR = 1.f;

        	// roughness between 0 and 1
        	float roughness = (y+0.5f) / float(height);
			std::shared_ptr<Texture> rough_texture = std::make_shared<TextureConstant>(Color(roughness));

			// dielectric bsdf
            std::shared_ptr<Bsdf> bsdf = std::make_shared<BsdfRoughDielectric>(
				intIOR, 
				extIOR, 
				nullptr,
				nullptr,
				rough_texture, 
				nullptr,
				distribution
			);

            for(int x = 0; x < width; ++x)
            {   
        		// cos theta between 0 and 1
            	float cosTheta = (x+0.5f) / float(width); 
                float sinTheta = std::sqrt(std::max(0.f, 1.f - cosTheta * cosTheta));

                // compute view direction from exterior with an incident angle of Theta
                Vector woExterior = normalize(Vector(sinTheta, 0, cosTheta));
                
                // compute view direction from interior with an incident angle of Theta
                Vector woInterior = normalize(-Vector(sinTheta, 0, cosTheta));
                
                // compute compensation term
                float essLowEta = 0, countLowEta = 0;
                float essHighEta = 0, countHighEta = 0;
                for(int i = 0; i < samples; ++i)
                {
                	// low eta from exterior to interior
                	{
		            	BsdfSample bs = bsdf->sample(sampler, woExterior, vec2(0.5,0.5));
		                if(bs.pdf > 0)
		                {
		                	float x = bs.weight.r;
		                	++countLowEta;
		                	// online mean
		                	essLowEta = essLowEta + (x - essLowEta) / countLowEta;
		                }
                	}

                	// high eta from interior to exterior
                	{
		            	BsdfSample bs = bsdf->sample(sampler, woInterior, vec2(0.5,0.5));
		                if(bs.pdf > 0)
		                {
		                	float x = bs.weight.r;
		                	++countHighEta;
		                	// online mean
		                	essHighEta = essHighEta + (x - essHighEta) / countHighEta;
		                }
                	}
                }

                // we have to reverse z order since extIOR/intIOR decreases from 1 to 1/3 
                compensationLowEta(x, y + (height-1-z) * height) = Color(essLowEta);

                // we do not have to reverse z order since intIOR/extIOR increases from 1 to 3
                compensationHighEta(x, y + z * height) = Color(essHighEta);
            }
        }
	}

    write_image(compensationLowEta, (std::string(filename) + std::string("_low.png")).c_str());
    write_image(compensationHighEta, (std::string(filename) + std::string("_high.png")).c_str());
}

static void ComputeRoughDielectricCompensationLUTBeckmann()
{
	ComputeRoughDielectricCompensationLUT("compensation_dielectric_beckmann_test", std::make_shared<MicroFacetBeckmann>());
} 

static void ComputeRoughDielectricCompensationLUTGGX()
{
	ComputeRoughDielectricCompensationLUT("compensation_dielectric_ggx_test", std::make_shared<MicroFacetGGX>());
}


///////////// ROUGH CONDUCTOR COMPENSATION /////////////

static void ComputeRoughConductorCompensationLUT( 
	std::string filename, 
	std::shared_ptr<MicroFacet> distribution, 
	int samples = 10000 )
{
	int width = 32, height = 32;

	// stores compensation term
    Image compensation(width, height);

    Sampler sampler;
	
	#pragma omp parallel
    {
    	Sampler sampler;

    	#pragma omp for
	    for(int y = 0; y < height; ++y)
	    for(int x = 0; x < width; ++x)
	    {   
	    	// linear roughness
	    	float roughness = (y+0.5f) / float(height);
	    	
	    	// remapped roughness
	    	float alpha = std::max(1e-6f, roughness * roughness);
	        
	        // compute view direction
	    	float cosTheta = (x+0.5f) / float(width); 
	        float sinTheta = std::sqrt(std::max(0.f, 1.f - cosTheta * cosTheta));
	        Vector wo = normalize(Vector(sinTheta, 0, cosTheta));
	        
	        // compute compensation term
	        float ess = 0, count = 0;
	        for(int i = 0; i < samples; ++i)
	        {
	            // sample microfacet normal
	            Vector wh = distribution->sample(alpha, alpha, sampler.sample(), sampler.sample(), wo);
	            float pdf_wh = distribution->pdf(alpha, alpha, wh, wo);
	            if (pdf_wh < 1e-10f) continue;
	            
	            // construct reflected direction
	            Vector wi = Reflect(wo, wh);

	            // eval bsdf
	            float G2 = distribution->G2(alpha, alpha, wh, wi, wo);
	            float G1 = distribution->G1(alpha, alpha, wh, wo);
	            float x = G1 > 0 ? G2 / G1 : 0;
	            
	            // online mean
	            ++count;
	            ess = ess + (x - ess) / count;
	        }
	        compensation(x, y) = Color(ess);
	    }
	}

    write_image(compensation, (std::string(filename) + std::string(".png")).c_str());
}

static void ComputeRoughConductorCompensationLUTBeckmann()
{
	ComputeRoughConductorCompensationLUT("compensation_conductor_beckmann_test", std::make_shared<MicroFacetBeckmann>());
} 

static void ComputeRoughConductorCompensationLUTGGX()
{
	ComputeRoughConductorCompensationLUT("compensation_conductor_ggx_test", std::make_shared<MicroFacetGGX>());
}


///////////// ENERGY SINGLE SCATTERING /////////////

static float EnergySingleScattering( std::shared_ptr<Bsdf> bsdf, Sampler& sampler, const float cosTheta, const int samples )
{
    // compute view direction
	float sinTheta = std::sqrt(std::max(0.f, 1.f - cosTheta * cosTheta));
    Vector wo = Vector(sinTheta, 0, cosTheta);

    // compute compensation term
    float ess = 0, count = 0;
    for(int i = 0; i < samples; ++i)
    {
    	BsdfSample bs = bsdf->sample(sampler, wo, vec2(0.5,0.5));
        if(bs.pdf <= 0)
        	continue;

        // online mean
        ++count;
        ess = ess + (bs.weight.r - ess) / count;
    }

    return ess;
}

///////////// ROUGH DIFFUSE COMPENSATION /////////////

static void ComputeRoughDiffuseCompensationLUT( 
	std::string filename, 
	std::shared_ptr<MicroFacet> distribution, 
	int samples = 10000 )
{
	int width = 32, height = 32;

	// stores compensation term
    Image compensation(width, height);

	#pragma omp parallel
    {
    	Sampler sampler;

    	#pragma omp for
	    for(int y = 0; y < height; ++y)
	    for(int x = 0; x < width; ++x)
	    {   
	    	// linear roughness
	    	float roughness = (y+0.5f) / float(height);

	    	// rough diffuse bsdf
	    	std::shared_ptr<Bsdf> bsdf = std::make_shared<BsdfRoughDiffuse>(
				std::make_shared<TextureConstant>(White()), // reflectance
				std::make_shared<TextureConstant>(Color(roughness)), // roughness
				nullptr, // anisotropy
	        	distribution // distribution
	        );
	        
	        // compute single scattering energy 
	    	float cosTheta = (x+0.5f) / float(width); 
	        float ess = EnergySingleScattering(bsdf, sampler, cosTheta, samples);
	        compensation(x, y) = Color(ess);
	    }
	}

    write_image(compensation, (std::string(filename) + std::string(".png")).c_str());
}

static void ComputeRoughDiffuseCompensationLUTBeckmann()
{
	ComputeRoughDiffuseCompensationLUT("compensation_rough_diffuse_beckmann_test", std::make_shared<MicroFacetBeckmann>());
} 

static void ComputeRoughDiffuseCompensationLUTGGX()
{
	ComputeRoughDiffuseCompensationLUT("compensation_rough_diffuse_ggx_test", std::make_shared<MicroFacetGGX>());
}


///////////// OREN-NAYAR COMPENSATION /////////////

static void ComputeOrenNayarCompensationLUT( 
	std::string filename, 
	int samples = 10000 )
{
	int width = 32, height = 32;

	// stores compensation term
    Image compensation(width, height);

	#pragma omp parallel
    {
    	Sampler sampler;

    	#pragma omp for
	    for(int y = 0; y < height; ++y)
	    for(int x = 0; x < width; ++x)
	    {   
	    	// linear roughness
	    	float roughness = (y+0.5f) / float(height);

	    	// oren-nayar bsdf
			std::shared_ptr<Bsdf> bsdf = std::make_shared<BsdfOrenNayar>(
				std::make_shared<TextureConstant>(White()), 
				std::make_shared<TextureConstant>(Color(roughness))
	        );
	        
	        // compute single scattering energy 
	    	float cosTheta = (x+0.5f) / float(width); 
	        float ess = EnergySingleScattering(bsdf, sampler, cosTheta, samples);
	        compensation(x, y) = Color(ess);
	    }
	}

    write_image(compensation, (std::string(filename) + std::string(".png")).c_str());
}

static void ComputeOrenNayarCompensationLUT()
{
	ComputeOrenNayarCompensationLUT("compensation_oren_nayar_test");
}