#pragma once

#include "vec.h"

///////////// MICROFACET INTERFACE /////////////

class MicroFacet
{
public:
    MicroFacet(){}
    virtual ~MicroFacet() {}
    virtual float Lambda( float alpha_x, float alpha_y, const Vector& w ) const = 0;
    virtual float D( float alpha_x, float alpha_y, const Vector& wh, const Vector& wo ) const = 0;
    virtual Vector sample( float alpha_x, float alpha_y, const float u, const float v, const Vector& wo ) const = 0;

    float pdf( float alpha_x, float alpha_y, const Vector& wh, const Vector& wo ) const
    {
        // visible normal distribution
        return  G1(alpha_x, alpha_y, wh, wo) * 
                D(alpha_x, alpha_y, wh, wo) *
                std::abs(dot(wo, wh)) / 
                std::abs(wo.z);
    }

    float G1( float alpha_x, float alpha_y, const Vector& wh, const Vector& w ) const
    { 
        // G1 = 1 / (1 + Lambda(w, wh))
        if(dot(wh, w) * w.z <= 0.0f) return 0.f;
        return 1.f / (1.f + Lambda(alpha_x, alpha_y, w));
    }

    float G2( float alpha_x, float alpha_y, const Vector& wh, const Vector& wi, const Vector& wo ) const
    {
        // G1 = 1 / (1 + Lambda(wi, wh) + Lambda(wo, wh))
        if(dot(wh, wi) * wi.z <= 0.0f) return 0.f;
        if(dot(wh, wo) * wo.z <= 0.0f) return 0.f;
        return 1.f / (1.f + Lambda(alpha_x, alpha_y, wi) + Lambda(alpha_x, alpha_y, wo));
    }
};

///////////// GGX MICROFACET MODEL /////////////

static float LambdaGGX(float, float, const Vector&);
static float DistributionGGX(float, float, const Vector&);
static Vector SampleVisibleNormalDistributionGGX(float, float, const float, const float, const Vector&);

class MicroFacetGGX : public MicroFacet
{
public:
    MicroFacetGGX(){}
    ~MicroFacetGGX(){}

    float Lambda( float alpha_x, float alpha_y, const Vector& w ) const
    {
        return LambdaGGX(alpha_x, alpha_y, w);
    }
    
    float D( float alpha_x, float alpha_y, const Vector& wh, const Vector& wo ) const
    {
        return DistributionGGX(alpha_x, alpha_y, wh);
    }

    Vector sample( float alpha_x, float alpha_y, const float u, const float v, const Vector& wo ) const
    {
        return SampleVisibleNormalDistributionGGX(alpha_x, alpha_y, u, v, wo);
    }
};

///////////// BECKMANN MICROFACET MODEL /////////////

static float LambdaBeckmann(float, float, const Vector&);
static float DistributionBeckmann(float, float, const Vector&);
static Vector SampleVisibleNormalDistributionBeckmann(float, float, const float, const float, const Vector&);

class MicroFacetBeckmann : public MicroFacet
{
public:
    MicroFacetBeckmann(){}
    ~MicroFacetBeckmann(){}
    
    float Lambda( float alpha_x, float alpha_y, const Vector& w ) const
    {
        return LambdaBeckmann(alpha_x, alpha_y, w);
    }
    
    float D( float alpha_x, float alpha_y, const Vector& wh, const Vector& wo ) const
    {
        return DistributionBeckmann(alpha_x, alpha_y, wh);
    }

    Vector sample( float alpha_x, float alpha_y, const float u, const float v, const Vector& wo ) const
    {
        return SampleVisibleNormalDistributionBeckmann(alpha_x, alpha_y, u, v, wo);
    }
};

///////////// GGX MICROFACET UTILITY FUNCTIONS /////////////

// Microfacet Models for Refraction through Rough Surfaces
// Bruce Walter 
// 2007 EGSR
//
// Importance Sampling Microfacet-Based BSDFs usingthe Distribution of Visible Normals
// Eric Heitz and Eugene d’Eon
// 2014 EGSR
//
// Understanding the Masking-Shadowing Functionin Microfacet-Based BRDFs
// Eric Heitz
// 2014 JCGT
//
// Sampling the GGX Distribution of Visible Normals
// Eric Heitz
// 2018 JCGT

static Vector SampleVisibleNormalDistributionGGX( float alpha_x, float alpha_y, const float u, const float v, const Vector& wo )
{
    // Switch hemisphere if the direction is in the lower hemisphere
    Vector w = wo.z > 0 ? wo : -wo;
    // Section 3.2: transforming the view direction to the hemisphere configuration 
    Vector Vh = normalize(Vector(alpha_x*w.x, alpha_y*w.y, w.z));
    // Section 4.1: orthonormal basis (with special case if cross product is zero)
    float lensq = Vh.x*Vh.x + Vh.y*Vh.y;
    Vector T1 = lensq > 0 ? Vector(-Vh.y, Vh.x, 0) / std::sqrt(lensq) : Vector(1,0,0); 
    Vector T2 = cross(Vh, T1);
    // Section 4.2: parameterization of the projected area
    float phi = 2.0*M_PI*u;
    float r = std::sqrt(v);
    float t1 = r*std::cos(phi);
    float t2 = r*std::sin(phi);
    float s = 0.5*(1.0 + Vh.z);
    t2 = (1.0 - s)*std::sqrt(1.0 - t1*t1) + s*t2;
    // Section 4.3: reprojection onto hemisphere
    Vector Nh = t1*T1 + t2*T2 + std::sqrt(std::max(0.0f, 1.0f - t1*t1 - t2*t2))*Vh;
    // Section 3.4: transforming the normal back to the ellipsoid configuration
    return normalize(Vector(alpha_x*Nh.x, alpha_y*Nh.y, Nh.z));
}

static float LambdaGGX( float alpha_x, float alpha_y, const Vector& w )
{   
    float cos2_theta = w.z * w.z;
    float tan2_theta = std::max(0.f, (1.f - cos2_theta)) / cos2_theta;

    // isotropic
    if( alpha_x == alpha_y )
    {
        return -0.5f + 0.5f * std::sqrt(1.f + alpha_x * alpha_x * tan2_theta);
    }

    // anisotropic
    float phi = std::atan2(w.y, w.x);
    float cos_phi = std::cos(phi);
    float sin_phi = std::sin(phi);

    float alpha2_0 = cos_phi * cos_phi * alpha_x * alpha_x + sin_phi * sin_phi * alpha_y * alpha_y;
    return -0.5f + 0.5f * std::sqrt(1.f + alpha2_0 * tan2_theta);
}

static float DistributionGGX( float alpha_x, float alpha_y, const Vector& wh )
{
    float cos2_theta = wh.z * wh.z;
    float tan2_theta = std::max(0.f, (1.f - cos2_theta)) / cos2_theta;

    // isotropic
    if( alpha_x == alpha_y )
    {
        float d = alpha_x * alpha_x + tan2_theta;
        return (alpha_x * alpha_x) / ((float) M_PI * cos2_theta * cos2_theta * d * d);
    }

    // anisotropic
    float phi = std::atan2(wh.y, wh.x);
    float cos_phi = std::cos(phi);
    float sin_phi = std::sin(phi);

    float alpha2_x = alpha_x * alpha_x; 
    float alpha2_y = alpha_y * alpha_y; 

    float d = 1.f + tan2_theta * (cos_phi * cos_phi / alpha2_x + sin_phi * sin_phi / alpha2_y);
    return 1.f / ((float) M_PI * alpha_x * alpha_y * cos2_theta * cos2_theta * d * d);
}

///////////// BECKMANN MICROFACET UTILITY FUNCTIONS /////////////

static void BeckmannSample11(float cos_theta, float U1, float U2, float *slope_x, float *slope_y);

static Vector SampleVisibleNormalDistributionBeckmann(const float alpha_x, const float alpha_y, const float u, const float v, const Vector &wo) 
{
    // 0. Switch hemisphere if the direction is in the lower hemisphere
    Vector w = wo.z > 0 ? wo : -wo;

    // 1. stretch w
    Vector wStretched = normalize(Vector(alpha_x * w.x, alpha_y * w.y, w.z));

    // 2. simulate P22_{w}(x_slope, y_slope, 1, 1)
    float slope_x, slope_y;
    BeckmannSample11(wStretched.z, u, v, &slope_x, &slope_y);

    // 3. rotate
    float phi = std::atan2(wStretched.y, wStretched.x);
    float cos_phi = std::cos(phi);
    float sin_phi = std::sin(phi);

    float tmp = cos_phi * slope_x - sin_phi * slope_y;
    slope_y = sin_phi * slope_x + cos_phi * slope_y;
    slope_x = tmp;

    // 4. unstretch
    slope_x = alpha_x * slope_x;
    slope_y = alpha_y * slope_y;

    // 5. compute normal
    return normalize(Vector(-slope_x, -slope_y, 1.f));
}

static float DistributionBeckmann( float alpha_x, float alpha_y, const Vector& wh )
{   
    float cos2_theta = wh.z * wh.z;
    float tan2_theta = std::max(0.f, (1.f - cos2_theta)) / cos2_theta;

    // isotropic
    if( alpha_x == alpha_y )
    {
        return std::exp(-tan2_theta / (alpha_x * alpha_x)) / ((float)M_PI * alpha_x * alpha_x * cos2_theta * cos2_theta);
    }

    // anisotropic
    float phi = std::atan2(wh.y, wh.x);
    float cos_phi = std::cos(phi);
    float sin_phi = std::sin(phi);

    return std::exp(-tan2_theta * (cos_phi * cos_phi / (alpha_x * alpha_x) + sin_phi * sin_phi / (alpha_y * alpha_y))) /
           ((float)M_PI * alpha_x * alpha_y * cos2_theta * cos2_theta);
}

static float LambdaBeckmann( float alpha_x, float alpha_y, const Vector& w )
{   
    float cos2_theta = w.z * w.z;
    float tan2_theta = std::max(0.f, (1.f - cos2_theta)) / cos2_theta;
    float tan_theta = std::sqrt(tan2_theta);

    // anisotropic
    float phi = std::atan2(w.y, w.x);
    float cos_phi = std::cos(phi);
    float sin_phi = std::sin(phi);

    float alpha = std::sqrt(cos_phi * cos_phi * alpha_x * alpha_x + sin_phi * sin_phi * alpha_y * alpha_y);
    float a = 1.f / (alpha * tan_theta);
    if (a >= 1.6f) return 0;
    return (1.f - 1.259f * a + 0.396f * a * a) / (3.535f * a + 2.181f * a * a);
}

///////////// BECKMANN MICROFACET UTILITY FUNCTIONS /////////////

// from PBRT / MITSUBA

static inline float ErfInv(float x) 
{
    float w, p;
    x = std::min(.99999f, std::max(-.99999f, x)); 
    // x = Clamp(x, -.99999f, .99999f);
    w = -std::log((1 - x) * (1 + x));
    if (w < 5) {
        w = w - 2.5f;
        p = 2.81022636e-08f;
        p = 3.43273939e-07f + p * w;
        p = -3.5233877e-06f + p * w;
        p = -4.39150654e-06f + p * w;
        p = 0.00021858087f + p * w;
        p = -0.00125372503f + p * w;
        p = -0.00417768164f + p * w;
        p = 0.246640727f + p * w;
        p = 1.50140941f + p * w;
    } else {
        w = std::sqrt(w) - 3;
        p = -0.000200214257f;
        p = 0.000100950558f + p * w;
        p = 0.00134934322f + p * w;
        p = -0.00367342844f + p * w;
        p = 0.00573950773f + p * w;
        p = -0.0076224613f + p * w;
        p = 0.00943887047f + p * w;
        p = 1.00167406f + p * w;
        p = 2.83297682f + p * w;
    }
    return p * x;
}

static inline float Erf(float x) 
{
    // constants
    float a1 = 0.254829592f;
    float a2 = -0.284496736f;
    float a3 = 1.421413741f;
    float a4 = -1.453152027f;
    float a5 = 1.061405429f;
    float p = 0.3275911f;

    // Save the sign of x
    int sign = 1;
    if (x < 0) sign = -1;
    x = std::abs(x);

    // A&S formula 7.1.26
    float t = 1 / (1 + p * x);
    float y =
        1 -
        (((((a5 * t + a4) * t) + a3) * t + a2) * t + a1) * t * std::exp(-x * x);

    return sign * y;
}

static void BeckmannSample11(float cos_theta, float U1, float U2, float *slope_x, float *slope_y) 
{
    /* Special case (normal incidence) */
    if (cos_theta > .9999) 
    {
        float r = std::sqrt(-std::log(1.0f - U1));
        float sinPhi = std::sin(2 * (float)M_PI * U2);
        float cosPhi = std::cos(2 * (float)M_PI * U2);
        *slope_x = r * cosPhi;
        *slope_y = r * sinPhi;
        return;
    }
    /* The original inversion routine from the paper contained
       discontinuities, which causes issues for Q integration
       and techniques like Kelemen-style MLT. The following code
       performs a numerical inversion with better behavior */
    float sin_theta = std::sqrt(std::max((float)0, (float)1 - cos_theta * cos_theta));
    float tan_theta = sin_theta / cos_theta;
    float cot_theta = 1.f / tan_theta;

    /* Search interval -- everything is parameterized in the Erf() domain */
    float a = -1, c = Erf(cot_theta);
    float sample_x = std::max(U1, (float)1e-6f);

    /* Start with a good initial guess */
    // float b = (1-sample_x) * a + sample_x * c;

    /* We can do better (inverse of an approximation computed in Mathematica) */
    float thetaI = std::acos(cos_theta);
    float fit = 1 + thetaI * (-0.876f + thetaI * (0.4265f - 0.0594f * thetaI));
    float b = c - (1 + c) * std::pow(1 - sample_x, fit);

    /* Normalization factor for the CDF */
    static const float SQRT_PI_INV = 1.f / std::sqrt((float)M_PI);
    float normalization = 1 / (1 + c + SQRT_PI_INV * tan_theta * std::exp(-cot_theta * cot_theta));

    int it = 0;
    while (++it < 10) 
    {
        /* Bisection criterion -- the oddly-looking
           Boolean expression are intentional to check
           for NaNs at little additional cost */
        if (!(b >= a && b <= c)) b = 0.5f * (a + c);

        /* Evaluate the CDF and its derivative (i.e. the density function) */
        float invErf = ErfInv(b);
        float value =   normalization * (1 + b + SQRT_PI_INV * tan_theta * std::exp(-invErf * invErf)) - sample_x;
        float derivative = normalization * (1 - invErf * tan_theta);
        if (std::abs(value) < 1e-5f) break;

        /* Update bisection intervals */
        if (value > 0)
            c = b;
        else
            a = b;

        b -= value / derivative;
    }

    /* Now convert back into a slope value */
    *slope_x = ErfInv(b);

    /* Simulate Y component */
    *slope_y = ErfInv(2.0f * std::max(U2, (float)1e-6f) - 1.0f);
}
