#pragma once

// std lib
#include <cstring>
#include <memory>

// gkit
#include "vec.h" 
#include "mesh.h"
#include "mesh_cache.h"
#include "wavefront.h"
#include "texturelib.h"

// vector / point
#include "renderer/ray.h"
#include "renderer/geometry.h"
#include "renderer/lightdistribution.h"
#include "renderer/interpolation.h"
#include "renderer/media/medium.h"

static float Epsilon(const Point& p, const float t )
{
    return 32.0f * 1.19209e-07f * std::max(std::max(std::abs(p.x),std::abs(p.y)),std::max(std::abs(p.z), t));
    // return 32.0f * 1.19209e-07f * std::max(std::max(std::abs(p.x),std::abs(p.y)),std::abs(p.z));
    // return 10.0f * std::max(std::max(std::abs(p.x),std::abs(p.y)),std::abs(p.z)) / float(1 << 23);
}

struct Interpolation
{
    bool normals = false;
    bool texcoords = false;
    bool dynamic = false;
    Vector n, n0, n1;
    vec2 t;
    Point p, dp, p0, p1;
    float t0, t1;

    const Bsdf * bsdf = nullptr;
};

struct MediumData
{
    std::string volume_name;
    std::string material_name;
    std::string nanovdb_file;
    std::string nanovdb_grid = "density";
    std::vector<std::string> inside;
    std::vector<std::string> outside;
    bool use_bsdf = false;
    bool noise = false;
    bool anim = false;
    bool infinite = false;
    std::vector<Transform> tr;
    // volume absorption and scattering
    Color sigmaa = Black(), sigmas = Black();           
    // phase 
    float phaseg = 0;                                   
    // noise params
    float noise_freq = 10.f;
    int noise_seed = 1234;
    DensityFunctionFastNoise::NoiseType noise_type; 

    // return the i-est instance of the medium
    std::shared_ptr<Medium> get(int i = 0) const
    {
        std::shared_ptr<Medium> medium = nullptr;
        if( noise )
        {
            printf("\tdensity fast noise\n");
            medium = std::make_shared<MediumHeterogeneous>(
                sigmaa,
                sigmas,
                std::make_shared<DensityFunctionFastNoise>(noise_seed, noise_type, noise_freq, 1, (i < (int)tr.size() ? tr[i] : Identity())),
                std::make_shared<BsdfHenyeyGreenstein>(phaseg)
            );
        }
        else if( !nanovdb_file.empty() )
        {
            printf("\tdensity NanoVDB\n");
            medium = std::make_shared<MediumHeterogeneous>(
                sigmaa,
                sigmas,
                std::make_shared<DensityFunctionNanoVDB>(nanovdb_file, nanovdb_grid, (i < (int)tr.size() ? tr[i] : Identity())),
                std::make_shared<BsdfHenyeyGreenstein>(phaseg)
            );
        }
        else if( anim )
        {
            printf("\tdensity animated\n");
            medium = std::make_shared<MediumHeterogeneous>(
                sigmaa,
                sigmas,
                std::make_shared<DensityFunctionFastNoiseAnimated>(noise_seed, noise_freq, 1),
                std::make_shared<BsdfHenyeyGreenstein>(phaseg)
            );
        }
        else
        {
            printf("\tdensity homogeneous\n");
            medium = std::make_shared<MediumHomogeneous>(
                sigmaa,
                sigmas,
                1,
                std::make_shared<BsdfHenyeyGreenstein>(phaseg)
            );
        }
        return medium;
    }
};

// builder : EMBREE STBVH
class Builder
{
public:
    // scene data
    std::vector<Geometry> geometries;
    std::vector<MediumData> mediumData;
    std::shared_ptr<Medium> mediumDefault = nullptr;
    int max_steps = 0;

    // light sources
    LightDistribution lights;  

    // embree scene
    RTCDevice device;
    RTCScene scene;
    RTCIntersectContext coherent;
    RTCIntersectContext incoherent;  

    bool primary( Ray& primary_ray )   
    { 
        rtcIntersect1(scene, &coherent, &primary_ray.data);       
        return primary_ray.hit();
    }

    bool intersect( Ray& secondary_ray ) 
    { 
        rtcIntersect1(scene, &incoherent, &secondary_ray.data);   
        return secondary_ray.hit();
    }
    
    bool visible( Ray& shadow_ray )    
    { 
        rtcOccluded1(scene, &incoherent, &shadow_ray.data.ray);   
        return shadow_ray.data.ray.tfar > 0;
    }

    bool intersect( Ray& ray, MediumRecords& records, bool primary )
    {
        // init intersection context
        auto& context = primary ? coherent : incoherent;

        // init medium records
        records.distances.clear();
        records.mediums.clear();
        records.ray = ray;

        if( mediumDefault ) ray.medium = mediumDefault.get();

        // if we start in a medium init the first interval
        if( ray.medium )
        {
            records.distances.push_back(ray.tmin());
            records.mediums.emplace_back(ray.medium);
        }

        // start intersection loop
        do
        {
            // intersect the geometry
            rtcIntersect1(scene, &context, &ray.data);
            
            // if an intersection was found
            if( ray.hit() )
            {
                // if we are within a medium end the medium interval
                if( ray.medium )
                    records.distances.push_back( ray.hitt() );

                // check if we hit a light source
                auto& material = get_material( ray.hitobject(), ray.hitid() );
                if( material.emission.luminance() > 0 )
                    break;

                // check if we hit a surface
                const Bsdf * bsdf_ptr = get_bsdf( ray.hitobject(), ray.hitid() );
                if( bsdf_ptr )
                    break;

                // else the hit corresponds to a volume boundary
                const Medium * const medium =
                    dot(ray.d(), ray.hitn()) > 0 ?
                    // ray is going from inside to outside
                    get_exterior_medium(ray.hitobject(), ray.hitid()):
                    // ray is going from outside to inside
                    get_interior_medium(ray.hitobject(), ray.hitid());

                // add a small epsilon
                float t = ray.hitt() + Epsilon(ray.hitp(), ray.hitt());
                
                // starts the next medium interval
                if( medium )
                {
                    // records.distances.push_back(t);
                    records.distances.push_back(ray.hitt());
                    records.mediums.emplace_back(medium);
                }

                // init next ray
                // warning 
                // update correctly tnear with the last intersection distance + epsilon 
                // and tfar with the initial ray tfar 
                // not to have surprise for visibility tests
                ray = Ray(ray.o(), ray.d(), ray.time(), t, records.ray.tmax());
                ray.medium = medium;
            }
            else // no intersection
            {
                // if we are still within a medium
                if( ray.medium )
                {
                    // if the ray was bounded (shadow rays) the extremity of the ray is in the current medium
                    if( records.ray.tmax() != std::numeric_limits<float>::infinity() )
                    {
                        // sometimes the sections are epsilon sized and cause issues hence we ignore them
                        if( records.ray.tmax() <= records.distances.back() )
                        {
                            #ifndef GK_RELEASE
                            printf("Builder::intersect - broken record back %.20f front %.20f section id %d\n", 
                                records.ray.tmax(), 
                                records.distances.back(), 
                                records.size()/2
                            );
                            #endif
                            records.distances.pop_back();
                            records.mediums.pop_back();
                        }
                        else
                            records.distances.push_back(records.ray.tmax());
                    }
                    // infinite mediums are not considered yet 
                    else if( mediumDefault )
                    {
                        // printf("(%f %f %f) primary %d\n", ray.o().x, ray.o().y, ray.o().z, (int)primary);
                        // records.distances.pop_back();
                        // records.mediums.pop_back();
                        records.distances.push_back(records.ray.tmax());
                    }
                }

                break;
            }
        }   
        while(true);

        return records.size() || ray.hit();
    }

    Builder()
    {
        // embree device configuration
        #if DEBUG
            const char * config = "verbose=3";
        #else
            // const char * config = "verbose=3";
            const char * config = "";
        #endif

        // embree init
        device = rtcNewDevice( config );
        scene = rtcNewScene( device );
        rtcSetSceneBuildQuality( scene, RTC_BUILD_QUALITY_HIGH );
        coherent = {RTC_INTERSECT_CONTEXT_FLAG_COHERENT, 0, {}}; 
        incoherent = {RTC_INTERSECT_CONTEXT_FLAG_INCOHERENT, 0, {}};
        rtcInitIntersectContext(&coherent);
        rtcInitIntersectContext(&incoherent);
    }

    ~Builder()
    {
        // release memory
        rtcReleaseScene( scene );
        rtcReleaseDevice( device );
    }

    int get_steps( const int object_id ) const {return (int)geometries[object_id].step_count();}

    const Material& get_material( const int object_id, const int triangle_id ) const 
    {
        assert( object_id >= 0 && object_id < (int)geometries.size() );
        assert( triangle_id >= 0 && triangle_id < (int)geometries[object_id].triangle_count() );
        int index = geometries[object_id].material_indices[triangle_id];
        assert( index >= 0 && index < (int)geometries[object_id].material_count() );
        return geometries[object_id].materials[index];
    }

    const Color get_emission( const int object_id, const int triangle_id, const vec2& uv) const 
    {
        assert( object_id >= 0 && object_id < (int)geometries.size() );
        assert( triangle_id >= 0 && triangle_id < (int)geometries[object_id].triangle_count() );
        int index = geometries[object_id].material_indices[triangle_id];
        assert( index >= 0 && index < (int)geometries[object_id].material_count() );
        int texture_id = geometries[object_id].materials[index].emission_texture;
        if(texture_id > 0)
        {
            auto texture = geometries[object_id].textures.textures[texture_id]; 
            return (*texture)(uv) * geometries[object_id].materials[index].emission;
        }
        return geometries[object_id].materials[index].emission;
    }

    const Bsdf * get_bsdf( const int object_id, const int triangle_id ) const 
    {
        assert( object_id >= 0 && object_id < (int)geometries.size() );
        assert( triangle_id >= 0 && triangle_id < (int)geometries[object_id].triangle_count() );
        int index = geometries[object_id].material_indices[triangle_id];
        assert( index >= 0 && index < (int)geometries[object_id].material_count() );
        return geometries[object_id].bsdfs[index] ? geometries[object_id].bsdfs[index].get() : nullptr;
    }

    const Medium * get_interior_medium( const int object_id, const int triangle_id ) const 
    {
        if( mediumDefault ) return mediumDefault.get();
        int index = geometries[object_id].material_indices[triangle_id];
        return geometries[object_id].interiors[index].get();
    }

    const Medium * get_exterior_medium( const int object_id, const int triangle_id ) const 
    {
        if( mediumDefault ) return mediumDefault.get();
        int index = geometries[object_id].material_indices[triangle_id];
        return geometries[object_id].exteriors[index].get();
    }

    const LightDistribution&  get_lights() const {return lights;}

    inline Interpolation interpolate( const Ray& ray )
    {
        return interpolate( 
            ray.hitobject(), 
            ray.hitid(), 
            ray.hitu(), 
            ray.hitv(), 
            ray.hitt(), 
            ray.time(), 
            ray.o()
        );
    }

    inline Interpolation interpolate( 
        const unsigned int& hitobject, 
        const unsigned int& hitid, 
        const float& hitu, 
        const float& hitv, 
        const float& hitt, 
        const float& time, 
        const Point& o
        )
    {
        Interpolation i;

        // non motion blur geometry
        if( get_steps(hitobject) == 1 )
        {
            i.t0 = i.t1 = 0;
            // normals / texcoords
            i.normals = (geometries[hitobject].normal_count() > 0);
            i.texcoords = (geometries[hitobject].texcoord_count() > 0);
            // triangle
            Triangle t = geometries[hitobject].triangle(0, hitid);
            // point
            i.p = t.point(hitu, hitv);
            // normal
            if( i.normals ) // shading
                i.n = t.normal(hitu, hitv);
            else // geometric
                i.n = t.normal();
            
            // flip normal if this is not a light
            // if( get_material(hitobject, hitid).emission.power() <= 0 && dot(i.n, Vector(i.p,o)) < 0 ) 
                // i.n = -i.n;

            // texcoord
            if( i.texcoords ) i.t = t.texcoord(hitu, hitv);

            // time frame 0 frame 1
            i.t0 = 0;
            i.t1 = 1;

            // point displacement for visibility tests and ray spawn
            // float eps = 0.0001;
            // float eps = 32.0f * 1.19209e-07f * std::max(std::max(std::abs(i.p.x),std::abs(i.p.y)),std::max(std::abs(i.p.z), hitt));
            float eps = Epsilon(i.p,hitt);

            i.dp = dot(i.n, Vector(i.p,o)) < 0 ? i.p + eps * -i.n : i.p + eps * i.n;
        }
        // motion blur geometry
        else
        {
            // frames id
            float nb = get_steps(hitobject) - 1; 
            int slot0 = time * nb;
            int slot1 = slot0 == nb ? slot0 : slot0 + 1; 

            // t normalized between frames
            float t = time * nb - slot0;

            // time frame 0 frame 1
            i.t0 = float(slot0) / nb;
            i.t1 = float(slot1) / nb;

            // normals / texcoords
            i.normals = (geometries[hitobject].normal_count() > 0);
            i.texcoords = (geometries[hitobject].texcoord_count() > 0);

            // triangles
            Triangle t0 = geometries[hitobject].triangle(slot0, hitid);
            Triangle t1 = geometries[hitobject].triangle(slot1, hitid);

            // point
            Point p0 = t0.point(hitu, hitv);
            Point p1 = t1.point(hitu, hitv);
            i.p = lerp(p0, p1, t);
            i.p0 = p0;
            i.p1 = p1;

            // normal
            i.n0 = i.normals ? t0.normal(hitu, hitv) : t0.normal();
            i.n1 = i.normals ? t1.normal(hitu, hitv) : t1.normal();
            i.n = normalize(lerp(i.n0, i.n1, t));
            
            // flip normal if this is not a light
            // if( get_material(hitobject, hitid).emission.power() <= 0 && dot(i.n, Vector(i.p,o)) < 0 ) 
                // i.n = -i.n;

            i.dynamic = (i.n0.x != i.n1.x) ||
                        (i.n0.y != i.n1.y) ||
                        (i.n0.z != i.n1.z) ||
                        (i.p0.x != i.p1.x) ||
                        (i.p0.y != i.p1.y) ||
                        (i.p0.z != i.p1.z);

            // texcoord
            if( i.texcoords ) i.t = lerp(t0.texcoord(hitu, hitv), t1.texcoord(hitu, hitv), t);

            // point displacement for visibility tests and ray spawn
            // float eps = 0.0001;
            // float eps = 32.0f * 1.19209e-07f * std::max(std::max(std::abs(i.p.x),std::abs(i.p.y)),std::max(std::abs(i.p.z), hitt));
            float eps = Epsilon(i.p,hitt);
            // i.dp = i.p + eps * i.n;
            i.dp = dot(i.n, Vector(i.p,o)) < 0 ? i.p + eps * -i.n : i.p + eps * i.n;
        }

        i.bsdf = get_bsdf(hitobject, hitid);
        return i;
    }

    Color diffuse( const Ray& ray, const Interpolation& i, const float lod = 0) const
    {
        if( !ray.hit() ) return Black();
        const Material& material = get_material(ray.hitobject(), ray.hitid()); 
        if(material.diffuse_texture == -1 || !i.texcoords ) return material.diffuse;
        assert(material.diffuse_texture < int(geometries[ray.hitobject()].textures.size()));
        return geometries[ray.hitobject()].textures.textures[material.diffuse_texture]->operator()(i.t.x, i.t.y);
    }

    Color specular( const Ray& ray, const Interpolation& i, const float lod = 0) const
    {
        if( !ray.hit() ) return Black();
        const Material& material = get_material(ray.hitobject(), ray.hitid()); 
        if(material.specular_texture == -1 || !i.texcoords ) return material.specular;
        assert(material.specular_texture < int(geometries[ray.hitobject()].textures.size()));
        return geometries[ray.hitobject()].textures.textures[material.specular_texture]->operator()(i.t.x, i.t.y);
    }

    Color emission( const Ray& ray, const Interpolation& i, const float lod = 0) const
    {
        if( !ray.hit() ) return Black();
        const Material& material = get_material(ray.hitobject(), ray.hitid()); 
        // if(material.emission_texture == -1 || !i.texcoords ) 
            return material.emission;
        // assert(material.emission_texture < int(geometries[ray.hitobject()].textures.size()));
        // return geometries[ray.hitobject()].textures.textures[material.emission_texture]->operator()(i.t.x, i.t.y);
    }

    // add objects to scene
    void add( const std::vector<std::string>& filenames )
    {       
        // set dynamic scene flags / build quality
        if( filenames.size() > 1 )
            printf("dynamic object\n");
        else
            printf("static object\n");
        printf("%lu keyframes.\n", filenames.size());

        // keyframes loading
        std::vector<MeshData> data(filenames.size());
        #pragma omp parallel for 
        for(unsigned int i = 0; i < filenames.size(); ++i)
        {
            // TODO : write obj cache
            data[i] = read_mesh_data( filenames[i].c_str() );
        }

        // mesh split
        Geometry g_static, g_dynamic;
        split_mesh( data, g_static, g_dynamic);

        // add objects to scene
        add( g_static );
        add( g_dynamic );
    }

    // add objects to scene
    void add( const Geometry& geometry )
    {       
        if( geometry.step_count() > 0 )
        {
            // printf("add geometry\n");
            geometries.push_back(geometry);

            RTCGeometry geom = rtcNewGeometry ( device, RTC_GEOMETRY_TYPE_TRIANGLE );
            rtcSetGeometryBuildQuality( geom, RTC_BUILD_QUALITY_HIGH );
            rtcSetGeometryTimeStepCount( geom, geometries.back().step_count() ); 

            for(int i = 0; i < geometries.back().step_count(); i++)
            {
                void *positions= rtcSetNewGeometryBuffer(geom, RTC_BUFFER_TYPE_VERTEX, i, RTC_FORMAT_FLOAT3, sizeof(vec3), geometries.back().vertex_count(i));
                assert(positions);
                memcpy(positions, geometries.back().vertex_buffer(i), geometries.back().vertex_buffer_size(i));
            }

            void *indices= rtcSetNewGeometryBuffer(geom, RTC_BUFFER_TYPE_INDEX, 0, RTC_FORMAT_UINT3, 3 * sizeof(unsigned int), geometries.back().triangle_count());
            assert(indices);
            memcpy(indices, geometries.back().vertex_index_buffer(), geometries.back().vertex_index_buffer_size());

            rtcCommitGeometry( geom );
            int geomID = rtcAttachGeometry( scene, geom );
            rtcReleaseGeometry( geom );
        }
    }

    bool find_material(std::string name, int& gid, int& mid)
    {
        gid = 0;
        for(; gid < (int)geometries.size(); ++gid )
        {
            const auto& geom = geometries[gid];
            mid = 0;
            for(; mid < (int)geom.materials.size(); ++mid )
            {
                const auto& mat = geom.materials[mid];
                if( mat.name == name )
                    return true;
            }
        }
        gid = -1; 
        mid = -1;
        return false;
    }

    void init_mediums()
    {
        printf("\nvolumes:\n");
        // read medium data
        for(const auto& mdata : mediumData )
        {
            printf("\tlink %s\n", mdata.volume_name.c_str());

            // one or several mediums transformed inside material proxies 
            if( mdata.tr.size() > 0 && mdata.inside.size() > 0 )
            {
                assert(mdata.inside.size() == mdata.tr.size() ); // check 
                for(int i =0; i < (int)mdata.tr.size(); ++i)
                {
                    // init medium
                    std::shared_ptr<Medium> medium = mdata.get(i);

                    // find material in geometries
                    int gid = 0, mid = 0;
                    if( !find_material(mdata.inside[i], gid, mid) )
                        continue;

                    // attach medium
                    geometries[gid].interiors[mid] = medium;

                    // handle bsdf interface if needed
                    if( !mdata.use_bsdf )
                    {
                        geometries[gid].bsdfs[mid] = nullptr;
                    }
                    else if( mdata.use_bsdf && !mdata.material_name.empty() )
                    {
                        int gid2 = 0, mid2 = 0;
                        if( find_material(mdata.material_name, gid2, mid2) )
                        {
                            geometries[gid].bsdfs[mid] = geometries[gid2].bsdfs[mid2];
                        }
                    }
                }   
            }
            // one medium inside material proxies 
            else if( mdata.inside.size() > 0 ) 
            {
                std::shared_ptr<Medium> medium = mdata.get();
                for(int i =0; i < (int)mdata.inside.size(); ++i)
                {
                    std::string mat = mdata.inside[i];
                    printf("\tinside %s\n", mat.c_str());

                    // find material in geometries
                    int gid = -1, mid = -1;
                    if( !find_material(mdata.inside[i], gid, mid) )
                        continue;

                    // attach medium
                    geometries[gid].interiors[mid] = medium;

                    // handle bsdf interface if needed
                    if( !mdata.use_bsdf )
                    {
                        geometries[gid].bsdfs[mid] = nullptr;
                        printf("\tbsdf set to null\n");
                    }
                    else if( mdata.use_bsdf && !mdata.material_name.empty() )
                    {
                        int gid2 = 0, mid2 = 0;
                        if( find_material(mdata.material_name, gid2, mid2) )
                        {
                            geometries[gid].bsdfs[mid] = geometries[gid2].bsdfs[mid2];
                            printf("\tbsdf set to %s\n", mdata.material_name.c_str());
                        }
                    }
                    else
                    {
                        printf("\tbsdf is %s\n", geometries[gid].materials[mid].name.c_str());
                    }

                    // printf("\tmedium set to %lu\n", medium);
                }
            }
            // one or several mediums inside box proxies 
            else if( mdata.tr.size() > 0 ) 
            {
                // Note here this is fucked, tr is used to place the box and in density function calls..  maybe not ?
                // multiple transformations
                for(int i =0; i < (int)mdata.inside.size(); ++i)
                {
                    // init medium
                    std::shared_ptr<Medium> medium = mdata.get(i);

                    // build geometry
                    Geometry g = make_cube(mdata.tr[i]);
                
                    // attach medium
                    g.interiors.back() = medium;

                    // print infos
                    g.print();

                    // attache to builder
                    add(g);

                    // handle bsdf interface if needed
                    if( !mdata.use_bsdf )
                    {
                        geometries.back().bsdfs.back() = nullptr;
                    }
                    else if( mdata.use_bsdf && !mdata.material_name.empty() )
                    {
                        int gid2 = 0, mid2 = 0;
                        if( find_material(mdata.material_name, gid2, mid2) )
                        {
                            geometries.back().bsdfs.back() = geometries[gid2].bsdfs[mid2];
                        }
                    }
                }
            }
            else if ( mdata.infinite ) // infinite medium
            {
                printf("\tinfinite\n");
                mediumDefault = mdata.get();
            }
        }
    }

    void build()
    {   
        // build light distribution with static geometries only !!
        // todo add support for dynamic area lights
        printf("\nmaterials:\n");
        std::vector<AreaLight> sources;
        for( int gid = 0; gid < int(geometries.size()); gid++)
        {
            // build bsdf
            // printf("init geometry %d bsdf\n", gid);
            geometries[gid].init_bsdf();
                        
            // printf("init geometry %d sources\n", gid);
            max_steps = std::max(max_steps, geometries[gid].step_count());
            if( geometries[gid].step_count() > 1 ) continue;
            for( int tid = 0; tid < geometries[gid].triangle_count(); tid++)
            {
                // recupere la matiere associee a chaque triangle de l'objet
                Triangle t = geometries[gid].triangle(0, tid);
                int index = geometries[gid].material_indices[tid];
                const Material& material= geometries[gid].materials[index];
                if(material.emission.power() > 0 && t.area() > 0 )
                    sources.push_back( AreaLight(t, material.emission, gid, tid) );
            }
        }

        printf("\n");
        lights.insert( sources );
        // build volumes
        init_mediums();

        // build scene
        rtcCommitScene( scene );
    }
};
