#ifndef CAMERA_H
#define CAMERA_H

#include <cassert>  
#include <cmath>
#include <fstream>

#include "vec.h"
#include "mat.h"
#include "orbiter.h"

#include "renderer/ray.h"
#include "renderer/film.h"
#include "renderer/interpolation.h"

class Lens
{
public:
    Lens(const float r = 0.f) : r(r), a( r > 0.f ? r * r * M_PI : 1.f ), p(1.f / a) {}
    inline Point sample( const vec2& u ) const { return sample(u.x, u.y); }
    inline Point sample( float u1, float u2 ) const
    {
        if (r <= 0) return Point();

        u1 = 2.f * (u1 - 0.5f);
        u2 = 2.f * (u2 - 0.5f);
        if( u1 == 0 && u2 == 0 ) return Point();

        float rho, theta;
        if(std::abs(u1) > std::abs(u2))
        {
            rho = u1;
            theta = 0.25f * M_PI * u2 / u1;
        }
        else
        {
            rho = u2;
            theta = M_PI * ( 0.5f - 0.25f * u1 / u2 );
        }
        return Point(r * rho * std::cos(theta), r * rho * std::sin(theta), 0.f); 
    }

    inline float area() const {return a;}
    inline float eval() const {return p;}
    inline float radius() const {return r;}
    inline float radius2() const {return r > 0 ? r * r : 1;}

private:
    float r, a, p;
};

struct CameraSample
{
    Ray ray;
    Color weight = Black();
    float pdf = 0;
    vec2 filmPosition;
    vec2 lensSample = vec2(0,0);
};

class Camera
{
public:
    // film size
    float width = -1.f;
    float height = -1.f;
    float fov = 45.f;

    // position on optical axis
    float z_film = 3.f;
    float z_focus = 1.f;
    float z_near = 0.01f;
    float z_far = 1000.f;
    float z_focal = 0.05f;

    // time range
    float open = 0.f;
    float close = 0.f;

    // transformations
    Transform tr_view, tr_proj, tr_film, w2c, c2w, c2r, r2c, w2r ,r2w;
    
    Lens lens;
    Film film;

    const Vector n_camera = Vector(0, 0, -1);
    Vector n_world;
    Vector dx_camera, dx_world; 
    Vector dy_camera, dy_world;
    float film_area = 0.f; 

    inline Point position() const {return c2w(Point(0,0,0));}
    inline Point from() const { Transform inv_view = tr_view.inverse(); return Point(inv_view.m[0][3], inv_view.m[1][3], inv_view.m[2][3]); }
    inline Point to() const { Transform inv_view = tr_view.inverse(); return from() - Vector(inv_view.m[0][2], inv_view.m[1][2], inv_view.m[2][2]); }
    inline Vector up() const { Transform inv_view = tr_view.inverse(); return Vector(inv_view.m[0][1], inv_view.m[1][1], inv_view.m[2][1]); }
    inline Vector normal() const {return n_world;}
    inline float area() const {return film_area;}

    Camera(){}

    Camera(const Point from, const Point to, const int width, const int height, const float fov = 45) :
            width(width), height(height), fov(fov), film(width, height)
    {
        // compute camera transformations
        // image space ---> camera space ---> world space
        // world space ---> camera space ---> image space

        z_far = 2.f * distance(from, to);
        tr_view = Lookat(from, to, Vector(0,1,0));
        tr_proj = Perspective(fov, float(width) / float(height), z_near, z_far);
        tr_film = Viewport(width, height); 
        init();      
    }

    Camera(const Orbiter& orbiter, const int width, const int height ) :
            width(width), height(height), film(width, height)
    {
        // compute camera transformations
        // image space ---> camera space ---> world space
        // world space ---> camera space ---> image space
        tr_view = orbiter.view();
        tr_proj = orbiter.projection(width, height, fov);
        tr_film = Viewport(width, height);

        // recompute z_near, z_far from projection matrix
        float a = tr_proj.m[2][2]; 
        float b = tr_proj.m[2][3];
        z_near = b / (a - 1.f);
        z_far = b / (a + 1.f);
        init();
    }

    Camera( const std::string& filename )
    {
        read(filename);
        init(); 
    }

    void init() 
    {
        // precompute camera transformations

        w2r = tr_film * tr_proj * tr_view;      // world to raster  (image space)
        r2w = w2r.inverse();                    // raster to world  (world space)
        
        c2r = tr_film * tr_proj;                // camera to raster (image space)
        r2c = c2r.inverse();                    // raster to camera (camera space)

        w2c = tr_view;                          // world to camera  (camera space)
        c2w = w2c.inverse();                    // camera to world  (world space)

        // precompute camera normal

        n_world = normalize(c2w(n_camera));

        // we need to find raster bounds in camera space at the specified film distance z_film
        //      
        //    (0,h)--------------(w,h)          py-----------------*
        //      |                  |            |                  |
        //      |                  |            |                  |
        //      |                  |     -->    |                  | 
        //      |                  |            |                  |
        //      |                  |            |                  |
        //    (0,0)--------------(w,0)          p0-----------------px
        //

        Point p0 = r2c(Point(0, 0, 1));
        Point px = r2c(Point(width, 0, 1));
        Point py = r2c(Point(0, height, 1));

        // compute directions between camera origin and raster bounds
        //      
        //         py-----------------*
        //     vy /|                  |
        //       / |                  | 
        //   O ./___                  | 
        //      \  |---___   vx       | 
        //       \ |      ---___      |
        //     v0 \|            ---___|
        //         p0-----------------px
        //
      
        Vector v0 = normalize(Vector(p0));
        Vector vx = normalize(Vector(px));
        Vector vy = normalize(Vector(py));
        
        // compute raster bounds in camera space at distance z_film
        //
        //                    film
        //                      |
        //                      |
        //      O .-------------|-------
        //                      |
        //         <- z_film -> |
        //

        p0 = Point(v0 * z_film / std::abs(v0.z));
        px = Point(vx * z_film / std::abs(vx.z));
        py = Point(vy * z_film / std::abs(vy.z));

        // compute film area
        film_area = length(cross(Vector(p0, px), Vector(p0, py)));

        // compute ray differentials in camera space
        dx_camera = Vector(p0, px) / float(width);
        dy_camera = Vector(p0, py) / float(height);
    }

    void print() const
    {
        printf("camera:\n\tlens radius\t%f\n\tlens area\t%f\n\tfocus distance\t%f\n\tfocal distance\t%f\n\tfilm distance\t%f\n\tfilm area\t%f\n\topen\t\t%f\n\tclose\t\t%f\n\twidth\t\t%f\n\theight\t\t%f\n", 
            lens.radius(), lens.area(), z_focus, z_focal, z_film, film_area, open, close, float(width), float(height));
    }

    // sample image plane coordinates
    vec2 sampleFilm(const vec2& v, const int px, const int py) const
    {   
        return sampleFilm(v.x, v.y, px, py);
    }

    vec2 sampleFilm(const float u, const float v, const int px, const int py) const
    {   
        return vec2( float(px) + u - 0.5f, float(py) + v - 0.5f ); 
    }

    // check the coordinates given lie within the image plane
    bool valid(const vec2& p) const 
    { 
        return ( p.x >= -0.5f && p.x <= width - 0.5f && p.y >= -0.5f && p.y <= height - 0.5f ); 
    }

    // check the time given lie within the time window
    bool in_time_window(const float t) const 
    { 
        return t <= close && t >= open; 
    }

    // triangle time filter
    float time_filter(const float t) const 
    { 
        if( close == open ) return 1.f;
        float t01 = (t - open) / (close - open);
        float tminus11 = 2.f * t01 - 1.f;
        return std::max(0.f, 1 - std::abs(tminus11));
    }

    // return pixel coordinates
    vec2 pixel(const Ray& ray) const 
    { 
        // check if ray point out of the camera
        Point p_lens = w2c(ray.o());
        Vector w_lens = w2c(ray.d());
        // find point of focus
        float t_focus = std::abs(z_focus / w_lens.z);
        Point p_focus = p_lens + t_focus * w_lens;
        // find pixel position 
        Point p_raster = c2r(p_focus);
        return vec2(p_raster.x, p_raster.y);
    }

    // compute importance for a given ray and return the associated pixel
    float we(const Ray& ray, vec2& pixel) const 
    {
        // check if ray belongs to the time window
        if( ray.time() > close || ray.time() < open  ) return 0.f;

        // check if ray point out of the camera
        Point p_lens = w2c(ray.o());
        Vector w_lens = w2c(ray.d());
        if (w_lens.z >= 0.f) return 0.f;
        
        // find point of focus
        float t_focus = std::abs(z_focus / w_lens.z);
        Point p_focus = p_lens + t_focus * w_lens;

        // find pixel position 
        Point p_raster = c2r(p_focus);
        pixel = vec2(p_raster.x, p_raster.y);

        // return zero importance for out of bounds points
        if( !valid(pixel) ) return 0.f;
        
        // compute importance value
        //
        //  We(p_lens, w_lens) = p(w_lens) * p(p_lens) / cos_theta_lens
        //
        //      !!! theta_lens is the angle between the camera normal and the ray leaving the pupil 
        //      (the pupil is merged with the lens aperture for the thin lens model)
        //
        //  p(w_lens) = d^2(p_film, p_lens) / (A_film * cos_theta_film)
        //            = z_film * z_film / (A_film * cos_theta_film * cos_theta_film * cos_theta_film)
        //
        //      !!! theta_film is the angle between the camera normal and the direction from film to lens
        //  
        //  p(p_lens) = 1 / A_lens
        //  
        //      !!! note that theta_film and theta_lens are equals if and only if z_film = z_focus
        //

        // direction between camera origin, film point and point of focus 
        Vector w_focus = normalize(Vector(p_focus));

        // find true film point in camera space
        float t_film = std::abs(z_film / w_focus.z); 
        Point p_film(t_film * w_focus);

        // lens to film direction in camera space
        Vector w_film = normalize(p_lens - p_film);

        // compute probability densities
        float pdf_p_lens = lens.eval();
        float pdf_w_lens = z_film * z_film / (film_area * w_film.z * w_film.z * w_film.z );
        float pdf_time = close-open > 0 ? 1.f/std::abs(close-open) : 1;
        
        // return importance value
        return std::abs(pdf_time * pdf_w_lens * pdf_p_lens / w_lens.z);
    }

    // compute pdf of emmited importance for a given ray
    float pdf_we( const Ray& ray ) const
    {
        // check if ray belongs to the time window
        if( ray.time() > close || ray.time() < open  ) return 0.f;

        // check if ray point out of the camera
        Point p_lens = w2c(ray.o());
        Vector w_lens = w2c(ray.d());
        if (w_lens.z >= 0.f) return 0.f;
        
        // find point of focus
        float t_focus = std::abs(z_focus / w_lens.z);
        Point p_focus = p_lens + t_focus * w_lens;

        // find pixel position 
        Point p_raster = c2r(p_focus);

        // return zero for out of bounds points
        if( !valid(vec2(p_raster.x, p_raster.y)) ) return 0.f;
        
        // compute pdf
        //
        //  p(p_lens, w_lens) = p(w_lens) * p(p_lens)
        //
        //  p(w_lens) = d^2(p_film, p_lens) / (A_film * cos_theta_film)
        //            = z_film * z_film / (A_film * cos_theta_film * cos_theta_film * cos_theta_film)
        //
        //      !!! theta_film is the angle between the camera normal and the direction from film to lens
        //  
        //  p(p_lens) = 1 / A_lens
        //  
        //      !!! note that theta_film and theta_lens are equals if and only if z_film = z_focus
        //

        // direction between camera origin, film point and point of focus 
        Vector w_focus = normalize(Vector(p_focus));

        // find true film point in camera space
        float t_film = std::abs(z_film / w_focus.z); 
        Point p_film(t_film * w_focus);

        // lens to film direction in camera space
        Vector w_film = normalize(p_lens - p_film);
        
        // compute probability densities
        float pdf_p_lens = lens.eval();
        float pdf_w_lens = z_film * z_film / (film_area * w_film.z * w_film.z * w_film.z );
        float pdf_time = close-open > 0 ? 1.f/std::abs(close-open) : 1;

        // return pdf of ray
        return std::abs(pdf_time * pdf_w_lens * pdf_p_lens);     
    }

    CameraSample sample( Sampler& sampler, const int px, const int py, bool computePdf = false ) const
    {
        CameraSample cs;

        // image plane coordinates
        cs.filmPosition = sampleFilm(sampler.sample(), sampler.sample(), px, py); 

        // origin in camera space
        Point o(0,0,0);

        // ray direction in camera space
        Point p_film = r2c(Point(cs.filmPosition.x, cs.filmPosition.y, 1.f));
        Vector d = normalize(Vector(p_film));
        
        // depth of field
        float t_focus = -1;
        Point p_focus;
        if( lens.radius() ) 
        {
            // find focus point in camera space
            t_focus = std::abs(z_focus / d.z);
            p_focus = Point(t_focus * d);

            // sample point on lens
            cs.lensSample = vec2(sampler.sample(), sampler.sample());
            o = lens.sample(cs.lensSample);

            // update ray for effect of lens
            d = normalize(p_focus - o);
        }

        cs.ray = Ray(c2w(o), c2w(d), lerp(open, close, sampler.sample()));
        cs.weight = White();
        cs.pdf = 1;

        // compute pdf
        //
        //  p(p_lens, w_lens) = p(w_lens) * p(p_lens)
        //
        //  p(w_lens) = d^2(p_film, p_lens) / (A_pixel * cos_theta_film)
        //            = z_film * z_film / (A_pixel * cos_theta_film * cos_theta_film * cos_theta_film)
        //            = z_film * z_film / (A_film * cos_theta_film * cos_theta_film * cos_theta_film) * (width * height)
        //
        //      !!! theta_film is the angle between the camera normal and the direction from film to lens
        //      !!! A_pixel is not directly stored to keep the floating point precision alive
        //  
        //  p(p_lens) = 1 / A_lens
        //  
        //      !!! note that theta_film and theta_lens are equals if and only if z_film = z_focus
        //
        
        if( computePdf )
        {
            // lens point and direction in camera space
            Point p_lens = o;
            Vector w_lens = d;

            // find point of focus
            float t_focus = std::abs(z_focus / w_lens.z);
            Point p_focus = p_lens + t_focus * w_lens;

            // direction between camera origin, film point and point of focus 
            Vector w_focus = normalize(Vector(p_focus));

            // find true film point in camera space
            float t_film = std::abs(z_film / w_focus.z); 
            Point p_film(t_film * w_focus);

            // lens to film direction in camera space
            Vector w_film = normalize(p_lens - p_film);
            
            // compute probability densities
            float pdf_p_lens = lens.eval();
            float pdf_w_lens = z_film * z_film / (film_area * w_film.z * w_film.z * w_film.z ) * (width * height);
            float pdf_time = close-open > 0 ? 1.f/std::abs(close-open) : 1;

            // compute pdf of ray
            cs.pdf = std::abs(pdf_w_lens * pdf_p_lens * pdf_time);   
            if( std::isnan(cs.pdf) )
                printf("Camera::sample pdf %f %f %f %f\n", pdf_p_lens, pdf_w_lens, film_area, w_film.z);  
        }

        return cs;
    }

    CameraSample connect( const Point& pWorld, const vec2& lensSample, const float time = 0) const
    {
        CameraSample cs;
        cs.lensSample = lensSample;
        
        // sample point on lens
        Point p_lens = cs.lensSample == vec2(0,0) ? Point(0,0,0) : lens.sample(cs.lensSample);
        
        // lens point in world space
        Point p_lens_world = c2w(p_lens);
        
        // init connection ray
        cs.ray = Ray(p_lens_world, normalize(pWorld - p_lens_world), time);

        // check if ray points in the same direction as the camera axis (0,0,1) (+z)
        // return invalid sample if not
        Vector w_lens = w2c(cs.ray.d());
        if(w_lens.z >= 0.f) 
            return cs;
        
        // find point of focus
        float t_focus = std::abs(z_focus / w_lens.z);
        Point p_focus = p_lens + t_focus * w_lens;

        // find pixel position 
        Point p_raster = c2r(p_focus);
        cs.filmPosition = vec2(p_raster.x, p_raster.y);

        // return invalid sample for out of bounds points
        if( !valid(cs.filmPosition) ) 
            return cs;
        
        // sample is valid -> weight of 1
        cs.weight = White();

        // compute pdf
        //
        //  p(p_lens, w_lens) = p(w_lens) * p(p_lens)
        //
        //  p(w_lens) = d^2(p_film, p_lens) / (A_pixel * cos_theta_film)
        //            = z_film * z_film / (A_pixel * cos_theta_film * cos_theta_film * cos_theta_film)
        //            = z_film * z_film / (A_film * cos_theta_film * cos_theta_film * cos_theta_film) * (width * height)
        //
        //      !!! theta_film is the angle between the camera normal and the direction from film to lens
        //      !!! A_pixel is not directly stored to keep the floating point precision alive
        //  
        //  p(p_lens) = 1 / A_lens
        //  
        //      !!! note that theta_film and theta_lens are equals if and only if z_film = z_focus
        //

        // direction between camera origin, film point and point of focus 
        Vector w_focus = normalize(Vector(p_focus));

        // find true film point in camera space
        float t_film = std::abs(z_film / w_focus.z); 
        Point p_film(t_film * w_focus);

        // lens to film direction in camera space
        Vector w_film = normalize(p_lens - p_film);
        
        // compute probability densities
        float pdf_p_lens = lens.eval();
        float pdf_w_lens = z_film * z_film / (film_area * w_film.z * w_film.z * w_film.z ) * (width * height);
        float pdf_time = close-open > 0 ? 1.f/std::abs(close-open) : 1;

        // return pdf of ray
        cs.pdf = std::abs(pdf_w_lens * pdf_p_lens * pdf_time);
        if( std::isnan(cs.pdf) )
            printf("Camera::sample pdf %f %f %f %f\n", pdf_p_lens, pdf_w_lens, film_area, w_film.z);  

        return cs;
    }

    // void ray_differentials( Ray& ray )
    // {
    //     // check if ray point out of the camera
    //     Point p_lens = w2c(ray.o());
    //     Vector w_lens = w2c(ray.d());
        
    //     // find point of focus
    //     float t_focus = std::abs(z_focus / w_lens.z);
    //     Point p_focus = p_lens + t_focus * w_lens;

    //     // direction between camera origin, film point and point of focus 
    //     Vector w_focus = normalize(Vector(p_focus));

    //     // find true film point in camera space
    //     float t_film = std::abs(z_film / w_focus.z); 
    //     Point p_film(t_film * w_focus);

    //     // init differentials
    //     Vector drx = normalize(Vector(p_film + dx_camera));
    //     Vector dry = normalize(Vector(p_film + dy_camera));

    //     if( lens.radius() )
    //     {
    //         // update differentials for effect of lens
    //         float tx = std::abs(z_focus / drx.z);
    //         Point px_focus(tx * drx);
    //         drx = normalize(px_focus - p_lens);

    //         float ty = std::abs(z_focus / dry.z);
    //         Point py_focus(ty * dry);
    //         dry = normalize(py_focus - p_lens);
    //     }

    //     // update differentials
    //     ray.orx = ray.ory = ray.o();
    //     ray.drx = c2w(drx);
    //     ray.dry = c2w(dry);
    // }

    void save( const std::string& filename )
    {   
        Transform inv_view = tr_view.inverse();

        Point from(inv_view.m[0][3], inv_view.m[1][3], inv_view.m[2][3]);
        Point to = from - Vector(inv_view.m[0][2], inv_view.m[1][2], inv_view.m[2][2]);
        Vector up(inv_view.m[0][1], inv_view.m[1][1], inv_view.m[2][1]);

        std::ofstream out(filename, std::ios::out);
        if( out )
        {   
            std::cout << "writing camera " << filename << std::endl;

            out << "# camera description" << std::endl;
            out << "# required" << std::endl;
            out << "res " << width << " " << height << std::endl;
            out << "fov " << fov << std::endl;
            out << "from " << from.x << " " << from.y << " " << from.z << std::endl;
            out << "to " << to.x << " " << to.y << " " << to.z << std::endl;
            out << "up " << up.x << " " << up.y << " " << up.z << std::endl;
            out << "# optional" << std::endl;
            out << "z_far " << z_far << std::endl;
            out << "z_near " << z_near << std::endl;
            out << "z_film " << z_film << std::endl;
            out << "z_focus " << z_focus << std::endl;
            out << "radius " << lens.radius() << std::endl;
            out << "shutter " << open << " " << close << std::endl;
        }
        else
        {
            std::cerr << "error writing camera " << filename << std::endl;
        }
    }

    void read( const std::string& filename )
    {
        Point from, to;
        Vector up;
        std::ifstream in(filename, std::ios::in);
        if( in )
        {
            std::cout << "reading camera " << filename << std::endl;
            
            // parse file
            std::string line;
            while(std::getline(in, line))
            {
                // parse line 
                std::istringstream iss(line);
                std::string header; 
                iss >> header;
                if( header == "res" )
                    iss >> width >> height;                        
                else if( header == "fov" )
                    iss >> fov;                                   
                else if( header == "from" )
                    iss >> from.x >> from.y >> from.z;                        
                else if( header == "to" )
                    iss >> to.x >> to.y >> to.z;                        
                else if( header == "up" )
                    iss >> up.x >> up.y >> up.z;   
                else if( header == "z_far" )
                    iss >> z_far;                  
                else if( header == "z_near" )
                    iss >> z_near;                  
                else if( header == "z_film" )
                    iss >> z_film;                  
                else if( header == "z_focus" )
                    iss >> z_focus;                  
                else if( header == "radius" )
                {
                    float radius;
                    iss >> radius;
                    lens = Lens(radius);                    
                }
                else if( header == "shutter" )
                    iss >> open >> close;
            } 
            film = Film(width, height); 
            tr_view = Lookat(from, to, up);
            tr_proj = Perspective(fov, float(width) / float(height), z_near, z_far);
            tr_film = Viewport(width, height); 
        }
        else
        {
            std::cerr << "error reading camera " << filename << std::endl;
            exit(1);
        }
    }
};

#endif
