#ifndef DEEPIMAGE_H
#define DEEPIMAGE_H

#include <vector>
#include <iostream>

// From Tamy Boubekeur's BCD

// ------------------------ PIXEL POSITION ------------------------

class PixelPosition
{
public:
	PixelPosition(){};
	PixelPosition(int row, int col) : row(row), col(col) {};
	PixelPosition(const PixelPosition& pos) = default;
	PixelPosition& operator=(const PixelPosition& pos) = default;

	inline void get(int& o_row, int& o_col) const 						{ o_row = row; o_col = col; }
	inline PixelPosition operator+(const PixelPosition& pos) const 	{ return PixelPosition(row + pos.row, col + pos.col); }
	inline PixelPosition operator-(const PixelPosition& pos) const 	{ return PixelPosition(row - pos.row, col - pos.col); }
	inline PixelPosition operator+(const int radius) const 			{ return PixelPosition(row + radius, col + radius); }
	inline PixelPosition operator-(const int radius) const 			{ return PixelPosition(row - radius, col - radius); }
	inline bool operator==(const PixelPosition& pos) const 			{ return (row == pos.row) && (col == pos.col); }
	inline bool operator!=(const PixelPosition& pos) const 			{ return (row != pos.row) || (col != pos.col); }

	inline PixelPosition min_corner(const int radius, const int border) const 
	{ 
		return PixelPosition(std::max(border, row - radius), std::max(border, col - radius)); 
	}

	inline PixelPosition max_corner(const int radius, const int border, const int width, const int height) const 
	{ 
		return PixelPosition(std::min(height - 1 - border, row + radius),std::min(width - 1 - border, col + radius)); 
	}

	int row = 0;
	int col = 0;
};

typedef PixelPosition PixelVector;
typedef PixelPosition PixelWindowSize;
typedef PixelPosition PatchSize;
typedef PixelPosition ImageWindowSize;

// ------------------------ PIXEL WINDOW ------------------------

class PixelWindow
{
public:
	class iterator
	{
	public:
		inline iterator() : min_corner(), max_corner(), current(){}
		inline iterator(PixelPosition center, int radius) : min_corner(center - radius), max_corner(center + radius), current(min_corner){}
		inline iterator(int width, int height, PixelPosition center, int radius, int border){reset(width, height, center, radius, border);}
		inline iterator( PixelPosition min_corner, PixelPosition max_corner, PixelPosition current) : min_corner(min_corner), max_corner(max_corner), current(current){}

		inline void reset(PixelPosition center, int radius) { min_corner = center - radius; max_corner = center + radius; current = min_corner; }

		inline void reset(int width, int height, PixelPosition center, int radius, int border)
		{
			min_corner = center.min_corner(radius, border);
			max_corner = center.max_corner(radius, border, width, height);
			current = min_corner;
		}

		inline void reset() {current = min_corner;}

		inline PixelWindowSize size() const {return PixelWindowSize(1, 1) + max_corner - min_corner; }
		inline const PixelPosition& operator*() const { return current; }
		inline const PixelPosition* operator->() const { return &current; }
		inline bool operator!=(const iterator& it) const { return current != it.current; }
		inline iterator& operator++()
		{
			if(current.col == max_corner.col)
			{
				current.row++;
				current.col = min_corner.col;
			}
			else
			{
				current.col++;
			}
			return *this;
		}

		inline bool has_ended() const { return current.row > max_corner.row; }

	private:
		PixelPosition min_corner;
		PixelPosition max_corner;
		PixelPosition current;
	};

public:
	inline PixelWindow() : w(0), h(0), min_corner(), max_corner() {}
	inline PixelWindow(int width, int height, PixelPosition center, int radius, int border) 
	{
		reset(width, height, center, radius, border);
	}

	inline PixelWindow(int width, int height, PixelPosition min_corner, PixelPosition max_corner ) : 
		w(width), h(height), min_corner(min_corner), max_corner(max_corner) {}

	inline void reset(int width, int height, PixelPosition center, int radius, int border)
	{
		w = width;
		h = height;
		min_corner = center.min_corner(radius, border);
		max_corner = center.max_corner(radius, border, w, h);
	}

	inline PixelWindowSize size() const { return PixelWindowSize(1, 1) + max_corner - min_corner; }
	inline iterator begin() const { return iterator(min_corner, max_corner, min_corner); }
	inline iterator end() const { return iterator(min_corner, max_corner, PixelPosition(max_corner.row + 1, min_corner.col));}
	inline const PixelPosition& min() const {return min_corner;} 
	inline const PixelPosition& max() const {return max_corner;} 

private:
	int w = 0, h = 0;
	PixelPosition min_corner;
	PixelPosition max_corner;
};

typedef PixelWindow 			PixelPatch;
typedef PixelWindow::iterator PixWinIt;
typedef PixelPatch::iterator 	PixPatchIt;

// ------------------------ DEEP IMAGE ------------------------

template <typename T = float>
class DeepImage
{
public:

	// ------------------------ CONSTRUCTORS ------------------------

	DeepImage(){}; 
	DeepImage(int w, int h, int d) : w(w), h(h), d(d), wd(w*d), data(w*h*d){};
	DeepImage(int w, int h, int d, T value) : w(w), h(h), d(d), wd(w*d), data(w*h*d, value){};
	DeepImage(const DeepImage<T>&) = default;
	DeepImage& operator=(const DeepImage<T>&) = default;
	~DeepImage() = default;

	// ------------------------ TYPEDEF ------------------------

	typedef DeepImage<T> Self; 

	// ------------------------ ITERATOR ------------------------
	
	class iterator
	{
	public:
		inline iterator(){}
		inline iterator(T* data_ptr, int data_size) : data_ptr(data_ptr), data_size(data_size){}
		inline iterator(const iterator&) = default;

		inline T* operator*() const 						{ return data_ptr; }
		inline iterator& operator++()						{ data_ptr += data_size;return *this; }
		inline T& operator[](int depth) const 				{ return data_ptr[depth]; }
		inline bool operator!=(const iterator& it) const 	{ return data_ptr != it.data_ptr; }

	private:
		T* data_ptr = nullptr;
		int data_size = 0;
	};

	// ------------------------ CONST ITERATOR ------------------------
	
	class const_iterator
	{
	public:
		inline const_iterator(){}
		inline const_iterator(T* data_ptr, int data_size) : data_ptr(data_ptr), data_size(data_size){}
		inline const_iterator(const const_iterator&) = default;

		inline T* operator*() const 							{ return data_ptr; }
		inline const_iterator& operator++()						{ data_ptr += data_size;return *this; }
		inline T& operator[](int depth) const 					{ return data_ptr[depth]; }
		inline bool operator!=(const const_iterator& it) const 	{ return data_ptr != it.data_ptr; }

	private:
		const T* data_ptr = nullptr;
		int data_size = 0;
	};

	// ------------------------ DATA ------------------------

	inline void resize(int width, int height, int depth)
	{
		w = width;
		h = height;
		d = depth;
		wd = width * depth;
		data.resize(width * height * depth);
	}

	inline void copy_data_from(const T* data_ptr)	{ std::copy(data_ptr, data_ptr + get_size(), data.begin()); }
	inline void copy_data_to(T* data_ptr) const 	{ std::copy(data.begin(), data.end(), data_ptr); }
	inline T* get_data_ptr() 						{ return data.data(); }
	inline const T* get_data_ptr() const 			{ return data.data(); }

	// ------------------------ CONST ------------------------

	inline int get_width() const 	{ return w; }
	inline int get_height() const 	{ return h; }
	inline int get_depth() const 	{ return d; }
	inline int get_size() const 	{ return data.size(); }

	// ------------------------ OFFSET ------------------------

	inline int offset( int row, int col, int depth) const						{ return (row * w + col) * d + depth; }
	inline void split_offset( int& row, int& col, int& depth, int index) const 	{ row = index / wd; col = (index / d) % w; depth = index % d; }

	// ------------------------ GETTER ------------------------

	inline const T& get(int row, int col, int depth) const 			{ return data[offset(row, col, depth)]; }
	inline T& get(int row, int col, int depth) 						{ return data[offset(row, col, depth)]; }
	inline const T& get(PixelPosition pos, int depth) const 		{ return data[offset(pos.row, pos.col, depth)]; }
	inline T& get(PixelPosition pos, int depth) 					{ return data[offset(pos.row, pos.col, depth)]; }
	inline const T& get(int index) const 							{ return data[index]; }
	inline T& get(int index)										{ return data[index]; }
	inline const T get_value(PixelPosition pos, int depth) const 	{ return data[offset(pos.row, pos.col, depth)]; }
	inline T get_value(PixelPosition pos, int depth)				{ return data[offset(pos.row, pos.col, depth)]; }
	
	// ------------------------ SETTER ------------------------

	inline void set(int row, int col, int depth, T value) 		{ data[offset(row, col, depth)] = value; }
	inline void set(PixelPosition pos, int depth, T value) 	{ data[offset(pos.row, pos.col, depth)] = value; }
	inline void set(int index, T value) 						{ data[index] = value; }
	inline void set(int row, int col, const T* data_ptr) 		{ std::copy(data_ptr, data_ptr + d, &(get(row, col, 0))); }
	inline void set(PixelPosition pos, const T* data_ptr) 	{ std::copy(data_ptr, data_ptr + d, &(get(pos))); }

	// ------------------------ SCALE ------------------------

	inline void isotropicalScale(T scale_factor) 
	{
		for(auto it = data.begin(); it != data.end(); it++)
			*it *= scale_factor; 
	}

	inline void anisotropicalScale(const T* scale_factor) 
	{
		for(int d = 0; d < get_size(); d++)
			data[d] *= scale_factor[d % d]; 
	}

	// ------------------------ MEMORY ------------------------

	inline void fill(T f) { std::fill(data.begin(), data.end(), f); }
	inline bool is_empty() const { return w == 0 || h == 0 || d == 0; }
	inline void release() { w = h = d = wd = 0; std::vector< T >().swap(data); }

	// ------------------------ DATA ACCESS ------------------------

	inline iterator begin() { return iterator(data.data(), d); }
	inline iterator end() { return iterator(data.data() + wd * h, d); }
	inline const_iterator begin() const { return const_iterator(data.data(), d); }
	inline const_iterator end() const { return const_iterator(data.data() + wd * h, d); }

	// ------------------------ DEEP IMAGE OPERATORS ------------------------

	inline Self& operator+=(const Self& image)
	{
		typename std::vector<T>::const_iterator it = image.data.cbegin();
		for(T& value : data)
			value += *it++;
		return *this;
	}

	inline Self& operator-=(const Self& image)
	{
		typename std::vector<T>::const_iterator it = image.data.cbegin();
		for(T& value : data)
			value -= *it++;
		return *this;
	}

	inline Self& operator/=(const Self& image)
	{
		if( image.get_depth() == 0 )
		{
			for(int i = 0; i < w; i++)
			for(int j = 0; j < h; j++)
			{	
				float w = image.get(i,j,0);
				if( w > 0 )
					for(int k = 0; k < d; k++)
						data[offset(i,j,k)] /= w;
			}
		}
		else if( image.get_depth() == d )
		{
			typename std::vector<T>::const_iterator it = image.data.cbegin();
			for(T& value : data)
			{
				if(*it) value /= *it;
				it++;
			}
		}

		return *this;
	}

	inline Self& operator*=(const Self& image)
	{
		if( image.get_depth() == 0 )
		{
			for(int i = 0; i < w; i++)
			for(int j = 0; j < h; j++)
			{	
				float w = image.get(i,j,0);
				if( w > 0 )
					for(int k = 0; k < d; k++)
						data[offset(i,j,k)] *= w;
			}
		}
		else if( image.get_depth() == d )
		{
			typename std::vector<T>::const_iterator it = image.data.cbegin();
			for(T& value : data)
			{
				if(*it) value *= *it;
				it++;
			}
		}

		return *this;
	}

	inline Self operator/( const Self& image ){Self out(*this); out /= image;return out;}
	inline Self operator+( const Self& image ){Self out(*this); out += image;return out;}
	inline Self operator-( const Self& image ){Self out(*this); out -= image;return out;}

private:
	int w = 0, h = 0, d = 0, wd = 0;
	std::vector<T> data;	
};

// ------------------------ DEEP IMAGE TYPEDEFS ------------------------

typedef DeepImage<float> 			DeepImf;
typedef DeepImf::iterator 		DeepImfIt;
typedef DeepImf::const_iterator 	DeepImfConstIt;

// ------------------------ IMAGE WINDOW ------------------------

template <typename T = float>
class ImageWindow
{
public:

	// ------------------------ TYPEDEFS ------------------------

	typedef DeepImage<T> 		DeepImT;
	typedef ImageWindow<T> 	Self;

	// ------------------------ ITERATORS ------------------------

	class iterator
	{
	public:
		inline iterator() : min_corner(), max_corner(), current(), data_ptr(nullptr) {}
		inline iterator(DeepImT& image, PixelPosition center, int radius, int border) { reset(image, center, radius, border); }
		inline iterator(int w, int h, int d, PixelPosition min_corner, PixelPosition max_corner, PixelPosition current, T* data_ptr) :
				w(w), h(h), d(d), min_corner(min_corner), max_corner(max_corner), current(current), data_ptr(data_ptr){}

		inline void reset(DeepImT& image, PixelPosition center, int radius, int border)
		{
			w = image.get_width();
			h = image.get_height();
			d = image.get_depth();
			min_corner = center.min_corner(radius, border);
			max_corner = center.max_corner(radius, border, w, h);
			current = min_corner;
			data_ptr = &(image.get(current, 0));
		}

		inline ImageWindowSize get_size() const { return ImageWindowSize(1, 1) + max_corner - min_corner; }
		inline T* operator*() const { return data_ptr; }
		inline T& operator[](int index) const { return data_ptr[index]; }
		inline bool operator!=(const iterator& it) const { return data_ptr != it.data_ptr; }

		inline iterator& operator++()
		{
			if(current.col == max_corner.col)
			{
				current.row++;
				current.col = min_corner.col;
				data_ptr += d * (w + min_corner.col - max_corner.col);
			}
			else
			{
				current.col++;
				data_ptr += d;
			}
			return *this;
		}

		inline bool has_ended() const { return current.row > max_corner.row; }
		inline PixelPosition position() const { return current; }

	private:
		int w;
		int h;
		int d;
		PixelPosition min_corner;
		PixelPosition max_corner;
		PixelPosition current;
		T* data_ptr;
	};


public:
	inline ImageWindow() : min_corner(), max_corner(), min_corner_data_ptr(nullptr) {}
	inline ImageWindow(DeepImT& image, PixelPosition center, int radius, int border) { reset(image, center, radius, border); }

	inline void reset(DeepImT& image, PixelPosition center, int radius, int border)
	{
		w = image.get_width();
		h = image.get_height();
		d = image.get_depth();
		min_corner = center.min_corner(radius, border);
		max_corner = center.max_corner(radius, border, w, h);
		min_corner_data_ptr = &(image.get(min_corner, 0));
	}

	inline ImageWindowSize get_size() const { return ImageWindowSize(1, 1) + max_corner - min_corner; }

	inline iterator begin() const 	{ return iterator(w, h, d, min_corner, max_corner, min_corner, min_corner_data_ptr); }
	inline iterator end() const 	
	{ 
		return iterator( 	w, h, d, 
							min_corner, 
							max_corner, 
							PixelPosition(max_corner.row + 1, min_corner.col), 
							min_corner_data_ptr + (1 + max_corner.row - min_corner.row) * d * w); 
	}

private:
	int w = 0, h = 0, d= 0;
	PixelPosition min_corner;
	PixelPosition max_corner;
	T* min_corner_data_ptr;
};

typedef ImageWindow<float> 	Windowf;
typedef Windowf::iterator 	WindowfIt;
typedef Windowf 				Tilef;
typedef WindowfIt 			TilefIt;

// ------------------------ CONST IMAGE WINDOW ------------------------

template <typename T = float>
class ConstImageWindow
{
public:

	// ------------------------ TYPEDEFS ------------------------

	typedef DeepImage<T> 			DeepImT;
	typedef ConstImageWindow<T> 	Self;

	// ------------------------ ITERATORS ------------------------

	class iterator
	{
	public:
		inline iterator() : min_corner(), max_corner(), current(), data_ptr(nullptr) {}
		inline iterator(const DeepImT& image, PixelPosition center, int radius, int border) { reset(image, center, radius, border); }
		inline iterator(int w, int h, int d, PixelPosition min_corner, PixelPosition max_corner, PixelPosition current, const T* data_ptr) :
				w(w), h(h), d(d), min_corner(min_corner), max_corner(max_corner), current(current), data_ptr(data_ptr){}

		inline void reset(const DeepImT& image, PixelPosition center, int radius, int border)
		{
			w = image.get_width();
			h = image.get_height();
			d = image.get_depth();
			min_corner = center.min_corner(radius, border);
			max_corner = center.max_corner(radius, border, w, h);
			current = min_corner;
			data_ptr = &(image.get(current, 0));
		}

		inline ImageWindowSize get_size() const { return ImageWindowSize(1, 1) + max_corner - min_corner; }
		inline const T* operator*() const { return data_ptr; }
		inline const T& operator[](int index) const { return data_ptr[index]; }
		inline bool operator!=(const iterator& it) const { return data_ptr != it.data_ptr; }

		inline iterator& operator++()
		{
			if(current.col == max_corner.col)
			{
				current.row++;
				current.col = min_corner.col;
				data_ptr += d * (w + min_corner.col - max_corner.col);
			}
			else
			{
				current.col++;
				data_ptr += d;
			}
			return *this;
		}

		inline PixelPosition position() const { return current; }

		inline bool has_ended() const { return current.row > max_corner.row; }

	private:
		int w;
		int h;
		int d;
		PixelPosition min_corner;
		PixelPosition max_corner;
		PixelPosition current;
		const T* data_ptr;
	};


public:
	inline ConstImageWindow() : min_corner(), max_corner(), min_corner_data_ptr(nullptr) {}
	inline ConstImageWindow(const DeepImT& image, PixelPosition center, int radius, int border) { reset(image, center, radius, border); }

	inline void reset(const DeepImT& image, PixelPosition center, int radius, int border)
	{
		w = image.get_width();
		h = image.get_height();
		d = image.get_depth();
		min_corner = center.min_corner(radius, border);
		max_corner = center.max_corner(radius, border, w, h);
		min_corner_data_ptr = &(image.get(min_corner, 0));
	}

	inline ImageWindowSize get_size() const { return ImageWindowSize(1, 1) + max_corner - min_corner; }

	inline iterator begin() const 	{ return iterator(w, h, d, min_corner, max_corner, min_corner, min_corner_data_ptr); }
	inline iterator end() const 	
	{ 
		return iterator( 	w, h, d, 
							min_corner, 
							max_corner, 
							PixelPosition(max_corner.row + 1, min_corner.col), 
							min_corner_data_ptr + (1 + max_corner.row - min_corner.row) * d * w); 
	}

private:
	int w = 0, h = 0, d= 0;
	PixelPosition min_corner;
	PixelPosition max_corner;
	const T* min_corner_data_ptr;
};

typedef ConstImageWindow<float> 	ConstWindowf;
typedef ConstWindowf::iterator 	ConstWindowfIt;
typedef ConstWindowf 				ConstTilef;
typedef ConstWindowfIt 			ConstTilefIt;

#endif