#ifndef FILM_H
#define FILM_H

#include <sstream>

#include "src/gKit/image.h"
#include "src/gKit/image_io.h"
#include "src/gKit/image_hdr.h"

#include "src/gKit/orbiter.h"

#include "renderer/deepimage.h"
#include "renderer/sampler.h"

//#define USE_OMP_LOCKS

class Film
{
public:
    int w, h;
    float invw, invh;

    DeepImf samples_mean;
    DeepImf sq_samples_mean;
    DeepImf sample_map;
    DeepImf image;
    DeepImf error;
    DeepImf variance;
    DeepImf sq_image;
    DeepImf reconstruction;
    
    std::vector<DeepImf> layers;

#ifdef USE_OMP_LOCKS
    std::vector<omp_lock_t> locks;
#endif


    float   mean_error, 
            mean_variance, 
            mean_error_variations,
            mean_weights, 
            mean_samples;
    
    bool normalized = true;
    
    Film(){}
    Film( const int w, const int h ) : 
        w(w), h(h), invw(1.f / w), invh(1.f / h),
        // samples_mean(w, h, 4, 0.f), 
        // sq_samples_mean(w, h, 4, 0.f), 
        // sample_map(w, h, 4, 0.f),
        image(w, h, 4, 0.f),
        error(w, h, 4, 0.f), 
        variance(w, h, 4, 0.f), 
        sq_image(w, h, 4, 0.f),
        reconstruction(w, h, 4, 0.f)
#ifdef USE_OMP_LOCKS
        ,locks(w*h)
#endif
        {
#ifdef USE_OMP_LOCKS
            for(int i = 0; i < w*h; i++)
                omp_init_lock(&(locks[i]));
#endif
        }
#ifdef USE_OMP_LOCKS
    ~Film()
        {
            for(int i = 0; i < w*h; i++)
                omp_destroy_lock(&(locks[i]));
        }
#endif

    int get_width()     const { return w; }
    int get_height()    const { return h; }
    int get_size()      const { return samples_mean.get_size(); }

    void init_layers( int nb_cameras )
    {
        layers = std::vector<DeepImf>(nb_cameras, DeepImf(w,h,4,0.f));
    }

    // for adaptive rendering purpose active the sample map layer
    void init_sample_map() 
    {
        sample_map = DeepImf(w,h,1,0.f);
    }

    void unset_normalized() { normalized = false; }
    void set_normalized() { normalized = true; }

    void reset( const vec2& pixel )
    {
        PixelPosition pos = offset(pixel);
        for(int k = 0; k < samples_mean.get_depth(); ++k)
        {
            samples_mean.set(pos, k, 0);
            sq_samples_mean.set(pos, k, 0);
            error.set(pos, k, 0);
            variance.set(pos, k, 0);
            sample_map.set(pos, k, 0);
        }
    }

    inline PixelPosition offset( const vec2& pixel ) const 
    {
        return  PixelPosition(
                    std::min(h - 1, std::max(0, static_cast<int>(pixel.y+0.5f))), 
                    std::min(w - 1, std::max(0, static_cast<int>(pixel.x+0.5f)))
                );
    }

    // filtre Bspline 3
    inline float b0( const float x ) const
    {
        float x2= x * x;
        float x3= x2 * x;
        return 1.f/6.f * x3;
    }
    
    inline float b1( const float x ) const
    {
        float x2= x * x;
        float x3= x2 * x;
        return 1.f/6.f * (-3.f*x3 + 3.f*x2 + 3.f*x + 1.f);
    }
    
    inline float b3( const float x ) const
    {
        if(x < -2)          return 0;
        else if(x <= -1)    return b0(x + 2);
        else if(x <= 0)     return b1(x + 1);
        else if(x <= 1)     return b1(1 - x);
        else if(x <= 2)     return b0(2 - x);
        else                return 0;
    }

    inline float gaussian1D(float x, float x0, float sigma) const 
    {
        return sigma > 0.f ? std::max(0.f, std::exp(-0.5f * (x - x0) * (x - x0) / (sigma * sigma))) : 0.f;
    }

    inline float isotropic_gaussian2D(const float x, const float y, const float radius, const float alpha) const 
    {   
        const float expv = std::exp(-alpha * radius * radius);
        const float expx = std::exp(-alpha * x * x);
        const float expy = std::exp(-alpha * y * y);
        return std::max(0.f, (expx - expv) * (expy - expv));
    }

    inline float anisotropic_gaussian2D(const float x, const float y, const vec2 radius) const 
    {   
        const float alpha = 0.5;
        const float expvx = std::exp(-alpha * radius.x * radius.x);
        const float expvy = std::exp(-alpha * radius.y * radius.y);
        const float expx = std::exp(-alpha * x * x);
        const float expy = std::exp(-alpha * y * y);
        return std::max(0.f, (expx - expvx) * (expy - expvy));
    }

    // t in range [-1;1] 
    inline float triangle1D( const float x ) const
    {
        return 1.f - std::max(x, 0.f); 
    }

    void add( const vec2& pixel, const Color& color, float weight = 1)
    {
        // check color
        if( !color.valid() || weight == 0 ) return;
        const float sample[3] = { color.r, color.g, color.b };
        const float sq_sample[3] = { color.r * color.r, color.g * color.g, color.b * color.b };

        // current pixel
        PixelPosition pos = offset(pixel);

#ifdef USE_OMP_LOCKS
        const unsigned int pixelId = pos.row*image.get_width() + pos.col;
        omp_set_lock(&(locks[pixelId]));
#endif

        image.get(pos, 3) += weight;
        sq_image.get(pos, 3) += weight;
        variance.get(pos, 3) += 1;
        // error.get(pos, 3) = 1;

        // update color
        if(normalized)
            for(int channel = 0; channel < 3; ++channel)
            {
                image.get(pos, channel) += (sample[channel] - image.get(pos, channel)) * weight / image.get(pos, 3);
                sq_image.get(pos, channel) += (sq_sample[channel] - sq_image.get(pos, channel)) * weight / image.get(pos, 3);
                variance.get(pos, channel) = sq_image.get(pos, channel) - image.get(pos, channel) * image.get(pos, channel);
                // error.get(pos, channel) = srcs(image.get(pos, channel), sq_image.get(pos, channel), image.get(pos, 3));
                error.get(pos, channel) = srcs(image.get(pos, channel), sq_image.get(pos, channel), variance.get(pos, 3));
            }
        else
            for(int channel = 0; channel < 3; ++channel)
            {
                image.get(pos, channel) += sample[channel];
                sq_image.get(pos, channel) += sq_sample[channel];
                variance.get(pos, channel) = sq_image.get(pos, channel) - image.get(pos, channel) * image.get(pos, channel);
                // error.get(pos, channel) = srcs(image.get(pos, channel), sq_image.get(pos, channel), image.get(pos, 3));
                error.get(pos, channel) = srcs(image.get(pos, channel), sq_image.get(pos, channel), variance.get(pos, 3));
            }

#ifdef USE_OMP_LOCKS
        omp_unset_lock(&(locks[pixelId]));
#endif
    }

    void add_reconstruction( const vec2& pixel, const Color& color, float weight = 1)
    {
        // check color
        if( !color.valid() || weight == 0 ) return;
        const float sample[3] = { color.r, color.g, color.b };
        const float sq_sample[3] = { color.r * color.r, color.g * color.g, color.b * color.b };

        // current pixel
        PixelPosition pos = offset(pixel);

#ifdef USE_OMP_LOCKS
        const unsigned int pixelId = pos.row*image.get_width() + pos.col;
        omp_set_lock(&(locks[pixelId]));
#endif

        reconstruction.get(pos, 3) += weight;

        // update color
        for(int channel = 0; channel < 3; ++channel)
        {
            reconstruction.get(pos, channel) += (sample[channel] - reconstruction.get(pos, channel)) * weight / reconstruction.get(pos, 3);
        }

#ifdef USE_OMP_LOCKS
        omp_unset_lock(&(locks[pixelId]));
#endif
    }

    void add( const int layer, const vec2& pixel, const Color& color, float weight = 1)
    {
        if( layers.size() == 0 )
            return;

        // check color
        if( !color.valid() || weight == 0 ) return;
        const float sample[3] = { color.r, color.g, color.b };

        // current pixel
        PixelPosition pos = offset(pixel), min_corner, max_corner;

#ifdef USE_OMP_LOCKS
        const unsigned int pixelId = pos.row*image.get_width() + pos.col;
        omp_set_lock(&(locks[pixelId]));
#endif

        layers[layer].get(pos, 3) += weight;

        // update color
        for(int channel = 0; channel < 3; ++channel)
        {
            layers[layer].get(pos, channel) += (sample[channel] - layers[layer].get(pos, channel)) * weight / layers[layer].get(pos, 3);
        }

#ifdef USE_OMP_LOCKS
        omp_unset_lock(&(locks[pixelId]));
#endif
    }

    void splat( const vec2& pixel, const Color& color, float weight = 1)
    {
        // check color
        if( !color.valid() || weight == 0 ) return;
        const float sample[3] = { color.r, color.g, color.b };
        const float sq_sample[3] = { color.r * color.r, color.g * color.g, color.b * color.b };

        // current pixel
        PixelPosition nearest = offset(pixel), min_corner, max_corner;

        // 4x4 window for bicubic spline filter
        if( pixel.y - nearest.row > 0 )
        {
            min_corner.row = std::max(0, nearest.row - 1);
            max_corner.row = std::min(h - 1, nearest.row + 2);
        }
        else
        {
            min_corner.row = std::max(0, nearest.row - 2);
            max_corner.row = std::min(h - 1, nearest.row + 1);
        }

        if( pixel.x - nearest.col > 0 )
        {
            min_corner.col = std::max(0, nearest.col - 1);
            max_corner.col = std::min(w - 1, nearest.col + 2);
        }
        else
        {
            min_corner.col = std::max(0, nearest.col - 2);
            max_corner.col = std::min(w - 1, nearest.col + 1);
        }

        PixelWindow window(w, h, min_corner, max_corner);

        for( auto it = window.begin(); it != window.end(); ++it )
        {
            // filter value
            float s = pixel.x - it->col;
            float t = pixel.y - it->row;
            float ws = 4 * b3(s) * b3(t); 
            float wp = ws * weight;
            
            // printf("%f ", ws);

            if( wp == 0 ) continue;

            image.get(*it, 3) += wp;
            sq_image.get(*it, 3) += wp;
            variance.get(*it, 3) += ws;
            error.get(*it, 3) = ws;

            // update color
            for(int channel = 0; channel < 3; ++channel)
            {
                image.get(*it, channel) += (sample[channel] - image.get(*it, channel)) * wp / image.get(*it, 3);
                sq_image.get(*it, channel) += (sq_sample[channel] - sq_image.get(*it, channel)) * wp / image.get(*it, 3);
                variance.get(*it, channel) = sq_image.get(*it, channel) - image.get(*it, channel) * image.get(*it, channel);
                error.get(*it, channel) = srcs(image.get(*it, channel), sq_image.get(*it, channel), image.get(*it, 3));
            }
        }
        // printf("\n");
    }

    void splat_importance( const vec2& pixel, float weight, int initial, float pdf_initial_to_me, int current = -1 , int nb = -1)
    {
        // current pixel
        PixelPosition nearest = offset(pixel), min_corner, max_corner;

        // 4x4 window for bicubic spline filter
        if( pixel.y - nearest.row > 0 )
        {
            min_corner.row = std::max(0, nearest.row - 1);
            max_corner.row = std::min(h - 1, nearest.row + 2);
        }
        else
        {
            min_corner.row = std::max(0, nearest.row - 2);
            max_corner.row = std::min(h - 1, nearest.row + 1);
        }

        if( pixel.x - nearest.col > 0 )
        {
            min_corner.col = std::max(0, nearest.col - 1);
            max_corner.col = std::min(w - 1, nearest.col + 2);
        }
        else
        {
            min_corner.col = std::max(0, nearest.col - 2);
            max_corner.col = std::min(w - 1, nearest.col + 1);
        }

        PixelWindow window(w, h, min_corner, max_corner);

        for( auto it = window.begin(); it != window.end(); ++it )
        {
            // filter value
            float s = pixel.x - it->col;
            float t = pixel.y - it->row;
            float ws = 4 * b3(s) * b3(t); 
            float wp = ws * weight;

            if( wp == 0 ) continue;

            if( initial >= 0 )
            {
                layers[initial].get(*it, 3) += wp;
                layers[initial].get(*it, 0) += (pdf_initial_to_me - layers[initial].get(*it, 0)) * wp / layers[initial].get(*it, 3);
                layers[initial].get(*it, 1) += (pdf_initial_to_me * pdf_initial_to_me - layers[initial].get(*it, 1)) * wp / layers[initial].get(*it, 3);
                layers[initial].get(*it, 2) = std::sqrt(std::max(0.f, layers[initial].get(*it, 1) - layers[initial].get(*it, 0) * layers[initial].get(*it, 0))) / layers[initial].get(*it, 0);
                
                double sum = 0, sq_sum = 0, count = 0;

                for( int i = std::max(0, current-2); i <= std::min(nb-1, current+2); i++)
                {
                    double value = layers[i].get(*it, 0);
                    double w = layers[i].get(*it, 3);

                    if( w > 0 ) 
                    {
                        count += w;
                        sum += (value - sum) * w / count;
                        sq_sum += (value * value - sq_sum) * w / count;
                    }
                }
                double var = std::max(0.0, sq_sum - sum * sum);
                double dev = std::sqrt(var);

                int index = (current+3)%nb;
                layers[index].get(*it, 0) = sum > 0 ? dev / sum : 0;
                layers[index].get(*it, 1) = sum > 0 ? dev / sum : 0;
                layers[index].get(*it, 2) = sum > 0 ? dev / sum : 0;
                layers[index].get(*it, 3) = 1;
            }
        }
    }

    // square root of chi square distance between uniform distribution and another distribution
    // samples_mean     = mean 
    // sq_samples_mean  = mean of squares 
    // samples_count    = samples number
    inline float srcs( const float& samples_mean, const float& sq_samples_mean, const float& samples_count ) const
    {   
        return samples_mean > 0 && samples_count > 0 ? std::sqrt(samples_count * sq_samples_mean - samples_mean * samples_mean) / (samples_count * samples_count) : 0; 
    }   

    Color get_tonemapped( const float x, const float y, const DeepImf& image, bool normalize = false ) const 
    { 
        return get_tonemapped(vec2(x, y), image, normalize); 
    }

    Color get_tonemapped( const vec2 pos, const DeepImf& image, bool normalize = false ) const
    {
        // tone mapping
        const float rgby[3] = {0.3, 0.59, 0.11};
        const float saturation = 1.f;
        const float compression = 2.2f; 
        const float power = 1.f / compression;
        const float k1 = 1.f / std::pow(saturation, power); 

        // pixel position
        PixelPosition pixel = offset(pos);

        Color color;

        // #pragma omp atomic read
        color.r = image.get(pixel, 0);
        // #pragma omp atomic read
        color.g = image.get(pixel, 1);
        // #pragma omp atomic read
        color.b = image.get(pixel, 2);
        // #pragma omp atomic read
        color.a = image.get(pixel, 3);

        if( normalize ) color =  color.a > 0 ? color / color.a : Black(); 

        float y = color.r * rgby[0] + color.g * rgby[1] + color.b * rgby[2];
        float v = y > 0 ? k1 * std::pow(y, power) / y : 0;
        color = Color(v * color, 1);

        return color.valid() ? color : Black();
    }

    Color tonemapping( const Color& color, const float saturation = 1.f, const float compression = 2.2f ) const
    {
        const float rgby[3] = {0.3, 0.59, 0.11};
        const float power = 1.f / compression;
        const float k1 = 1.f / std::pow(saturation, power); 

        float y = color.r * rgby[0] + color.g * rgby[1] + color.b * rgby[2];
        float v = y > 0 ? k1 * std::pow(y, power) / y : 0;

        return Color(v * color, 1);
    }

    Color get_color( const float x, const float y, const DeepImf& image ) const 
    { 
        return get_color(vec2(x, y), image); 
    }
    
    Color get_color( const vec2 pos, const DeepImf& image ) const
    {
        PixelPosition pixel = offset(pos);
        return Color(image.get(pixel,0), image.get(pixel,1), image.get(pixel,2), image.get(pixel, 3));
    }

    float get_criterion( const float x, const float y ) const { return get_criterion(vec2(x, y)); }
    float get_criterion( const vec2 pos ) const 
    { 
        return sample_map.get( offset(pos), 0 );
    }
    
    // cf : https://research.nvidia.com/publication/2018-05_Combining-Analytic-Direct
    // compute a total variation noise estimate of the given pixel
    // Estimate desired radius from the second derivative 
    // of the signal itself in source0 relative to baseline,
    // which is the noisier image because it contains shadows
    float estimateNoise( Sampler& sampler, const float x, const float y ) const { return estimateNoise(sampler, vec2(x, y)); }
    float estimateNoise( Sampler& sampler, const vec2 pos ) const 
    {
        float u1 = sampler.sample();
        float u2 = sampler.sample(); 

        // base angle
        float angle = u1 * 2.f * M_PI;

        // radius
        const float max_radius = 10.f;
        float radius = u2 * max_radius;

        // noise estimator
        float result = 0;
        const int n = 4;
        const float step = M_PI / float(n);
        for(float t = 0; t < 4; ++t, angle += step) 
        {   
            vec2 pixel2(std::floor(pos.x + radius * std::cos(angle)), std::floor(pos.y + radius * std::sin(angle)));
            vec2 pixel1(std::floor(pos.x + (1.f-radius) * std::cos(angle)), std::floor(pos.y + (1.f-radius) * std::sin(angle)));
            Color e2 = tonemapping(get_color(pixel2, error));
            Color e1 = tonemapping(get_color(pixel1, error));

            float d2mag = 0;
            // The first two points are accounted for above
            for (int r = -radius + 2.f; r <= radius; ++r) 
            {
                vec2 pixel0(std::floor(pos.x + r * std::cos(angle)), std::floor(pos.y + r * std::sin(angle)));
                Color e0 = tonemapping(get_color(pixel0, error));
                
                // Second derivative
                Color d2 = e2 - e1 * 2.f + e0;
               
                d2mag += length(d2);

                // Shift weights in the window
                e2 = e1; e1 = e0;
            }

            // Scaled value by 1.5 *before* clamping to visualize going out of range
            // It is clamped again when applied.
            // float clamp = std::min(1.f, std::max(0.f, std::sqrt(d2mag * (1.f / max_radius)) * (1.f / 2.f)));
            float clamp = std::min(1.f, std::max(0.f, std::sqrt(d2mag * (1.f / max_radius)) * (1.f / 1.5f)));
            result = std::max(clamp, result);
        }

        return result;
    }

    void save( std::string filename, float scale = 1.f, bool tonemap = false ) const
    {
        Image out(w, h, Black()), out_hdr(w, h, Black());
        #pragma omp parallel for 
        for (int i = 0; i < w; ++i)
        for (int j = 0; j < h; ++j)
        {   
            out(i,j) = tonemap ? Color(get_tonemapped(i,j,samples_mean) * scale, 1.f) : Color(get_color(i,j,samples_mean) * scale, 1.f);
            out_hdr(i,j) = Color(get_color(i,j,samples_mean) * scale, 1.f);
        }   

        write_image_hdr(out_hdr, (filename+std::string(".hdr")).c_str());
        write_image(out, (filename+std::string(".png")).c_str());
    }

    void save_layers( std::string filename, float scale = 1.f, bool tonemap = false ) const
    {
        for (int k = 0; k < (int)layers.size(); ++k)
        {   
            Image out(w, h, Black()), out_hdr(w, h, Black());
            #pragma omp parallel for 
            for (int j = 0; j < h; ++j)
            for (int i = 0; i < w; ++i)
            {
                out(i,j) = tonemap ? Color(get_tonemapped(i,j,layers[k]) * scale, 1.f) : Color(get_color(i,j,layers[k]) * scale, 1.f);
                out_hdr(i,j) = Color(get_color(i,j,layers[k]), 1.f);
            }
            std::stringstream ss;
            ss << filename << "-layer-" << k;
            write_image_hdr(out_hdr, (ss.str()+std::string(".hdr")).c_str());
            write_image(out, (ss.str()+std::string(".png")).c_str());
        }   
    }

    void save2( std::string filename, double scale = 1., bool tonemap = false, bool samples = false , bool png = true ) const
    {
        Image out(w, h, Black()), out_hdr(w, h, Black()), im_weights(w, h, Black()), im_samples(w, h, Black());
        #pragma omp parallel for 
        for (int i = 0; i < w; ++i)
        for (int j = 0; j < h; ++j)
        {   
            // out(i,j) = tonemap ? Color(get_tonemapped(i,j,image) * scale, 1.f) : Color(get_color(i,j,image) * scale, 1.f);
            const Color& inColor = get_color(i,j,image);
            const Color outColor = Color(inColor.r * scale, inColor.g * scale, inColor.b * scale, 1.f);
            out(i,j) = tonemap ? tonemapping(outColor) : outColor;
            out_hdr(i,j) = outColor;

            //if(i == 640 && j == 640)
            //{
            //    std::cout << std::setprecision(20) << "save2 pixel color[0] before = " << get_color(i,j,image) << std::endl;
            //    std::cout << std::setprecision(20) << "save2 scale = " << scale << std::endl;
            //    std::cout << std::setprecision(20) << "save2 pixel color[0] after = " << out(i,j) << std::endl;
            //}

            if( samples ) im_weights(i, j) = Color(image.get(j, i, 3),image.get(j, i, 3),image.get(j, i, 3));
            if( samples ) im_samples(i, j) = Color(variance.get(j, i, 3),variance.get(j, i, 3),variance.get(j, i, 3));
        }   

        write_image_hdr(out_hdr, (filename+std::string(".hdr")).c_str());
        if(png)     write_image(out, (filename+std::string(".png")).c_str());
        if(samples) write_image_hdr(im_weights, (filename+std::string("-weights.hdr")).c_str());
        if(samples) write_image_hdr(im_samples, (filename+std::string("-samples.hdr")).c_str());
    }

    void save_reconstruction( std::string filename, float scale = 1.f, bool tonemap = false ) const
    {
        Image out(w, h, Black()), out_hdr(w, h, Black()), im_weights(w, h, Black()), im_samples(w, h, Black());
        #pragma omp parallel for 
        for (int i = 0; i < w; ++i)
        for (int j = 0; j < h; ++j)
        {   
            out(i,j) = tonemap ? Color(get_tonemapped(i,j,reconstruction) * scale, 1.f) : Color(get_color(i,j,reconstruction) * scale, 1.f);
            out_hdr(i,j) = Color(get_color(i,j,reconstruction) * scale, 1.f);
        }   
        write_image_hdr(out_hdr, (filename+std::string(".hdr")).c_str());
        write_image(out, (filename+std::string(".png")).c_str());
    }

    float merge( const float bin_weight )
    {
        if( bin_weight == 0 ) return 0;

        // merge
        #pragma omp parallel for 
        for (int x = 0; x < w; ++x)
        for (int y = 0; y < h; ++y)
        { 
            float wn = samples_mean.get(y, x, 3);
            if( wn == 0 ) continue;

            image.get(y, x, 3) += bin_weight;
            sq_image.get(y, x, 3) += wn;

            for(int channel = 0; channel < 3; ++channel)
            {
                float v = samples_mean.get(y,x,channel);
                // maj color
                image.get(y, x, channel) += (v - image.get(y, x, channel)) * bin_weight / image.get(y, x, 3);
                sq_image.get(y, x, channel) += (v * v - sq_image.get(y, x, channel)) * bin_weight / image.get(y, x, 3);
                
                // maj error
                variance.get(y, x, channel) = sq_image.get(y, x, channel) - image.get(y, x, channel) * image.get(y, x, channel);
                
                // Square root chi2 : mean, mean of squares, samples number
                error.get(y, x, channel) = srcs(image.get(y, x, channel), sq_image.get(y, x, channel), image.get(y, x, 3));

                // reset running layers
                samples_mean.get(y, x, channel) = 0;
                sq_samples_mean.get(y, x, channel) = 0;
            }

            samples_mean.get(y, x, 3) = 0;
            sq_samples_mean.get(y, x, 3) = 0;
        }

        return 0;
    }

    float get_error(int x, int y) const
    {
        return tonemapping(get_color(x, y, error)).luminance();
    }

    float get_variance(int x, int y) const
    {
        return tonemapping(get_color(x, y, variance)).luminance();
    }

    float get_std_deviation(int x, int y) const
    {
        Color stddev= sqrt(get_color(x, y, variance));
        return tonemapping(stddev).luminance();
    }

    float get_std_deviation_over_mean(int x, int y) const
    {
        Color stddev = sqrt(get_color(x, y, variance));
        Color mean = get_color(x, y, image);
        Color c = stddev / mean;
        return tonemapping(c).luminance();
    }

    ////////////////////////////////////////////////////////////////
    ///////////////////// for adaptive purposes ////////////////////
    ////////////////////////////////////////////////////////////////
    enum AdaptiveCriterion {VARIANCE, F_DIVERGENCE, TV_F_DIVERGENCE, STD_DEVIATION, STD_DEVIATION_OVER_MEAN};
    // todo: STD_DEVIATION
    // todo: STD_DEVIATION / MEAN
    float ada_get_error(Sampler& sampler, AdaptiveCriterion criterion=TV_F_DIVERGENCE)
    {
        float e = 0;
        #pragma omp parallel for collapse(2) reduction(+:e)
        for (int x = 0; x < w; ++x)
        for (int y = 0; y < h; ++y)
        {
            float value = 0;
            switch(criterion)
            {
                case VARIANCE: value = get_variance(x, y); break;
                case F_DIVERGENCE: value = get_error(x, y); break;
                case TV_F_DIVERGENCE: value = estimateNoise(sampler, x, y); break;
                case STD_DEVIATION: value = get_std_deviation(x, y); break;
                case STD_DEVIATION_OVER_MEAN: value = get_std_deviation_over_mean(x, y); break;
            }
            PixelPosition pixel(y,x);
            if( std::isnan(value) || std::isinf(value) ) 
            {
                sample_map.get(pixel, 0) = 0;
            }
            else
            {
                sample_map.get(pixel, 0) = value;
                e += value;
            }
        }
        return e/(w*h);
    }

    void ada_update_error(Sampler& sampler)
    {
        mean_error = ada_get_error(sampler);
    }

    void ada_update_sample_map( Sampler& sampler, const float avg_error, const float total_samples )
    {
        float sum = 0;

        #pragma omp parallel for 
        for (int x = 0; x < w; ++x)
        for (int y = 0; y < h; ++y)
        {
            // sample_map.get(y, x, 0) = std::max(0.f, sample_map.get(y, x, 0) / avg_error - 1.f);
            sample_map.get(y, x, 0) = sample_map.get(y, x, 0) / avg_error;
            #pragma omp atomic
            sum += sample_map.get(y, x, 0);
        }   

        #pragma omp parallel for 
        for (int x = 0; x < w; ++x)
        for (int y = 0; y < h; ++y)
        {
            // sample_map.get(y, x, 0) *= total_samples / sum;
            sample_map.get(y, x, 0) = 1+sample_map.get(y, x, 0) * total_samples / sum;
            float fsamples_intr = 0;
            float fsamples_frac = std::modf(sample_map.get(y, x, 0), &fsamples_intr);
            // int samples = static_cast<int>(fsamples_intr); 
            // Russian Roulette to add a sample stochastically
            if( sampler.sample() < fsamples_frac ) fsamples_intr+=1;
            sample_map.get(y, x, 0) = fsamples_intr;
        }
    }

    float ada_get_error_criterion(int x, int y) const
    {
        PixelPosition pixel(y,x);
        return sample_map.get(pixel, 0);
    }
    ////////////////////////////////////////////////////////////////
    ///////////////////// for adaptive purposes ////////////////////
    ////////////////////////////////////////////////////////////////

    void stats()
    {
        mean_error = 0; 
        mean_variance = 0; 
        mean_weights = 0; 
        mean_samples = 0;
        mean_error_variations = 0;
        // reset
        #pragma omp parallel
        {
            Sampler sampler;
            #pragma omp for schedule(dynamic, 8)
            for (int x = 0; x < w; ++x)
            for (int y = 0; y < h; ++y)
            {
                // error update
                float pixel_error = tonemapping(get_color(x, y, error)).power();
                float pixel_variance = tonemapping(get_color(x, y, variance)).power();
                float pixel_weight = image.get(y, x, 3);
                float pixel_samples = variance.get(y, x, 3);
                float pixel_error_variations = estimateNoise(sampler, x, y);

                sample_map.get(y, x, 0) = std::isnan(pixel_error_variations) ? 0 : pixel_error_variations;
                // sample_map.get(y, x, 1) = std::isnan(pixel_error) ? 0 : pixel_error;

                #pragma omp atomic
                mean_error += std::isnan(pixel_error) ? 0 : pixel_error;
                #pragma omp atomic
                mean_error_variations += sample_map.get(y, x, 0);
                #pragma omp atomic
                mean_variance += pixel_variance;
                #pragma omp atomic
                mean_weights += pixel_weight;
                #pragma omp atomic
                mean_samples += pixel_samples;
            }
        }
        
        const float inv_wh = invw * invh; 
        mean_error *= inv_wh;
        mean_variance *= inv_wh;
        mean_error_variations *= inv_wh;
        mean_weights *= inv_wh;
        mean_samples *= inv_wh;
    }

    float compute_sum( const float avg_error_variations )
    {
        float sum = 0;

        #pragma omp parallel for 
        for (int x = 0; x < w; ++x)
        for (int y = 0; y < h; ++y)
        {
            sample_map.get(y, x, 0) = std::max(0.f, sample_map.get(y, x, 0) / avg_error_variations - 1.f);
            #pragma omp atomic
            sum += sample_map.get(y, x, 0);
        }   

        return sum;
    }

    void update_sample_map( const float nb_samples, const float sum )
    {
        #pragma omp parallel for 
        for (int x = 0; x < w; ++x)
        for (int y = 0; y < h; ++y)
        {
            sample_map.get(y, x, 0) /= sum;
            sample_map.get(y, x, 0) *= nb_samples;
        }
    }
};

#endif
