#include "renderer/geometry.h"
#include "renderer/triangle.h"

#include "vec.h"
#include "mesh.h"
#include "wavefront.h"

// from several keyframes separate static geometry and dynamic geometry
void split_mesh( std::vector<MeshData>& data, Geometry& static_geometry, Geometry& dynamic_geometry )
{
    // input infos
    int steps = data.size();
    int triangles = data[0].position_indices.size() / 3;
    int positions = data[0].positions.size();
    int normals = data[0].normals.size();
    int texcoords = data[0].texcoords.size();
    int materials = data[0].materials.size();

    size_t size = 0;
    for(int step = 0; step < steps; ++step )
    {
        size += data[step].positions.size() * sizeof(vec3);
        size += data[step].texcoords.size() * sizeof(vec2);
        size += data[step].normals.size() * sizeof(vec3);

        size += data[step].position_indices.size() * sizeof(int);
        size += data[step].texcoord_indices.size() * sizeof(int);
        size += data[step].normal_indices.size() * sizeof(int);

        size += data[step].materials.size() * sizeof(Material);
        size += data[step].material_indices.size() * sizeof(unsigned int);
    }

    printf("mesh infos:\n");
    printf("\tsteps    \t%d\n", steps);
    printf("\ttriangles\t%d\n", triangles);
    printf("\tpositions\t%d\n", positions);
    printf("\tnormals  \t%d\n", normals);
    printf("\ttexcoords\t%d\n", texcoords);
    printf("\tmaterials\t%d\n", materials);
    printf("\tsize\t\t%lu Mo\n", size / 1000000);
    
    // frame unique
    if( data.size() <= 1 )
    { 
        static_geometry = Geometry( data[0] );
        return;
    }

    // reindexation des triangles statiques
    static_geometry = Geometry(1, positions / 2, normals / 2, texcoords / 2, triangles / 3, materials);
    HashMap static_positions_remap;
    HashMap static_materials_remap;
    HashMap3 static_normals_remap;
    HashMap2 static_texcoords_remap;
    
    // reindexation des triangles dynamiques
    dynamic_geometry = Geometry(steps, positions / 2, normals / 2, texcoords / 2, triangles / 3, materials);
    HashMap dynamic_positions_remap;
    HashMap dynamic_materials_remap;
    HashMap3 dynamic_normals_remap;
    HashMap2 dynamic_texcoords_remap;
    
    // pour chaque triangle 
    for(int index = 0; index < triangles; index++)
    {
        // statique par défaut
        bool dynamic = false;

        // pour chaqe frame
        for(int frame = 0; frame < steps - 1; frame++)
        {
            // tester l'equivalence des 3 sommets
            if( data[frame].positions[data[frame].position_indices[3 * index]] != data[frame + 1].positions[data[frame].position_indices[3 * index]] ||
                data[frame].positions[data[frame].position_indices[3 * index + 1]] != data[frame + 1].positions[data[frame].position_indices[3 * index + 1]] ||
                data[frame].positions[data[frame].position_indices[3 * index + 2]] != data[frame + 1].positions[data[frame].position_indices[3 * index + 2]] )
            {
                // sinon le triangle est dynamique
                dynamic = true;
                break;
            }
        }

        // tri des sommets et reindexation
        for(int k = 0; k < 3; k++)
        {
            // unique index
            int vertex = data[0].position_indices[3 * index + k];
            
            // dynamic triangles reindexation
            if( dynamic )
            {
                // positions
                auto pfound = dynamic_positions_remap.insert( std::make_pair(vertex, dynamic_positions_remap.size()) );
                for(int frame = 0; frame < steps; frame++)
                    if(pfound.second) dynamic_geometry.positions[frame].push_back( data[frame].positions[vertex] );
                dynamic_geometry.position_indices.push_back( pfound.first->second );

                // normals / texcoords
                for(int frame = 0; frame < steps; frame++)
                {
                    if(normals)
                    {
                        // get normal
                        int inormal = data[frame].normal_indices[3 * index + k];
                        vec3 n = data[frame].normals[inormal];
                        // insert normal
                        auto nfound = dynamic_normals_remap.insert( std::make_pair(n, dynamic_normals_remap.size()) );
                        if(nfound.second) dynamic_geometry.normals.push_back( n );
                        // reindexation
                        dynamic_geometry.normal_indices[frame].push_back( nfound.first->second );
                    }
                    
                    if(texcoords)
                    {
                        // get texcoord
                        int itexcoord = data[frame].texcoord_indices[3 * index + k];
                        vec2 t = data[frame].texcoords[itexcoord];
                        // insert texcoords
                        auto tfound = dynamic_texcoords_remap.insert( std::make_pair(t, dynamic_texcoords_remap.size()) );
                        if(tfound.second) dynamic_geometry.texcoords.push_back( t );
                        // reindexation
                        dynamic_geometry.texcoord_indices[frame].push_back( tfound.first->second );
                    }
                }   
            }
            // static triangles reindexation
            else 
            {
                // positions
                auto found = static_positions_remap.insert( std::make_pair(vertex, static_positions_remap.size()) );
                if(found.second) static_geometry.positions[0].push_back( data[0].positions[vertex] );
                static_geometry.position_indices.push_back( found.first->second );

                if( normals )
                {
                    // get normal
                    int inormal = data[0].normal_indices[3 * index + k];
                    vec3 n = data[0].normals[inormal];
                    // insert normal
                    auto nfound = static_normals_remap.insert( std::make_pair(n, static_normals_remap.size()) );
                    if(nfound.second) static_geometry.normals.push_back( n );
                    // reindexation
                    static_geometry.normal_indices[0].push_back( nfound.first->second );
                }
                
                if( texcoords )
                {
                    // get texcoord
                    int itexcoord = data[0].texcoord_indices[3 * index + k];
                    vec2 t = data[0].texcoords[itexcoord];
                    // insert texcoord
                    auto tfound = static_texcoords_remap.insert( std::make_pair(t, static_texcoords_remap.size()) );
                    if(tfound.second) static_geometry.texcoords.push_back( t );
                    // reindexation
                    static_geometry.texcoord_indices[0].push_back( tfound.first->second );
                }
            }
        }

        // materials
        if(materials)
        {
            // get material
            int imaterial = data[0].material_indices[index];
            
            if( dynamic )
            {
                auto tfound = dynamic_materials_remap.insert( std::make_pair(imaterial, dynamic_materials_remap.size()) );
                if(tfound.second) 
                {
                    const Material& material = data[0].materials[imaterial];
                    dynamic_geometry.materials.push_back( material );
                }
                // reindexation
                dynamic_geometry.material_indices.push_back( tfound.first->second );
            }
            else
            {
                auto tfound = static_materials_remap.insert( std::make_pair(imaterial, static_materials_remap.size()) );
                if(tfound.second) 
                {
                    const Material& material = data[0].materials[imaterial];
                    static_geometry.materials.push_back( material );
                }
                // reindexation
                static_geometry.material_indices.push_back( tfound.first->second );
            }
        }
    }

    static_geometry.print();
    dynamic_geometry.print();
}

template< typename T>
static void write( const T& n, std::ofstream& out )
{
    try 
    {
        out.write(reinterpret_cast<const char*>(&n), sizeof(T));
        out.exceptions(std::ofstream::failbit);
    } 
    catch(std::ios_base::failure& fail) 
    {
        printf("erreur d'écriture ...\n");
    }
}

template< typename T >
static void write( const std::vector<T>& buffer, std::ofstream& out )
{
    try 
    {
        write(int(buffer.size()), out);
        out.write(reinterpret_cast<const char*>(buffer.data()), sizeof(T) * buffer.size());
        out.exceptions(std::ofstream::failbit);
    } 
    catch(std::ios_base::failure& fail) 
    {
        printf("erreur d'écriture ...\n");
    }
}

void write_geometry_cache( const Geometry& geometry, std::string filename )
{
    if(!geometry.triangle_count()) return;
    
    std::string cache_filename = filename + ".geometry";
    std::ofstream out(cache_filename, std::ios::out | std::ios::binary);
    if(!out) return;
    
    printf("writing geometry cache '%s'...\n", cache_filename.c_str());
    
    // write number of steps
    write(geometry.step_count(), out);
    
    // write geometric data
    for(int i = 0; i < geometry.step_count(); i++)
    {
        write(geometry.positions[i], out);
        write(geometry.texcoord_indices[i], out);
        write(geometry.normal_indices[i], out);
    }
    
    // write geometric data
    write(geometry.texcoords, out);
    write(geometry.normals, out);
    write(geometry.position_indices, out);
    write(geometry.material_indices, out);
    out.close();
    
    // materials
    cache_filename = filename + ".materials";
    out = std::ofstream(cache_filename, std::ios::out);
    if(!out) return;
    
    printf("writing materials cache '%s'...\n", cache_filename.c_str());
    
    for(int i = 0; i < geometry.material_count(); i++)
    {
        const Material& m = geometry.materials[i];
        out << "newmtl " << m.name << std::endl;
        if(m.diffuse.power())
            out << "    Kd " << m.diffuse.r << " " << m.diffuse.g << " " << m.diffuse.b << std::endl;
        if(m.specular.power())
            out << "    Ks " << m.specular.r << " " << m.specular.g << " " << m.specular.b << std::endl;
        if(m.emission.power())
            out << "    Ke " << m.emission.r << " " << m.emission.g << " " << m.emission.b << std::endl;
        if(m.ns)
            out << "    Ns " << m.ns << std::endl;
        out << "    Ni " << m.ni << std::endl;
        out << "    illum " << m.illum << std::endl;
        if(!m.diffuse_filename.empty())
            out << "    map_Kd " << m.diffuse_filename << std::endl;
        if(!m.specular_filename.empty())
            out << "    map_Ks " << m.specular_filename << std::endl;
        if(!m.emission_filename.empty())
            out << "    map_Ke " << m.emission_filename << std::endl;
        if(!m.ns_filename.empty())
            out << "    map_Ns " << m.ns_filename << std::endl;
    }
    out.close();
}

template< typename T>
static void read( T& n, std::ifstream& in )
{
    try 
    {
        in.read(reinterpret_cast<char*>(&n), sizeof(T));
        in.exceptions(std::ifstream::failbit);
    } 
    catch(std::ios_base::failure& fail) 
    {
        printf("erreur de lecture ...\n");
    }
}

template< typename T >
static void read( std::vector<T>& buffer, std::ifstream& in )
{
    try 
    {
        int n;
        read(n, in);
        if( n )
        {
            buffer.resize(n);
            in.read(reinterpret_cast<char*>(buffer.data()), sizeof(T) * n);
            in.exceptions(std::ifstream::failbit);
        }
    } 
    catch(std::ios_base::failure& fail) 
    {
        printf("erreur de lecture ...\n");
    }
}

int read_geometry_cache( std::string filename, Geometry& geometry )
{   
    std::string cache_filename = filename + ".geometry";
    std::ifstream in(cache_filename, std::ios::in | std::ios::binary);
    if( !in ) 
    {
        printf("error reading geometry cache '%s'...\n", cache_filename.c_str());
        return 0;
    }

    printf("reading geometry cache '%s'...\n", cache_filename.c_str());
    
    // read number of steps
    read(geometry.steps, in);
    // printf("steps %d\n", geometry.steps);
    geometry.positions.resize(geometry.step_count());
    geometry.texcoord_indices.resize(geometry.step_count());
    geometry.normal_indices.resize(geometry.step_count());
    
    // read geometric data
    for(int i = 0; i < geometry.step_count(); i++)
    {
        read(geometry.positions[i], in);
        read(geometry.texcoord_indices[i], in);
        read(geometry.normal_indices[i], in);
    }
    
    // read geometric data
    read(geometry.texcoords, in);
    read(geometry.normals, in);
    read(geometry.position_indices, in);
    read(geometry.material_indices, in);
    in.close();
    
    // read materials
    cache_filename = filename + ".materials";
    MaterialLib lib = read_materials(cache_filename.c_str());
    geometry.materials = lib.data;

    geometry.print();

    return 1;
}

Geometry make_cube(const Transform& matrix)
{
    Geometry g;
    g.steps = 1;
    g.positions.resize(1);
    g.texcoord_indices.resize(1);
    g.normal_indices.resize(1);
    g.positions[0] = {
        vec3(matrix(Point( 0.5, 0.5, -0.5))),
        vec3(matrix(Point( 0.5, -0.5, -0.5))),
        vec3(matrix(Point( 0.5, 0.5, 0.5))),
        vec3(matrix(Point( 0.5, -0.5, 0.5))),
        vec3(matrix(Point( -0.5, 0.5, -0.5))),
        vec3(matrix(Point( -0.5, -0.5, -0.5))),
        vec3(matrix(Point( -0.5, 0.5, 0.5))),
        vec3(matrix(Point( -0.5, -0.5, 0.5)))
    };
    g.texcoords = { // cubemap texcoords
        vec2( 0.875000, 0.500000 ),
        vec2( 0.625000, 0.750000 ),
        vec2( 0.625000, 0.500000 ),
        vec2( 0.375000, 1.000000 ),
        vec2( 0.375000, 0.750000 ),
        vec2( 0.625000, 0.000000 ),
        vec2( 0.375000, 0.250000 ),
        vec2( 0.375000, 0.000000 ),
        vec2( 0.375000, 0.500000 ),
        vec2( 0.125000, 0.750000 ),
        vec2( 0.125000, 0.500000 ),
        vec2( 0.625000, 0.250000 ),
        vec2( 0.875000, 0.750000 ),
        vec2( 0.625000, 1.000000 )
    };
    g.normals = {
        vec3(normalize(matrix(Vector( 0, 1, 0)))),
        vec3(normalize(matrix(Vector( 0, 0, 1)))),
        vec3(normalize(matrix(Vector( -1, 0, 0)))),
        vec3(normalize(matrix(Vector( 0, -1, 0)))),
        vec3(normalize(matrix(Vector( 1, 0, 0)))),
        vec3(normalize(matrix(Vector( 0, 0, -1))))
    };
    g.position_indices =    { 4, 2, 0, 2, 7, 3, 6, 5, 7, 1, 7, 5, 0, 3, 1, 4, 1, 5, 4, 6, 2, 2, 6, 7, 6, 4, 5, 1, 3, 7, 0, 2, 3, 4, 0, 1 };
    g.texcoord_indices[0] = { 0, 1, 2, 1, 3, 4, 5, 6, 7, 8, 9, 10, 2, 4, 8, 11, 8, 6, 0, 12, 1, 1, 13, 3, 5, 11, 6, 8, 4, 9, 2, 1, 4, 11, 2, 8 };
    g.normal_indices[0] =   { 0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5, 0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5 };
    g.materials.resize(1);
    g.material_indices.resize(g.triangle_count(), 0);
    return g;
}

Geometry make_cube(const Point& pmin, const Point& pmax)
{
    Vector scale = pmax - pmin;
    Vector mid((pmin + pmax)*0.5f);
    return make_cube(Translation(mid) * Scale(scale.x, scale.y, scale.z));
}

Geometry make_quad(const Point& a, const Point& b, const Point& c, const Point& d)
{
    Geometry g;
    g.steps = 1;
    g.positions.resize(1);
    g.texcoord_indices.resize(1);
    g.normal_indices.resize(1);
    g.positions[0] = { vec3(a), vec3(b), vec3(c), vec3(d) };
    g.texcoords = { vec2( 0, 0 ), vec2( 0, 1 ), vec2( 1, 1 ), vec2( 1, 0 ) };
    g.normals = { vec3(normalize(cross(Vector(a, b), Vector(a, d)))) };

    g.position_indices =    { 0, 1, 3, 1, 2, 3 };
    g.texcoord_indices[0] = { 0, 1, 3, 1, 2, 3 };
    g.normal_indices[0] =   { 0, 0, 0, 0, 0, 0 };
    
    g.materials.resize(1);
    g.material_indices.resize(g.triangle_count(), 0);
    return g;
}