#pragma once

#ifndef GEOMETRY_H
#define GEOMETRY_H

#include <vector>
#include <map>
#include <unordered_map>
#include <functional>
#include <iostream>
#include <fstream>
#include <memory>

#include "renderer/triangle.h"
#include "renderer/texturelib.h"
#include "renderer/sampler.h"
#include "renderer/bsdf/bsdf.h"
#include "renderer/media/medium.h"

#include "vec.h"
#include "wavefront.h"

struct Geometry
{
    int steps = 0;

    const vec3 *vertex_buffer( const int step ) const { return positions[step].data(); }
    std::size_t vertex_buffer_size( const int step ) const { return positions[step].size() * sizeof(vec3); }
    const unsigned int *vertex_index_buffer( ) const { return position_indices.data(); }
    std::size_t vertex_index_buffer_size( ) const { return position_indices.size() * sizeof(unsigned int); }

    int vertex_count( const int step ) const { return (int) positions[step].size(); }
    int vertex_index_count( ) const { return (int) position_indices.size(); }

    int normal_count( ) const { return (int) normals.size(); }
    int normal_index_count( const int step ) const { return (int) normal_indices[step].size(); }

    int texcoord_count( ) const { return (int) texcoords.size(); }
    int texcoord_index_count( const int step ) const { return (int) texcoord_indices[step].size(); }

    int material_count( ) const { return (int) materials.size(); }
    int material_index_count( const int step ) const { return (int) material_indices.size(); }

    int step_count( ) const { return (int) steps; }
    int triangle_count( ) const { return (int) position_indices.size() / 3; }

    Triangle triangle( const int step, const int id ) const
    {
        Triangle t;

        int i0 = 3 * id;
        int i1 = 3 * id + 1;
        int i2 = 3 * id + 2;

        t.a = Point(positions[step][position_indices[i0]]);
        t.b = Point(positions[step][position_indices[i1]]);
        t.c = Point(positions[step][position_indices[i2]]);

        if( normal_count() )
        {
            int n0 = normal_indices[step][i0];
            int n1 = normal_indices[step][i1];
            int n2 = normal_indices[step][i2];
            if( n0 >= 0 ) t.na = Vector(normals[n0]);
            if( n1 >= 0 ) t.nb = Vector(normals[n1]);
            if( n2 >= 0 ) t.nc = Vector(normals[n2]);
        }

        if( texcoord_count() )
        {
            int t0 = texcoord_indices[step][i0];
            int t1 = texcoord_indices[step][i1];
            int t2 = texcoord_indices[step][i2];
            if( t0 >= 0 ) t.ta = texcoords[t0];
            if( t1 >= 0 ) t.tb = texcoords[t1];
            if( t2 >= 0 ) t.tc = texcoords[t2];
        }

        return t;
    }

    // vertices
    std::vector< std::vector<vec3> > positions;
    std::vector<unsigned int> position_indices;

    // normals
    std::vector<vec3> normals;
    std::vector< std::vector<unsigned int> > normal_indices;

    // texcoords
    std::vector<vec2> texcoords;
    std::vector< std::vector<unsigned int> > texcoord_indices;

    // materials
    std::vector<Material> materials;
    std::vector<unsigned int> material_indices;
    std::vector<std::shared_ptr<Bsdf>> bsdfs;
    std::vector<std::shared_ptr<Medium>> interiors;
    std::vector<std::shared_ptr<Medium>> exteriors;

    TextureLib textures;

    Geometry(){}

    Geometry( int nb_steps, int nb_vertices, int nb_normals, int nb_texcoords, int nb_indices, int nb_materials ) :
        steps(nb_steps),
        positions(nb_steps),
        normal_indices(nb_steps),
        texcoord_indices(nb_steps)
    {
        for( int i = 0; i < steps; i++ )
        {
            positions[i].reserve(nb_vertices);
            normal_indices[i].reserve(nb_indices);
            texcoord_indices[i].reserve(nb_indices);
        }

        position_indices.reserve(nb_indices);
        material_indices.reserve(nb_indices);
        materials.reserve(nb_materials);
        normals.reserve(nb_normals);
        texcoords.reserve(nb_texcoords);
    }

    // static geometry only: copy data
    Geometry( MeshData& data ) :
        steps(1),
        positions(1),
        position_indices(data.position_indices.size()),
        normal_indices(1, std::vector<unsigned int>(data.position_indices.size())),
        texcoord_indices(1, std::vector<unsigned int>(data.position_indices.size())),
        material_indices(data.position_indices.size())
    {
        // input infos
        int nb_normals = data.normals.size();
        int nb_texcoords = data.texcoords.size();
        int nb_materials = data.materials.size();

        positions[0] = data.positions;
        if( nb_normals ) normals = data.normals;
        if( nb_texcoords ) texcoords = data.texcoords;
        if( nb_materials )
        {
            materials = data.materials;
            material_indices = data.material_indices;
        }

        for(unsigned int i = 0; i < data.position_indices.size(); i++)
        {
            position_indices[i] = data.position_indices[i];
            if( nb_normals ) normal_indices[0][i] = data.normal_indices[i];
            if( nb_texcoords ) texcoord_indices[0][i] = data.texcoord_indices[i];
        }
    }

    // TODO
    // dynamic geometry only: copy positions & materials, remap normals & texcoords 
    Geometry( std::vector<MeshData>& data ) :
        steps(data.size()),
        positions(data.size()),
        position_indices(data[0].position_indices.size()),
        normal_indices(data.size(), std::vector<unsigned int>(data[0].position_indices.size())),
        texcoord_indices(data.size(), std::vector<unsigned int>(data[0].position_indices.size())),
        material_indices(data[0].position_indices.size())
    {

    }

    void init_bsdf()
    {
        interiors.resize(materials.size(), nullptr);
        exteriors.resize(materials.size(), nullptr);
        bsdfs.reserve(materials.size());
        for( auto& mat : materials )
        {
            // light source
            if( mat.emission.power() > 0 || !mat.emission_filename.empty() )
            {
                // ne rien faire
                bsdfs.push_back(std::make_shared<BsdfLambert>(nullptr));
                printf("\tlight source - bsdf null\n");
            }
            else
            {
                // new bsdf
                std::shared_ptr<Bsdf> bsdf;
                float roughness = mat.ns > 0 ? ShininessToRoughness(mat.ns) : 0;
                // illumination models
                switch( mat.illum )
                {
                case 2: // plastic
                    // make smooth plastic
                    if( roughness <= 1e-2f )
                    {
                        bsdf = std::make_shared<BsdfPlastic>(
                            (mat.diffuse_texture >= 0 ? textures.textures[mat.diffuse_texture] : std::make_shared<TextureConstant>(mat.diffuse)),
                            (mat.specular_texture >= 0 ? textures.textures[mat.specular_texture] : std::make_shared<TextureConstant>(mat.specular)),
                            1.5,
                            1.f,
                            true
                        );
                        printf("\tbsdf plastic\n");
                    }
                    // make rough plastic 
                    else
                    {
                        bsdf = std::make_shared<BsdfRoughPlastic>(
                            (mat.diffuse_texture >= 0 ? textures.textures[mat.diffuse_texture] : std::make_shared<TextureConstant>(mat.diffuse)),
                            (mat.specular_texture >= 0 ? textures.textures[mat.specular_texture] : std::make_shared<TextureConstant>(mat.specular)),
                            (mat.ns_texture >= 0 ? textures.textures[mat.ns_texture] : std::make_shared<TextureConstant>(Color(roughness))),
                            nullptr,
                            std::make_shared<MicroFacetGGX>(),
                            1.5,
                            1.f,
                            true
                        );
                        printf("\tbsdf rough plastic\texp %f roughness %f\n", mat.ns, roughness);
                    }
                    break;
                case 5: // conductor
                    // make smooth conductor
                    if( roughness <= 1e-2f )
                    {
                        bsdf = std::make_shared<BsdfConductor>(
                            // mat.specular
                            (mat.specular_texture >= 0 ? textures.textures[mat.specular_texture] : std::make_shared<TextureConstant>(mat.specular))
                        );
                        printf("\tbsdf conductor\n");
                    }
                    // make rough conductor
                    else
                    {
                        bsdf = std::make_shared<BsdfRoughConductor>(
                            // mat.specular,
                            (mat.specular_texture >= 0 ? textures.textures[mat.specular_texture] : std::make_shared<TextureConstant>(mat.specular)),
                            (mat.ns_texture >= 0 ? textures.textures[mat.ns_texture] : std::make_shared<TextureConstant>(Color(roughness)))
                        );
                        printf("\tbsdf rough conductor\texp %f roughness %f\n", mat.ns, roughness);
                    }
                    break;
                case 7: // dielectric
                    // make smooth dielectric
                    if( roughness <= 1e-2f )
                    {
                        bsdf = std::make_shared<BsdfDielectric>(
                            1.5f,
                            1.f,
                            (mat.specular_texture >= 0 ? textures.textures[mat.specular_texture] : std::make_shared<TextureConstant>(mat.specular)),
                            (mat.specular_texture >= 0 ? textures.textures[mat.specular_texture] : std::make_shared<TextureConstant>(mat.specular))
                        );
                        printf("\tbsdf dielectric\n");
                    }
                    // make rough dielectric 
                    else
                    {
                        bsdf = std::make_shared<BsdfRoughDielectric>(
                            1.5f,
                            1.f,
                            (mat.specular_texture >= 0 ? textures.textures[mat.specular_texture] : std::make_shared<TextureConstant>(mat.specular)),
                            (mat.specular_texture >= 0 ? textures.textures[mat.specular_texture] : std::make_shared<TextureConstant>(mat.specular)),
                            (mat.ns_texture >= 0 ? textures.textures[mat.ns_texture] :  std::make_shared<TextureConstant>(Color(roughness)))
                        );
                        printf("\tbsdf rough dielectric\texp %f roughness %f\n", mat.ns, roughness);
                    }
                    break;
                case 11: // volumes :) :)
                {
                    bsdf = nullptr;
                    // homogeneous tests

                    // interiors[bsdfs.size()] = std::make_shared<MediumHomogeneous>(
                    //      Color(0, 0, 0), 
                    //      Color(1, 1, 1) * 5
                    // );

                    interiors[bsdfs.size()] = std::make_shared<MediumHeterogeneous>(
                         Color(0, 0, 0),
                         Color(1, 1, 1) * 30 * BsdfRoughDiffuse::rand1D(),
                         std::make_shared<DensityFunctionConstant>(0.5, 0.5 + BsdfRoughDiffuse::rand1D())
                    );

                    // printf("\tmedium - bsdf null\tsigmaA (%f %f %f)\tsigmaS (%f %f %f)\n",
                    //     interiors[bsdfs.size()]->mSigmaA.r,
                    //     interiors[bsdfs.size()]->mSigmaA.g,
                    //     interiors[bsdfs.size()]->mSigmaA.b,
                    //     interiors[bsdfs.size()]->mSigmaS.r,
                    //     interiors[bsdfs.size()]->mSigmaS.g,
                    //     interiors[bsdfs.size()]->mSigmaS.b
                    // );

                    // interiors[bsdfs.size()] = std::make_shared<MediumHeterogeneous>(
                    //             Black(),
                    //             Color(1, 1, 1) * 50
                    //     );

                    // interiors[bsdfs.size()] = std::make_shared<MediumHomogeneous>(
                    //      Color(0, 0, 0), 
                    //      Color(0.4, 1, 0.7)* 4
                    // );

                    // interiors[bsdfs.size()] = std::make_shared<MediumHeterogeneous>(
                    //      Color(0, 0, 0), 
                    //      Color(0.4, 1, 0.7)* 4,
                    //      std::make_shared<DensityFunctionConstant>()
                    // );

                    // heterogeneous tests

                    #define TRUE_HOMOGENEOUS    false
                    #define FALSE_HOMOGENEOUS   false
                    #define HETEROGENEOUS       true
                    #define NANOVDB             false

                    const Color sigmaS = White();
                    //const Color sigmaS = Color(0.5);
                    // const Color sigmaS = Color(0.5, 0.5, 1.);
                    //const Color sigmaS = 10000.*Color(0.0000058, 0.0000135, 0.0000331);
                    const Color sigmaA = Black();
                    //const Color sigmaA = 0.5*sigmaS;
                    const float maxDensity = 1.f;

                    if( TRUE_HOMOGENEOUS )
                    {
                        interiors[bsdfs.size()] = std::make_shared<MediumHomogeneous>(
                             sigmaA,
                             sigmaS,
                             maxDensity
                        );

                        printf("\tmedium - bsdf null\tsigmaA (%f %f %f)\tsigmaS (%f %f %f)\n",
                            interiors[bsdfs.size()]->getMuA(Point(0,0,0)).r,
                            interiors[bsdfs.size()]->getMuA(Point(0,0,0)).g,
                            interiors[bsdfs.size()]->getMuA(Point(0,0,0)).b,
                            interiors[bsdfs.size()]->getMuS(Point(0,0,0)).r,
                            interiors[bsdfs.size()]->getMuS(Point(0,0,0)).g,
                            interiors[bsdfs.size()]->getMuS(Point(0,0,0)).b
                        );
                    }
                    else if( FALSE_HOMOGENEOUS )
                    {
                        interiors[bsdfs.size()] = std::make_shared<MediumHeterogeneous>(
                            sigmaA*maxDensity,
                            sigmaS*maxDensity,
                            std::make_shared<DensityFunctionConstant>(1.)
                            //sigmaA*2.*maxDensity,
                            //sigmaS*2.*maxDensity,
                            //std::make_shared<DensityFunctionConstant>(0.5)
                            //std::make_shared<DensityFunctionConstant>(maxDensity, 2.*maxDensity)
                        );

                        printf("\tmedium - bsdf null\tsigmaA (%f %f %f)\tsigmaS (%f %f %f)\n",
                            interiors[bsdfs.size()]->getMuA(Point(0,0,0)).r,
                            interiors[bsdfs.size()]->getMuA(Point(0,0,0)).g,
                            interiors[bsdfs.size()]->getMuA(Point(0,0,0)).b,
                            interiors[bsdfs.size()]->getMuS(Point(0,0,0)).r,
                            interiors[bsdfs.size()]->getMuS(Point(0,0,0)).g,
                            interiors[bsdfs.size()]->getMuS(Point(0,0,0)).b
                        );
                    }
                    else if( HETEROGENEOUS )
                    {
                        interiors[bsdfs.size()] = std::make_shared<MediumHeterogeneous>(
                            sigmaA*2.*maxDensity,
                            sigmaS*2.*maxDensity,
                            std::make_shared<DensityFunctionFastNoise>()
                            // std::make_shared<DensityFunctionFastNoiseAnimated>()
                        );

                        printf("\tmedium - bsdf null\tsigmaA (%f %f %f)\tsigmaS (%f %f %f)\n",
                            interiors[bsdfs.size()]->getMuA(Point(0,0,0)).r,
                            interiors[bsdfs.size()]->getMuA(Point(0,0,0)).g,
                            interiors[bsdfs.size()]->getMuA(Point(0,0,0)).b,
                            interiors[bsdfs.size()]->getMuS(Point(0,0,0)).r,
                            interiors[bsdfs.size()]->getMuS(Point(0,0,0)).g,
                            interiors[bsdfs.size()]->getMuS(Point(0,0,0)).b
                        );
                    }
                    else if( NANOVDB )
                    {
                        interiors[bsdfs.size()] = std::make_shared<MediumHeterogeneous>(
                            Black(),
                            3.*White()*4.*maxDensity,
                            std::make_shared<DensityFunctionNanoVDB>("wdas_cloud_quarter.nvdb")
                        );
                    }
                    else
                    {
                        Color transmittanceColor = Color(0.9, 1, 0.95);
                        float thickness = 0.01f;
                        Color sigmaA = -log(transmittanceColor) / thickness;
                        Color scaledSigmaA = thickness * sigmaA;
                        for(int i = 0; i < 3; ++i)
                            sigmaA[i] = std::max(0.f, sigmaA[i]);
                        interiors[bsdfs.size()] = std::make_shared<MediumHeterogeneous>(
                                Black(),
                                scaledSigmaA * 50.f
                        );
                    }

                    // interiors[bsdfs.size()] = std::make_shared<MediumHomogeneous>(Black(), Color(0.01f));
                    break;
                }
                case 12: // thin dielectrics
                    bsdf = std::make_shared<BsdfThinDielectric>(
                        1.5f,
                        1.f,
                        (mat.specular_texture >= 0 ? textures.textures[mat.specular_texture] : std::make_shared<TextureConstant>(mat.specular)),
                        (mat.specular_texture >= 0 ? textures.textures[mat.specular_texture] : std::make_shared<TextureConstant>(mat.specular))
                    );
                    printf("\tbsdf thin dielectric\n");
                    break;
                case 13: // diffuse transmission
                    bsdf = std::make_shared<BsdfDiffuseTransmission>(
                        (mat.diffuse_texture >= 0 ? textures.textures[mat.diffuse_texture] : std::make_shared<TextureConstant>(mat.diffuse)),
                        // (mat.specular_texture >= 0 ? textures.textures[mat.specular_texture] : std::make_shared<TextureConstant>(mat.specular))
                        std::make_shared<TextureConstant>(0.5)
                    );
                    printf("\tbsdf diffuse transmission\n");
                    break;
                default: // 0-1 and others : diffuse (Lambert)
                    bsdf = std::make_shared<BsdfLambert>(
                        (mat.diffuse_texture >= 0 ? textures.textures[mat.diffuse_texture] : std::make_shared<TextureConstant>(mat.diffuse))
                    );
                    printf("\tbsdf lambert\n");
                    break;
                }
                bsdfs.push_back(bsdf);
            }
        }
        assert(bsdfs.size() == materials.size());
        // int cmp = 0;
        // for(auto i : material_indices)
        // {
        //     auto bsdf_ptr = bsdfs[i];
            
        //     if( bsdf_ptr )
        //         std::cout << cmp << " " << i << " " << typeid(*bsdf_ptr).name() << std::endl;
        //     else
        //         std::cout << cmp << " " << i << " null material" << std::endl;
        //     ++cmp;
        // }
    }

    void print()
    {
        size_t size = 0;
        for(int step = 0; step < step_count(); ++step )
        {
            size += positions[step].size() * sizeof(vec3);
            size += texcoord_indices[step].size() * sizeof(int);
            size += normal_indices[step].size() * sizeof(int);
        }

        size += texcoords.size() * sizeof(vec2);
        size += normals.size() * sizeof(vec3);
        size += position_indices.size() * sizeof(int);
        size += materials.size() * sizeof(Material);
        size += material_indices.size() * sizeof(int);

        printf("\ngeometry:\n");
        printf("\tsteps    \t%d\n", step_count());
        printf("\ttriangles\t%d\n", triangle_count());
        printf("\tpositions\t%d (x%d) \tindices\t%d (x%d)\n", vertex_count(0), step_count(), vertex_index_count(), 1);
        printf("\tnormals  \t%d (x%d) \tindices\t%d (x%d)\n", normal_count(), 1, normal_index_count(0), step_count());
        printf("\ttexcoords\t%d (x%d) \tindices\t%d (x%d)\n", texcoord_count(), 1, texcoord_index_count(0), step_count());
        printf("\tmaterials\t%d (x%d) \tindices\t%d (x%d)\n", material_count(), 1, material_index_count(0), step_count());
        printf("\tsize\t\t%lu Mo\n", size / 1000000);
        printf("\n");
    }
};


struct Hash2
{
    size_t operator()(const vec2& s) const
    {
        return std::hash<float>()(s.x) ^ (std::hash<float>()(s.y) << 1);
    }
};

struct Hash3
{
    size_t operator()(const vec3& s) const
    {
        return std::hash<float>()(s.x) ^ (std::hash<float>()(s.y) << 1) ^ (std::hash<float>()(s.z) << 2);
    }
};

typedef std::unordered_map<vec2, int, Hash2> HashMap2;
typedef std::unordered_map<vec3, int, Hash3> HashMap3;
typedef std::map<int, int> HashMap;

void split_mesh( std::vector<MeshData>& data, Geometry& static_geometry, Geometry& dynamic_geometry );
void write_geometry_cache( const Geometry& geometry, std::string filename );
int read_geometry_cache( std::string filename, Geometry& geometry );

Geometry make_cube(const Transform& matrix = Identity());
Geometry make_cube(const Point& pmin, const Point& pmax);
Geometry make_quad(const Point& a, const Point& b, const Point& c, const Point& d);

#endif
