#include "abstractintegrator.h"

#include <iomanip>

AbstractIntegrator::AbstractIntegrator(
    std::vector<Camera>&  aCameras,
    Builder&              aBuilder,
    const IntegratorConfig& aConfig
):
    cameras(aCameras),
    mBuilder(aBuilder),
    config(aConfig),
    mNumThreads(omp_get_max_threads())
{}

void AbstractIntegrator::InitSamplers()
{
    // init one sampler per thread
    for( int i = 0; i < mNumThreads; ++i )
    {
        // void* a = this;
        // uint64_t seed = uint64_t(a) + i;
        uint64_t seed = 12345 + i;
        mSamplers.emplace_back(seed);
    }
}

void AbstractIntegrator::InitTiles()
{
    // prepare cameras tiles
    const unsigned int nbCameras = cameras.size();
    for(unsigned int camId = 0u; camId < nbCameras; camId++)
    {
        const unsigned int width = cameras[camId].width;
        const unsigned int height = cameras[camId].height;
        const unsigned int tilex = std::ceil(float(width) / float(mTileSize));
        const unsigned int tiley = std::ceil(float(height) / float(mTileSize));
        const unsigned int nbTiles = tilex * tiley;
        for(unsigned int tileId = 0u; tileId < nbTiles; tileId++)
        {
            const unsigned int tymin = (tileId / tilex) * mTileSize;
            const unsigned int tymax = std::min(tymin + mTileSize, height);
            const unsigned int txmin = (tileId % tilex) * mTileSize;
            const unsigned int txmax = std::min(txmin + mTileSize, width);
            mTiles.push_back({camId, txmin, txmax, tymin, tymax});
        }
    }
}

void AbstractIntegrator::Quit()
{
    #pragma omp critical
    {
        mStop = true;
    }
}

bool AbstractIntegrator::IsRendering()
{
    return !mStop;
}

void AbstractIntegrator::Progress() const
{
#ifndef PROFILE
    float v = 0;
    float ti = (float)mTimer.sec().count() / (float)mIteration;
    if( config.mSampleLimit > 0 )
        v = (float)mIteration / (float)config.mSampleLimit;
    else if( config.mTimeLimit > 0 )
        v = (float)mTimer.sec().count() / (float)config.mTimeLimit;
    else
        v = 1;
#if 0
    int barWidth = 70;
    std::cout << "[";
    int pos = barWidth * v;
    for (int i = 0; i < barWidth; ++i) {
        if (i < pos) std::cout << "=";
        else if (i == pos) std::cout << ">";
        else std::cout << " ";
    }
    std::cout << "] " << int(100.f * v) << " % " << (int)mTimer.sec().count() << " s\r";
#else
    std::cout << "Progress " << int(100.f * v) << " % " << (int)mTimer.sec().count() << " s " << ti << "s/iter\r";
#endif
    std::cout.flush();
#endif
}

void AbstractIntegrator::SaveRenders() const
{
    if( !config.mSave ) return;

#ifdef GK_RELEASE
    int time = (int)mTimer.sec().count();
    int samples = mIteration;
    for (int i = 0; i < (int)cameras.size(); i++)
    {
        std::stringstream ss;
        ss  << config.mFilePath
            << config.mOutputName
            << "-"
            << std::setfill('0')
            << std::setw(3)
            << i
            << "-time-"
            << std::setfill('0')
            << std::setw(6)
            << time
            << "-samples-"
            << std::setfill('0')
            << std::setw(6)
            << samples;

        cameras[i].film.save2(ss.str(), 1.f, true, config.mSaveSamples, config.mSavePng );

        if( config.mSaveNatives )
            cameras[i].film.save_reconstruction( ss.str()+std::string("-natives"), 1.f, true );

        if( config.mLayers )
            cameras[i].film.save_layers(ss.str(), 1.f, true);
    }
#endif
}

#include "renderer/integrators/ambientintegrator.h"
#include "renderer/integrators/normalintegrator.h"
#include "renderer/integrators/volumetricpathtracingintegrator.h"
#include "renderer/integrators/multiviewpathtracingintegrator.h"
#include "renderer/integrators/adaptivemultiviewpathtracingintegrator.h"
#include "renderer/integrators/mismultiviewpathtracingintegrator.h"

AbstractIntegrator* AbstractIntegrator::CreateIntegrator(
    std::vector<Camera>& aCameras,
    Builder& aBuilder,
    const IntegratorConfig& aConfig
    )
{
    using Algorithm = IntegratorConfig::Algorithm;

    switch(aConfig.mAlgorithm)
    {
    case Algorithm::kNormal:
        return new NormalIntegrator(aCameras, aBuilder, aConfig);
    case Algorithm::kAmbient:
        return new AmbientIntegrator(aCameras, aBuilder, aConfig);
    case Algorithm::kVolumetricPathTracing:
        return new VolumetricPathTracingIntegrator(aCameras, aBuilder, aConfig);
    case Algorithm::kMultiViewPathTracing:
        return new MultiViewPathTracingIntegrator(aCameras, aBuilder, aConfig);
    case Algorithm::kAdaptiveMultiViewPathTracing:
        return new AdaptiveMultiViewPathTracingIntegrator(aCameras, aBuilder, aConfig);
    case Algorithm::kMISMultiViewPathTracing:
        return new MISMultiViewPathTracingIntegrator(aCameras, aBuilder, aConfig);
    default:
        return nullptr;
    };
}
