#ifndef _ABSTRACT_INTEGRATOR_H
#define _ABSTRACT_INTEGRATOR_H

#include <omp.h>
#include <unordered_set>

#include "renderer/camera.h"
#include "renderer/builder.h"
#include "renderer/world.h"
#include "renderer/sampler.h"
#include "renderer/timer.h"
#include "renderer/bsdf/bsdfutils.h"

static const std::string algorithmNames[] = {
    "normal",
    "ambient occlusion",
    "volumetric path tracing",
    "multi view path tracing",
    "adaptive multi view path tracing",
    "mis multi view path tracing",
};

static const std::string algorithmAcronyms[] = {
    "nor",
    "ao",
    "vpt",
    "mvpt",
    "amvpt",
    "mismvpt",
};

struct IntegratorConfig
{
    enum Algorithm
    {
        kNormal=0,
        kAmbient,
        kVolumetricPathTracing,
        kMultiViewPathTracing,
        kAdaptiveMultiViewPathTracing,
        kMISMultiViewPathTracing,
        kAlgorithmMax
    };

    static std::string GetName(Algorithm aAlgorithm)
    {
        if(aAlgorithm < 0 || aAlgorithm >= kAlgorithmMax)
        {
            int value = aAlgorithm;
            return "unknown algorithm";
        }
        return algorithmNames[aAlgorithm];
    }

    static std::string GetAcronym(Algorithm aAlgorithm)
    {
        if(aAlgorithm < 0 || aAlgorithm >= kAlgorithmMax)
            return "unknown";
        return algorithmAcronyms[aAlgorithm];
    }

    static Algorithm GetAlgorithm(std::string aAcronym)
    {
        for(int i = kNormal; i < kAlgorithmMax; ++i)
            if( algorithmAcronyms[i] == aAcronym )
                return Algorithm(i);
        return kAlgorithmMax;
    }

    Algorithm   mAlgorithm;
    std::string mOutputName;
    std::string mFilePath;
    bool        mUseColors               = false;
    bool        mUseTextures             = false;
    int         mMaxPathLength           = -1;
    int         mMinPathLength           = 0;
    int         mBaseSeed                = 1234;
    bool        mUseAdaptive             = false;
    int         mAdaptiveMaxSamples      = 32;
    int         mAdaptiveMinSamples      = 1;
    float       mSampleLimit             = -1;
    float       mTimeLimit               = -1;
    bool        mViewer                  = true;
    bool        mRoughnessClamping       = true;
    bool        mSave                    = true;
    bool        mSavePartial             = false;
    bool        mLayers                  = false;
    bool        mReuseLimit              = false;
    bool        mSavePng                 = true;
    bool        mSaveSamples             = false;
    bool        mSaveNatives             = false;
    std::unordered_set<int> mPartialTimes;
    std::unordered_set<int> mPartialSamples;

    // mvpt specific
    bool        mUseMixture              = true;
    bool        mUseSelection            = true;
    bool        mUseSelectionProbability = false;

    void Print() const
    {
        printf("\nintegrator:\n");
        printf("    algorithm           %s (%s)\n", GetName(mAlgorithm).c_str(), GetAcronym(mAlgorithm).c_str());
        printf("    adaptive            %s\n", ( mUseAdaptive ? "On" : "Off" ));
        printf("    textures            %s\n", ( mUseTextures ? "On" : "Off" ));
        printf("    time limit          %d (s)\n", int(mTimeLimit));
        printf("    sample limit        %d (spp)\n", int(mSampleLimit));
        if( mPartialTimes.size() )
        {
            printf("    partial times       ");
            for( auto i : mPartialTimes )
                printf("%d ", i);
            printf("\n");
        }
        if( mPartialSamples.size() )
        {
            printf("    partial samples     ");
            for( auto i : mPartialSamples )
                printf("%d ", i);
            printf("\n");
        }
        // integrator specific print
        if(
            mAlgorithm == kMultiViewPathTracing ||
            mAlgorithm == kAdaptiveMultiViewPathTracing ||
            mAlgorithm == kMISMultiViewPathTracing
        )
        {
            printf("    selection             %s\n", ( mUseSelection ? "On" : "Off" ));
            printf("    selection probability %s\n", ( mUseSelectionProbability ? "On" : "Off" ));
            printf("    mixture               %s\n", ( mUseMixture ? "On" : "Off" ));
        }
        printf("\n");
    }
};

struct Tile {
    const unsigned int camId;
    const unsigned int txmin, txmax;
    const unsigned int tymin, tymax;
};

class AbstractIntegrator
{
public:
    std::vector<Camera>&  cameras;
    Builder&              mBuilder;
    IntegratorConfig      config;
    int                   mNumThreads;
    std::vector<Sampler>  mSamplers;
    Timer                 mTimer;
    int                   mTileSize = 64;
    std::vector<Tile>     mTiles;
    bool                  mStop = false;
    int                   mTile = -1; //used by volumetricfeaturesintegrator
    int                   mIteration = -1;
    float                 mMeanError = 0; //used by amvpt;
    float                 mCount = 0; //used by amvpt

public:
    AbstractIntegrator(
        std::vector<Camera>& aCameras,
        Builder& aBuilder,
        const IntegratorConfig& aConfig
    );
    virtual ~AbstractIntegrator(){}

    void InitSamplers();
    void InitTiles();

    virtual void Render() = 0;
    virtual int GetTile() const { return mTile; }
    virtual int GetTileSize() const { return mTileSize; }
    virtual int GetIteration() const { return mIteration; }
    virtual bool IsRendering();
    virtual void Quit();
    virtual void Progress() const;
    virtual void SaveRenders() const;

    static AbstractIntegrator* CreateIntegrator(
        std::vector<Camera>& aCameras,
        Builder& aBuilder,
        const IntegratorConfig& aConfig
    );
};


#endif
