#include "ambientintegrator.h"

AmbientIntegrator::AmbientIntegrator( 
    std::vector<Camera>&      aCameras, 
    Builder&                  aBuilder, 
    const IntegratorConfig&   aConfig 
) : AbstractIntegrator(aCameras, aBuilder, aConfig) {}

void AmbientIntegrator::Render()
{
    int mNumThreads = omp_get_max_threads();
    std::vector<Sampler> mSamplers(mNumThreads);

    // while integrator is not notified to stop
    mTimer.reset();
    while( !mStop )
    {
        // loop over cameras
        for( int CamID = 0, CameraNumber = cameras.size(); CamID < CameraNumber; ++CamID )
        {
            int width = cameras[CamID].width;
            int height = cameras[CamID].height;
            Camera& camera = cameras[CamID];

            #pragma omp parallel
            {      
                // sampler selection
                Sampler& sampler = mSamplers[omp_get_thread_num()]; 

                // render pass 
                #pragma omp for schedule(dynamic, 8)
                for( int px = 0; px < width; ++px )
                for( int py = 0; py < height; ++py )
                {
                    CameraSample cs = camera.sample(sampler, px, py, true);
                    Ray& ray = cs.ray;
                    vec2& pixel = cs.filmPosition;

                    Color color = Black();
                    if( mBuilder.primary( ray ) )
                    {   
                        // path vertex data
                        Interpolation interpolation = mBuilder.interpolate( ray );
                        Vector hitn = interpolation.n;
                        
                        // p = triangle.point(u, v) more precise than p = o + t * d
                        // Point hitp = interpolation.p;
                        Point dp = interpolation.dp;

                        // local orthonormal 
                        Frame frame(hitn);

                        // sample direction
                        Vector wi_local = SampleCosineWeightedHemisphere( sampler.sample(), sampler.sample() );
                        float pdf = PdfCosineWeightedHemisphere(wi_local);
                        Vector wi_world = frame.ToWorld(wi_local);

                        // shadow ray
                        Ray shadow( dp, dp + wi_world * 4, ray.time() );
                        float cos_theta = dot(hitn, shadow.d()); 
                        bool visible = cos_theta > 0 && mBuilder.visible( shadow );

                        if( visible )
                        {
                            color = Color(cos_theta, cos_theta, cos_theta);

                            if( config.mUseColors )
                            {
                                const Material& material = mBuilder.get_material( ray.hitobject(), ray.hitid() );
                                Color diffuse = mBuilder.diffuse( ray , interpolation );
                                color = color * diffuse;
                            }
                        }

                    }
                    camera.film.splat(pixel, color);
                }
            }
        }
    }
}
