#include "mismultiviewpathtracingintegrator.h"
#include <iomanip>

// #define DOUBLE_SIDED_FACES

MISMultiViewPathTracingIntegrator::MISMultiViewPathTracingIntegrator(
    std::vector<Camera>& aCameras,
    Builder& aBuilder,
    const IntegratorConfig& aConfig
) : VolumetricPathTracingIntegrator(aCameras, aBuilder, aConfig) {}

#define BSDF_ONLY 0                 // only bsdf sources sampling                       - state OK
#define LIGHT_ONLY 0                // only light sources sampling                      - state OK
#define DIRECT_MIS_2_STRATEGIES 1   // MIS light sources and bsdf sampling              - state OK
#define DIRECT_MIS_3_STRATEGIES 0   // MIS light sources, bsdf and equiangular sampling - state TODO

void MISMultiViewPathTracingIntegrator::Render()
{
    InitSamplers();
    InitTiles();
    const unsigned int nbCameras = cameras.size();

    // init layers for debug and savings
    if(config.mLayers)
        for(unsigned int i = 0; i < nbCameras; ++i)
            cameras[i].film.init_layers(nbCameras);

    // set unnormalized film
    for(unsigned int i = 0; i < nbCameras; ++i)
        cameras[i].film.unset_normalized();

    // multi view prefix and visibility data
    std::vector<std::vector<PathPrefixData>> threadsPrefixData(mNumThreads);
    std::vector<std::vector<int>> threadsVisibleData(mNumThreads);

    for(auto& prefixData : threadsPrefixData)
        prefixData.resize(nbCameras);

    for(auto& visibleData : threadsVisibleData)
        visibleData.reserve(nbCameras);

    // reset timer and iteration count
    mTimer.reset();

    mIteration = 0;

    #ifndef SINGLE_THREAD
    #pragma omp parallel shared(mSamplers, mTiles, threadsPrefixData, threadsVisibleData)
    #endif
    {
        // int iteration = 0;

        // sampler selection
        Sampler& sampler = mSamplers[omp_get_thread_num()];

        // thread local data for multi view integration 
        std::vector<PathPrefixData>& prefixData = threadsPrefixData[omp_get_thread_num()];
        std::vector<int>& visibleData = threadsVisibleData[omp_get_thread_num()];

        // while integrator is not notified to stop
        while( !mStop )
        {
            // loop over cameras tiles
            #ifndef SINGLE_THREAD
            #ifdef USE_OMP_LOCKS
            //#pragma omp for schedule(dynamic, 1) nowait
            #pragma omp for schedule(static, 1)
            #else
            #pragma omp for schedule(dynamic, 1)
            #endif
            #endif
            for(unsigned int tileId = 0; tileId < mTiles.size(); tileId++)
            {
                const Tile& tile = mTiles[tileId];
                const unsigned int CamID = tile.camId;
                Camera& camera = cameras[CamID];

                for( unsigned int px = tile.txmin; px < tile.txmax; ++px )
                for( unsigned int py = tile.tymin; py < tile.tymax; ++py )
                {
                    // clear temporary data
                    std::fill(prefixData.begin(), prefixData.end(), PathPrefixData());
                    visibleData.clear();

                    // sample camera ray
                    prefixData[CamID].cameraSample = camera.sample(sampler, px, py, true);

                    // init path data
                    int pathLength = 0;

                    ////////////// first hit sampling //////////////

                    // references to sampled camera prefix data
                    Ray& ray = prefixData[CamID].cameraSample.ray;
                    MediumRecords& records = prefixData[CamID].records;
                    std::vector<MediumSample>& mediumSamplesChain = prefixData[CamID].mediumSamplesChain;
                    prefixData[CamID].pdf = prefixData[CamID].cameraSample.pdf;
                    prefixData[CamID].cameraPdf = prefixData[CamID].cameraSample.pdf;
                    prefixData[CamID].weight = prefixData[CamID].cameraSample.weight;
                    visibleData.push_back( CamID );
                    float sumImageWeights = prefixData[CamID].imageWeight;
                    //float sumSimilarity = 1;


                    // useless loop to keep using break in the code
                    for(;;++pathLength)
                    {
                        // gather medium sections and geometry intersection
                        mBuilder.intersect( ray, records, true );

                        bool hitSurface = ray.hit();
                        bool hitMedium = records.size();

                        if( !hitMedium && !hitSurface )
                            break;

                        // interaction data
                        Point hitp;
                        Vector hitn, wo;
                        vec2 uv;
                        float eps = 0;
                        Frame frame;
                        const Bsdf * bsdf = nullptr;
                        const Medium * medium = nullptr;

                        // sample medium interaction
                        const bool realInteraction = SampleChainMediumRecordsMiller(sampler, records, mediumSamplesChain);
                        // if ( !hitMedium && !hitSurface )
                            // break;

                        const MediumSample& mediumSample = mediumSamplesChain.back();
                        // if( mediumSample.pdf <= 0)
                            // break;
                        if( !mediumSample.w.valid() || mediumSample.w.power() <= 0 )
                            break;

                        // hit medium
                        if( realInteraction )
                        {
                            hitSurface = false;

                            // assert(mediumSample.index >= 0 && mediumSample.index < records.size() );

                            // update path throughput
                            // pathWeight *= mediumSample.w;
                            // prefixData[CamID].weight *= mediumSample.w;

                            // event infos
                            hitp = mediumSample.p;
                            eps = Epsilon(mediumSample.p, mediumSample.t);
                            hitn = Vector(0);
                            medium = records.mediums[mediumSample.index];
                            bsdf = medium->phaseFunction(hitp);
                            frame = Frame(-ray.nd());
                            wo = Vector(0, 0, 1); //frame.ToLocal(-ray.nd()); // wo has to be in local coordinates system
                            uv = vec2(0,0);

                            // absorbtion + emission : accounted for later to allow reprojection
                            //if( mediumSample.type == MediumSample::InteractionType::Absorbtion )
                        }
                        else 
                        {
                            if(!hitSurface)
                               break;
                        }

                        // update prefix pdf with the probability of hitting a particle in case of medium hit,
                        // or the combined null collision pdf in case we hit the surface through the medium 
                        prefixData[CamID].mediumPdf = mediumSample.pdf;
                        prefixData[CamID].pdf *= mediumSample.pdf;
                        prefixData[CamID].weight *= mediumSample.w;

                        if( hitSurface )
                        {
                            // hit surface
                            hitMedium = false;

                            // interpolation of local data
                            Interpolation i = mBuilder.interpolate(ray);

                            // event infos
                            hitp = i.p;
                            eps = Epsilon(hitp, ray.hitt());
                            hitn = i.n;
#ifdef DOUBLE_SIDED_FACES
                            if( dot(hitn, -records.ray.d()) < 0 )
                                hitn = -hitn;
#endif
                            medium = ray.medium;
                            bsdf = mBuilder.get_bsdf( ray.hitobject(), ray.hitid() );
                            frame = Frame(hitn);
                            wo = frame.ToLocal(-ray.nd());
                            uv = i.t;

                            // changes epsilon side depending on whether the ray is inside the object or outside
                            bool inside = wo.z < 0;
                            if( inside )
                                eps = -eps;

                            // directly hit some light, lights do not reflect, 
                            // and we do not reuse directly hit light
                            const Material& material = mBuilder.get_material( ray.hitobject(), ray.hitid() );
                            if( material.emission.power() > 0 )
                            {
                                if(pathLength < config.mMinPathLength)
                                    break;

#ifndef DOUBLE_SIDED_FACES
                                // if light is pointing towards the camera
                                float cosThetaS = std::max(0.f, dot(hitn, -ray.nd()));
                                if( cosThetaS <= 0 )
                                    break;
#endif

                                prefixData[CamID].color += prefixData[CamID].weight * material.emission;
                                break;
                            }
                        }

                        if( config.mMaxPathLength >= 0 && pathLength >= config.mMaxPathLength )
                            break;

                        // Update local outgoing direction
                        prefixData[CamID].wo = wo;
                        // slightly move the point away from the surface to avoid self intersections
                        Point dp = hitMedium ? hitp : hitp + eps * hitn;

                        // float cos_theta = hitMedium ? 1 : std::abs(prefixData[CamID].wo.z);
                        float cos_theta = hitMedium ? 1 : std::abs(dot(prefixData[CamID].cameraSample.ray.nd(), hitn));
                        float d2 = distance2(hitp, ray.o());
                        prefixData[CamID].pdf *= cos_theta / d2;
                        prefixData[CamID].geoTermPdf = cos_theta / d2;

                        // required for MIS
                        prefixData[CamID].imageWeight = 1.f;
                        sumImageWeights = prefixData[CamID].imageWeight;

                        // Select a subset of camera prefixes that can see / sample the hit point
                        if( mediumSample.type == MediumSample::InteractionType::Absorbtion || (bsdf && !bsdf->isDelta()) )
                        {
                            // Loop over opened cameras
                            for( unsigned int k = 0; k < nbCameras; ++k )
                            {
                                // don't process sampled viewer
                                if( k == CamID )
                                    continue;

                                // if camera is not opened continue
                                if( !cameras[k].in_time_window(ray.time()) )
                                    continue;

                                // connect camera k to hitpoint
                                prefixData[k].cameraSample = cameras[k].connect(hitp, prefixData[CamID].cameraSample.lensSample, ray.time());

                                // check geometry orientation and camera orientation
                                if( prefixData[k].cameraSample.pdf <= 0 )
                                    continue;

                                prefixData[k].pdf = prefixData[k].cameraSample.pdf;

                                // update local outgoing direction
                                prefixData[k].wo = frame.ToLocal(-prefixData[k].cameraSample.ray.nd());

                                // if we hit a surface and outgoing directions are not on the same hemisphere reject projection
                                if( !hitMedium && prefixData[k].wo.z * prefixData[CamID].wo.z <= 0 )
                                    continue;

                                if( config.mUseSelection )
                                {
                                    // compute similarity between observers w.r.t the BSDF
                                    // - the similarity has to be normalized between 0 (different) and 1 (equal distributions)
                                    // - the similarity has to be symmetric S(0 -> 1) = S(1 -> 0);
                                    float similarity = bsdf->TotalVariationApprox(prefixData[CamID].wo, prefixData[k].wo, uv);

                                    // if( bsdf->isObserverDependent() )
                                        // printf("similarity %f\n", similarity);

                                    // Stochastic selections between the initial prefix and the others.
                                    if( similarity < 1.f && sampler.get1D() > similarity )
                                        continue;

                                    // store probability
                                    prefixData[k].alpha = similarity;
                                }

                                // shadow ray
                                Ray shadow( prefixData[k].cameraSample.ray.o(), dp, ray.time() );
                                // shadow.medium = medium; 
                                mBuilder.intersect( shadow, prefixData[k].records, false );
                                if( shadow.hit() )
                                    continue;

                                std::vector<MediumSample>& shiftedMediumSamplesChain = prefixData[k].mediumSamplesChain;
                                float shiftedChainWeight = 1;
                                float J = 1;
                                const bool validShift = ShiftMediumRecordsChain(
                                    sampler, 
                                    records, 
                                    mediumSamplesChain, 
                                    prefixData[k].records, 
                                    shiftedMediumSamplesChain, 
                                    shiftedChainWeight,
                                    J
                                );

                                if( !validShift ) continue;
                                if( std::isnan(shiftedChainWeight) ) continue;
                                if( std::isinf(shiftedChainWeight) ) continue;
                                if( shiftedChainWeight <= 0 ) continue;

                                const MediumSample& msproj = shiftedMediumSamplesChain.back();

                                if( !msproj.w.valid() ) continue;
                                if( msproj.pdf <= 0 ) continue;

                                // update weight with camera sample jacobian
                                // float cos_theta = hitMedium ? 1 : std::abs(prefixData[k].wo.z);
                                float cos_theta = hitMedium ? 1 : std::abs(dot(prefixData[k].cameraSample.ray.nd(), hitn));
                                float d2 = distance2(hitp, prefixData[k].cameraSample.ray.o());
                                prefixData[k].geoTermPdf = cos_theta / d2;
                                prefixData[k].pdf = prefixData[k].cameraSample.pdf * cos_theta / d2;

                                // jacobian with geometric terms
                                float pdfK = prefixData[k].cameraSample.pdf * prefixData[k].geoTermPdf;
                                float pdfCamId = prefixData[CamID].cameraSample.pdf * prefixData[CamID].geoTermPdf;
                                // update Jacobian with random replayed pdf
                                shiftedChainWeight *= pdfK / pdfCamId;

                                // contrib as evaluated from camera K : f_k(x)/p_k(x)
                                prefixData[k].weight = msproj.w * prefixData[k].cameraSample.weight;

                                // new MIS 
                                // shiftedChainWeight = pk / pi
                                // sumImageWeights = sum_j pj / pi
                                prefixData[k].imageWeight = shiftedChainWeight;

                                if( config.mUseSelection && config.mUseSelectionProbability )
                                {
                                    // this factor is the probability of connecting to the camera k 
                                    // w.r.t the bsdf of the initial camera
                                    // it should multipy with the jacobian 
                                    // as a part of the relative number of samples for strategy k 

                                    // taking into account this rejection probability
                                    // seems to add artifacts in MVPT - MWIS.
                                    // removing it seems to better converge

                                    prefixData[k].imageWeight /= prefixData[k].alpha;
                                }

                                sumImageWeights += prefixData[k].imageWeight;

                                // set the prefix as visible
                                visibleData.push_back( k );
                            } // end camera selection
                        }

                        // absorbtion + isotropic emission
                        if( mediumSample.type == MediumSample::InteractionType::Absorbtion )
                        {
                            const Color Le = medium->getLe(mediumSample.p, ray.time());
                            for(int k : visibleData)
                            {
                                prefixData[k].color += prefixData[k].weight * Le * 4.f * (float)M_PI ;
                            }
                            break;
                        }

                        // Next event estimation
                        if( bsdf && !bsdf->isDelta() && pathLength + 1 >= config.mMinPathLength && !mBuilder.lights.empty() )
                        {
                            // direct light sample
                            LightSample light = mBuilder.lights.sample( vec2(sampler.sample(), sampler.sample()) );

                            // slightly move the point away from the surface to avoid self intersections
                            Point ds = light.s + Epsilon(light.s, distance(light.s, hitp)) * light.sn;
                            // shadow ray
                            Ray shadow( dp, ds, ray.time() );
                            // shadow ray spawn in the current medium
                            shadow.medium = medium;
                            mBuilder.intersect( shadow, records, false );

                            // strategy 1 : direct light sample
                            if( !shadow.hit() )
                            {
                                Vector wi_world = normalize(light.s - hitp);
                                Vector wi = frame.ToLocal(wi_world);
                                float cos_theta = hitMedium ? 1.f : std::max(0.f, wi.z);
                                float cos_theta_s = std::max(0.f, -dot(light.sn, wi_world));
                                float g = cos_theta_s / distance2(hitp, light.s);

                                if( cos_theta * g > 0 )
                                {
                                #if LIGHT_ONLY
                                    // NO direct MIS weight
                                    Color transmittance = White();
                                    EvalTransmittanceMiller(sampler, records, transmittance);
                                    float wmis = 1;
                                    // common quantities
                                    Color common = wmis * transmittance * light.emission * cos_theta * g / light.pdf;
                                    // direct contribution 1
                                    for(int k : visibleData)
                                    {
                                        prefixData[k].color += prefixData[k].weight * bsdf->eval(wi, prefixData[k].wo, uv) * common;
                                    }
                                #elif DIRECT_MIS_2_STRATEGIES
                                    // WIP add direct MIS weights with Miller framework
                                    // Here we have 2 strategies : 
                                    // - sample depth with delta tracking and find light after bsdf sampling                (DT + BSDF + DT)
                                    // - sample depth with delta tracking and connect to light with explicit light sampling (DT + RT + LIGHT)
                                    Color transmittance = White();
                                    float ratio_DT_RT;
                                    EvalTransmittanceMiller(sampler, records, transmittance, &ratio_DT_RT);

                                    // compute mixture bsdf pdf with alpha uniform (1/k)
                                    float bsdf_pdf = 0;
                                    if( config.mUseMixture )
                                    {
                                        // compute mixture bsdf pdf with alpha uniform (1/k)
                                        float sum_pdf = 0, count = 0;
                                        for( int k : visibleData )
                                        {
                                            float pdf = bsdf->pdf(wi, prefixData[k].wo, uv);
                                            sum_pdf += pdf;
                                            count += pdf > 0 ? 1 : 0;
                                        }
                                        bsdf_pdf = count > 0 ? sum_pdf / count : 0;
                                    }
                                    else
                                    {
                                        bsdf_pdf = bsdf->pdf(wi, prefixData[CamID].wo, uv);
                                    }

                                    // MIS weight 
                                    // w = pdf1 / (pdf1 + pdf2) = 1 / (1 + pdf2/pdf1)
                                    // pdf2 / pdf1 = (pdf_DT1 * pdf_BSDF * pdf_DT2 * G) / (pdf_DT1 * pdf_RT2 * pdf_LIGHT)
                                    //             = (pdf_BSDF * G / pdf_LIGHT) * (pdf_DT2 / pdf_RT2)
                                    float b = bsdf_pdf * g / light.pdf * ratio_DT_RT;
                                    float wmis = 1 / (1+b);
                                    // common quantities
                                    Color common = wmis * transmittance * light.emission * cos_theta * g / light.pdf;
                                    // direct contribution 1
                                    for(int k : visibleData)
                                    {
                                        const float pjOverPi = prefixData[k].imageWeight / prefixData[CamID].imageWeight;
                                        const float misWeight = prefixData[CamID].imageWeight / sumImageWeights;
                                        prefixData[k].color += misWeight * pjOverPi * prefixData[k].weight * bsdf->eval(wi, prefixData[k].wo, uv) * common;
                                    }
                                #elif DIRECT_MIS_3_STRATEGIES
                                    // TODO add direct MIS weights with Miller framework and equiangular sampling
                                    // Here we have 3 strategies : 
                                    // - sample depth with delta tracking and find light after bsdf sampling                                    (DT + BSDF + DT)
                                    // - sample depth with delta tracking and connect to light with explicit light sampling and ratio tracking  (DT + RT + LIGHT)
                                    // - sample light with explicit light sampling and sample depth with equiangular sampling 
                                    //   then connect from obs with RT and connect to light with RT                                             (RT + EQUI + RT + LIGHT)
                                #endif
                                }
                            }
                        }

                        // BSDF sampling and continue random walk
                        {
                            // bsdf sample from the mixture of all observers BSDF
                            int pick = sampler.sample() * visibleData.size();
                            int pickid = config.mUseMixture ? visibleData[pick] : CamID;

                            // sample bsdf from the picked observer
                            BsdfSample bsdfSample = bsdf->sample(sampler, prefixData[pickid].wo, uv);
                            if( bsdfSample.pdf == 0 || length2(bsdfSample.wi) == 0 )
                                break;

                            // compute mixture pdf with alpha uniform (1/k)
                            float bsdf_pdf = 0;
                            if( config.mUseMixture )
                            {
                                // compute mixture bsdf pdf with alpha uniform (1/k)
                                float sum_pdf = 0, count = 0;
                                for( int k : visibleData )
                                {
                                    float pdf = bsdf->pdf(bsdfSample.wi, prefixData[k].wo, uv);
                                    sum_pdf += pdf;
                                    count += pdf > 0 ? 1 : 0;
                                }
                                bsdf_pdf = count > 0 ? sum_pdf / count : 0;
                            }
                            else
                            {
                                bsdf_pdf = bsdfSample.pdf;
                                // reset 
                                sumImageWeights = 0;
                                // update mis weights with bsdf pdf
                                for( int k : visibleData )
                                {
                                    prefixData[k].imageWeight *= bsdf->pdf(bsdfSample.wi, prefixData[k].wo, uv) / bsdf_pdf;
                                    sumImageWeights += prefixData[k].imageWeight;
                                }
                            }

                            Vector wi_world = frame.ToWorld(bsdfSample.wi);

                            if( bsdf_pdf <= 0 )
                                break;

                            // hit medium
                            if( hitMedium )
                            {
                                ray = Ray( hitp, wi_world, ray.time() );
                                ray.medium = medium;
                            }
                            // hit surface
                            else
                            {
                                // the normal is always pointing outwards the objects
                                // hence a negative cosine between the normal and the sampled direction
                                // means the direction is inside an object
                                bool next_inside = bsdfSample.wi.z < 0;

                                // if we should not be inside the object return
                                if( next_inside && !bsdf->isTransmissive() )
                                    break;

                                // select the next crossed medium 
                                medium = next_inside ? 
                                            mBuilder.get_interior_medium(ray.hitobject(), ray.hitid()):
                                            mBuilder.get_exterior_medium(ray.hitobject(), ray.hitid());

                                // we already switched the epsilon sign if the direction wo 
                                // was inside the object. Hence we test if the sampled bsdf 
                                // direction is in the same hemisphere to know if we have 
                                // to reverse the epsilon sign again
                                bool same_hemisphere = bsdfSample.wi.z * prefixData[pickid].wo.z > 0;

                                // slightly move the point away from the surface to avoid self intersections
                                Point dp = same_hemisphere ? hitp + eps * hitn : hitp - eps * hitn; 
                                ray = Ray( dp, wi_world, ray.time() );
                                ray.medium = medium;
                            }

                            // trace ray and gather indirect illumination
                            Color pathWeight = White();
                            Color indirect;
                            if( SampleAndEvalPath(indirect, ray, sampler, pathWeight, records, 1, bsdf_pdf) )
                            {
                                float cos_theta = hitMedium ? 1.f : std::abs(bsdfSample.wi.z);
                                for(int k : visibleData)
                                {
                                    const float pjOverPi = prefixData[k].imageWeight / prefixData[CamID].imageWeight;
                                    const float misWeight = prefixData[CamID].imageWeight / sumImageWeights;
                                    prefixData[k].color += misWeight * pjOverPi * prefixData[k].weight * bsdf->eval(bsdfSample.wi, prefixData[k].wo, uv) * cos_theta / bsdf_pdf * indirect;
                                }
                            }
                        }

                        // quit loop
                        break;
                    }

                    // add samples in respective images and layers 
                    for( int k : visibleData )
                    {
                        // new mis
                        // const float pjOverPi = prefixData[k].imageWeight / prefixData[CamID].imageWeight;
                        // const float misWeight = prefixData[CamID].imageWeight / sumImageWeights;
                        // Color contribution = misWeight * pjOverPi * prefixData[k].color;// * (static_cast<float>(visibleData.size())/nbCameras);
                        const Color& contribution = prefixData[k].color;
                        
                        // shorter version
                        // const float weight = prefixData[k].imageWeight / sumImageWeights;
                        // const float weight = prefixData[k].misWeight / prefixData[k].alpha;
                        // Color contribution = weight * prefixData[k].color;
                        // Color contribution = Color(weight,weight,weight);

                        // l'image est utilisée comme un accumulateur,
                        // donc la normalisation sur N samples demandés 
                        // en parametre est faite a la main ici:
                        // if( config.mSampleLimit > 0 )
                            // contribution /= config.mSampleLimit;

                        //// accumulation a la main dans l'image (plutot que moyenne via la fonction film.add )
                        //if( !contribution.valid() ) continue;
                        //const float sample[3] = { contribution.r, contribution.g, contribution.b };
                        //const float sq_sample[3] = { contribution.r * contribution.r, contribution.g * contribution.g, contribution.b * contribution.b };

                        //// current pixel
                        //PixelPosition pos = cameras[k].film.offset(prefixData[k].cameraSample.filmPosition);
                        //DeepImf& image = cameras[k].film.image; 
                        //DeepImf& sq_image = cameras[k].film.sq_image; 
                        //DeepImf& variance = cameras[k].film.variance; 
                        //DeepImf& error = cameras[k].film.error;

                        //image.get(pos, 3) += 1;
                        //sq_image.get(pos, 3) += 1;
                        //variance.get(pos, 3) += 1;

                        //// update color
                        //for(int channel = 0; channel < 3; ++channel)
                        //{
                        //    image.get(pos, channel) += sample[channel];
                        //    sq_image.get(pos, channel) += sq_sample[channel];
                        //    variance.get(pos, channel) = sq_image.get(pos, channel) - image.get(pos, channel) * image.get(pos, channel);
                        //    error.get(pos, channel) = cameras[k].film.srcs(image.get(pos, channel), sq_image.get(pos, channel), variance.get(pos, 3));
                        //}

                        const auto& pos = prefixData[k].cameraSample.filmPosition;
                        cameras[k].film.add(pos, contribution);
                        if(config.mLayers)
                            cameras[k].film.add(CamID, pos, contribution);

                    }
                    // add native sample in reconstruction image
                    cameras[CamID].film.add_reconstruction(prefixData[CamID].cameraSample.filmPosition, prefixData[CamID].color);
                }
            }

            #ifndef SINGLE_THREAD
            #pragma omp single
            #endif
            {
                mIteration++;
                
                int time = mTimer.sec().count();
                int samples = mIteration;
                if( config.mPartialTimes.size() > 0 )
                {
                    auto it = std::min_element(config.mPartialTimes.begin(), config.mPartialTimes.end());
                    if( time > *it )
                    {
                        SaveRenders();
                        config.mPartialTimes.erase(it);
                    }
                }

                if( config.mPartialSamples.size() > 0 )
                {
                    if( config.mPartialSamples.find(samples) != config.mPartialSamples.end() )
                    {
                        SaveRenders();
                    }
                }

                Progress();

                //#pragma omp critical
                //std::cout << "mIteration " << mIteration << " finished!" << std::endl;
                if( config.mSampleLimit > 0 && mIteration >= config.mSampleLimit )
                    mStop = true;

                if( config.mTimeLimit > 0 && mTimer.sec().count() >= config.mTimeLimit )
                    mStop = true;
            }
        }
    }

    std::cout <<  std::endl << mTimer << std::endl;

    SaveRenders();

    int sec = mTimer.sec().count();
    int min = mTimer.min().count();
    float ti = float(sec) / float(mIteration);
    if( std::isinf(sec) || std::isnan(sec) ) sec = -1;
    if( std::isinf(min) || std::isnan(min) ) min = -1;
    printf("\ntime %d min / %d s %f s/iter", min, sec, ti);
    printf("\ndone.\n");
}

void MISMultiViewPathTracingIntegrator::SaveRenders() const
{
    if( !config.mSave ) return;

#ifdef GK_RELEASE
    int time = (int)mTimer.sec().count();
    int samples = mIteration;
    for (int i = 0; i < (int)cameras.size(); i++)
    {
        std::stringstream ss;
        ss  << config.mFilePath
            << config.mOutputName
            << "-"
            << std::setfill('0')
            << std::setw(3)
            << i
            << "-time-"
            << std::setfill('0')
            << std::setw(6)
            << time
            << "-samples-"
            << std::setfill('0')
            << std::setw(6)
            << samples;

        // normalize required due to MIS
        const double scale = 1. / mIteration;
        //std::cout << "Saving render of view " << i << " at iteration " << mIteration << " with scale " << scale << std::endl;
        cameras[i].film.save2(ss.str(), scale, true, config.mSaveSamples, config.mSavePng );

        if( config.mSaveNatives )
            cameras[i].film.save_reconstruction( ss.str()+std::string("-natives"), 1.f, true );

        if( config.mLayers )
            cameras[i].film.save_layers(ss.str(), 1.f, true);
    }
#endif
}
