#include "volumetricpathtracingintegrator.h"


VolumetricPathTracingIntegrator::VolumetricPathTracingIntegrator(
    std::vector<Camera>& aCameras,
    Builder& aBuilder,
    const IntegratorConfig& aConfig
) : AbstractIntegrator(aCameras, aBuilder, aConfig) {}

void VolumetricPathTracingIntegrator::Render()
{
    InitSamplers();
    InitTiles();
    const unsigned int nbTiles = mTiles.size();

    // reset timer and iteration count
    mTimer.reset();
    mIteration = 0;

    #ifndef SINGLE_THREAD
    //#pragma omp parallel shared(mSamplers, mStop, mIteration)
    #pragma omp parallel shared(mSamplers, mTiles)
    #endif
    {
        // sampler selection
        Sampler& sampler = mSamplers[omp_get_thread_num()];

        // while integrator is not notified to stop
        while( !mStop )
        {
            // loop over cameras tiles
            #pragma omp for schedule(dynamic, 1) nowait
            for(unsigned int tileId = 0; tileId < nbTiles; tileId++)
            {
                const Tile& tile = mTiles[tileId];
                const unsigned int CamID = tile.camId;
                Camera& camera = cameras[CamID];

                // medium records unique allocation 
                MediumRecords records;

                for( unsigned int px = tile.txmin; px < tile.txmax; ++px )
                for( unsigned int py = tile.tymin; py < tile.tymax; ++py )
                {
                    // const PixelPosition pos(cameras[0].film.h-44, 63);
                    // if(pos != PixelPosition(py,px))
                        // continue;
                    // sampler.index((py*camera.width+px)*mIteration, mIteration);

                    // sample camera ray
                    CameraSample cameraSample = camera.sample(sampler, px, py);

                    // init path data
                    Color pathWeight = cameraSample.weight;
                    Color color;
                    bool validPath = SampleAndEvalPath(color, cameraSample.ray, sampler, pathWeight, records);

                    // splat sample 
                    // camera.film.splat(cameraSample.filmPosition, color);
                    if( validPath )
                        camera.film.add(cameraSample.filmPosition, color);
                }
            }
            
            #ifndef SINGLE_THREAD
            #pragma omp single
            #endif
            {
                mIteration++;
                
                int time = mTimer.sec().count();
                int samples = mIteration;
                if( config.mPartialTimes.size() > 0 )
                {
                    auto it = std::min_element(config.mPartialTimes.begin(), config.mPartialTimes.end());
                    if( time > *it )
                    {
                        SaveRenders();
                        config.mPartialTimes.erase(it);
                    }
                }

                if( config.mPartialSamples.size() > 0 )
                {
                    if( config.mPartialSamples.find(samples) != config.mPartialSamples.end() )
                    {
                        SaveRenders();
                    }
                }

                Progress();
            }
            
            if( config.mSampleLimit > 0 && mIteration >= config.mSampleLimit )
                mStop = true;

            if( config.mTimeLimit > 0 && mTimer.sec().count() >= config.mTimeLimit )
                mStop = true;

            #ifndef USE_OMP_LOCKS //see film.h
            #pragma omp barrier
            #endif
        }
    }

    SaveRenders();

    int sec = mTimer.sec().count();
    int min = mTimer.min().count();
    float ti = float(sec) / float(mIteration);
    if( std::isinf(sec) || std::isnan(sec) ) sec = -1;
    if( std::isinf(min) || std::isnan(min) ) min = -1;
    printf("\ntime %d min / %d s %f s/iter", min, sec, ti);
    printf("\ndone.\n");

    // PixelPosition pos(cameras[0].film.h-44, 63);
    // float r = cameras[0].film.image.get(pos, 0);
    // float g = cameras[0].film.image.get(pos, 1);
    // float b = cameras[0].film.image.get(pos, 2);
    // printf("pixel (%d,%d,%d) value (%.24f,%.24f,%.24f)\n", 0, cameras[0].film.h-44, 63, r,g,b);
}

// #define DOUBLE_SIDED_FACES

#define BSDF_ONLY 0                 // only bsdf sources sampling                       - state OK
#define LIGHT_ONLY 0                // only light sources sampling                      - state OK
#define EA_ONLY 0                   // only equiangular sampling                        - state WIP
#define DIRECT_MIS_LIGHT_BSDF 1     // MIS light sources and bsdf sampling              - state OK
#define DIRECT_MIS_LIGHT_EA 0       // MIS light sources and equiangular sampling       - state WIP
#define DIRECT_MIS_EA_BSDF 0        // MIS light sources and bsdf sampling              - state TODO - but painful
#define DIRECT_MIS_ALL 0            // MIS light sources, bsdf and equiangular sampling - state TODO - but painful

bool VolumetricPathTracingIntegrator::SampleAndEvalPath(
    Color& color,
    Ray& ray,
    Sampler& sampler,
    Color& pathWeight,
    MediumRecords& records,
    int pathLength,
    float lastPdfW,
    float last_ratio_DT_RT,
    float last_pdf_EA
    )
{
    color = Black();
    Color pathWeightOld;
    MediumRecords shadow_records;

    // unidirectional path construction loop 
    for(;; ++pathLength)
    {
        // gather medium sections and geometry intersection
        mBuilder.intersect( ray, records, (pathLength == 0 ? true : false) );

        bool hitSurface = ray.hit();
        bool hitMedium = records.size();

        if( !hitMedium && !hitSurface )
            break;

        // interaction data
        Point hitp;
        Vector hitn, wo;
        vec2 uv;
        float eps;
        Frame frame;
        const Bsdf * bsdf = nullptr;
        const Medium * medium = nullptr;

        // #1 - Equiangular sampling
        // if( records.size() && pathLength + 1 >= config.mMinPathLength && !mBuilder.lights.empty()  )
        // {
        //     color += pathWeight * SampleEquiangularDirect(sampler, records);
        // }

        // #2 - Ratio Tracking 
        float ratio_DT_RT1 = 1; // computed for direct mis weights
        MediumSample mediumSample;
        hitMedium = SampleMediumRecordsMiller(sampler, records, mediumSample, &ratio_DT_RT1);

        // std::vector<MediumSample> chain;
        // hitMedium = SampleChainMediumRecordsMiller(sampler, records, chain);
        // const MediumSample& mediumSample = chain.back();

        if( !mediumSample.w.valid() || mediumSample.w.power() <= 0 )
            break;

        // update path throughput
        pathWeightOld = pathWeight;
        pathWeight *= mediumSample.w;

        if( hitMedium )
        {
            medium = records.mediums[mediumSample.index];

            // absorbtion + isotropic emission
            if( mediumSample.type == MediumSample::InteractionType::Absorbtion )
            {
                color += pathWeight * medium->getLe(mediumSample.p, ray.time()) * 4.f * (float)M_PI;
                break;
            }
            // hit medium
            hitSurface = false;

            // event infos
            hitp = mediumSample.p;
            eps = Epsilon(mediumSample.p, mediumSample.t);
            hitn = Vector(0);
            bsdf = medium->phaseFunction(hitp);
            frame = Frame(-records.ray.nd());
            wo = Vector(0, 0, 1); //frame.ToLocal(-records.ray.nd()); // wo has to be in local coordinates system
            uv = vec2(0,0);
        }
        else if(!hitSurface)
            break;

        if( hitSurface )
        {
            // hit surface
            hitMedium = false;

            // interpolation of local data
            Interpolation i = mBuilder.interpolate(ray);

            // event infos
            hitp = i.p;
            eps = Epsilon(hitp, ray.hitt());
            hitn = i.n;

#ifdef DOUBLE_SIDED_FACES
            if( dot(hitn, -records.ray.d()) < 0 )
                hitn = -hitn;
#endif

            medium = ray.medium;
            bsdf = mBuilder.get_bsdf( ray.hitobject(), ray.hitid() );
            frame = Frame(hitn);
            wo = frame.ToLocal(-records.ray.nd()); // wo has to be in local coordinates system
            uv = i.t;

            // changes epsilon side depending on whether the ray is inside the object or outside
            bool inside = wo.z < 0;
            if( inside )
                eps = -eps;

            // directly hit some light, lights do not reflect
            const Color emission = mBuilder.get_emission( ray.hitobject(), ray.hitid(), uv );
            if( emission.power() > 0 )
            {
                if(pathLength < config.mMinPathLength)
                    break;

#ifndef DOUBLE_SIDED_FACES
                // if light is pointing towards the camera
                float cosThetaS = std::max(0.f, dot(hitn, -ray.nd()));
                if( cosThetaS <= 0.f )
                    break;
#else
                float cosThetaS = std::abs(dot(hitn, -ray.nd()));
#endif

                // no MIS possible for direct (zero order) illumination
                if( pathLength == 0 )
                {
                    color += pathWeight * emission;
                    break;
                }

            #if BSDF_ONLY
                color += pathWeight * emission;
            // #elif LIGHT_ONLY
                // do nothing
            // #elif EA_ONLY
                // do nothing
            #elif DIRECT_MIS_LIGHT_BSDF
                // - bsdf sampling      (DT1 + BSDF + DT2)
                // - light sampling     (DT1 + RT2 + LIGHT)
                // MIS weight 
                // w = pdf1 / (pdf1 + pdf2) = 1 / (1 + pdf2/pdf1) = 1 / (1 + a)
                // a = pdf2 / pdf1 = (pdf_DT1 * pdf_RT2 * pdf_LIGHT) / (pdf_DT1 * pdf_BSDF * pdf_DT2 * G)
                //                 = (pdf_LIGHT / pdf_BSDF * G) * (pdf_RT2 / pdf_DT2)
                //
                float bsdf_pdf = lastPdfW * cosThetaS / distance2(hitp, records.ray.o());
                float light_pdf = mBuilder.lights.eval();
                float a = light_pdf / bsdf_pdf / ratio_DT_RT1;        
                float w = 1 / (1 + a);
                color += pathWeight * w * emission;
            #elif DIRECT_MIS_EA_BSDF
                // - bsdf sampling          (DT1 + BSDF + DT2)
                // - equiangular sampling   (RT1 + EQUI + RT2 + LIGHT)
            #elif DIRECT_MIS_ALL
                // - bsdf sampling          (DT1 + BSDF + DT2)
                // - light sampling         (DT1 + RT2 + LIGHT)
                // - equiangular sampling   (RT1 + EQUI + RT2 + LIGHT)
            #endif
                break;
            }
        }

        if( config.mMaxPathLength >= 0 && pathLength >= config.mMaxPathLength )
            break;

        // update last pdfs values for direct mis weight with equiangular
        // if( DIRECT_MIS_EA_BSDF || DIRECT_MIS_ALL )
        // {
        //     last_ratio_DT_RT = ratio_DT_RT1;
        //     last_pdf_EA = PdfEquiangular(sampler, light, records, distance(ray.o(), hitp));
        // } 

        // Next event estimation
        if( bsdf && !bsdf->isDelta() && pathLength + 1 >= config.mMinPathLength && !mBuilder.lights.empty() )
        {
            // direct light sample
            LightSample light = mBuilder.lights.sample( vec2{sampler.sample(), sampler.sample()} );
            // ca c'est faux si les records sont + grand que l'evenement echantillonné
            // j'ai peut etre ecrit des betises ...
            float pdf_EA = hitMedium ? PdfEquiangular(light, records, distance(hitp, records.ray.o())) : 0; 

            // shadow ray
            // slightly move the point away from the hit point surface to avoid self intersections
            Point dp =  hitMedium ? hitp : hitp + eps * hitn;
            // slightly move the point away from the light source surface to avoid self intersections
            Point ds = light.s + Epsilon(light.s, distance(light.s, hitp)) * light.sn;
            Ray shadow( dp, ds, ray.time() );
            shadow.medium = medium;
            mBuilder.intersect( shadow, shadow_records, false );

            // strategy 1 : direct light sample
            if( !shadow.hit() )
            {
                Vector wi_world = normalize(light.s - hitp);
                Vector wi = frame.ToLocal(wi_world);
                float cos_theta = hitMedium ? 1.f : std::max(0.f, wi.z);
                float cos_theta_s = std::max(0.f, -dot(light.sn, wi_world));
                float g = cos_theta_s / distance2(hitp, light.s);

                if( cos_theta * g > 0.f )
                {
                #if BSDF_ONLY
                    // nothing to do
                #elif LIGHT_ONLY
                    // - light sampling     (DT + RT + LIGHT)
                    Color tmaj2 = White();
                    EvalTransmittanceMiller(sampler, shadow_records, tmaj2);
                    float w = 1.f;
                    color += pathWeight * w * tmaj2 * light.emission * bsdf->eval(wi, wo, uv) * cos_theta * g / light.pdf;
                #elif EA_ONLY
                    // - light sampling     (DT + RT + LIGHT)
                    // if( hitSurface )
                    // {
                    //     Color tmaj2 = White();
                    //     EvalTransmittanceMiller(sampler, shadow_records, tmaj2);
                    //     float w = 1;
                    //     color += pathWeight * w * tmaj2 * light.emission * bsdf->eval(wi, wo, uv) * cos_theta * g / light.pdf;
                    // }
                #elif DIRECT_MIS_LIGHT_BSDF
                    // - light sampling     (DT + RT + LIGHT)
                    // - bsdf sampling      (DT + BSDF + DT)
                    Color tmaj2 = White();
                    float ratio_DT_RT2;
                    EvalTransmittanceMiller(sampler, shadow_records, tmaj2, &ratio_DT_RT2);
                    // MIS weight 
                    // w = pdf1 / (pdf1 + pdf2) = 1 / (1 + pdf2/pdf1) = 1 / (1 + a)
                    // a = pdf2 / pdf1 = (pdf_DT1 * pdf_BSDF * pdf_DT2 * G) / (pdf_DT1 * pdf_RT2 * pdf_LIGHT)
                    //                 = (pdf_BSDF * G / pdf_LIGHT) * (pdf_DT2 / pdf_RT2)
                    float a = bsdf->pdf(wi, wo, uv) * g / light.pdf * ratio_DT_RT2;        
                    float w = 1 / (1 + a);
                    color += pathWeight * w * tmaj2 * light.emission * bsdf->eval(wi, wo, uv) * cos_theta * g / light.pdf;
                #elif DIRECT_MIS_LIGHT_EA
                    // - light sampling         (DT + RT + LIGHT)
                    // - equiangular sampling   (RT + EQUI + RT + LIGHT)
                    Color tmaj2 = White();
                    EvalTransmittanceMiller(sampler, shadow_records, tmaj2);
                    // MIS weight 
                    // w = pdf1 / (pdf1 + pdf2) = 1 / (1 + pdf2/pdf1) = 1 / (1 + a)
                    // a = pdf2 / pdf1 = (pdf_RT1 * pdf_EA * pdf_RT2 * pdf_LIGHT) / (pdf_DT1 * pdf_RT2 * pdf_LIGHT)
                    //                 = (pdf_RT1 / pdf_DT1) * pdf_EA
                    float a = pdf_EA / ratio_DT_RT1;        
                    float w = 1 / (1 + a);
                    Color dcontrib = pathWeight * w * tmaj2 * light.emission * bsdf->eval(wi, wo, uv) * cos_theta * g / light.pdf;
                    color += dcontrib;
                    // if( hitMedium )
                    //     printf("w_light %f tmaj2 %f contrib %f le %f G %f bsdf %f pw %f pdf %f 1/pdf %f\n", 
                    //         w, 
                    //         tmaj2.luminance(), 
                    //         dcontrib.luminance(),
                    //         light.emission.luminance(),
                    //         cos_theta * g,
                    //         bsdf->eval(wi, wo, uv).luminance(),
                    //         pathWeight.luminance(),
                    //         (light.pdf),
                    //         1/ (light.pdf)
                    //     );
                #elif DIRECT_MIS_ALL
                    // - light sampling         (DT + RT + LIGHT)
                    // - bsdf sampling          (DT + BSDF + DT)
                    // - equiangular sampling   (RT + EQUI + RT + LIGHT)
                #endif
                }
            }
        }

        // continue random walk
        {
            // bsdf sample
            BsdfSample bsdfSample = bsdf->sample(sampler, wo, uv);
            if( bsdfSample.pdf == 0.f || length2(bsdfSample.wi) == 0.f )
                break;
            lastPdfW = bsdfSample.pdf;
            Vector wi_world = frame.ToWorld(bsdfSample.wi);

            if( hitMedium )
            {
                ray = Ray( hitp, wi_world, ray.time() );
                ray.medium = medium;
            }
            else
            {
                // the normal is always pointing outwards the objects
                // hence a negative cosine between the normal and the sampled direction
                // means the direction is inside an object
                bool next_inside = bsdfSample.wi.z < 0;

                // if we should not be inside the object
                if( next_inside && !bsdf->isTransmissive() )
                    break;

                // select the next crossed medium 
                medium = next_inside ?
                    mBuilder.get_interior_medium(ray.hitobject(), ray.hitid()):
                    mBuilder.get_exterior_medium(ray.hitobject(), ray.hitid());

                // we already switched the epsilon sign if the direction wo 
                // was inside the object. Hence we test if the sampled bsdf 
                // direction is in the same hemisphere to know if we have 
                // to reverse the epsilon sign again
                bool same_hemisphere = bsdfSample.wi.z * wo.z > 0;

                // slightly move the point away from the surface to avoid self intersections
                Point dp = same_hemisphere ? hitp + eps * hitn : hitp - eps * hitn;
                ray = Ray( dp, wi_world, ray.time() );
                ray.medium = medium;
            }

            // update path troughput
            Color newPathWeight = pathWeight * bsdfSample.weight;

            // Arnold trick RR (2012)
            // TODO: BSDF based continuation prob
            float h1 = pathWeight.luminance(), h2 = newPathWeight.luminance();

            // ugly hack to avoid infinite path with thin dielectric bsdf ??
            // float pp = bsdf->isThinDielectric() ? 1.f / pathLength : 1;
            // float p = std::min( 1.f, std::sqrt(h2 / h1 * pp) );

            // RR prob
            const float hmin = 1.19209e-15f;
            // const float hmin = h1;
            float p = std::min( 1.f, std::sqrt(h2 / std::max(h1, hmin)));
            if( p < 0.1 && config.mMaxPathLength < 0 && pathLength > 3 ) // russian roulette
            {
                if( sampler.get1D() > p )
                    break;

                // continue path
                newPathWeight = newPathWeight / p;
                lastPdfW = bsdfSample.pdf * p;
            }

            pathWeight = newPathWeight;

            // if( pathLength > 1e3)
                // printf("pathlength %d %s %f %f %f\n", pathLength, (hitSurface ? "surf" : "med"), h1, h2, p);
        }
        
        // hack in case infinite specular paths ... until I find better solution
        if( pathLength > 1e4 ) break;
    }

    return true;
}

Color VolumetricPathTracingIntegrator::SampleEquiangularDirect(
    Sampler& sampler,
    const MediumRecords& records
    )
{
    Color color = Black();
    MediumRecords shadow_records;

    // direct light sample
    LightSample light = mBuilder.lights.sample( vec2(sampler.sample(),sampler.sample()) );

    // sample equiangular
    float t, pdf_ea;
    const bool foundInteraction = SampleEquiangular(sampler, light, records, t, pdf_ea);

    // compute direct contribution
    if( foundInteraction )
    {   
        // clamp medium records to desired size
        MediumRecords crecords = records.clamped(t);

        // eval transmittance on the first segment 
        float ratio_DT_RT1_EA = 1;
        Color tmaj1 = White();
        EvalTransmittanceMiller(sampler, crecords, tmaj1, &ratio_DT_RT1_EA);

        // sampled position
        Point pea = records.ray(t);

        // slightly move the point away from the light source surface to avoid self intersections
        Point ds = light.s + Epsilon(light.s, distance(light.s, pea)) * light.sn;
        Ray shadow( pea, ds, records.ray.time() );
        int index = crecords.size()-1; 
        shadow.medium = crecords.mediums[index];
        assert(records.depthIndex(t) == crecords.size()-1);
        // shadow.medium = crecords.mediums[crecords.size()-1];
        mBuilder.intersect( shadow, shadow_records, false );

        // strategy 1 : direct light sample
        if( !shadow.hit() )
        {
            Frame frame(-crecords.ray.nd());
            Vector wi_world = normalize(light.s - pea);
            float cos_theta = 1;
            float cos_theta_s = std::max(0.f, -dot(light.sn, wi_world));
            float g = cos_theta_s / distance2(pea, light.s);

            if( cos_theta * g > 0 )
            {
                // get phase and scattering extiction coefficient
                auto bsdf = crecords.mediums[index]->phaseFunction(pea);

            #if EA_ONLY
                // Equiangular sampling only
                Color tmaj2 = White();
                EvalTransmittanceMiller(sampler, shadow_records, tmaj2);
                Vector wo = frame.ToLocal(-crecords.ray.nd());
                Vector wi = frame.ToLocal(wi_world);
                Color mus = records.mediums[index]->getMuS(pea, records.ray.time());
                // compute contribution
                color += tmaj1 * bsdf->eval(wi, wo) * mus * tmaj2 * cos_theta * g * light.emission / (light.pdf * pdf_ea);
            #elif DIRECT_MIS_LIGHT_EA
                // - equiangular sampling   (RT1 + EQUI + RT2 + LIGHT)
                // - light sampling         (DT1 + RT2 + LIGHT)
                Color tmaj2 = White();
                EvalTransmittanceMiller(sampler, shadow_records, tmaj2);
                // MIS weight 
                // w = pdf1 / (pdf1 + pdf2) = 1 / (1 + pdf2/pdf1) = 1 / (1 + a)
                // a = pdf2 / pdf1 = (pdf_DT1 * pdf_RT2 * pdf_LIGHT) / (pdf_RT1 * pdf_EA * pdf_RT2 * pdf_LIGHT)
                //                 = (pdf_DT1 / pdf_RT1) / pdf_EA
                float a = ratio_DT_RT1_EA / pdf_ea;        
                float w = 1 / (1 + a);
                Vector wo = frame.ToLocal(-crecords.ray.nd());
                Vector wi = frame.ToLocal(wi_world);
                Color mus = records.mediums[index]->getMuS(pea, records.ray.time());
                // compute contribution
                color += w * tmaj1 * bsdf->eval(wi, wo) * mus * tmaj2 *  cos_theta * g * light.emission / (light.pdf * pdf_ea);
            #elif DIRECT_MIS_EA_BSDF
                // - equiangular sampling   (RT1 + EQUI + RT2 + LIGHT)
                // - bsdf sampling          (DT1 + BSDF + DT2)
            #elif DIRECT_MIS_ALL
                // - equiangular sampling   (RT1 + EQUI + RT2 + LIGHT)
                // - bsdf sampling          (DT1 + BSDF + DT2)
                // - light sampling         (DT1 + RT2 + LIGHT)
            #endif
            }
        }
    }

    return color;
}
