#ifndef INTERPOLATION_H
#define INTERPOLATION_H

#include "vec.h"

#include <math.h>

// spherical angle interpolation between 3D vectors
inline Vector slerp(const Vector& from, const Vector& to, const float step)
{   
    if ( step <= 0.f ) return from;
    if ( (from.x == to.x && from.y == to.y && from.z == to.z) || step == 1.f ) return to;

    float cos_theta = dot(normalize(from), normalize(to));
    if( cos_theta >= 1.f ) return to;
    if( cos_theta <= -1.f ) return to;

    float theta = std::acos( cos_theta );
    theta *= step;
    Vector relative = normalize(to - from * cos_theta);

    return from * std::cos(theta) + relative * std::sin(theta);
}

// linear interpolation between 3D vectors
inline Vector lerp(const Vector& from, const Vector& to, const float step)
{   
    if ( step <= 0.f ) return from;
    if ( (from.x == to.x && from.y == to.y && from.z == to.z) || step == 1.f ) return to;
    return from + step * (to - from);
}

// linear interpolation between 3D vectors with normalization
inline Vector nlerp(const Vector& from, const Vector& to, const float step)
{   
    return normalize(lerp(from, to, step));
}

// linear interpolation between 3D points
inline Point lerp(const Point& from, const Point& to, const float step)
{   
    if ( step <= 0.f ) return from;
    if ( (from.x == to.x && from.y == to.y && from.z == to.z) || step == 1.f ) return to;
    return from + step * (to - from);
}

// linear interpolation between 2D vectors
inline vec2 lerp(const vec2& from, const vec2& to, const float step)
{   
    if ( step <= 0.f ) return from;
    if ( (from.x == to.x && from.y == to.y) || step == 1.f ) return to;
    return vec2(from.x + step * (to.x - from.x), from.y + step * (to.y - from.y));
}

// linear interpolation between floats
inline float lerp(const float from, const float to, const float step)
{   
    if ( step <= 0.f ) return from;
    if ( (from == to) || (step == 1.f) ) return to;
    return from + step * (to - from);
}

inline float capsuleSDF(float px, float py, float ax, float ay, float bx, float by, float r) 
{
    float pax = px - ax, pay = py - ay, bax = bx - ax, bay = by - ay;
    float h = fmaxf(fminf((pax * bax + pay * bay) / (bax * bax + bay * bay), 1.0f), 0.0f);
    float dx = pax - bax * h, dy = pay - bay * h;
    return sqrtf(dx * dx + dy * dy) - r;
}

static void lineSDFAABB(Image& image, const Color& cline, float ax, float ay, float bx, float by, float r) 
{
    int x0 = std::min(image.width()-1, std::max(0, (int)floorf(fminf(ax, bx) - r)));
    int x1 = std::min(image.width()-1, std::max(0, (int) ceilf(fmaxf(ax, bx) + r)));
    int y0 = std::min(image.height()-1, std::max(0, (int)floorf(fminf(ay, by) - r)));
    int y1 = std::min(image.height()-1, std::max(0, (int) ceilf(fmaxf(ay, by) + r)));
    for(int y = y0; y <= y1; y++)
    for(int x = x0; x <= x1; x++)
    {
        Color c = image(x, y);
        float t = fmaxf(fminf(0.5f - capsuleSDF(x, y, ax, ay, bx, by, r), 1.0f), 0.0f);
        float r = lerp(c.r, cline.r, t);
        float g = lerp(c.g, cline.g, t);
        float b = lerp(c.b, cline.b, t);
        image(x, y) = Color(r, g, b) ;
    }
}


#endif