#pragma once

#include "renderer/arealight.h"

#define UNIFORM_LIGHT_SAMPLING 1
#define POWER_LIGHT_SAMPLING 0
#define RIS_LIGHT_SAMPLING 0

class LightSample
{
public:
    Point s;
    Vector sn;
    Color emission = Black();
    int object_id = -1;
    int triangle_id = -1;
    float pdf = 0;
    float u = 0,v = 0;

    LightSample( ){}
    LightSample( const Point& _s, const Vector& _sn, const Color& _emission, const int object_id = -1, const int triangle_id = -1, const float pdf = 0.f, const float u = 0, const float v = 0)
        : s(_s), sn(_sn), emission(_emission), object_id(object_id), triangle_id(triangle_id), pdf(pdf), u(u), v(v) {}

};

// choisit une source proportionellement à son aire
class LightDistribution
{
public:
    LightDistribution(){}
    LightDistribution( const std::vector<AreaLight>& sources ) : sources(sources) {}

    bool empty() const 
    {
        return !sources.size();
    }
    
    void insert(const std::vector<AreaLight>& lights)
    {
        for(unsigned int i = 0; i < lights.size(); i++)
        {
            // float v = lights[i].area() * lights[i].emission.power();
            float v = lights[i].area();
            assert(v > 0.f);
            total += v;
            cdf.push_back(total);
            sources.push_back(lights[i]);
        }
        std::cout << "distribution:\n\tligths\t\t" << cdf.size() << "\n\ttotal area\t" << total << std::endl;
    }

    LightSample sample( const vec2 v ) const { return sample(v.x, v.y);}

    LightSample sample( const float u1, const float u2 ) const
    {
        // choisit une source en fonction de son aire et de ses emissions
        int id = -1;

        // recherche dichotomique
        float r = u1 * total;
        int p = 0;
        int q = int(cdf.size()) - 1;
        while(p < q)
        {
            int m = (p+q) / 2;
            if(cdf[m] < r)
               p = m + 1;
            else
               q = m;
        }
        assert(p >= 0 && p < (int) cdf.size());
        id = p;

        // renormalise u1 : cdf[p-1] < r < cdf[p]
        float pp = 0;
        if(id > 0) pp = cdf[id-1];
        float u = (r - pp) / (cdf[id] - pp);
        // debug only
        // assert(u >= 0 && u <= 1.f);
     
    #if 0
        // choisit un point sur la source
        // square root sampling
        // cf: Greg Turk. Generating random points in triangles. 1990.
        float su = (1.f - u) * std::sqrt(u2);
        float sv = u * std::sqrt(u2);
        Point s = sources[id].point(su,sv);
        vec2 uv = sources[id].texcoord(su,sv);
    #endif

    #if 0
        // choisit un point sur la source (u, u2)
        // square root sampling
        // cf: Greg Turk. Generating random points in triangles. 1990.

        float r1 = std::sqrt(u);
        float r2 = u2;

        float b0 = (1.f - r1);
        float b1 = (1.f - r2) * r1;
        float b2 = r2 * r1;

        Point a = sources[id].a;
        Point b = sources[id].b;
        Point c = sources[id].c;
        vec2 ta = sources[id].ta;
        vec2 tb = sources[id].tb;
        vec2 tc = sources[id].tc;
        float fa = distance(a, b) * distance(a, c);
        float fb = distance(b, a) * distance(b, c);
    
        // choisir le coté qui deforme le moins les samples
        if(fb < fa)
        {
            a = sources[id].b;
            b = sources[id].c;
            c = sources[id].a;
            ta = sources[id].tb;
            tb = sources[id].tc;
            tc = sources[id].ta;
        }
        float fc = distance(c, a) * distance(c, b);
        if(fc < fb && fc < fa)
        {
            a = sources[id].c;
            b = sources[id].a;
            c = sources[id].b;
            ta = sources[id].tc;
            tb = sources[id].ta;
            tc = sources[id].tb;
        }
        
        Point s =  b0 * a + b1 * b + b2 * c;
        vec2 uv(   b0 * ta.x + b1 * tb.x + b2 * tc.x, 
                   b0 * ta.y + b1 * tb.y + b2 * tc.y);
    #endif

    #if 1
        // choisit un point sur la source
        // cf: A Low-Distortion Map Between Triangle and Square. Eric Heitz. 2019

        float su = u;
        float sv = u2;
        if( sv > su ) 
        {
            su *= 0.5f;
            sv -= su;
        } 
        else 
        {
            sv *= 0.5f;
            su -= sv;
        }

        Point s = sources[id].point(su, sv);
        vec2 uv = sources[id].texcoord(su, sv);

    #endif

        // densité
        // float pdf = sources[id].emission.power() / total;
        float pdf = 1 / total;

        // light sample
        Vector sn = sources[id].normal();
        // Vector sn = sources[id].normal(uv.x, uv.y);
        return LightSample(s, sn, sources[id].emission, sources[id].object_id, sources[id].triangle_id, pdf, uv.x, uv.y);
    }
    
    float eval() const
    {
        return 1.f / total;
    }

public:
    std::vector<AreaLight> sources; 
    std::vector<float> cdf;
    float total = 0;
};
