// std lib
#include <iostream>
#include <chrono>
#include <cstring>
#include <sstream>
#include <omp.h>
#include <iomanip>

// gkit
#include "image.h"
#include "image_io.h"
#include "app.h"
#include "program.h"
#include "widgets.h"
#include "uniforms.h"
#include "texture.h"

// renderer
#include "renderer/builder.h"
#include "renderer/viewer.h"
#include "renderer/scene.h"
#include "renderer/integrators/abstractintegrator.h"

int main(int argc, char* argv[])
{
    // options
    if( argc < 2 )
    {
        std::cout << "Usage bin/embree <scene>" << std::endl;
        exit(1);
    }

    // image definition
    std::vector<Camera> cameras;
    Builder builder;
    IntegratorConfig config;

    // read scene file
    std::string scenefile(argv[1]);
    std::string path, name;
    const size_t last_slash_index = scenefile.find_last_of("/");
    if(std::string::npos != last_slash_index)
    {
        path = scenefile.substr(0,last_slash_index+1);
        name = scenefile.substr(last_slash_index+1);
    }

    config.mFilePath = path;

    read_scene(scenefile, cameras, builder, config);

    // float width = cameras[0].width;
    // float height = cameras[0].height;
    float height = 1080;
    float width = height * cameras[0].width / cameras[0].height;

    // print config
    config.Print();

    // init integrator
    AbstractIntegrator* integrator = AbstractIntegrator::CreateIntegrator(cameras, builder, config);
    if(integrator == nullptr)
    {
        printf("Unknown algorithm!!\n");
        exit(1);
    }

    // enable nested parallelism
    omp_set_nested(1);

    if( config.mViewer )
    {
        // open 2 concurrent sections
        #pragma omp parallel sections num_threads(2)
        {
            // live viewer
            #pragma omp section
            {
                Viewer viewer(width, height, integrator);
                viewer.run();
            }
            // live render
            #pragma omp section
            {
                integrator->Render();
            }
        }
    }
    else
    {
        integrator->Render();
    }

    delete integrator;

    return 0;
}
