#pragma once

#include <memory>
#include <vector>

#include "color.h"
#include "renderer/bsdf/bsdf.h"
#include "renderer/sampler.h"
#include "renderer/ray.h"
#include "renderer/media/mediumutils.h"
#include "renderer/lightdistribution.h"

#include "FastNoise.h"

#include <nanovdb/NanoVDB.h>
#define NANOVDB_USE_ZIP 1
#include <nanovdb/util/IO.h>
#include <nanovdb/util/GridHandle.h>
#include <nanovdb/util/SampleFromVoxels.h>

//#define LOG_PDFS
#define OPTICAL_DEPTH_SHIFT

#define DEBUG_EPS 0.001f

class Ray;

class DensityFunction
{
public:
    DensityFunction(){}
    virtual ~DensityFunction() {}
    virtual float max() const = 0;
    virtual float eval(const Point& p, float time = 0) const = 0;
};

class DensityFunctionConstant : public DensityFunction
{
public:
    DensityFunctionConstant(float d = 1, float maxd = 1) : d(d), maxd(maxd) {}
    float max() const { return maxd; }
    float eval(const Point& p, float time = 0) const { return d; }
protected:
    float d, maxd;
};

class DensityFunctionFastNoise : public DensityFunction
{
public:
    enum NoiseType
    {
        Value = 0,
        Simplex,
        Perlin,
        Cellular,
    };

    DensityFunctionFastNoise(int seed = 1234, NoiseType type = Value, float frequency = 10.f, float maxd = 1, const Transform& tr = Identity()) :
        noise(seed),
        type(type),
        frequency(frequency),
        maxd(maxd)
    {
        noise.SetFrequency(frequency);
    }

    float max() const { return maxd; };
    float eval(const Point& p, float time = 0) const
    {
        float n = 0;
        switch(type)
        {
            case NoiseType::Value:    n = noise.GetValueFractal(p.x, p.y, p.z);   break;
            case NoiseType::Simplex:  n = noise.GetSimplexFractal(p.x, p.y, p.z); break;
            case NoiseType::Perlin:   n = noise.GetPerlinFractal(p.x, p.y, p.z);  break;
            case NoiseType::Cellular: n = noise.GetCellular(p.x, p.y, p.z);       break;
        }
        return std::min(std::max(0.f, n), 1.f);
    }

protected:
    FastNoise noise;
    NoiseType type;
    float frequency;
    float maxd;
};

class DensityFunctionFastNoiseAnimated : public DensityFunction
{   
public:
    DensityFunctionFastNoiseAnimated(int seed = 1234, float frequency = 5.f, float maxd = 1) :
        noise(seed),
        frequency(frequency),
        maxd(maxd)
    {
        noise.SetFrequency(frequency);
        // noise.SetFractalLacunarity(2.0);
        // noise.SetFractalGain(0.5);
        // noise.SetFractalOctaves(7);
    }

    float max() const { return maxd; }

    static float clamp(float x, float a, float b) 
    {
        if (x < a) x = a;
        if (x > b) x = b;
        return x;
    }

    static float smootherstep(float a, float b, float x) 
    {
        // Scale, and clamp x to 0..1 range
        x = clamp((x - a) / (b - a), 0.f, 1.f);
        // Evaluate polynomial
        return x * x * x * (x * (x * 6.f - 15.f) + 10.f);
    }

    // mémo : 
    // si scale vaut 1, alors un changement d'une unité sur p (x, y ou z) 
    // donnera grosso modo 2 blobs de noise, donc le double 4 en mode ridge
    // octaves de 3 a 9
    // persistence smooth 0.5 - 0.7 bruité
    float fractalRidgePerlin(const Point& p, float time, int octaves = 7, float scale = 1.f, float persistence = 0.65f, float lacunarity = 2.f ) const
    {
        float total = 0.f;
        float norma = 0.f;
        float amp = 1.f;
        float f = scale;
        for (int i=0; i < octaves; ++i)
        {
            float ridge = 1.f - abs(noise.GetPerlin(f * p.x, f * p.y, f * p.z));
            // float ridge = 1.f - abs(noise.GetSimplex(f * p.x, f * p.y, f * p.z, f * time * 80));
            total += ridge * amp;
            norma += amp;
            f *= lacunarity;
            amp *= persistence;
        }
        return total / norma;
    }

    float eval(const Point& p, float time = 0) const
    {
        // bouger le point p si besoin sur le temps
        const Vector v(0,-10,0);
        const Point dp = p + time * v;
        // const Point dp = p ;
        // calculer le bruit
        const float nv = fractalRidgePerlin(dp, time);
        // faire des trous
        return smootherstep(0.85, 1.0, nv);
    }
    
protected:
    FastNoise noise;
    float frequency;
    float maxd;
};


class DensityFunctionNanoVDB : public DensityFunction
{
public:
    DensityFunctionNanoVDB(const std::string& fileName, const std::string& gridName = "density", const Transform& matrix = Identity()) :
        toMedium(matrix),
        //toWorld(matrix.inverse())
        densityScaling(length(matrix(normalize(Vector(1., 1., 1.)))))
    {
        //std::cout << "nanovdb custom tranformation:" << std::endl << matrix << std::endl;
        //const auto wanted =
        //    Translation(-9.984, 73.008, -42.64)
        //    *
        //    Scale(206.544, 140.4, 254.592)
        //    *
        //    Scale(2.5, 2.5*2., 2.5)
        //    *
        //    Translation(0., -1.05, 0.)
        //;
        //std::cout << "wanted transformation:" << std::endl << wanted << std::endl;
        try {
            handle = nanovdb::io::readGrid(fileName, gridName);
            auto* grid = handle.grid<float>(); // get a (raw) pointer to the first NanoVDB grid of value type float
            assert(grid != nullptr);
        }
        catch (const std::exception& e) {
            std::cerr << "An exception occurred: \"" << e.what() << "\"" << std::endl;
            assert(false);
        }

        float mind;
        handle.grid<float>()->tree().extrema(mind, maxd);

        std::cout << "Read NanoVDB density grid:" << std::endl;
        std::cout << "\tfile: " << fileName << std::endl;
        std::cout << "\tgrid name: " << gridName << std::endl;
        std::cout << "\textrema: " << mind << ", " << maxd << std::endl;
        std::cout << "\tdensity scaling: " << densityScaling << std::endl;

        mind *= densityScaling;
        maxd *= densityScaling;

        const auto bbox = handle.grid<float>()->worldBBox();
        const Point a(bbox.min()[0], bbox.min()[1], bbox.min()[2]);
        const Point b(bbox.max()[0], bbox.max()[1], bbox.max()[2]);
        std::cout << "\tbounding box:" << std::endl << "\t\t" << a << std::endl << "\t\t" << b << std::endl;
        std::cout << "\tfitted bounding box:" << std::endl << "\t\t" << toMedium.inverse()(a) << std::endl << "\t\t" << toMedium.inverse()(b) << std::endl;
    }
    float max () const override{ return maxd; }
    float eval(const Point& p, float time = 0) const override
    {
        using tAccessor = nanovdb::FloatGrid::AccessorType;
        using tSampler = nanovdb::SampleFromVoxels<tAccessor, 1, true>;
        static thread_local const DensityFunctionNanoVDB * instance = this;
        static thread_local const auto grid = handle.grid<float>();
        static thread_local tAccessor accessor = grid->getAccessor();
        static thread_local tSampler sampler = tSampler(accessor);
        assert(instance == this);
        //if(instance != this)
        //{
        //    instance = this;
        //    accessor = handle.grid<float>()->getAccessor();
        //    //sampler = tSampler(accessor);
        //}
        //tSampler sampler = tSampler(accessor);
        const Point q = toMedium(p);
        const nanovdb::Vec3<float> vp(q.x, q.y, q.z);
        //const nanovdb::Vec3<float> pIndex = handle.grid<float>()->worldToIndexF(vp);
        const nanovdb::Vec3<float> pIndex = grid->worldToIndexF(vp);
        //using Sampler = nanovdb::SampleFromVoxels<nanovdb::FloatGrid::TreeType, 1, false>;
        const float density = sampler(pIndex) * densityScaling;
        return density;
    }

protected:
    const Transform toMedium;
    const float densityScaling = 1.f;
    float maxd;
    nanovdb::GridHandle<nanovdb::HostBuffer> handle;
};

// class DensityFunctionGrid;

struct MediumSample
{
    enum InteractionType { PassThrough, NullInteraction, RealInteraction, Absorbtion, Scattering };

    MediumSample(const Point& p, float t, const Color& w, const int channel, const float channelPdf, float pdf, int index, InteractionType type) :
        p(p), t(t), w(w), channel(channel), channelPdf(channelPdf), pdf(pdf), index(index), type(type)
    {
    }

    MediumSample(){}

    Point p;
    float t = 0.f;
    Color w = White();
    int channel = -1;
    float channelPdf = 0.f;
    //Color wp = White();
    //Color ws = White();
    float pdf = 1.;
#ifdef LOG_PDFS
    Color lf = Black();
    Color lpdfs = Black();
#endif
    int index = -1;
    InteractionType type = PassThrough;
};

class Medium
{
public:
    Medium() {}
    virtual ~Medium(){}

    //virtual bool isHomogeneous() const = 0;
    //virtual bool isHeterogeneous() const = 0;

    //Color albedo() const { return mSigmaS / mSigmaT; }
    //Color sigmaS() const { return mSigmaS; }
    //Color sigmaA() const { return mSigmaA; }
    //Color sigmaT() const { return mSigmaT; }
    virtual float density(const Point& p, float t = 0) const {return 1;}
    virtual float maxDensity() const {return 1;}

    virtual const Bsdf * phaseFunction(const Point& p) const = 0;
    virtual Color getMuA(const Point& p, float t = 0) const = 0;
    virtual Color getMuS(const Point& p, float t = 0) const = 0;
    virtual std::pair<Color, Color> getMu(const Point& p, float t = 0) const { return {getMuA(p,t), getMuS(p,t)}; }
    virtual Color getMuT(const Point& p, float t = 0) const { return getMuA(p, t) + getMuS(p, t); }
    virtual Color getMuN(const Point& p, float t = 0) const { return maxMuT() - getMuT(p, t); }
    virtual Color maxMuT() const = 0;
    virtual Color getLe(const Point& p, float t = 0) const { return Black(); };


    Color maxMuTChannelsCDF() const
    {
        Color cdf = Black();
        float sum = 0.;
        for(unsigned int i = 0u; i < 3u; i++)
        {
            sum += maxMuT()[i] > 0.f ? 1.f : 0.f;
            // sum += maxMuT()[i] > 0.f ? maxMuT()[i] : 0.f;
            // sum += maxMuT()[i] > 0.f ? std::exp(-maxMuT()[i]) : 0.f;
            cdf[i] = sum;
        }
        assert(sum >= 0.f);
        return sum > 0.f ? cdf/sum : Black();
    }
    std::pair<unsigned int, float> sampleMaxMuTChannel(const float u) const
    {
        const Color cdf = maxMuTChannelsCDF();
        unsigned int i;
        for(i = 0u; i < 3u && u > cdf[i]; i++);
        assert(i < 3u);
        const float pdf = cdf[i] - (i > 0u ? cdf[i - 1] : 0.);
        return {i, pdf};
    }
    float maxMuTChannelPdf(const unsigned int channel) const
    {
        assert(channel < 3u);
        const Color cdf = maxMuTChannelsCDF();
        return cdf[channel] - (channel > 0u ? cdf[channel - 1] : 0.);
    }

    // sample interval [tmin; tmax] on the given ray (delta tracking Miller)
    virtual bool sample(Sampler& sampler, const Ray& ray, float tmin, float tmax, MediumSample& ms, std::pair<unsigned int, float>& channelSample, float * ratio = nullptr) const
    {
        // compute Miller pdf ratio for MIS between direct light sampling + delta tracking and bsdf sampling + delta tracking
        // since we are going spectral this is a bit tricky :
        // we need ratio        = pdf_DT / pdf_RT 
        //                      = 3 * (pdf_DT^R+pdf_DT^G+pdf_DT^B) / (3 * (pdf_RT^R+pdf_RT^G+pdf_RT^B))
        //                      = pdf_DT^R / (pdf_RT^R+pdf_RT^G+pdf_RT^B) + pdf_DT^G / (pdf_RT^R+pdf_RT^G+pdf_RT^B) + pdf_DT^B / (pdf_RT^R+pdf_RT^G+pdf_RT^B)
        //                      = 1 / (rr+gr+br) + 1 / (rg+gg+bg) + 1 / (rb+gb+bb)
        //                      where ij = pdf_RT^i / pdf_DT^j
        // hence we need 9 floats to compute the ratio
        float ij[9] = {1,1,1,1,1,1,1,1,1};
        // recall the difference between ratio tracking and delta tracking:
        // - RT sample tmaj + force null interaction to happen the resulting pdf is 
        //      pdf_RT = mubar * tbar
        // - DT sample tmaj + explicit event sampling between abs, null, scat, but here we evaluate only null collisions
        //      pdf_DT = mubar * tbar * muN / mubar = muN * tbar
        // hence the resulting ij ratios to compute are (maybe we are wrong here since tbar may be spectral too...)
        //      ij = mubar_i / muN_i

        // compute the extinction majorant (constant over the medium so we compute only once)
        const Color barMuT = maxMuT();

        //Color w = White(); // frgb/Prgb not consistent with spectral MIS

        Color ws = White(); // frgb/Pc
        Color wp = White(); // Prgb/Pc

        Color pdfs = White(); // Prgb

#ifdef LOG_PDFS
        Color lf = Black(); // frgb in log domain
        Color lpdfs = Black(); //Prgb in log domain
#endif

        // force valid channel to be sampled (mubar > 0)
        channelSample = sampleMaxMuTChannel(sampler.sample());
        const unsigned int& channel = channelSample.first;
        const float& channelPdf = channelSample.second;

        // start at tmin
        float t = tmin;
        MediumSample::InteractionType type = MediumSample::InteractionType::PassThrough;
        while(true)
        {
            // force valid channel to be sampled (mubar > 0)
            //channelSample = sampleMaxMuTChannel(sampler.sample());

            // sample distance
            const float u = sampler.sample();
            const float step = -std::log(1.f - u) / barMuT[channel];

            // update pdf for free flight sampling
            const Color opticalDepth = std::min(step, tmax - t) * barMuT;
            const Color barT = exp(-opticalDepth);

            ////w *= barT/barT;
            ws *= barT/barT[channel];
            wp *= barT/barT[channel];
            pdfs *= barT;
#ifdef LOG_PDFS
            lf -= opticalDepth;
            lpdfs -= opticalDepth;
#endif

            // update distance
            t += step;

            // if the step falls out of the medium section (no interaction)
            if( t >= tmax )
                break;

            // get current point
            const Point p = ray(t);

            // compute event probabilities (scattering, null)
            const std::pair<Color, Color> mu = getMu(p, ray.time());
            const Color& muA = mu.first;
            const Color& muS = mu.second;
            const Color muT = muA + muS;
            const Color muN = barMuT - muT;
            // todo: if no emission do not sample absorption 
            // const Color le = getLe(p, ray.time());

            // const float pA = getLe(p, ray.time()).power() > 0 ? muA[channel] / barMuT[channel] : 0; // absorbtion prob
            const float pA = muA[channel] / barMuT[channel]; // absorbtion prob
            const float pS = pA > 0 ? muS[channel] / barMuT[channel] : muT[channel] / barMuT[channel]; // scattering prob
            const float pT = pA + pS; // real interaction prob
            const float pN = muN[channel] / barMuT[channel]; // null collision prob

            // sample event (scattering, null) and update transmittance and pdfs for event type sampling
            const float pp = sampler.sample();
            // absorbtion
            if( pp < pA )
            {
                ////w *= muA/muA;
                ws *= muA/muA[channel];
                wp *= muA/muA[channel];
                pdfs *= muA;
#ifdef LOG_PDFS
                lf += log(muA);
                lpdfs += log(muA);
#endif

                type = MediumSample::InteractionType::Absorbtion;
                break;
            }
            // scattering
            else if( pp < pT )
            {
                ////w *= muS/muS;
                ws *= muS/muS[channel];
                wp *= muS/muS[channel];
                pdfs *= muS;
#ifdef LOG_PDFS
                lf += log(muS);
                lpdfs += log(muS);
#endif

                type = MediumSample::InteractionType::Scattering;
                break;
            }
            // null interaction
            else
            {
                ////w *= muN/muN;
                ws *= muN/muN[channel];
                wp *= muN/muN[channel];
                pdfs *= muN;
#ifdef LOG_PDFS
                lf += log(muN);
                lpdfs += log(muN);
#endif

                // update pdfs ratio for direct MIS weight
                if( ratio )
                {
                    for(int i = 0; i < 3; ++i)
                        for(int j = 0; j < 3; ++j)
                            ij[3*i+j] *= barMuT[j] / muN[i];
                }
            }
        }

        // update pdfs ratio for direct MIS weight
        // Note: we do not consider spectral channel pdf since they cancel out 
        if( ratio )
        {
            float ratio_DT_RT = 0;
            for(int i = 0; i < 3; ++i)
            {
                float sum = 0;
                for(int j = 0; j < 3; ++j)
                    sum += ij[3*i+j];
                ratio_DT_RT += 1.f / sum;
            }
            *ratio *= ratio_DT_RT;
        }

        const float pdf = pdfs.rgbsum() * channelPdf; // spectral MIS case
        t = std::min(t, tmax);
        ms = MediumSample(
            ray(t),
            t,
#ifndef LOG_PDFS
            // w, // not consistent with spectral MIS to change later
            ws/(channelPdf*wp.rgbsum()),
#else
            exp(lf)/(channelPdf*exp(lpdfs).rgbsum()), //log to real w
#endif
            channel,
            channelPdf,
            pdf,
            ms.index,
            type
        );
        //ms.wp = wp;
#ifdef LOG_PDFS
        ms.lf += lf;
        ms.lpdfs = lpdfs;
#endif

        return type != MediumSample::InteractionType::PassThrough;
    }

    virtual bool sampleChain(
        Sampler& sampler, const Ray& ray, float tmin, float tmax, std::vector<MediumSample>& chain, Color& w, Color& ws, Color& wp, Color& pdfs,
#ifdef LOG_PDFS
        Color& lf, Color& lpdfs,
#endif
        int index, float& opticalDepthStepRemainder, std::pair<unsigned int, float>& channelSample
    ) const
    {
        // compute the extinction majorant (constant over the medium so we compute only once)
        const Color barMuT = maxMuT();

        // force valid channel to be sampled (mubar > 0)
        if(opticalDepthStepRemainder == 0.)
            channelSample = sampleMaxMuTChannel(sampler.sample());
        const unsigned int& channel = channelSample.first;
        const float& channelPdf = channelSample.second;

        float t = tmin;
        MediumSample::InteractionType type = MediumSample::InteractionType::PassThrough;
        while(true)
        {
            // sample distance
            const float opticalDepthStep = (opticalDepthStepRemainder > 0. ? opticalDepthStepRemainder : -std::log(1 - sampler.sample()));
            const float step = opticalDepthStep / barMuT[channel];

            // update pdf for free flight sampling
            const Color opticalDepth = std::min(step, tmax - t) * barMuT;
            const Color barT = exp(-opticalDepth);
            //w *= barT/barT;
            ws *= barT/barT[channel];
            wp *= barT/barT[channel];
            pdfs *= barT;
#ifdef LOG_PDFS
            lf -= opticalDepth;
            lpdfs -= opticalDepth;
#endif

            // update distance
            t += step;

            // if the step falls out of the medium section (no interaction)
            if( t >= tmax )
            {
                opticalDepthStepRemainder = std::max(0.f, (t - tmax) * barMuT[channel]); //keep up for next records
                break;
            }
            opticalDepthStepRemainder = 0.f;

            // get current point
            const Point p = ray(t);

            // compute event probabilities (scattering, null)
            const std::pair<Color, Color> mu = getMu(p, ray.time());
            const Color& muA = mu.first;
            const Color& muS = mu.second;
            const Color muT = muA + muS;
            const Color muN = barMuT - muT;

            // const float pA = getLe(p, ray.time()).power() > 0 ? muA[channel] / barMuT[channel] : 0; // absorbtion prob
            const float pA = muA[channel] / barMuT[channel]; // absorbtion prob
            const float pS = pA > 0 ? muS[channel] / barMuT[channel] : muT[channel] / barMuT[channel]; // scattering prob
            const float pT = pA + pS; // real interaction prob
            const float pN = muN[channel] / barMuT[channel]; // null collision prob

            // sample event (scattering, null) and update pdfs
            const float pp = sampler.sample();
            // absorbtion
            if( pp < pA )
            {
                //w *= muA/muA;
                ws *= muA/muA[channel];
                wp *= muA/muA[channel];
                pdfs *= muA;
#ifdef LOG_PDFS
                lf += log(muA);
                lpdfs += log(muA);
#endif

                type = MediumSample::InteractionType::Absorbtion;
            }
            // scattering
            else if( pp < pT )
            {
                //w *= muS/muS;
                ws *= muS/muS[channel];
                wp *= muS/muS[channel];
                pdfs *= muS;
#ifdef LOG_PDFS
                lf += log(muS);
                lpdfs += log(muS);
#endif

                type = MediumSample::InteractionType::Scattering;
            }
            // null interaction
            else
            {
                //w *= muN / muN; 
                ws *= muN/muN[channel];
                wp *= muN/muN[channel];
                pdfs *= muN;
#ifdef LOG_PDFS
                lf += log(muN);
                lpdfs += log(muN);
#endif

                type = MediumSample::InteractionType::NullInteraction;
            }

            const float pdf = pdfs.rgbsum() * channelPdf; // spectral MIS case
            pdfs = White();
            //std::cout << chain.size() << ": " << pdf << std::endl;
            //std::cout << chain.size() << ": " << w << std::endl;
            chain.emplace_back(
                ray(t),
                t,
#ifndef LOG_PDFS
                //w, // not consistent with spectral MIS to change later
                ws/(channelPdf*wp.rgbsum()),
#else
                //exp(lf - lpdfs), //log to real w
                exp(lf)/(channelPdf*exp(lpdfs).rgbsum()), //log to real w
#endif
                channel,
                channelPdf,
                pdf,
                index,
                type
            );
#ifdef LOG_PDFS
            chain.back().lf = lf;
            chain.back().lpdfs = lpdfs;
#endif

            if( type == MediumSample::InteractionType::Absorbtion || type == MediumSample::InteractionType::Scattering || type == MediumSample::InteractionType::RealInteraction )
                return true;
        }

        return false;
    }

    // explicit ratio tracking estimator using Miller et al. framework
    Color transmittanceIndependentChannels(Sampler& sampler, const Ray &ray, float tmin, float tmax) const
    {
        // compute the extinction majorant (constant over the medium so we compute only once)
        const Color barMu = maxMuT();

        // init f and pdf with barT
        Color w = White();

        for(int channel = 0; channel < 3; channel++)
        {
            // sample null medium interaction
            float t = tmin;
            while(true)
            {
                // sample distance
                const float u = sampler.sample();
                const float step = -std::log(1 - u) / barMu[channel];

                // update distance
                t += step;

                // if the step falls out of the medium section, stop evaluation
                if( t >= tmax )
                    break;

                // get current point
                const Point p = ray(t);

                // compute coefficients
                const float muT = getMuT(p, ray.time())[channel];
                const float muN = barMu[channel] - muT;
                
                // if the density is 0 continue
                if( muT <= 0.)
                    continue;

                // forced null interaction
                w[channel] *= muN/barMu[channel];
            }
            const float barT = std::exp(-std::min(t, tmax) * barMu[channel]);
            //w[channel] *= barT/barT;
        }

        return w;
    }

    // explicit ratio tracking estimator using Miller et al. framework
    Color transmittanceSpectralMIS(Sampler& sampler, const Ray &ray, float tmin, float tmax, float * ratio = nullptr) const
    {
        // compute the extinction majorant (constant over the medium so we compute only once)
        const Color barMu = maxMuT();
        //const float barbarMu = maxMuT().max();

        // compute Miller pdf ratio for MIS between direct light sampling + delta tracking and bsdf sampling + delta tracking
        // since we are going spectral this is a bit tricky :
        // we need ratio        = pdf_DT / pdf_RT 
        //                      = 3 * (pdf_DT^R+pdf_DT^G+pdf_DT^B) / (3 * (pdf_RT^R+pdf_RT^G+pdf_RT^B))
        //                      = pdf_DT^R / (pdf_RT^R+pdf_RT^G+pdf_RT^B) + pdf_DT^G / (pdf_RT^R+pdf_RT^G+pdf_RT^B) + pdf_DT^B / (pdf_RT^R+pdf_RT^G+pdf_RT^B)
        //                      = 1 / (rr+gr+br) + 1 / (rg+gg+bg) + 1 / (rb+gb+bb)
        //                      where ij = pdf_RT^i / pdf_DT^j
        // hence we need 9 floats to compute the ratio
        float ij[9] = {1,1,1,1,1,1,1,1,1};
        // recall the difference between ratio tracking and delta tracking:
        // - RT sample tmaj + force null interaction to happen the resulting pdf is 
        //      pdf_RT = mubar * tbar
        // - DT sample tmaj + explicit event sampling between abs, null, scat, but here we evaluate only null collisions
        //      pdf_DT = mubar * tbar * muN / mubar = muN * tbar
        // hence the resulting ij ratios to compute are (maybe we are wrong here since tbar may be spectral too...)
        //      ij = mubar_i / muN_i

        // init f and pdf with barT
        // Color w = White();
        Color ws = White(); // frgb/Pc
        Color wp = White(); // Prgb/Pc

        // force valid channel to be sampled (mubar > 0)
        std::pair<unsigned int, float> channelSample;
        const unsigned int& channel = channelSample.first;
        const float& channelPdf = channelSample.second;
        channelSample = sampleMaxMuTChannel(sampler.sample());

        // sample null medium interaction
        float t = tmin;
        while(true)
        {
            //channelSample = sampleMaxMuTChannel(sampler.sample());

            // sample distance
            const float u = sampler.sample();
            const float step = -std::log(1 - u) / barMu[channel];
            //const float step = -std::log(1 - u) / barbarMu;

            const Color barT = exp(-std::min(tmax - t, step) * barMu);
            //w *= barT/barT;
            ws *= barT/barT[channel];
            wp *= barT/barT[channel];

            // update distance
            t += step;

            // if the step falls out of the medium section, stop evaluation
            if( t >= tmax )
                break;

            // get current point
            const Point p = ray(t);

            // compute coefficients
            const Color muT = getMuT(p, ray.time());
            const Color muN = barMu - muT;
            //const Color muN = Color(barbarMu) - muT;

            if( muT.rgbsum() <= 0. )
                continue;

            // forced null interaction
            const Color f = muN;
            const Color pdfs = barMu;
            //const Color pdfs = Color(barbarMu);
            //w *= f/pdfs;
            //w *= muN/barbarMu;
            ws *= f/pdfs[channel];
            wp *= pdfs/pdfs[channel];

            // update pdfs ratio for direct MIS weight
            if( ratio )
            {
                for(int i = 0; i < 3; ++i)
                    for(int j = 0; j < 3; ++j)
                        ij[3*i+j] *= barMu[j] / muN[i];
            }
        }

        // update pdfs ratio for direct MIS weight
        if( ratio )
        {
            float ratio_DT_RT = 0;
            for(int i = 0; i < 3; ++i)
            {
                float sum = 0;
                for(int j = 0; j < 3; ++j)
                    sum += ij[3*i+j];
                ratio_DT_RT += 1.f / sum;
            }
            *ratio *= ratio_DT_RT;
        }

        //const Color barT = exp(-std::min(t, tmax) * barMu);
        //w *= barT/barT;
        //return w;
        //ws *= barT/barT[channel];
        //wp *= barT/barT[channel];
        return ws /(channelPdf * wp.rgbsum());
    }


    // An unbiased ray-marching transmittance estimatorn, markus kettunen, eugene d’eon, jacopo pantaleoni, jan novak
    // does not match reference for now ...
    Color transmittanceBiasedRayMarching(Sampler& sampler, const Ray &ray, float tmin, float tmax, float * ratio = nullptr) const
    {
        if( ratio ) *ratio = 1;
        // compute the extinction majorant (constant over the medium so we compute only once)
        const Color barMu = maxMuT();
        const float l = tmax - tmin;
        const Color tau = l * barMu;
        const Color tmp = (0.015f+tau)*(0.65f+tau)*(60.3f+tau);
        const float u = sampler.sample();

        Color x = Black();
        for(int c = 0; c < 3; ++c)
        {
            const int m = std::ceil(std::cbrt(tmp[c]));
            const float r = l / m;
            for(int i = 0; i < m; ++i)
            {
                const float t = tmin + r * (u + i);
                x[c] -= r * density(ray(t), ray.time());
            }
            // optionnal
            x[c] -= r * (0.5f - u) * (density(ray(tmax), ray.time()) - density(ray(tmin), ray.time()));
        }
        return exp(x);
    }

    virtual Color transmittance(Sampler& sampler, const Ray &ray, float tmin, float tmax, float * ratio = nullptr) const
    {
        // return transmittanceIndependentChannels(sampler, ray, tmin, tmax);
        return transmittanceSpectralMIS(sampler, ray, tmin, tmax, ratio);
        // return transmittanceBiasedRayMarching(sampler, ray, tmin, tmax, ratio);
    }
};

class MediumUniform : public Medium
{
public:
    MediumUniform(
        std::shared_ptr<Bsdf> phase = std::make_shared<BsdfHenyeyGreenstein>(),
        const Color& sigmaA = Black(),
        const Color& sigmaS = Black(),
        const Color& Le = Black()
    ) :
        mPhase(phase),
        mSigmaA(sigmaA),
        mSigmaS(sigmaS),
        mSigmaT(sigmaA + sigmaS)//,
        //mLe(Le)
    {}
    virtual ~MediumUniform(){}

    virtual const Bsdf * phaseFunction(const Point& p) const override
    {
        return mPhase.get();
    }
    virtual Color getMuA(const Point& p, float time = 0) const override
    {
        return density(p, time) * mSigmaA;
    }
    virtual Color getMuS(const Point& p, float time = 0) const override
    {
        return density(p, time) * mSigmaS;
    }
    virtual std::pair<Color, Color> getMu(const Point& p, float time = 0) const override
    {
        const float d = density(p, time);
        return {d * mSigmaA, d * mSigmaS};
    }
    virtual Color getMuT(const Point& p, float time = 0) const override
    {
        return density(p, time) * mSigmaT;
    }
    virtual Color maxMuT() const override
    {
        return maxDensity() * mSigmaT;
    }
    //virtual Color getLe(const Point& p,  float t = 0) const override
    //{
    //    return 0.01 * Color(1., 0.1, 0.1);
    //    //return mLe;
    //}

    // virtual float density(const Point& p, float t = 0) const override;
    // virtual float maxDensity() const override;

protected:
    std::shared_ptr<Bsdf> mPhase = nullptr;
    const Color mSigmaA, mSigmaS, mSigmaT;
    //const Color mLe;
};

class MediumHomogeneous : public MediumUniform
{
public:
    MediumHomogeneous(
        const Color& sigmaA,
        const Color& sigmaS,
        float d = 1.,
        std::shared_ptr<Bsdf> phase = std::make_shared<BsdfHenyeyGreenstein>(),
        const Color& Le = Black()
    ) :
        MediumUniform(phase, sigmaA, sigmaS),
        mDensity(d),
        mLe(Le)
    {}

    virtual Color getLe(const Point& p,  float time = 0)  const override
    {
        return mLe;
    }

    //bool isHomogeneous() const { return true; }
    //bool isHeterogeneous() const { return false; }

    virtual float density(const Point& p, float time = 0) const override
    {
        return mDensity;
    }

    virtual float maxDensity() const override
    {
        return mDensity;
    }

    //to fix for spectral case
    //bool sample(Sampler& sampler, const Ray &ray, float tmin, float tmax, MediumSample& ms) const override
    //{
    //    const float d = density(Point());
    //    // if the medium density is 0 (no interaction)
    //    if ( d <= 0.)
    //        return false;

    //    // compute the extinction coefficient (constant)
    //    float t = tmin;
    //    Point p = ray(t);

    //    // compute majorant
    //    Color muT = d * mSigmaT;

    //    // compute spectral number of strategies for MIS weight 
    //    float strategies = 0;
    //    for(int i = 0; i < 3; ++i)
    //        if( muT[i] > 0 )
    //            strategies++;
    //    assert(strategies > 0);

    //    // force valid channel to be sampled (mubar > 0)
    //    int channel = sampler.sample() * 3;
    //    while(muT[channel] <= 0)
    //        channel = sampler.sample() * 3;

    //    // sample distance through the medium
    //    float u = sampler.sample();
    //    float step = -std::log(1 - u) / muT[channel];
    //    t += step;

    //    // the sample go through the whole medium section (no interaction)
    //    if(t >= tmax)
    //    {
    //        ms = MediumSample(
    //            ray(tmax),
    //            tmax,
    //            White(),
    //            exp(-(tmax-tmin) * muT).rgbsum() / strategies,
    //            ms.index,
    //            MediumSample::InteractionType::PassThrough
    //        );
    //        return false; // no medium interaction in the section (tmin, tmax)
    //    }

    //    // the sample is inside the medium section (real interaction)
    //    p = ray(t);
    //    Color muS = d * mSigmaS;
    //    Color barT = exp(-step * muT);
    //    float pdf = (muT * barT).rgbsum() / strategies; // spectral MIS

    //    // a medium interaction has been sampled
    //    ms = MediumSample(
    //        p,
    //        t,
    //        barT * muS / pdf,
    //        pdf,
    //        ms.index,
    //        MediumSample::InteractionType::RealInteraction
    //    );
    //    return true;
    //}

    virtual Color transmittance(Sampler& sampler, const Ray &ray, float tmin, float tmax, float * ratio = nullptr) const override
    {
        const float dt = std::max(0.f, tmax - tmin);

        // compute the extinction coefficient (constant)
        const Color muT = mDensity * mSigmaT;

        // compute transmittance at distance tmax
        return exp(-dt * muT);
    }

private:
    const float mDensity;
    const Color mLe;
};

//https://github.com/mmp/pbrt-v4/blob/master/src/pbrt/util/spectrum.h#L66
static inline float Blackbody(float lambda, float T) {
    if (T <= 0)
        return 0;
    const float c = 299792458;
    const float h = 6.62606957e-34;
    const float kb = 1.3806488e-23;
    // Return emitted radiance for blackbody at wavelength _lambda_
    float l = lambda * 1e-9f;
    float Le = (2 * h * c * c) / (std::pow(l, 5.f) * (std::exp((h * c) / (l * kb * T)) - 1));
    assert(!std::isnan(Le));
    //if(Le > 0.) std::cout << "Le(" << lambda << "):" << Le << std::endl;
    return Le;
}

static inline float NormalizedBlackbody(float lambda, float T) {
    // Compute blackbody normalization constant for given temperature
    float lambdaMax = 2.8977721e-3f / T;
    float normalizationFactor = 1. / Blackbody(lambdaMax * 1e9f, T);
    return Blackbody(lambda, T) * normalizationFactor;
}

class MediumHeterogeneous : public MediumUniform
{
public:
    MediumHeterogeneous(
        const Color& sigmaA,
        const Color& sigmaS,
        std::shared_ptr<DensityFunction> density = std::make_shared<DensityFunctionFastNoise>(),
        std::shared_ptr<Bsdf> phase = std::make_shared<BsdfHenyeyGreenstein>(),
        std::shared_ptr<DensityFunction> temperature = nullptr//std::make_shared<DensityFunctionConstant>(0.)
    ) :
        MediumUniform(phase, sigmaA, sigmaS),
        mDensity(density),
        mTemperature(temperature)
    {
        //mDensity = std::make_shared<DensityFunctionNanoVDB>("fire.nvdb", "temperature", Scale(40., 40., 40.));
        //mTemperature = std::make_shared<DensityFunctionNanoVDB>("fire.nvdb", "temperature", Scale(40., 40., 40.));
        //mTemperature = nullptr;
    }

    //bool isHomogeneous() const { return false; }
    //bool isHeterogeneous() const { return true; }

    virtual Color getLe(const Point& p, float time = 0) const override
    {
        if(mTemperature == nullptr)
            return Black();

        const float T = 100. * (mTemperature->eval(p, time) - 1.);
        //todo: do that better
        //const Color color = 0.0000001 * Color(
        const Color color = 5. * Color(
            NormalizedBlackbody(800., T),
            NormalizedBlackbody(600., T),
            NormalizedBlackbody(400., T)
        );
        //if(color.g > 0.)
        //    std::cout << "temperature " << t << " color " << color << std::endl;
        return color;
    }

    virtual float density(const Point& p, float time = 0) const override
    {
        return mDensity->eval(p, time);
    }

    virtual float maxDensity() const override
    {
        return mDensity->max();
    }

private:
    std::shared_ptr<DensityFunction> mDensity = nullptr;
    std::shared_ptr<DensityFunction> mTemperature = nullptr;
};

struct DiscreteNearestDistribution
{
    DiscreteNearestDistribution(std::vector<float> xi, std::vector<float> yi):xi(xi),yi(yi)
    {
        assert(xi.size());
        assert(yi.size() == xi.size());
    } 

    float operator()(float x) const
    {
        if(xi.size()==1) return yi[0];

        if( x < (xi[1]+xi[0])*0.5f ) return yi[0];
        for(int i = 1; i < (int)xi.size()-1; ++i)
            if( x < (xi[i+1]+xi[i])*0.5f ) 
                return yi[i];
        // if( x > (xi[1]+xi[0])*0.5f ) return yi[xi.size()-1];
        return yi[xi.size()-1];
    }

    std::vector<float> xi, yi;   
    std::vector<float> distances, values;   
};

struct MediumRecords
{
    std::vector<float> distances;
    std::vector<Color> barCOD; //cumulated optical depths of majorants
    std::vector<const Medium *> mediums;
    Ray ray;
    int size() const { return mediums.size(); }

    // i-th section length
    float length(int i) const
    {
        assert( i >= 0 );
        assert( i < size() );
        return distances[2*i+1] - distances[2*i];
    }

    // total sections length
    float length() const
    {
        float l = 0.f;
        for(int i = 0; i < size(); ++i)
            l += length(i);
        return l;
    }

    int depthIndex(const float t, int i = 0) const
    {
        assert(i >= 0);
        assert(i < size());

        assert(t >= distances[2*i] || distances[2*i] - t < DEBUG_EPS);
        assert(t <= distances.back() || t - distances.back() < DEBUG_EPS);

        // if(t == distances.front()) return 0;
        // if(t == distances.back()) return size()-1;

        while(distances[2*i + 1] < t)
            ++i;

        i = std::min(i, size());

        //for(; i < size(); ++i)
        //    if( t >= distances[2*i] && t <= distances[2*i+1] )
        //        break;
        //// if( i >= size() ) printf("Error MediumSample::index %f %i %f %f\n", t, i, distances.front(), distances.back());
        //assert(i < size());

        //std::cout << "DEBUG t:" << t << " " << distances[2*i+1] << std::endl;
        assert(distances[2*i] < t || distances[2*i] - t < DEBUG_EPS);
        assert(t < distances[2*i+1] || t - distances[2*i+1] < DEBUG_EPS);
        return i;
    }

    // return normalized ratio in [0,1] from real depth 
    float depthToRatio(const float t, const float len) const
    {
        assert(t >= distances.front());
        assert(t <= distances.back());
        float lt = 0.f;
        for(int i = 0; i < size(); ++i)
        {
            const float a = distances[2*i];
            const float b = distances[2*i+1];
            // cumulate section length until we reach t
            lt += std::min(t,b) - a;
            if( t <= b )
                break;
        }
        // float r = std::min(1.f, std::max(0.f, lt / length()));
        float r = std::min(1.f, std::max(0.f, lt / len));
        return r;
    }

    // return real depth from normalized ratio
    float ratioToDepth(const float ratio, const float len) const
    {
        // assert(ratio >= 0.f);
        // assert(ratio <= 1.f);

        if( ratio == 1.f ) return distances.back();
        if( ratio == 0.f ) return distances.front();

        // const float total = length();
        const float lt = ratio * len;   // length in medium to real position
        float l = 0.f;                  // passed length in medium
        float t = distances.front();
        for(int i = 0; i < size(); ++i)
        {
            const float a = distances[2*i];
            const float b = distances[2*i+1];
            l += b - a;
            // the real position is in this section
            if( lt <= l )
            {
                // from the end of the section we have to move backwards of
                // delta = (l - lt) to find the real depth
                t = b - (l - lt);
                break;
            }
        }

        // assert(t >= distances.front());
        // assert(t <= distances.back());
        return t;
    }

    // i-th section mod
    float mod(int i, const int channel) const
    {
        assert( i >= 0 );
        assert( i < size() );
        return (distances[2*i+1] - distances[2*i]) * mediums[i]->maxMuT()[channel];
    }

    // total sections mod
    float mod(const int channel) const
    {
        float m = 0.f;
        for(int i = 0; i < size(); ++i)
            m += mod(i, channel);
        return m;
    }

    // return normalized majorant optical depth in [0,1] from real depth 
    float depthToModRatio(const float t, const float modmax, const int channel) const
    {
        assert(t >= distances.front());
        assert(t <= distances.back());
        float modt = 0.f;
        for(int i = 0; i < size(); ++i)
        {
            const float a = distances[2*i];
            const float b = distances[2*i+1];
            // cumulate section length until we reach t
            modt += (std::min(t,b) - a) * mediums[i]->maxMuT()[channel];
            if( t <= b )
                break;
        }
        // float r = std::min(1.f, std::max(0.f, modt / length()));
        float r = std::min(1.f, std::max(0.f, modt / modt));
        return r;
    }

    // return real depth from normalized ratio of majorant optical depth
    float modRatioToDepth(const float ratio, const float modmax, const int channel) const
    {
        // assert(ratio >= 0.f);
        // assert(ratio <= 1.f);

        if( ratio == 1.f ) return distances.back();
        if( ratio == 0.f ) return distances.front();

        // const float total = length();
        const float modt = ratio * modmax; // majorant optical depth to reach to find the real position
        float mod = 0.f;                   // passed optical depth in medium
        float t = distances.front();
        for(int i = 0; i < size(); ++i)
        {
            const float a = distances[2*i];
            const float b = distances[2*i+1];
            mod += (b - a) * mediums[i]->maxMuT()[channel];
            // the real position is in this section
            if( modt <= mod )
            {
                // from the end of the section we have to move backwards of
                // delta = (mod - modt) / barMu to find the real depth
                t = b - (mod - modt) / mediums[i]->maxMuT()[channel];
                break;
            }
        }

        // assert(t >= distances.front());
        // assert(t <= distances.back());
        return t;
    }

    void buildBarCOD()
    {
        barCOD.clear();

        Color od = Black();
        for(int i = 0; i < size(); ++i)
        {
            od += length(i) * mediums[i]->maxMuT();
            barCOD.push_back(od);
        }
    }

    // i-th section optical depth of majorant
    Color barTau(int i) const
    {
        assert(int(barCOD.size()) == size());
        if(i == 0) return barCOD[i];
        assert( i > 0 );
        assert( i < size() );
        return barCOD[i] - barCOD[i - 1];
    }

    // total sections optical depth of majorants
    Color barTau() const
    {
        assert(int(barCOD.size()) == size());
        return size() > 0 ? barCOD[size() - 1] : Black();
    }

    int barTauIndex(const float tOD, const unsigned int channel, int i = 0) const
    {
        assert(size() > 0);
        assert(tOD >= 0.);
        assert(tOD <= barCOD.back()[channel]);

        for(; tOD > barCOD[i][channel]; ++i);
        //std::cout << "DEBUG i: " << i << " " << size() << std::endl;
        //std::cout << "DEBUG tOD: " << tOD << " " << barCOD.back()[channel] << std::endl;
        i = std::min(i, size());
        assert(tOD < barCOD[i][channel] || abs(tOD - barCOD[i][channel]) < DEBUG_EPS);
        return i;
    }

    Color nextBarTau(const float t, int& i) const
    {
        assert(i >= 0);
        assert(i < size());
        assert(t >= distances.at(i));
        assert(t <= distances.back());

        //Find section/segment associated to t.
        const int oldI = i;
        i = depthIndex(t, i);
        const float a = distances[2*i];
        const float b = distances[2*i+1];

        if(a == b)
        {
            std::cout << "WARNING: null length section!" << std::endl;
            return barCOD[i];
        }
        assert(length(i) > 0.);

        //std::cout << "DEBUG section ratio: [" << a << ", " << b << "] t=" << t << std::endl;
        //std::cout << "DEBUG i i': " << oldI << " " << i << std::endl;
        assert(a <= t || a - t < DEBUG_EPS);
        assert(t <= b || t - b < DEBUG_EPS);
        //const float sectionRatio = (b - std::min(b, std::max(a, t)))/(b - a);
        const float sectionRatio = (t - a)/length(i);
        //std::cout << "DEBUG section ratio= " << sectionRatio << std::endl;
        assert(0.f < sectionRatio  || 0.f - sectionRatio < DEBUG_EPS);
        assert(sectionRatio < 1.f || sectionRatio - 1.f < DEBUG_EPS);
        return (i > 0.f ? barCOD[i - 1] : Black()) + barTau(i) * sectionRatio;
    }
    // return majorant opticalDepth between t0 and t1
    Color barTau(const float t0, const float t1) const
    {
        if( t0 == t1 ) return Black();

        assert(t0 < t1);

        int i = 0;
        const Color barTau0 = nextBarTau(t0, i);
        return nextBarTau(t1, i) - barTau0;
    }
    Color barTau(const float t) const
    {
        return barTau(distances[0], t);
    }
    //Color barTauBackup(const float t0, const float t1) const
    //{
    //    assert(t0 >= distances.front());
    //    assert(t0 <= distances.back());

    //    assert(t1 >= distances.front());
    //    assert(t1 <= distances.back());

    //    assert(t0 <= t1);

    //    if( t0 == t1 ) return Black();

    //    Color barTau = Black();
    //    for(int i = 0; i < size(); ++i)
    //    {
    //        const float a = distances[2*i];
    //        const float b = distances[2*i+1];

    //        if( t0 > b || t1 < a )
    //            continue;
    //        else
    //        {
    //            const float l = std::min(t1, b) - std::max(t0, a);
    //            barTau += l * mediums[i]->maxMuT();
    //        }
    //    }
    //    return barTau;
    //}

    float barTauRatio(const Color& barTau, const unsigned int channel) const
    {
        const float tOD = barTau[channel];
        assert(tOD >= 0.f);
        const float tMax = this->barTau()[channel];
        assert(tOD <= tMax);

        if(tMax == 0.f)
        {
            std::cout << "WARNING: null optical depth section!" << std::endl;
            return 0.f;
        }

        const float r = tOD/tMax;
        //std::cout << "DEBUG ratio: " << "tOD=" << tOD << " tMax=" << tMax << " r=" << r << std::endl;
        //const int i = barTauIndex(tOD, channel);
        //assert(i < size());
        //const float r = (tOD - barCOD[i][channel]) / (this->barTau(i)[channel]);
        ////std::cout << "DEBUG i: " << i << " " << size() << std::endl;
        ////std::cout << "DEBUG r: " << r << std::endl;
        //////std::cout << "DEBUG tOD: " << tOD << " " << barCOD.back()[channel] << std::endl;
        assert(r > 0.f || abs(r) < DEBUG_EPS);
        assert(r < 1.f || abs(r - 1.f) < DEBUG_EPS);
        return std::min(1.f, std::max(0.f, r));
    }
    //real depth to optical depth majorant ratio
    float depthToBarTauRatio(const float t, const unsigned int channel) const
    {
        return barTauRatio(barTau(t), channel);
    }
    //real depth from optical depth majorant ratio
    float barTauRatioToDepth(const float ratio, const unsigned int channel) const
    {
        const float tOD = ratio * barTau()[channel];
        const int i = barTauIndex(tOD, channel);

        if(barTau(i)[channel] == 0.f)
        {
            std::cout << "WARNING: null optical depth section!" << std::endl;
            return distances[2*i + 1];
        }

        const float sectionRatio = (tOD - (i > 0 ? barCOD[i - 1][channel] : 0.f))/(barTau(i)[channel]);
        //std::cout << "DEBUG section:" << "tOD=" << tOD << " i=" << i << " sectionRatio=" << sectionRatio << std::endl;
        const float a = distances[2*i];
        const float b = distances[2*i + 1];
        //return (1.f - sectionRatio)*a + sectionRatio*b;
        return a + sectionRatio*length(i);
    }

    Color barT(const float t0, const float t1) const
    {
        return exp(-barTau(t0, t1));
    }

    MediumRecords clamped(const float tmax) const
    {
        MediumRecords clampedRecords;
        clampedRecords.ray = ray;

        const unsigned int n = size();
        for(unsigned int i = 0u; i < n && tmax >= distances[2*i]; i++)
        {
            clampedRecords.mediums.push_back(mediums[i]);
            clampedRecords.distances.push_back(distances[2*i]);
            clampedRecords.distances.push_back(std::min(tmax, distances[2*i+1]));

            //if( tmax < distances[2*i+1] )
            //    break;
        }

        return clampedRecords;
    }
};

static bool SampleEquiangular(Sampler& sampler, const LightSample& light, const MediumRecords& records, float& distance, float& pdf)
{
    // test if there is any medium section 
    if( !records.size() )
        return false;

    // cf : Importance sampling techniques for path tracing in participating media, Kulla et al, EGSR 2012
    // cf : Monte Carlo Methods for Volumetric Light Transport Simulation 2018

    // get coord of closest point to light along (infinite) ray
    float th = std::abs(dot(light.s - records.ray.o(), records.ray.nd()));

    // get distance this point is from light
    float h = length(records.ray.o() + th * records.ray.nd() - light.s);
   
    // get angle of endpoints for each medium section and
    // compute piecewise angular cdf and total angular extent
    float total = 0;
    std::vector<float> cdf;
    cdf.reserve(records.size());
    for(int i = 0; i < records.size(); ++i)
    {
        float thetaA = std::atan2(records.distances[2*i] - th, h);
        float thetaB = std::atan2(records.distances[2*i+1] - th, h);
        total += thetaB - thetaA;
        cdf.push_back(total);
    }

    // binary search
    float u = sampler.sample();
    float r = u * total;
    int id = -1, p = 0, q = (int)cdf.size()-1;
    while(p < q)
    {
        int m = (p+q) / 2;
        if(cdf[m] < r)  p = m + 1;
        else            q = m;
    }
    assert(p >= 0 && p < (int)cdf.size());
    id = p;

    // normalize u : cdf[p-1] < r < cdf[p]
    float pp = 0;
    if(id > 0) pp = cdf[id-1];
    u = (r - pp) / (cdf[id] - pp);

    // sample distance
    float thetaA = std::atan2(records.distances[2*id] - th, h);
    float thetaB = std::atan2(records.distances[2*id+1] - th, h);
    float t = h * std::tan(thetaA + u*(thetaB-thetaA));
    
    // return sampled distance and pdf
    distance = th + t;
    pdf = h/(total*(h*h + t*t));

    return true;
}

static float PdfEquiangular(const LightSample& light, const MediumRecords& records, const float distance)
{   
    // test if there is any medium section 
    if( !records.size() )
        return 0;

    // cf : Importance sampling techniques for path tracing in participating media, Kulla et al, EGSR 2012
    // cf : Monte Carlo Methods for Volumetric Light Transport Simulation 2018

    // get coord of closest point to light along (infinite) ray
    float th = std::abs(dot(light.s - records.ray.o(), records.ray.nd()));
    
    // get distance this point is from light
    float h = length(records.ray.o() + th * records.ray.nd() - light.s);
      
    // get total angular extent
    float total = 0;
    for(int i = 0; i < records.size(); ++i)
    {
        float thetaA = std::atan2(records.distances[2*i] - th, h);
        float thetaB = std::atan2(records.distances[2*i+1] - th, h);
        total += thetaB - thetaA;
    }

    // distance (positive or negative) from the projected light position
    float t = distance - th;

    // return pdf
    return h / (total*(h*h + t*t));
}

// Unbiased spectral MIS + unbiased pdf evaluation [Miller et al. 2019]
static void EvalTransmittanceMiller(Sampler& sampler, const MediumRecords& records, Color& tr, float * ratio = nullptr)
{
    tr = White();
    if( ratio ) *ratio = 1;
    const Ray& ray = records.ray;
    for(int i = 0; i < records.size(); ++i)
    {
        float tmin = records.distances[2*i];
        float tmax = records.distances[2*i+1];
        const Medium * const medium = records.mediums[i];

        tr *= medium->transmittance(sampler, ray, tmin, tmax, ratio);
    }
    return;
}

// Unbiased spectral MIS + unbiased pdf evaluation [Miller et al. 2019]
static bool SampleMediumRecordsMiller(Sampler& sampler, const MediumRecords& records, MediumSample& ms, float * ratio = nullptr)
{
    if( records.size() == 0 )
        return false;

    if( ratio ) *ratio = 1;

    const Ray& ray = records.ray;
    Color w = White();
    float pdf = 1.;
#ifdef LOG_PDFS
    Color lf = Black();
    Color lpdfs = Black();
#endif
    std::pair<unsigned int, float> channelSample = {0u, 1./3.};
    for(int i = 0; i < records.size(); ++i)
    {
        const float tmin = records.distances[2*i];
        const float tmax = records.distances[2*i+1];
        const Medium * const medium = records.mediums[i];

        const bool realInteraction = medium->sample(sampler, ray, tmin, tmax, ms, channelSample, ratio);

        // if smthing went wrong in sampling
        // if( ms.pdf <= 0 )
            // return false;

        w *= ms.w;
        pdf *= ms.pdf;
#ifdef LOG_PDFS
        lf += ms.lf;
        lpdfs += ms.lpdfs;
#endif

        if(realInteraction)
        {
            ms = MediumSample(
                ms.p,
                ms.t,
#ifndef LOG_PDFS
                w,
#else
                exp(lf)/(channelSample.second*exp(lpdfs).rgbsum()), //log to real w
#endif
                ms.channel,
                ms.channelPdf,
                pdf,
                i,
                ms.type
            );
#ifdef LOG_PDFS
            ms.lf = lf;
            ms.lpdfs = lpdfs;
#endif
            return true;
        }
    }

    // Passthrough.
    ms = MediumSample(
        ray(records.distances.back()),
        records.distances.back(),
#ifndef LOG_PDFS
        w,
#else
        //exp(lf - lpdfs),
        exp(lf)/(channelSample.second*exp(lpdfs).rgbsum()), //log to real w
#endif
        channelSample.first,
        channelSample.second,
        pdf,
        records.size() - 1,
        MediumSample::InteractionType::PassThrough
    );
#ifdef LOG_PDFS
    ms.lf = lf;
    ms.lpdfs = lpdfs;
#endif
    return false;
}

// Unbiased spectral MIS + unbiased pdf evaluation [Miller et al. 2019]
static bool SampleChainMediumRecordsMiller(Sampler& sampler, const MediumRecords& records, std::vector<MediumSample>& chain)
{
    const Ray& ray = records.ray;

    Color w = White();

    Color ws = White();
    Color wp = White();
    Color pdfs = White();
#ifdef LOG_PDFS
    Color lf = Black();
    Color lpdfs = Black();
#endif
    float opticalDepthStepRemainder = 0.;
    std::pair<unsigned int, float> channelSample = {0u, 1./3.};
    for(int i = 0; i < records.size(); ++i)
    {
        const float tmin = records.distances[2*i];
        const float tmax = records.distances[2*i+1];
        const Medium * const medium = records.mediums[i];

        if(medium->sampleChain(
            sampler, ray, tmin, tmax, chain, w, ws, wp, pdfs,
#ifdef LOG_PDFS
            lf, lpdfs,
#endif
            i, opticalDepthStepRemainder, channelSample
        ))
            return true;
    }

    // Final passthrough.
    const unsigned int& channel = channelSample.first;
    const float& channelPdf = channelSample.second;
    const float pdf = pdfs.rgbsum() * channelPdf; // spectral MIS case
    const float tmax = records.size() > 0 ? records.distances.back() : 0.;
    chain.emplace_back(
        ray(tmax),
        tmax,
#ifndef LOG_PDFS
        //w, //not consistent with spectral MIS
        ws / (channelPdf*wp.rgbsum()),
#else
        exp(lf)/(channelSample.second*exp(lpdfs).rgbsum()), //log to real w
#endif
        channel,
        channelPdf,
        pdf,
        records.size() - 1,
        MediumSample::InteractionType::PassThrough
    );
#ifdef LOG_PDFS
    chain.back().lf = lf;
    chain.back().lpdfs = lpdfs;
#endif
    return false;
}
/*
struct ShiftMappingUtils
{
    ShiftMappingUtils(
        const std::vector<MediumSample>& primaryChain,
        const MediumRecords& primaryRecords,
        const MediumRecords& secondaryRecords )
    {
        // source
        std::vector<float> src_distances;
        std::vector<float> src_ratios;
        std::vector<Color> src_mu;
        std::vector<Color> src_mun;
        std::vector<float> src_depth;

        // start
        src_distances.push_back(primaryRecords.distances.front());
        src_ratios.push_back(0);
        // mid
        const float primaryTotalLength = primaryRecords.length();
        for(int i = 1; i < int(chain.size()-1); i++)
        {
            const float depth = (primaryChain[i-1].t+primaryChain[i].t)*0.5f;
            const float ratio = primaryRecords.depthToRatio(depth, primaryTotalLength);
            src_depth.push_back(primaryChain[i-1].t);
            src_distances.push_back(depth);
            src_distances.push_back(depth);
            src_ratios.push_back(ratio);
            src_ratios.push_back(ratio);

            const int index = primaryRecords.depthIndex(primaryChain[i-1].t);
            std::shared_ptr<Medium> medium = primaryRecords.mediums[index];
            src_mu.push_back(medium->getMuT(primaryChain[i-1].t, primaryRecords.ray.time()));
            src_mun.push_back(medium->getMuN(primaryChain[i-1].t, primaryRecords.ray.time()));
        }
        // end
        src_depth.push_back(primaryChain[chain.size()-1].t);
        src_distances.push_back(primaryRecords.distances.back());
        src_ratios.push_back(1);
        const int index = primaryRecords.depthIndex(primaryChain[chain.size()-1].t);
        std::shared_ptr<Medium> medium = primaryRecords.mediums[index];
        src_mu.push_back(medium->getMuT(primaryChain[chain.size()-1].t, primaryRecords.ray.time()));
        src_mun.push_back(medium->getMuN(primaryChain[chain.size()-1].t, primaryRecords.ray.time()));

        // destination
        std::vector<float> dst_distances;
        std::vector<float> dst_ratios;
        std::vector<float> dst_depth;
        std::vector<Color> dst_mu;
        std::vector<Color> dst_mun;
        // start
        dst_distances.push_back(secondaryRecords.distances.front());
        dst_ratios.push_back(0);
        // mid
        const float secondaryTotalLength = secondaryRecords.length();
        for(int i = 1; i < int(chain.size()-1); i++)
        {
            const float depth = src_depth[i];
            const float ratio = src_ratios[2*i];
            dst_distances.push_back(depth);
            dst_distances.push_back(depth);
            dst_ratios.push_back(ratio);
            dst_ratios.push_back(ratio);

            const int index = secondaryRecords.depthIndex(depth);
            std::shared_ptr<Medium> medium = secondaryRecords.mediums[index];
            dst_mu.push_back(medium->getMuT(depth, secondaryRecords.ray.time()));
            dst_mun.push_back(medium->getMuN(depth, secondaryRecords.ray.time()));
        }
        // end
        dst_distances.push_back(secondaryRecords.distances.back());
        dst_ratios.push_back(1);
        const int index = secondaryRecords.depthIndex(secondaryChain[chain.size()-1].t);
        std::shared_ptr<Medium> medium = secondaryRecords.mediums[index];
        dst_mu.push_back(medium->getMuT(secondaryChain[chain.size()-1].t, secondaryRecords.ray.time()));
        dst_mun.push_back(medium->getMuN(secondaryChain[chain.size()-1].t, secondaryRecords.ray.time()));

    }
};
*/
#ifdef SINGLE_THREAD
    #define DEBUG_SHIFT(I) I
#else
    #define DEBUG_SHIFT(I)
#endif

static std::size_t chainlen = 0;
static std::size_t chaincmp = 0;


static bool ShiftMediumRecordsChain(
    Sampler& sampler,
    MediumRecords primaryRecords,
    const std::vector<MediumSample>& chain,
    MediumRecords secondaryRecords,
    std::vector<MediumSample>& shiftedChain,
    float& shiftedChainWeight,
    float& J
    )
{
    // Clamping medium records to area of interest.
    // This is needed since the last event sampled may be between start and end points.
    // On contrary the secondary records are built by construction until the last event known thus no clamp is necessary.
    if(chain.size() && chain.back().type != MediumSample::InteractionType::PassThrough && primaryRecords.size() > 0)
    {
        // Primary records.
        const float tmax = chain.back().t;
        primaryRecords = primaryRecords.clamped(tmax);
    }

    // shiftedChainWeight stores the ratio of pdfs that could generate the path 
    // strategy 1. primary obs + shift
    // strategy 2. secondary obs
    shiftedChainWeight = 1;

    // 1. hit surface no medium sections - 2. hit surface no medium sections
    if( primaryRecords.size() == 0 && secondaryRecords.size() == 0 )
    {
        const MediumSample& ms = chain.back();
        assert(ms.type == MediumSample::InteractionType::PassThrough);
        shiftedChain.emplace_back(ms);
        return true;
    }
    // 1. hit surface no medium sections - 2. hit surface with medium sections
    else if( primaryRecords.size() == 0 )
    {
        // corner case, for now pdfs are not comparable maybe later
        return false;
    }
    // 1. hit surface with medium sections - 2. hit surface no medium sections
    else if( secondaryRecords.size() == 0 )
    {
        // corner case, for now pdfs are not comparable maybe later
        return false;
    }

    // Remove surface interactions seen throught mediums
    // if( chain.back().type == MediumSample::InteractionType::PassThrough )
        // return false;

    // Keep only surface interactions seen throught mediums
    // if( chain.back().type != MediumSample::InteractionType::PassThrough )
       // return false;

    // Keep all reprojections except those which involve 2 medium sections
    // This converges OK
    // if( primaryRecords.size() == 2 || secondaryRecords.size() == 2 )
       // return false;

    // Remove all reprojections except those which involve 2 medium sections
    // if( primaryRecords.size() != 2 && secondaryRecords.size() != 2 )
       // return false;

    // Remove all reprojections except those which involve 2 medium sections for secondary 
    //if( secondaryRecords.size() != 2 )
    //    return false;

    primaryRecords.buildBarCOD();
    secondaryRecords.buildBarCOD();

    const float primaryTotalLength = primaryRecords.length();
    const float secondaryTotalLength = secondaryRecords.length();
    const Ray& ray(secondaryRecords.ray);
    int secondaryIndex = 0;
    //Color w = White();
    Color ws = White();
    Color wp = White();
    Color pdfs = White();
#ifdef LOG_PDFS
    Color lf = Black();
    Color lpdfs = Black();
#endif

    // jacobian of primary shifted chain
#ifdef OPTICAL_DEPTH_SHIFT
    float jacobian;
#else
    const float jacobian = secondaryTotalLength/primaryTotalLength;
#endif

    float t = secondaryRecords.distances.front();
    for(int i = 0; i < int(chain.size()); i++)
    {
        const MediumSample& ms = chain[i];
        const unsigned int channel = ms.channel;

        //switch(ms.type)
        //{
        //    case MediumSample::InteractionType::PassThrough: DEBUG_SHIFT(std::cout << "pass through";) break;
        //    case MediumSample::InteractionType::NullInteraction: DEBUG_SHIFT(std::cout << "null interaction";) break;
        //    case MediumSample::InteractionType::RealInteraction: DEBUG_SHIFT(std::cout << "real interaction";) break;
        //    case MediumSample::InteractionType::Absorbtion: DEBUG_SHIFT(std::cout << "absorbtion interaction";) break;
        //    case MediumSample::InteractionType::Scattering: DEBUG_SHIFT(std::cout << "scattering interaction";) break;
        //}

        // Shift 1 : lengths ratio.
        // Compute shifted depth and 
        // check utility functions in MediumRecords class for details
        // 1. compute the depth ratio on the primary segment
        // 2. compute the depth on the secondary segment
        // 3. compute the index of the medium section containing the point
        // 4. compute the maj transmittance form last depth to current 

#ifdef OPTICAL_DEPTH_SHIFT
        const float barTauRatio = primaryRecords.depthToBarTauRatio(ms.t, channel);
        if( barTauRatio < 0.f || barTauRatio > 1.f ) return false;

        const float nextT = secondaryRecords.barTauRatioToDepth(barTauRatio, channel);
        if( nextT < secondaryRecords.distances.front() || nextT > secondaryRecords.distances.back() ) return false;

        //const float t1 = ms.t;
        //std::cout << "t1: " << ms.t << std::endl;
        //std::cout << "barTauRatio: " << barTauRatio << std::endl;
        //const float t1bis = primaryRecords.barTauRatioToDepth(barTauRatio, channel);
        //std::cout << "t1': " << t1bis << std::endl;
        //assert(abs(t1 - t1bis) < DEBUG_EPS);
        //const float t2 = nextT;
        //std::cout << "t2: " << t2 << std::endl;
        //const float barTauRatioBis = secondaryRecords.depthToBarTauRatio(nextT, channel);
        //std::cout << "barTauRatio': " << barTauRatioBis << std::endl;
        //const float t1ter = primaryRecords.barTauRatioToDepth(barTauRatioBis, channel) ;
        //std::cout << "t1'': " << t1ter << std::endl;
        //assert(abs(t1 - t1ter) < DEBUG_EPS);

        // float primaryTotalMod = primaryRecords.mod(channel);
        // float secondaryTotalMod = secondaryRecords.mod(channel);
        // const float depthRatio = primaryRecords.depthToModRatio(ms.t, primaryTotalMod, channel);
        // if( depthRatio < 0.f || depthRatio > 1.f ) return false;

        // const float nextT = secondaryRecords.modRatioToDepth(depthRatio, secondaryTotalMod, channel);
        // if( nextT < secondaryRecords.distances.front() || nextT > secondaryRecords.distances.back() ) return false;

#else
        const float depthRatio = primaryRecords.depthToRatio(ms.t, primaryTotalLength);
        if( depthRatio < 0.f || depthRatio > 1.f ) return false;

        const float nextT = secondaryRecords.ratioToDepth(depthRatio, secondaryTotalLength);
        if( nextT < secondaryRecords.distances.front() || nextT > secondaryRecords.distances.back() ) return false;
#endif

        secondaryIndex = secondaryRecords.depthIndex(nextT, secondaryIndex);
        const Color barTau = secondaryRecords.barTau(t, nextT);
        const Medium * const medium = secondaryRecords.mediums[secondaryIndex];
        const float channelPdf = medium->maxMuTChannelPdf(channel);
        const Color barT = exp(-barTau);
        //w *= barT/barT;
        ws *= barT/barT[channel];
        wp *= barT/barT[channel];
        pdfs *= barT;
#ifdef LOG_PDFS
        lf -= barTau;
        lpdfs -= barTau;
#endif
        t = nextT;

        const Color barMu = medium->maxMuT();
        const Point p = ray(t);
        if( ms.type != MediumSample::InteractionType::PassThrough )
        {
            const std::pair<Color, Color> mu = medium->getMu(p, ray.time());
            const Color& muA = mu.first;
            const Color& muS = mu.second;
            const Color muT = muA + muS;
            const Color muN = barMu - muT;

            //const float pA = muA[channel] / barMuT[channel]; // absorbtion prob
            //const float pS = muS[channel] / barMuT[channel]; // scattering prob
            //const float pT = pA + pS; // real interaction prob
            //const float pN = muN[channel] / barMu[channel]; // null collision prob

            // const float pA = medium->getLe(p, ray.time()).power() > 0 ? muA[channel] / barMu[channel] : 0; // absorbtion prob
            const float pA = muA[channel] / barMu[channel]; // absorbtion prob
            const float pS = pA > 0 ? muS[channel] / barMu[channel] : muT[channel] / barMu[channel]; // scattering prob
            const float pT = pA + pS; // real interaction prob
            const float pN = muN[channel] / barMu[channel]; // null collision prob

            DEBUG_SHIFT(std::cout << "\tmuN: " << muN << std::endl;)
            DEBUG_SHIFT(std::cout << "\tmuS: " << muS << std::endl;)
            DEBUG_SHIFT(std::cout << "\tmuT: " << muT << std::endl;)

            if( ms.type == MediumSample::InteractionType::Absorbtion )
            {
                //w *= muA/muA;
                ws *= muA/muA[channel];
                wp *= muA/muA[channel];
                pdfs *= muA;
#ifdef LOG_PDFS
                lf += log(muA);
                lpdfs += log(muA);
#endif
            }
            else if( ms.type == MediumSample::InteractionType::Scattering )
            {
                //w *= muS/muS;
                ws *= pA > 0 ? muS/muS[channel] : muS/muT[channel];
                wp *= pA > 0 ? muS/muS[channel] : muT/muT[channel];
                pdfs *= pA > 0 ? muS : muT;
#ifdef LOG_PDFS
                lf += log(muS);
                lpdfs += log(muS);
#endif
            }
            else if( ms.type == MediumSample::InteractionType::NullInteraction )
            {
                //w *= muN/muN;
                ws *= muN/muN[channel];
                wp *= muN/muN[channel];
                pdfs *= muN;
#ifdef LOG_PDFS
                lf += log(muN);
                lpdfs += log(muN);
#endif
            }
            // mixed real interaction (scattering + absorbtion) : here appears the albedo muS/muT
            else if( ms.type == MediumSample::InteractionType::RealInteraction )
            {
                //w *= muS / muT; // not consistent with spectral MIS TODO change later
                ws *= muS/muT[channel];
                wp *= muT/muT[channel];
                pdfs *= muT;
#ifdef LOG_PDFS
                lf += log(muS);
                lpdfs += log(muT);
#endif
            }
            else assert(false);
        }

        const float pdf = pdfs.rgbsum() * channelPdf; // spectral MIS
        pdfs = White();
        shiftedChain.emplace_back(
            p,
            t,
#ifndef LOG_PDFS
            //w, //not consistent with spectral MIS
            ws/(channelPdf*wp.rgbsum()),
#else
            exp(lf)/(channelSample.second*exp(lpdfs).rgbsum()), //log to real w
#endif
            channel,
            channelPdf,
            pdf,
            secondaryIndex,
            ms.type
        );
        // WARNING HERE IMPORTANT FORMULAE
        // chainWeight = jointPdfSecondary / (jointPdfPrimary * inverseShiftJacobian)
#ifndef LOG_PDFS
        shiftedChainWeight *= pdf/ms.pdf;
#else
        shiftedChain.back().lf = lf;
        shiftedChain.back().lpdfs = lpdfs;
        assert(!std::isinf(lf[0]));
        assert(!std::isnan(lf[0]));
        assert(!std::isinf(lpdfs[0]));
        assert(!std::isnan(lpdfs[0]));
#endif

#ifdef OPTICAL_DEPTH_SHIFT
        jacobian = secondaryRecords.barTau()[channel] / primaryRecords.barTau()[channel] * primaryRecords.mediums[ms.index]->maxMuT()[channel] / barMu[channel];
        assert(ms.index != -1);
#endif

        // WARNING ONLY NULL INTERACTION ARE SHIFTED HENCE NEED A JACOBIAN NOT REAL ONES
        if( chain[i].type == MediumSample::InteractionType::NullInteraction ){
            shiftedChainWeight *= jacobian;
            J *= jacobian;
        }

        // DEBUG_SHIFT(std::cout << "\tpdf: " << pdf << std::endl;)
        // DEBUG_SHIFT(std::cout << "\tshiftw: " << shiftedChainWeight << std::endl;)


        if( std::isinf(shiftedChainWeight) )
            return false;
        if( std::isnan(shiftedChainWeight) )
            return false;
        //if( shiftedChainWeight <= 0. )
        //    return false;
        //assert(!std::isinf(shiftedChainWeight));
        //assert(!std::isnan(shiftedChainWeight));
    }

    if( false )
    {
        printf("primary chain records %d interaction %d\n", primaryRecords.size(), (int)chain.size());
        printf("total length %f\n", primaryTotalLength);
        for(int i = 0; i< (int)primaryRecords.size(); ++i)
            printf("section %d tmin %f tmax %f length %f\n",
                i,
                primaryRecords.distances[2*i],
                primaryRecords.distances[2*i+1],
                primaryRecords.distances[2*i+1] - primaryRecords.distances[2*i]
            );
        for(int i = 0; i< (int)chain.size(); ++i)
            printf("link %d wpow %f t %f p (%f %f %f) pdf %f record %d ratio %f tbis %f type %s\n",
                i,
                chain[i].w.power(),
                chain[i].t,
                chain[i].p.x,
                chain[i].p.y,
                chain[i].p.z,
                chain[i].pdf,
                chain[i].index,
                primaryRecords.depthToRatio(chain[i].t, primaryTotalLength),
                primaryRecords.ratioToDepth(primaryRecords.depthToRatio(chain[i].t, primaryTotalLength), primaryTotalLength),
                (chain[i].type == MediumSample::InteractionType::RealInteraction ? "Real" : "Other")
            );
        printf("secondary chain records %d interaction %d\n", secondaryRecords.size(), (int)shiftedChain.size());
        printf("total length %f\n", secondaryTotalLength);
        for(int i = 0; i< (int)secondaryRecords.size(); ++i)
            printf("section %d tmin %f tmax %f length %f\n",
                i,
                secondaryRecords.distances[2*i],
                secondaryRecords.distances[2*i+1],
                secondaryRecords.distances[2*i+1] - secondaryRecords.distances[2*i]
            );
        for(int i = 0; i< (int)shiftedChain.size(); ++i)
            printf("link %d wpow %f t %f p (%f %f %f) pdf %f record chain %d record %d ratio %f tbis %f type %s\n",
                i,
                shiftedChain[i].w.power(),
                shiftedChain[i].t,
                shiftedChain[i].p.x,
                shiftedChain[i].p.y,
                shiftedChain[i].p.z,
                shiftedChain[i].pdf,
                shiftedChain[i].index,
                secondaryRecords.depthIndex(shiftedChain[i].t),
                secondaryRecords.depthToRatio(shiftedChain[i].t, secondaryTotalLength),
                secondaryRecords.ratioToDepth(secondaryRecords.depthToRatio(shiftedChain[i].t, secondaryTotalLength), secondaryTotalLength),
                (shiftedChain[i].type == MediumSample::InteractionType::RealInteraction ? "Real" : "Other")
            );
        printf("jacobian %f\n", jacobian);
        printf("shiftw %f\n", shiftedChainWeight);
        if( chain.back().type == MediumSample::InteractionType::RealInteraction )
            printf("pivot point (%f %f %f)\n\n", chain.back().p.x, chain.back().p.y, chain.back().p.z);
    }

#ifdef LOG_PDFS
    shiftedChainWeight *= std::exp((shiftedChain.back().lpdfs - chain.back().lpdfs)[0]);
#endif


    // #pragma omp atomic
    // chainlen+=shiftedChain.size();
    // #pragma omp atomic
    // chaincmp++;

    // printf("mean shift length%f\n", double(chainlen)/double(chaincmp));

    return true;
}

