#pragma once

static void uniformDistanceSample( const float u, const float tmax, float& distance, float& pdf )
{
    distance = u * tmax;
    pdf = 1.f / tmax;
}

static float uniformDistancePdf(const float tmax)
{
    return 1.f / tmax;
}

static void transUnboundedDistanceSample( const float u, const float sigma, float& distance, float& pdf )
{
    // cf : Production Volume Rendering 2017
    // equation 15
    distance = -std::log(1 - u) / sigma;
    // equation 14
    // with : exp(-sigma*t) = exp(-sigma*-ln(1-u)/sigma) = exp(ln(1-u)) = 1-u
    pdf = sigma * (1 - u); 
}

static float transUnboundedDistancePdf( const float t, const float sigma )
{
    // cf : Production Volume Rendering 2017
    // with : exp(-sigma*t) = exp(-sigma*-ln(1-u)/sigma) = exp(ln(1-u)) = 1-u
    return sigma * exp(-sigma * t); 
}

static void transBoundedDistanceSample( const float u, const float tmax, const float sigma, float& distance, float& pdf )
{
    // cf : Production Volume Rendering 2017
    float transMin = std::exp(-sigma * tmax); 
    float trans = 1.f - u * (1.f - transMin);
    // equation 29
    distance = -std::log(trans)/sigma;
    // equation 28
    // pdf = sigma * exp(-sigma*t) / (1 - exp(-sigma*tmax))
    // with: exp(-sigma*t) = exp(-sigma*-ln(a)/sigma) = exp(ln(a)) = a
    pdf = sigma * trans / (1.f - transMin);
}

static float transBoundedDistancePdf( const float t, const float tmax, const float sigma )
{
    // cf : Production Volume Rendering 2017
    return sigma * std::exp(-sigma * t) / (1.f - std::exp(-sigma * tmax));
}

static void equiAngularSample( const float u, const float tmax, const Point& rayOrigin, const Vector& rayDir, const Point& lightPos, float& distance, float& pdf )
{
    // cf : Importance sampling techniques for path tracing in participating media, Kulla et al, EGSR 2012
    // cf : Monte Carlo Methods for Volumetric Light Transport Simulation 2018
    // get coord of closest point to light along (infinite) ray
    float th = std::abs(dot(lightPos - rayOrigin, rayDir));
    // get distance this point is from light
    float h = length(rayOrigin + th*rayDir - lightPos);
    // get angle of endpoints
    float thetaA = std::atan2(0.0 - th, h);
    float thetaB = std::atan2(tmax - th, h);
    // take sample
    float t = h*std::tan(lerp(thetaA, thetaB, u));
    distance = th + t;
    pdf = h/((thetaB - thetaA)*(h*h + t*t));
}

static float equiAngularPdf( const float distance, const float tmax, const Point& rayOrigin, const Vector& rayDir, const Point& lightPos )
{
    // cf : Importance sampling techniques for path tracing in participating media, Kulla et al, EGSR 2012
    // cf : Monte Carlo Methods for Volumetric Light Transport Simulation 2018
    // get coord of closest point to light along (infinite) ray
    float th = std::abs(dot(lightPos - rayOrigin, rayDir));
    // get distance this point is from light
    float h = length(rayOrigin + th*rayDir - lightPos);
    // get angle of endpoints
    float thetaA = std::atan2(0.0 - th, h);
    float thetaB = std::atan2(tmax - th, h);
    // distance (positive or negative) from the projected light position
    float t = distance - th;
    return h/((thetaB - thetaA)*(h*h + t*t));
}