#ifndef RAY_H
#define RAY_H

// vector / point
#include "vec.h"

#include <memory>

// embree core
#include <embree3/rtcore.h>

class Medium;
class Ray
{
public:
    RTCRayHit data;
    Point org;
    Vector dir;
    const Medium * medium = nullptr;

    Ray( ) : data({{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, RTC_INVALID_GEOMETRY_ID, RTC_INVALID_GEOMETRY_ID, { RTC_INVALID_GEOMETRY_ID }}}), org(0,0,0), dir(0,0,0) {}
    Ray( const Point& o, const Vector& d, const float time = 0, const float tnear = 0, const float tfar = INFINITY ) : 
        data({{o.x, o.y, o.z, tnear, d.x, d.y, d.z, time, tfar, ~0u, 0, 0}, {0, 0, 0, 0, 0, RTC_INVALID_GEOMETRY_ID, RTC_INVALID_GEOMETRY_ID, { RTC_INVALID_GEOMETRY_ID }}}), org(o), dir(d) {}
    Ray( const Point& o, const Point& e, const float time = 0 )
    {
        org = o;
        dir = e - o;
        float tnear = 0;
        float tfar = length(dir);
        dir = normalize(dir);
        data = 
        {
            {
                o.x, o.y, o.z,  // origin 
                tnear,          // tnear    
                dir.x, dir.y, dir.z,  // direction
                time,           // time
                tfar,           // tfar
                ~0u,            // mask
                0,              // ray id 
                0               // ray flags
            },
            
            {
                0, 0, 0,        // normal
                0, 0,           // u, v
                RTC_INVALID_GEOMETRY_ID, // primitive ID
                RTC_INVALID_GEOMETRY_ID, // geometry ID
                { RTC_INVALID_GEOMETRY_ID } // instance ID
            }
        };
    }
    
    // Point o()               const { return Point(data.ray.org_x, data.ray.org_y, data.ray.org_z); }
    // Vector d()              const { return Vector(data.ray.dir_x, data.ray.dir_y, data.ray.dir_z); }
    Point o()               const { return org; }
    Vector d()              const { return dir; }
    Vector nd()             const { return normalize(dir); }
    float time()            const { return data.ray.time; }
    float tmin()            const { return data.ray.tnear; }
    float tmax()            const { return data.ray.tfar; }
    
    bool hit()              const { return (data.hit.primID != RTC_INVALID_GEOMETRY_ID); }
    unsigned int hitid()    const { return  data.hit.primID; }
    unsigned int hitobject()const { return  data.hit.geomID; }
    Point hitp()            const { assert(hit()); return org + data.ray.tfar * dir; }
    Vector hitn()           const { assert(hit()); return normalize(Vector(data.hit.Ng_x, data.hit.Ng_y, data.hit.Ng_z)); }
    float hitu()            const { assert(hit()); return data.hit.u; }
    float hitv()            const { assert(hit()); return data.hit.v; }
    float hitt()            const { assert(hit()); return data.ray.tfar; }

    Point operator() (const float t) const { return org + t * dir; }
};

class ShadowRay
{
public:
    RTCRay data;
    
    ShadowRay( ) : data({0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}) {}
    ShadowRay( const Point& o, const Vector& d, const float time = 0, const float tnear = 0, const float tfar = INFINITY ) : 
        data({o.x, o.y, o.z, tnear, d.x, d.y, d.z, time, tfar, ~0u, 0, 0}) {}
    ShadowRay( const Point& o, const Point& e, const float time = 0, const float tnear = 0, const float tfar = 1 )
    {
        Vector d(o, e);
        data = {
            o.x, o.y, o.z,  // origin 
            tnear,          // tnear    
            d.x, d.y, d.z,  // direction
            time,           // time
            tfar,           // tfar
            ~0u,            // mask
            0,              // ray id 
            0               // ray flags
        };
    }
    
    Point o()               const { return Point(data.org_x, data.org_y, data.org_z); }
    Vector d()              const { return normalize(Vector(data.dir_x, data.dir_y, data.dir_z)); }
    float time()            const { return data.time; }
    bool occluded()         const { return data.tfar > 0; }
};

#endif
