#pragma once

#include <cstdlib>
#include <random>

#include "vec.h"

using G = std::default_random_engine;
using D = std::uniform_real_distribution<float>;

class Sampler
{
public:
    Sampler() : m_gen(std::random_device()()), m_dist(0.f, 1.f) {}
    Sampler( const uint64_t seed ) : m_gen(seed), m_dist(0.f, 1.f) {}
    Sampler( const Sampler& ) = default;
    Sampler& operator= ( const Sampler& ) = default;

    virtual ~Sampler() {}
    virtual float sample() { return m_dist(m_gen); }
    virtual void index( const uint64_t id, const uint64_t sample = 0 ) {}

    virtual float get1D() { return sample(); }
    virtual vec2 get2D() { return vec2(sample(), sample()); }
    virtual vec3 get3D() { return vec3(sample(), sample(), sample()); }

protected:
    G m_gen;
    D m_dist;
};

class SamplerPCG : public Sampler
{
public:
    SamplerPCG( const uint64_t seed = 1234, const uint64_t sequence = 0 ) : m_state(seed), m_sequence(sequence) {}
    SamplerPCG( const SamplerPCG& ) = default;
    SamplerPCG& operator= ( const SamplerPCG& ) = default;

    //////////////////////////
    // from Tungsten 
    inline float uintBitsToFloat(uint32_t i)
    {
        union { float f; uint32_t i; } unionHack;
        unionHack.i = i;
        return unionHack.f;
    }

    //////////////////////////
    // from Tungsten 
    // 2x-5x faster than i/float(UINT_MAX)
    inline float normalizedUint(uint32_t i)
    {
        return uintBitsToFloat((i >> 9u) | 0x3F800000u) - 1.0f;
    }

    //////////////////////////
    // PCG random number generator
    // See http://www.pcg-random.org/
    inline uint32_t uniformUint()
    {
        uint64_t oldState = m_state;
        m_state = oldState * 6364136223846793005ULL + (m_sequence | 1);
        uint32_t xorShifted = uint32_t(((oldState >> 18u) ^ oldState) >> 27u);
        uint32_t rot = oldState >> 59u;
        return (xorShifted >> rot) | (xorShifted << (uint32_t(-int32_t(rot)) & 31));
    }

    float sample() { return normalizedUint(uniformUint()); }
    virtual void index( const uint64_t id, const uint64_t sample = 0 ) { m_sequence = id; }

protected:
    uint64_t m_state;
    uint64_t m_sequence;
};

class SamplerRN : public Sampler
{
protected:
    double basis[1000];

public:
    static double phid(uint32_t dim)
    {
        double x = 2.0; 
        for(int i = 0; i < 20; ++i) 
            x = std::pow(1.0+x,1.0/(dim+1.0));
        return x;
    }

    SamplerRN( const uint64_t seed = 1234, const uint64_t sequence = 0 ) : m_state(seed), m_sequence(sequence) 
    {
        int dims = 1000;
        double phi = phid(dims);
        for(int i = 0; i < dims; ++i) 
            basis[i] =  1.0/std::pow(phi, i+1); 
    }
    SamplerRN( const SamplerRN& ) = default;
    SamplerRN& operator= ( const SamplerRN& ) = default;

    float sample() 
    { 
        float rnd = m_state <= 1000 ? std::fmod(m_seed + basis[m_state]*m_sample, 1.0) : m_dist(m_gen);  
        m_state+=1;
        return rnd;    
    }

    virtual void index( const uint64_t id, const uint64_t sample = 0 ) 
    { 
        m_sequence = id;
        m_sample = sample;
        m_gen=G(id);
        m_seed = m_dist(m_gen); 
        m_state = 0;
    }

protected:
    uint64_t m_sample;
    uint64_t m_state;
    uint64_t m_sequence;
    float m_seed;
};
