#pragma once

#include "vec.h"
#include "color.h"
#include "image.h"

///////////// TEXTURE SAMPLING METHODS /////////////

enum Sampling
{
    NEAREST = 0,
    // BIQUADRATIC,
    BILINEAR,
    BICUBIC,
    TRILINEAR,
    // TRIQUADRATIC,
    // TRICUBIC,
};

///////////// TEXTURE 2D INTERFACE /////////////

class Texture
{
public:
    Texture(){}
    virtual ~Texture() {}
    virtual Color operator() ( const vec2& uv ) const { return (*this)(uv.x, uv.y); }
    virtual Color operator() ( float, float) const = 0;
    virtual Color average() const = 0;
    virtual bool isConstant() const { return false; }
};

///////////// TEXTURE 2D CONSTANT /////////////

class TextureConstant : public Texture
{
public:
    TextureConstant() : m_color(0.f) {}
    TextureConstant(float value) : m_color(value) {}
    TextureConstant(const Color& color) : m_color(color) {}

    Color operator() ( float u, float v ) const { return m_color; }
    Color average() const { return m_color; }
    bool isConstant() const { return true; }

protected:
    Color m_color;
};

///////////// TEXTURE 2D CHECKERBOARD /////////////

class TextureChecker : public Texture
{
public:
    TextureChecker() : m_on(0.75f), m_off(0.25f), m_scalex(0.1), m_scaley(0.1) {}
    TextureChecker(const Color& on, const Color& off, const float scalex = 1, const float scaley = 1) : m_on(on), m_off(off), m_scalex(scalex), m_scaley(scaley) {}

    Color operator() ( float u, float v ) const 
    {
        // (u odd and u even) or (u even and u odd) -> color off
        // (u odd and u odd) or (u even and u even) -> color on
        int u_remap = std::floor(u * m_scalex);
        int v_remap = std::floor(v * m_scaley);
        return (u_remap + v_remap) % 2 ? m_on : m_off;
    }
    
    Color average() const 
    { 
        return (m_on + m_off) * 0.5f; 
    }

protected:
    Color m_on, m_off;
    float m_scalex, m_scaley;
};

///////////// TEXTURE 2D IMAGE /////////////

class TextureImage : public Texture
{
public:
    TextureImage(){}    
    TextureImage(
        const Image& image, 
        Sampling sampling = NEAREST, 
        bool repeat = true, 
        float scalex = 1.f, 
        float scaley = 1.f
    ) : 
        m_width(image.width()), 
        m_height(image.height()), 
        m_sampling(sampling),
        m_data(m_width * m_height),
        m_avg(0.f),
        m_repeat(repeat),
        m_scalex(scalex),
        m_scaley(scaley)
    {
        // copy data and compute average
        int index = 0;
        float norm = 1.f / float(m_width * m_height);
        for (int y = 0; y < m_height; ++y) 
        for (int x = 0; x < m_width; ++x) 
        {
            const Color& color = image(x, y); 
            m_data[index] = color;
            m_avg = m_avg + color * norm;
            ++index;
        }
    }   

    TextureImage(
        const std::vector<Color>& data,
        int width, 
        int height, 
        Sampling sampling = NEAREST, 
        bool repeat = true, 
        float scalex = 1.f, 
        float scaley = 1.f 
    ) : 
        m_width(width), 
        m_height(height), 
        m_sampling(sampling),
        m_data(data),
        m_avg(0.f),
        m_repeat(repeat),
        m_scalex(scalex),
        m_scaley(scaley)
    {
        float norm = 1.f / float(m_data.size());
        for(const Color& c : m_data)
            m_avg = m_avg + c * norm;
    }   

    Color operator() ( float x, float y ) const 
    { 
        Color ret(0.f);
        float xremap, yremap;
        if( !m_repeat && x * m_scalex >= 0 && x * m_scalex <= 1 && y * m_scaley >= 0 && y * m_scaley <= 1 )
        {
            xremap = x * (m_width - 1) * m_scalex;
            yremap = y * (m_height - 1) * m_scaley;
        }
        else if( m_repeat )
        {
            float u = x * m_scalex;
            float v = y * m_scaley;
            xremap = std::fmod(std::abs(std::floor(u)) + u, 1.f) * (m_width - 1);
            yremap = std::fmod(std::abs(std::floor(v)) + v, 1.f) * (m_height - 1);
        }
        else
            return Black();

        switch(m_sampling)
        {
            case BILINEAR:
            {
                float u = xremap - std::floor(xremap);
                float v = yremap - std::floor(yremap);
                int ix0 = std::floor(xremap);
                int iy0 = std::floor(yremap);
                int ix1 = std::min(ix0+1, m_width -1);
                int iy1 = std::min(iy0+1, m_height -1);
                
                ret = m_data[offset(ix0, iy0)] * ((1 - u) * (1 - v))
                    + m_data[offset(ix1, iy0)] * (u       * (1 - v))
                    + m_data[offset(ix0, iy1)] * ((1 - u) * v)
                    + m_data[offset(ix1, iy1)] * (u       * v);

                break;
            }
            case BICUBIC:
            {
                // current pixel
                int xmin, xmax, ymin, ymax;
                int px = offsetX(xremap);
                int py = offsetY(yremap);

                if( !m_repeat )
                {
                    // 4x4 window bounds for bicubic spline filter clamped at boundaries
                    xmin = ( xremap - px > 0 ) ? std::max(0, px - 1) : std::max(0, px - 2);
                    ymin = ( yremap - py > 0 ) ? std::max(0, py - 1) : std::max(0, py - 2);
                    xmax = ( xremap - px > 0 ) ? std::min(m_width - 1, px + 2) : std::min(m_width - 1, px + 1);
                    ymax = ( yremap - py > 0 ) ? std::min(m_height - 1, py + 2) : std::min(m_height - 1, py + 1);

                    // filtered value
                    float sum = 0;
                    for( int row = ymin; row <= ymax; ++row)
                    for( int col = xmin; col <= xmax; ++col)
                    {
                        float s = xremap - col;
                        float t = yremap - row;
                        float ws = 4 * b3(s) * b3(t); 
                        sum += ws;
                        ret = ret + m_data[offset(col, row)] * ws;
                    }
                    ret = ret / sum;
                }
                else
                {
                    // 4x4 window bounds for bicubic spline filter
                    xmin = ( xremap - px > 0 ) ? px - 1 : px - 2;
                    ymin = ( yremap - py > 0 ) ? py - 1 : py - 2;
                    xmax = ( xremap - px > 0 ) ? px + 2 : px + 1;
                    ymax = ( yremap - py > 0 ) ? py + 2 : py + 1;

                    // filtered value
                    float sum = 0;
                    for( int row = ymin; row <= ymax; ++row)
                    for( int col = xmin; col <= xmax; ++col)
                    {
                        float s = xremap - col;
                        float t = yremap - row;
                        float ws = 4 * b3(s) * b3(t); 
                        sum += ws;
                        ret = ret + m_data[offset(modplus(col, m_width), modplus(row, m_height))] * ws;
                    }
                    ret = ret / sum;

                }
                break;
            }
            default:
            {
                ret = m_data[offset(xremap, yremap)];
                break;
            }
        }

        return ret;
    }

    Color average() const { return m_avg; }

protected:
    int m_width, m_height;
    Sampling m_sampling;
    std::vector<Color> m_data;
    Color m_avg;
    bool m_repeat;
    float m_scalex, m_scaley;
    bool m_constant = false;

    inline int modplus(int i, int n) const
    {
        return (i % n + n) % n;
    }

    inline int offsetY( const float y ) const 
    {
        return std::min(m_height - 1, std::max(0, static_cast<int>(y + 0.5f)));
    }

    inline int offsetX( const float x ) const 
    {
        return std::min(m_width - 1, std::max(0, static_cast<int>(x + 0.5f)));
    }

    inline int offset( const float x, const float y ) const
    {
        return offsetY(y) * m_width + offsetX(x);
    }

    inline int offset( const int x, const int y ) const
    {
        return y * m_width + x;
    }

    // filtre Bspline 3
    inline float b0( const float x ) const
    {
        float x2= x * x;
        float x3= x2 * x;
        return 1.f/6.f * x3;
    }
    
    inline float b1( const float x ) const
    {
        float x2= x * x;
        float x3= x2 * x;
        return 1.f/6.f * (-3.f*x3 + 3.f*x2 + 3.f*x + 1.f);
    }
    
    inline float b3( const float x ) const
    {
        if(x < -2)          return 0;
        else if(x <= -1)    return b0(x + 2);
        else if(x <= 0)     return b1(x + 1);
        else if(x <= 1)     return b1(1 - x);
        else if(x <= 2)     return b0(2 - x);
        else                return 0;
    }
};

///////////// TEXTURE 3D INTERFACE /////////////

class Texture3D
{
public:
    Texture3D(){}
    virtual ~Texture3D() {}
    virtual Color operator() ( const vec3& uvw ) const { return (*this)(uvw.x, uvw.y, uvw.z); }
    virtual Color operator() ( float x, float y, float z ) const = 0;
    virtual Color average() const = 0;
    virtual bool isConstant() const { return false; }
};

///////////// TEXTURE 3D IMAGE /////////////

class TextureImage3D : public Texture3D
{
public:
    TextureImage3D(const std::vector<Color>& data, int width, int height, int depth) : 
        m_data(data), 
        m_width(width), 
        m_height(height), 
        m_depth(depth),
        m_sampling(Sampling::TRILINEAR)
    {}

    Color operator() ( float x, float y, float z ) const
    {
        Color ret(0.f);
        float xremap, yremap, zremap;
        if( x >= 0 && x <= 1 && y >= 0 && y <= 1 && z >= 0 && z <= 1 )
        {
            xremap = x * (m_width - 1);
            yremap = y * (m_height - 1);
            zremap = z * (m_depth - 1);
        }
        else
            return Black();

        switch(m_sampling)
        {
            case TRILINEAR:
            {
                float u = xremap - std::floor(xremap);
                float v = yremap - std::floor(yremap);
                float w = zremap - std::floor(zremap);

                int ix0 = std::floor(xremap);
                int iy0 = std::floor(yremap);
                int iz0 = std::floor(zremap);

                int ix1 = std::min(ix0+1, m_width -1);
                int iy1 = std::min(iy0+1, m_height -1);
                int iz1 = std::min(iz0+1, m_depth -1);
                
                ret =   m_data[offset(ix0, iy0, iz0)] * (1 - u) * (1 - v) * (1 - w) +
                        m_data[offset(ix1, iy0, iz0)] * u * (1 - v) * (1 - w) +
                        m_data[offset(ix0, iy1, iz0)] * (1 - u) * v * (1 - w) +
                        m_data[offset(ix0, iy0, iz1)] * (1 - u) * (1 - v) * w +
                        m_data[offset(ix1, iy0, iz1)] * u * (1 - v) * w +
                        m_data[offset(ix0, iy1, iz1)] * (1 - u) * v * w +
                        m_data[offset(ix1, iy1, iz0)] * u * v * (1 - w) +
                        m_data[offset(ix1, iy1, iz1)] * u * v * w; 

                break;
            }
            default:
            {
                ret = m_data[offset(xremap, yremap, zremap)];
                break;
            }
        }

        return ret;
    }

    Color average() const 
    {
        return Black();
    }

protected:

    inline int offset( const float v, const int vmax ) const 
    {
        return std::min(vmax - 1, std::max(0, static_cast<int>(v + 0.5f)));
    }

    inline int offset( const float x, const float y, const float z ) const
    {
        return  (offset(z, m_depth) * m_height + offset(y, m_height)) * m_width + offset(x, m_width);
    }

    inline int offset( const int x, const int y, const int z ) const
    {
        return  (z * m_height + y) * m_width + x;
    }

    std::vector<Color> m_data;
    int m_width, m_height, m_depth;
    Sampling m_sampling;
};
