
#ifndef _TRIANGLE_H
#define _TRIANGLE_H

#include "vec.h"
#include "mesh.h"

class Triangle
{
public:    
    Point a, b, c;
    Vector na, nb, nc;
    vec2 ta, tb, tc;
    
    Triangle( ) = default;
    
    Triangle( const TriangleData &data ) 
        : 
        a(Point(data.a)), b(Point(data.b)), c(Point(data.c)),
        na(Vector(data.na)), nb(Vector(data.nb)), nc(Vector(data.nc)),
        ta(data.ta), tb(data.tb), tc(data.tc)
    {}

    Triangle& operator=(const Triangle&) = default;
    
    //! renvoie l'aire du triangle
    float area( ) const
    {
        return .5f * length(cross(Vector(a, b), Vector(a, c)));
    }
    
    //! renvoie la normale du triangle
    Vector normal( ) const
    {
        return normalize(cross(Vector(a, b), Vector(a, c)));
    }
    
    //! renvoie un point a l'interieur du triangle connaissant ses coordonnees barycentriques.
    //! convention p(u, v)= (1 - u - v) * a + u * b + v * c
    Point point( const float u, const float v ) const
    {
        float w = 1.f - u - v;
        return Point(Vector(a) * w + Vector(b) * u + Vector(c) * v);
    }

    //! renvoie une normale a l'interieur du triangle connaissant ses coordonnees barycentriques.
    //! convention p(u, v)= (1 - u - v) * a + u * b + v * c
    Vector normal( const float u, const float v ) const
    {
        float w = 1.f - u - v;
        return normalize(Vector(na) * w + Vector(nb) * u + Vector(nc) * v);
    }

    //! renvoie les texcoord, connaissant les coordonnees barycentriques. 
    //! convention p(u, v)= (1 - u - v) * a + u * b + v * c
    vec2 texcoord( const float u, const float v ) const
    {
        float w = 1.f - u - v;
        return vec2(ta.x * w + tb.x * u + tc.x * v, ta.y * w + tb.y * u + tc.y * v);
    }

    bool intersect( const Point& o, const Vector& d, float& rt ) const
    {
        const Point p = a;
        const Vector e1(a, b);
        const Vector e2(a, c);

        Vector pvec= cross(d, e2);
        float det= dot(e1, pvec);
        
        float inv_det = 1.0f / det;
        Vector tvec(p, o);

        float u= dot(tvec, pvec) * inv_det;
        if(u < 0.0f || u > 1.0f) return false;

        Vector qvec= cross(tvec, e1);
        float v= dot(d, qvec) * inv_det;
        if(v < 0.0f || u + v > 1.0f) return false;

        rt= dot(e2, qvec) * inv_det;
        float ru= u;
        float rv= v;
        return rt >= 0.f;
    }
};

#endif
