#ifndef VIEWER_H
#define VIEWER_H

#include "renderer/integrators/abstractintegrator.h"

enum ViewerMode
{
    COLOR, SAMPLES, WEIGHTS, VARIANCE, ERROR, RECONSTRUCTION,
};

struct ViewerParams
{
    // mode
    ViewerMode mode = ViewerMode::COLOR;
    // tonemapping
    float compression = 2.2, saturation = 1;
    // zoom
    float zoom = 4;
    // screenchots count
    int screenshots = 0;
    // filtering
    int lod = 0;
};

class Viewer : public App
{
public:

    // start time 
    std::chrono::time_point<std::chrono::system_clock> start = std::chrono::high_resolution_clock::now();
    
    Widgets m_widgets;
    GLuint m_gl_texture;
    int m_width, m_height;

    GLuint m_program;
    GLuint m_vao;
    
    // render mode
    AbstractIntegrator* m_integrator;

    // active channels
    int m_red = 1, m_green = 1, m_blue = 1, m_alpha = 1, m_gray = 0;

    // tonemapping
    float m_compression = 2.2;
    float m_saturation = 1;
    float m_saturation_step = 0.1;
    float m_saturation_max = 1000;

    // active index / active layer
    int m_index = 0, m_layer = 0;

    // camera
    float m_radius = 0.001;
    float m_fov = 45;
    float m_focal = 1;
    float m_shutter_open = 0;
    float m_shutter_close = 0;

    // viewer params
    ViewerParams m_viewer_params;
    
    // screenshot counter
    int m_screenshots = 0;

    // key released handler
    bool released = true;

    bool m_show_tile = false;

public:
    Viewer( const int width, const int height, AbstractIntegrator* integrator ) : 
        App(width, height), m_width(width), m_height(height), m_integrator(integrator) 
    {
        // genere le vao
        glGenVertexArrays(1, &m_vao);
        glBindVertexArray(m_vao);   
        
        // initialise les textures pour chaque observateur
        init_render_texture( m_integrator->cameras[m_index].film );
        
        // charge les programmes
        m_program = read_program(smart_path("data/shaders/tonemap.glsl"));    // pour afficher les rendus 
        program_print_errors(m_program);

        // change le titre de la fenetre
        SDL_SetWindowTitle(m_window, "Renderer");

        // redminsionne la fenetre
        SDL_SetWindowSize(m_window, m_width, m_height);
        
        // creer les widgets interactifs
        m_widgets = create_widgets();
        
        // etat openGL par defaut
        glUseProgram(0);
        glBindVertexArray(0);
        glBindTexture(GL_TEXTURE_2D, 0);
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_CULL_FACE);

        // init
        glViewport(0, 0, window_width(), window_height());
    }
    
    void init_render_texture( const Film& film )
    {
        // format de la texture de rendu (r,g,b,n) 
        GLenum format = GL_RGBA;
        GLenum type = GL_FLOAT;
        GLenum texel_type = GL_RGBA32F;
        // cree la texture openGL
        glGenTextures(1, &m_gl_texture);
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, m_gl_texture);
        // fixe les parametres de filtrage par defaut
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 2);
        // transfere les donnees dans la texture
        glTexImage2D( GL_TEXTURE_2D, 0, texel_type, film.w, film.h, 0, format, type, film.samples_mean.get_data_ptr() );
        // prefiltre la texture
        glGenerateMipmap(GL_TEXTURE_2D);
        // reset binding
        glBindTexture( GL_TEXTURE_2D, 0 );
    }

    void update_render_texture( const Film& film )
    {
        // format de la texture de rendu (r,g,b,n) 
        GLenum format = GL_RGBA;
        GLenum type = GL_FLOAT;
        GLenum texel_type = GL_RGBA32F;
        // bind la texture openGL
        glBindTexture(GL_TEXTURE_2D, m_gl_texture);
        // fixe les parametres de filtrage par defaut
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 2);
        // transfere les donnees dans la texture
        switch( m_viewer_params.mode )
        {
            default:
                if( m_integrator->config.mLayers && m_layer > 0 )
                {
                    int i = m_layer-1;
                    glTexImage2D( GL_TEXTURE_2D, 0, texel_type, film.w, film.h, 0, format, type, film.layers[i].get_data_ptr() );
                }
                else
                glTexImage2D( GL_TEXTURE_2D, 0, texel_type, film.w, film.h, 0, format, type, film.image.get_data_ptr() );
                break;
            // case ViewerMode::WEIGHTS:
                // glTexImage2D( GL_TEXTURE_2D, 0, texel_type, film.w, film.h, 0, format, type, film.image.get_data_ptr() );
                // break;
            case ViewerMode::VARIANCE:
                glTexImage2D( GL_TEXTURE_2D, 0, texel_type, film.w, film.h, 0, format, type, film.variance.get_data_ptr() );
                break;
            case ViewerMode::ERROR:
                glTexImage2D( GL_TEXTURE_2D, 0, texel_type, film.w, film.h, 0, format, type, film.error.get_data_ptr() );
                break;
            case ViewerMode::SAMPLES:
                glTexImage2D( GL_TEXTURE_2D, 0, texel_type, film.w, film.h, 0, format, type, film.variance.get_data_ptr() );
                break;
            case ViewerMode::RECONSTRUCTION:
                glTexImage2D( GL_TEXTURE_2D, 0, texel_type, film.w, film.h, 0, format, type, film.reconstruction.get_data_ptr() );
                break;
        }
        // prefiltre la texture
        glGenerateMipmap(GL_TEXTURE_2D);
        // reset binding
        glBindTexture( GL_TEXTURE_2D, 0 );
    }

    int init(){return 1;}
    
    int quit()
    {
        m_integrator->Quit();
        glDeleteVertexArrays(1, &m_vao);
        glDeleteTextures(1, &m_gl_texture);
        release_program(m_program);
        release_widgets(m_widgets);
        return 0;
    }

    void save_camera()
    {
        #pragma omp critical
        {
            int time = m_integrator->mTimer.sec().count();
            for (unsigned int index = 0u; index < m_integrator->cameras.size(); index++)
            {
                std::stringstream ss;
                ss << "output-" << std::setfill('0') << std::setw(3) << index << "-time-" << time;
                m_integrator->cameras[index].film.save(ss.str(), 1.f, true); 
            }
        }
    }

    int run( )
    {
        if(init() < 0)
            return -1;

        // configure openGL
        glViewport(0, 0, window_width(), window_height());

        // gestion des evenements
        while(events(m_window))
        {
            if(update(global_time(), delta_time()) < 0)
                break;
            if(render() < 1)
                break;
            // presenter le resultat
            SDL_GL_SwapWindow(m_window);
            
            if( !m_integrator->IsRendering() )
                break;
        }

        if(quit() < 0)
            return -1;
        return 0;
    }

    int render( )
    {
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glClearColor(1, 1, 0, 1);

        if(key_state(SDLK_F7)) 
        { 
            clear_key_state(SDLK_F7); 
            m_viewer_params.mode = RECONSTRUCTION; 
            m_viewer_params.saturation = 1.f;
            m_red = m_green = m_blue = 1;
            // printf("mode RECONSTRUCTION\n"); 
        }

        if(key_state(SDLK_F8))  
        { 
            clear_key_state(SDLK_F8); 
            m_viewer_params.mode = COLOR;
            m_viewer_params.saturation = 1.f;
            m_red = m_green = m_blue = 1;
            // printf("mode COLOR\n");
        }
        if(key_state(SDLK_F9))  
        { 
            clear_key_state(SDLK_F9); 
            m_viewer_params.mode = WEIGHTS;
            m_viewer_params.saturation = 1;
            m_red = m_green = m_blue = 0;
            // printf("mode WEIGHTS\n");
        }
        if(key_state(SDLK_F10)) 
        { 
            clear_key_state(SDLK_F10); 
            m_viewer_params.mode = VARIANCE;
            m_viewer_params.saturation = 1;
            m_red = m_green = m_blue = 1;
            // printf("mode VARIANCE\n");
        }
        if(key_state(SDLK_F11)) 
        { 
            clear_key_state(SDLK_F11); 
            m_viewer_params.mode = ERROR;
            m_viewer_params.saturation = 1;
            m_red = m_green = m_blue = 1;
            // printf("mode ERROR\n");
        }
        if(key_state(SDLK_F12)) 
        { 
            clear_key_state(SDLK_F12); 
            m_viewer_params.mode = SAMPLES;
            m_viewer_params.saturation = 1;
            m_red = m_green = m_blue = 0;
            // printf("mode SAMPLES\n");
        }

        // change viewer
        if(key_state(SDLK_LEFT)) 
        { 
            clear_key_state(SDLK_LEFT); 
            m_index = (m_index - 1 + m_integrator->cameras.size()) % m_integrator->cameras.size(); 
        }

        if(key_state(SDLK_RIGHT)) 
        { 
            clear_key_state(SDLK_RIGHT); 
            m_index = (m_index + 1 + m_integrator->cameras.size()) % m_integrator->cameras.size(); 
        }

        // if(key_state(SDLK_UP)) 
        // { 
        //     clear_key_state(SDLK_UP); 
        //     int size = m_integrator->cameras.size() + 1;
        //     m_layer = (m_layer + 1 + size) % size; 
        // }

        // if(key_state(SDLK_DOWN)) 
        // { 
        //     clear_key_state(SDLK_DOWN); 
        //     int size = m_integrator->cameras.size() + 1;
        //     m_layer = (m_layer - 1 + size) % size; 
        // }

        // if(key_state(SDLK_UP)) 
        // { 
        //     clear_key_state(SDLK_UP); 
        //     m_viewer_params.lod--;
        //     if( m_viewer_params.lod < 0 ) m_viewer_params.lod = 0; 
        //     printf("lod %d\n", m_viewer_params.lod);
        // }

        // if(key_state(SDLK_DOWN)) 
        // { 
        //     clear_key_state(SDLK_DOWN); 
        //     m_viewer_params.lod++;
        //     if( m_viewer_params.lod > 1 ) m_viewer_params.lod = 1; 
        //     printf("lod %d\n", m_viewer_params.lod);
        // }

        // mouse movements
        int xmouse, ymouse;
        unsigned int bmouse = SDL_GetMouseState(&xmouse, &ymouse);

        if(bmouse & SDL_BUTTON(3))
        {
            SDL_MouseWheelEvent wheel= wheel_event();
            if(wheel.y != 0)
            {
                // clear_wheel_event();
                m_viewer_params.zoom= m_viewer_params.zoom + float(wheel.y) / 4.f;
                if(m_viewer_params.zoom < .1f) m_viewer_params.zoom= .1f;
                if(m_viewer_params.zoom > 30.f) m_viewer_params.zoom= 30.f;
            }
        }

        // update rendered image
        update_render_texture( m_integrator->cameras[m_index].film );
        
        // render pass
        glUseProgram(m_program);

        glUniform1f( glGetUniformLocation(m_program, "compression"), m_viewer_params.compression );
        glUniform1f( glGetUniformLocation(m_program, "saturation"), m_viewer_params.saturation );
        glUniform1i( glGetUniformLocation(m_program, "alpha"), int( m_viewer_params.mode == WEIGHTS || m_viewer_params.mode == SAMPLES) );
        glUniform1i( glGetUniformLocation(m_program, "lod"), m_viewer_params.lod );
        glUniform1i( glGetUniformLocation(m_program, "variance"), int( m_viewer_params.mode == VARIANCE ) );
        
        program_uniform(m_program, "channels", Color(m_red, m_green, m_blue, m_alpha));
        program_uniform(m_program, "gray", float(m_gray));
        
        vec2 center( float(xmouse) / float(window_width()), float(window_height() - ymouse +1) / float(window_height()));
        glUniform2fv( glGetUniformLocation(m_program, "center"), 1, &(center.x) );

        if(!key_state(SDLK_LCTRL) && (bmouse & SDL_BUTTON(3)) )
            glUniform1f( glGetUniformLocation(m_program, "zoom"), m_viewer_params.zoom );
        else
            glUniform1f( glGetUniformLocation(m_program, "zoom"), 1.f );
        
        if( m_show_tile )
        {
            // program_uniform(m_program, "width", m_integrator->cameras[m_index].width);
            // program_uniform(m_program, "height", m_integrator->cameras[m_index].height);
            program_uniform(m_program, "width", m_width);
            program_uniform(m_program, "height", m_height);
            program_uniform(m_program, "tile", m_integrator->GetTile());
            program_uniform(m_program, "tile_size", m_integrator->GetTileSize());
        }
        else
        {
            program_uniform(m_program, "tile", -1);
        }

        // used for MIS progressive previsualization
        int normalized = (m_viewer_params.mode == COLOR) ? (int)m_integrator->cameras[m_index].film.normalized : 1;
        program_uniform(m_program, "normalized", normalized);
        float scale = normalized ? 1.f : 1.f / (m_integrator->mIteration+1);
        program_uniform(m_program, "scale", scale);

        // selectionne l'unite de texture
        glActiveTexture(GL_TEXTURE0);
        // configure la texture
        glBindTexture(GL_TEXTURE_2D, m_gl_texture);
        // transmet l'indice de l'unite de texture au shader
        glUniform1i( glGetUniformLocation(m_program, "image"), 0);

        // dessine 1 triangle plein ecran
        glBindVertexArray(m_vao);
        glDrawArrays(GL_TRIANGLES, 0, 3);

        // screenshots
        if(key_state('s'))
        {
            clear_key_state('s');
            std::stringstream ss;
            ss << "output-" << std::setfill('0') << std::setw(3) << m_index << "-screen-" << std::setw(3) << m_viewer_params.screenshots <<".png";
            printf("writing '%s'...\n", ss.str().c_str());
            screenshot(ss.str().c_str());
            m_viewer_params.screenshots++;
        }

        if(key_state('t'))
        {
            clear_key_state('t');
            m_show_tile = !m_show_tile;
        }
        
     //    int time = m_integrator.timer.sec().count(); 
     //    if( time%3600 == 0 )
     //    {
     //         save_camera();
    	// }

        // full save
        if(key_state('i'))
        {
            clear_key_state('i');
            save_camera();
        }

        float steps = 0.1;
        if( m_viewer_params.saturation <= 1000000000 )
            steps = 100000000;
        if( m_viewer_params.saturation <= 100000000 )
            steps = 10000000;
        if( m_viewer_params.saturation <= 10000000 )
            steps = 1000000;
        if( m_viewer_params.saturation <= 1000000 )
            steps = 100000;
        if( m_viewer_params.saturation <= 100000 )
            steps = 10000;
        if( m_viewer_params.saturation <= 10000 )
            steps = 1000;
        if( m_viewer_params.saturation <= 1000 )
            steps = 100;
        if( m_viewer_params.saturation <= 100 )
            steps = 10;
        if( m_viewer_params.saturation <= 10 )
            steps = 1;
        if( m_viewer_params.saturation <= 1 )
            steps = 0.1;
        if( m_viewer_params.saturation <= 0.1 )
            steps = 0.01;
        if( m_viewer_params.saturation <= 0.01 )
            steps = 0.001;
        if( m_viewer_params.saturation <= 0.001 )
            steps = 0.0001;
        if( m_viewer_params.saturation <= 0.0001 )
            steps = 0.00001;
        if( m_viewer_params.saturation <= 0.00001 )
            steps = 0.000001;

        // tone mapping control
        begin(m_widgets);
            label( m_widgets, "camera %d", m_index );
        
        begin_line(m_widgets);
            value(m_widgets, "saturation", m_viewer_params.saturation, 0.f, std::numeric_limits<float>::max(), steps);
            value(m_widgets, "compression", m_viewer_params.compression, .1f, 10.f, .1f);
        end(m_widgets);

        begin_line(m_widgets);
            button(m_widgets, "R", m_red);
            button(m_widgets, "G", m_green);
            button(m_widgets, "B", m_blue);
            button(m_widgets, "A", m_alpha);
            button(m_widgets, "gray", m_gray);
        end(m_widgets);

        int iter = m_integrator->GetIteration();
        int sec = m_integrator->mTimer.sec().count();
        int min = m_integrator->mTimer.min().count();
        if( std::isinf(sec) || std::isnan(sec) )
            sec = -1;
        if( std::isinf(min) || std::isnan(min) )
            min = -1;
        float speri = iter > 0 ? (float)sec/(float)iter : -1;

        begin_line(m_widgets);
            label( m_widgets, "iteration %d, %.2f s/iter", iter, speri);
        end(m_widgets);

        begin_line(m_widgets);
            label( m_widgets, "time %d min / %d s", min, sec );
        end(m_widgets);

        int px = xmouse * m_integrator->cameras[m_index].film.w / m_width;
        int py = ymouse * m_integrator->cameras[m_index].film.h / m_height;
        begin_line(m_widgets);
            label( m_widgets, "pixel (%d,%d)", px, py );
        end(m_widgets);

        begin_line(m_widgets);
            label( m_widgets, "normalized (%d,%f)", normalized, scale );
        end(m_widgets);

        begin_line(m_widgets);
        end(m_widgets);

        begin_line(m_widgets);
            label( m_widgets, "right button + wheel: zoom / unzoom");
        end(m_widgets);
        begin_line(m_widgets);
            label( m_widgets, "left / right arrows: prec / next camera");
        end(m_widgets);
        begin_line(m_widgets);
            label(m_widgets, "F7: Base samples only"); 
        end(m_widgets);
        begin_line(m_widgets);
            label(m_widgets, "F8: Base + reused samples");
        end(m_widgets);
        begin_line(m_widgets);
            label(m_widgets, "F9: Weight integrals");
        end(m_widgets);
        begin_line(m_widgets);
            label(m_widgets, "F10: Variance");
        end(m_widgets);
        begin_line(m_widgets);
            label(m_widgets, "F11: Chi2-error");
        end(m_widgets);
        begin_line(m_widgets);
            label(m_widgets, "F12: Samples count");
        end(m_widgets);

        
        draw(m_widgets, window_width(), window_height());

        return 1;
    }
};


#endif
