
#ifndef _COLOR_H
#define _COLOR_H

#include <cassert>
#include <cmath>

static float luminanceToPerceived(float luminance) 
{
    // Send this function a luminance value between 0.0 and 1.0,
    // and it returns L* which is "perceptual lightness"
    // The CIE standard states 0.008856 but 216/24389 is the intent for 0.008856451679036
    if( luminance <= 216/24389 )
    {       
        // The CIE standard states 903.3, but 24389/27 is the intent, making 903.296296296296296
        return luminance * 24389 / 27;  
    } 
    else 
    {
        return std::pow(luminance,(1/3)) * 116 - 16;
    }
}


//! \addtogroup image
///@{

//! \file
//! manipulation de couleurs

//! representation d'une couleur (rgba) transparente ou opaque.
struct Color
{
    //! constructeur par defaut.
    Color( ) : r(0.f), g(0.f), b(0.f), a(1.f) {}
    explicit Color( const float _r, const float _g, const float _b, const float _a= 1.f ) : r(_r), g(_g), b(_b), a(_a) {}
    explicit Color( const float _value ) : r(_value), g(_value), b(_value), a(1.f) {}
    explicit Color( const float data[4] ) : r(data[0]), g(data[1]), b(data[2]), a(data[3]) {}
    
    //! cree une couleur avec les memes composantes que color, mais remplace sa composante alpha (color.r, color.g, color.b, alpha).
    Color( const Color& color, const float alpha ) : r(color.r), g(color.g), b(color.b), a(alpha) {}  // remplace alpha.
    
    float& operator[] (int i) 
    {
    	assert(i >= 0 && i < 4); 
    	if(i == 0 ) return r;
    	else if(i == 1 ) return g;
    	else if(i == 2 ) return b;
    	else return a;
    } 

    const float& operator[] (int i) const 
    {
    	assert(i >= 0 && i < 4); 
    	if(i == 0 ) return r;
    	else if(i == 1 ) return g;
    	else if(i == 2 ) return b;
    	else return a;
    } 

    float& operator() (int i) 
    {
        assert(i >= 0 && i < 4); 
        if(i == 0 ) return r;
        else if(i == 1 ) return g;
        else if(i == 2 ) return b;
        else return a;
    } 

    const float& operator() (int i) const 
    {
        assert(i >= 0 && i < 4); 
        if(i == 0 ) return r;
        else if(i == 1 ) return g;
        else if(i == 2 ) return b;
        else return a;
    } 

    bool valid( ) const
    {
        if(std::isinf(r) || std::isinf(g) || std::isinf(b) || std::isinf(a)) return false;
        if(std::isnan(r) || std::isnan(g) || std::isnan(b) || std::isnan(a)) return false;
        return true;
    }
    
    float r, g, b, a;

    float max() const { return std::max(r, std::max(g, b)); }
    float min() const { return std::min(r, std::min(g, b)); }
    float power( ) const { return (r+g+b)/3.f; }
    float rgbsum( ) const { return r+g+b; }
    float luminance( ) const { return r*0.212671f+g*0.715160f+b*0.072169f; }

    Color linear() const 
    {
        // Send this function a decimal sRGB gamma encoded color value
        // between 0.0 and 1.0, and it returns a linearized value.
        return Color(
            r <= 0.04045 ? r / 12.92 : std::pow((r+0.055)/1.055,2.4),
            g <= 0.04045 ? g / 12.92 : std::pow((g+0.055)/1.055,2.4),
            b <= 0.04045 ? b / 12.92 : std::pow((b+0.055)/1.055,2.4)
        );
    }

    Color& operator+= ( const Color& other )
    {
        r += other.r;
        g += other.g;
        b += other.b;
        return *this;
    }

    Color& operator+= ( const float k )
    {
        r += k;
        g += k;
        b += k;
        return *this;
    }

    Color& operator-= ( const Color& other )
    {
        r -= other.r;
        g -= other.g;
        b -= other.b;
        return *this;
    }

    Color& operator-= ( const float k )
    {
        r -= k;
        g -= k;
        b -= k;
        return *this;
    }

    Color& operator*= ( const Color& other )
    {
        r *= other.r;
        g *= other.g;
        b *= other.b;
        return *this;
    }

    Color& operator*= ( const float k )
    {
        r *= k;
        g *= k;
        b *= k;
        return *this;
    }

    Color& operator/= ( const Color& other )
    {
        r /= other.r;
        g /= other.g;
        b /= other.b;
        return *this;
    }

    Color& operator/= ( const float k )
    {
        r /= k;
        g /= k;
        b /= k;
        return *this;
    }
};

//! utilitaire. renvoie une couleur noire.
Color Black( );
//! utilitaire. renvoie une couleur blanche.
Color White( );
//! utilitaire. renvoie une couleur rouge.
Color Red( );
//! utilitaire. renvoie une couleur verte.
Color Green( );
//! utilitaire. renvoie une couleur bleue.
Color Blue( );
//! utilitaire. renvoie une couleur jaune.
Color Yellow( );

Color operator+ ( const Color& a, const Color& b );
Color operator- ( const Color& a, const Color& b );
Color operator- ( const Color& c );

Color operator* ( const Color& a, const Color& b );
Color operator* ( const Color& c, const float k );

Color operator* ( const float k, const Color& c );
Color operator/ ( const Color& a, const Color& b );

Color operator/ ( const float k, const Color& c );
Color operator/ ( const Color& c, const float k );

Color operator+ ( const Color& c, const float k );
Color operator+ ( const float k, const Color& c );

Color abs(const Color& c);
Color exp(const Color& c);
Color sqrt(const Color& c);
Color pow(const Color& c, float exp);
Color log(const Color& c);
Color clamp(const Color& c, float vmin = 0, float vmax = 1);
Color max(const Color& a, const Color& b);
Color min(const Color& a, const Color& b);
Color min(const Color& c, const float v);
Color max(const Color& c, const float v);
Color min(const float v, const Color& c);
Color max(const float v, const Color& c);

float length(const Color& c);

//
#include <iostream>

inline std::ostream& operator<<(std::ostream& o, const Color& c)
{
    o<<"c("<<c.r<<","<<c.g<<","<<c.b<<")";
    return o;
}

///@}
#endif
