
#include "mesh.h"
#include "wavefront.h"
#include "mesh_buffer.h"
#include "mesh_cache.h"

template< typename T>
static int read( FILE *in, std::vector<T>& data )
{
    data.clear();
    
    int n= 0;
    size_t code= fread(&n, sizeof(int), 1, in);
    if(code == 0 || n == 0)
        return 0;
    
    data.resize(n);
    code= fread(data.data(), sizeof(T), n, in);
    if(int(code) != n)
        return 0;
    
    return n;
}

Mesh read_mesh_cache( const char *filename )
{
    Mesh mesh;
    std::string cache_filename= std::string(filename) + ".mesh";

    FILE *in= fopen(cache_filename.c_str(), "r");
    if(in)
    {
        printf("loading mesh cache '%s'...\n", cache_filename.c_str());
        
        int primitives= mesh.primitives();
        if(fread(&primitives, sizeof(int), 1, in) != 1)
            return mesh;
        
        mesh= Mesh(primitives);
        
        std::vector<vec3> data3;
        if(read(in, data3)) 
            mesh.positions(data3);
        
        std::vector<vec2> data2;
        if(read(in, data2))
            mesh.texcoords(data2);
        if(read(in, data3))
            mesh.normals(data3);
        
        std::vector<vec4> data4;
        if(read(in, data4))
            mesh.colors(data4);
        
        std::vector<unsigned int> data;
        if(read(in, data))
            mesh.indices(data);
        if(read(in, data))
            mesh.materials(data);
        fclose(in);
        
        cache_filename= std::string(filename) + ".materials";
        MaterialLib materials= read_materials(cache_filename.c_str());
        mesh.mesh_materials(materials.data);
        
        printf("  mesh: %d positions, %d texcoords, %d normals, %d indices, %d materials\n", 
            (int) mesh.positions().size(), (int) mesh.texcoords().size(), (int) mesh.normals().size(), (int) mesh.indices().size(), (int) mesh.mesh_materials().size());
        
        //~ printf("    indices %dMB\n", int(mesh.indices().size() * sizeof(int) / 1024 / 1024));
        //~ printf("    positions %dMB\n", int(mesh.positions().size() * sizeof(vec3) / 1024 / 1024));
        //~ printf("          %dMB\n", int((mesh.indices().size() * sizeof(int) + mesh.positions().size() * sizeof(vec3)) / 1024 / 1024));
        //~ printf("    array %dMB\n", int(mesh.indices().size() * sizeof(vec3) / 1024 / 1024));
    }
    else
    {
        MeshData data= read_mesh_data(filename);
        data.texcoords.clear();
        data.texcoord_indices.clear();
        build_normals( data );
        mesh = build_linear_mesh(data);
        
        if(!mesh.triangle_count())
            return mesh;

        // printf("  mesh: %d positions, %d texcoords, %d normals, %d indices, %d materials\n", 
        //     (int) mesh.positions().size(), (int) mesh.texcoords().size(), (int) mesh.normals().size(), (int) mesh.indices().size(), (int) mesh.mesh_materials().size());
        
        write_mesh_cache(mesh, filename);
    }
    
    return mesh;
}


template< typename T >
static size_t write( const std::vector<T>& buffer, FILE *out )
{
    int n= int(buffer.size());
    size_t code= fwrite(&n, sizeof(int), 1, out);
    if(n == 0)
        return 1;
    
    code= fwrite(buffer.data(), sizeof(T), buffer.size(), out);
    return code;
}

int write_mesh_cache( const Mesh& mesh, const char *filename )
{
    if(!mesh.triangle_count())
        return -1;
    
    std::string cache_filename= std::string(filename) + ".mesh";
    FILE *out= fopen(cache_filename.c_str(), "w");
    if(!out)
        return -1;
    
    // mesh data
    printf("writing mesh cache '%s'...\n", cache_filename.c_str());
    
    int primitives= mesh.primitives();
    fwrite(&primitives, sizeof(int), 1, out);
    
    write(mesh.positions(), out);
    write(mesh.texcoords(), out);
    write(mesh.normals(), out);
    write(mesh.colors(), out);
    write(mesh.indices(), out);
    write(mesh.materials(), out);
    fclose(out);
    
    // materials
    cache_filename= std::string(filename) + ".materials";
    out= fopen(cache_filename.c_str(), "wt");
    if(!out)
        return -1;
    
    printf("writing materials '%s'...\n", cache_filename.c_str());
    
    const std::vector<Material>& materials= mesh.mesh_materials();
    for(int i= 0; i < int(materials.size()); i++)
    {
        const Material& m= materials[i];
        
        fprintf(out, "newmtl %d\n", i);
        if(m.diffuse.power())
            fprintf(out, "Kd %f %f %f\n", m.diffuse.r, m.diffuse.g, m.diffuse.b);
        if(m.specular.power())
            fprintf(out, "Ks %f %f %f\n", m.specular.r, m.specular.g, m.specular.b);
        if(m.emission.power())
            fprintf(out, "Ke %f %f %f\n", m.emission.r, m.emission.g, m.emission.b);
        if(m.ns)
            fprintf(out, "Ns %f\n", m.ns);
        if(m.ni)
            fprintf(out, "Ni %f\n", m.ni);
        if(m.illum)
            fprintf(out, "illum %d\n", m.illum);
        if(!m.diffuse_filename.empty())
            fprintf(out, "map_Kd %s\n", m.diffuse_filename.c_str());
        if(!m.specular_filename.empty())
            fprintf(out, "map_Ks %s\n", m.specular_filename.c_str());
        if(!m.emission_filename.empty())
            fprintf(out, "map_Ke %s\n", m.emission_filename.c_str());
        if(!m.ns_filename.empty())
            fprintf(out, "map_Ns %s\n", m.ns_filename.c_str());
    }
    fclose(out);
    
    return 0;
}

