
#include <cassert>
#include <cstdio>
#include <ctype.h>
#include <climits>

#include <string>
#include <algorithm>
#include <sstream>

#include "wavefront.h"


/*! renvoie le chemin d'acces a un fichier. le chemin est toujours termine par /
    pathname("path/to/file") == "path/to/"
    pathname("file") == "./"
 */
static
std::string pathname( const std::string& filename )
{
    std::string path= filename;
#ifndef WIN32
    std::replace(path.begin(), path.end(), '\\', '/');   // linux, macos : remplace les \ par /.
    size_t slash = path.find_last_of( '/' );
    if(slash != std::string::npos)
        return path.substr(0, slash +1); // inclus le slash
    else
        return "./";
#else
    std::replace(path.begin(), path.end(), '/', '\\');   // windows : remplace les / par \.
    size_t slash = path.find_last_of( '\\' );
    if(slash != std::string::npos)
        return path.substr(0, slash +1); // inclus le slash
    else
        return ".\\";
#endif
}


static
std::string normalize_path( std::string file )
{
#ifndef WIN32
    std::replace(file.begin(), file.end(), '\\', '/');   // linux, macos : remplace les \ par /.
#else
    std::replace(file.begin(), file.end(), '/', '\\');   // windows : remplace les / par \.
#endif
    return file;
}

static
bool relative_path( const std::string& filename )
{
    if(filename.empty()) 
        return false;
    
#ifndef WIN32
    return (filename[0] != '/');
#else
    return (filename[0] != '\\');
#endif
}

Mesh read_mesh( const char *filename )
{
    FILE *in= fopen(filename, "rt");
    if(in == NULL)
    {
        printf("[error] loading mesh '%s'...\n", filename);
        return Mesh::error();
    }
    
    Mesh data(GL_TRIANGLES);
    
    printf("loading mesh '%s'...\n", filename);
    
    std::vector<vec3> positions;
    std::vector<vec2> texcoords;
    std::vector<vec3> normals;
    MaterialLib materials;
    int default_material_id= -1;
    int material_id= -1;
    
    std::vector<int> idp;
    std::vector<int> idt;
    std::vector<int> idn;
    
    char tmp[1024];
    char line_buffer[1024];
    bool error= true;
    for(;;)
    {
        // charge une ligne du fichier
        if(fgets(line_buffer, sizeof(line_buffer), in) == NULL)
        {
            error= false;       // fin du fichier, pas d'erreur detectee
            break;
        }
        
        // force la fin de la ligne, au cas ou
        line_buffer[sizeof(line_buffer) -1]= 0;
        
        // saute les espaces en debut de ligne
        char *line= line_buffer;
        while(*line && isspace(*line))
            line++;
        
        if(line[0] == 'v')
        {
            float x, y, z;
            if(line[1] == ' ')          // position x y z
            {
                if(sscanf(line, "v %f %f %f", &x, &y, &z) != 3)
                    break;
                positions.push_back( vec3(x, y, z) );
                // printf("v %f %f %f\n", positions.back().x, positions.back().y, positions.back().z);
            }
            else if(line[1] == 'n')     // normal x y z
            {
                if(sscanf(line, "vn %f %f %f", &x, &y, &z) != 3)
                    break;
                normals.push_back( vec3(x, y, z) );
                // printf("vn %f %f %f\n", normals.back().x, normals.back().y, normals.back().z);
            }
            else if(line[1] == 't')     // texcoord x y
            {
                if(sscanf(line, "vt %f %f", &x, &y) != 2)
                    break;
                texcoords.push_back( vec2(x, y) );
                // printf("vt %f %f\n", texcoords.back().x, texcoords.back().y);
            }
        }
        
        else if(line[0] == 'f')         // triangle a b c, les sommets sont numerotes a partir de 1 ou de la fin du tableau (< 0)
        {
            idp.clear();
            idt.clear();
            idn.clear();
            
            int next= 0;
            for(line= line +1; ; line= line + next)
            {
                idp.push_back(0); 
                idt.push_back(0); 
                idn.push_back(0);         // 0: invalid index
                
                next= 0;
                if(sscanf(line, " %d/%d/%d %n", &idp.back(), &idt.back(), &idn.back(), &next) == 3) 
                {
                    // printf("face %d %d %d\n", idp.back(), idt.back(), idn.back());
                    continue;
                }
                else if(sscanf(line, " %d/%d %n", &idp.back(), &idt.back(), &next) == 2)
                {
                    // printf("face %d %d %d\n", idp.back(), idt.back(), idn.back());
                    continue;
                }
                else if(sscanf(line, " %d//%d %n", &idp.back(), &idn.back(), &next) == 2)
                {
                    // printf("face %d %d %d\n", idp.back(), idt.back(), idn.back());
                    continue;
                }
                else if(sscanf(line, " %d %n", &idp.back(), &next) == 1)
                {
                    // printf("face %d %d %d\n", idp.back(), idt.back(), idn.back());
                    continue;
                }
                else if(next == 0)      // fin de ligne
                {
                    // printf("face %d %d %d\n", idp.back(), idt.back(), idn.back());
                    break;
                }

            }
            
            // force une matiere par defaut, si necessaire
            if(material_id == -1)
            {
                if(default_material_id == -1)
                    // creer une matiere par defaut
                    default_material_id= data.mesh_material(Material());
                
                material_id= default_material_id;
                data.material(material_id);
                
                printf("usemtl default\n");
            }
            
            for(int v= 2; v +1 < (int) idp.size(); v++)
            {
                int idv[3]= { 0, v -1, v };
                for(int i= 0; i < 3; i++)
                {
                    int k= idv[i];
                    int p= (idp[k] < 0) ? int(positions.size()) + idp[k] : idp[k] -1;
                    int t= (idt[k] < 0) ? int(texcoords.size()) + idt[k] : idt[k] -1;
                    int n= (idn[k] < 0) ? int(normals.size())   + idn[k] : idn[k] -1;
                    
                    if(p < 0) break; // error
                    if(t >= 0) data.texcoord(texcoords[t]);
                    if(n >= 0) data.normal(normals[n]);
                    data.vertex(positions[p]);
                }
            }
        }
        
        else if(line[0] == 'm')
        {
           if(sscanf(line, "mtllib %[^\r\n]", tmp) == 1)
           {
               materials= read_materials( std::string(pathname(filename) + tmp).c_str() );
               // enregistre les matieres dans le mesh
               data.mesh_materials(materials.data);
           }
        }
        
        else if(line[0] == 'u')
        {
           if(sscanf(line, "usemtl %[^\r\n]", tmp) == 1)
           {
               material_id= -1;
               for(unsigned int i= 0; i < (unsigned int) materials.names.size(); i++)
                    if(materials.names[i] == tmp)
                        material_id= i;
                
                if(material_id == -1)
                {
                    // force une matiere par defaut, si necessaire
                    if(default_material_id == -1)
                        default_material_id= data.mesh_material(Material());
                    
                    material_id= default_material_id;
                }
                
                // selectionne une matiere pour le prochain triangle
                data.material(material_id);
           }
        }
    }
    
    fclose(in);
    
    if(error)
        printf("loading mesh '%s'...\n[error]\n%s\n\n", filename, line_buffer);
    
    return data;
}

Mesh read_mesh_cpp( const char *filename )
{
    FILE *in= fopen(filename, "rt");
    if(in == NULL)
    {
        printf("[error] loading mesh '%s'...\n", filename);
        return Mesh::error();
    }
    
    Mesh data(GL_TRIANGLES);
    printf("loading mesh '%s'...\n", filename);
    
    std::vector<vec3> positions;
    std::vector<vec2> texcoords;
    std::vector<vec3> normals;
    MaterialLib materials;
    int default_material_id= -1;
    int material_id= -1;
    
    std::vector<int> idp;
    std::vector<int> idt;
    std::vector<int> idn;
    idp.reserve(100);
    idt.reserve(100);
    idn.reserve(100);
    
    char tmp[1024];
    char line_buffer[1024];
    bool error= true;

    std::stringstream ss;
    std::string token;
    for(;;)
    {
        // charge une ligne du fichier
        if(fgets(line_buffer, sizeof(line_buffer), in) == NULL)
        {
            error= false;       // fin du fichier, pas d'erreur detectee
            break;
        }
        
        // force la fin de la ligne, au cas ou
        line_buffer[sizeof(line_buffer) -1]= 0;
        
        // saute les espaces en debut de ligne
        char *line= line_buffer;
        while(*line && isspace(*line))
            line++;

        // get token
        ss = std::stringstream(line);
        ss >> token;
        if( token == "v" )
        {
            vec3 p;
            ss >> p.x >> p.y >> p.z;
            positions.push_back( p );
            // printf("v %f %f %f\n", positions.back().x, positions.back().y, positions.back().z);
        }
        else if( token == "vn" )
        {
            vec3 n;
            ss >> n.x >> n.y >> n.z;
            normals.push_back( n );
            // printf("vn %f %f %f\n", normals.back().x, normals.back().y, normals.back().z);
        }
        else if( token == "vt" )
        {
            vec2 t;
            ss >> t.x >> t.y;
            texcoords.push_back( t );
            // printf("vt %f %f\n", texcoords.back().x, texcoords.back().y);
        }
        else if( token[0] == 'f' )         // triangle a b c, les sommets sont numerotes a partir de 1 ou de la fin du tableau (< 0)
        {
            idp.clear();
            idt.clear();
            idn.clear();

            while(ss)
            {
                idp.push_back(0); 
                idt.push_back(0); 
                idn.push_back(0);         // 0: invalid index
                char dumb;

                // read position id
                ss >> idp.back();

                // if slash is found after
                if( ss.get(dumb) && dumb == '/' )
                {
                    // if slash again -> no texcoords
                    if( ss.get(dumb) && dumb == '/' )
                    {
                        // read normal id
                        ss >> idn.back();
                    }
                    // texcoords given
                    else
                    {
                        // unget last char -> part of the texcoord id
                        ss.unget();

                        // read texcoords id
                        ss >> idt.back();

                        // if slash is found 
                        if( ss.get(dumb) && dumb == '/' )
                        {
                            // read normal id
                            ss >> idn.back();
                        }
                    }
                }
                // printf("face %d %d %d\n", idp.back(), idt.back(), idn.back());
            }

            // force une matiere par defaut, si necessaire
            if(material_id == -1)
            {
                if(default_material_id == -1)
                    // creer une matiere par defaut
                    default_material_id= data.mesh_material(Material());
                
                material_id= default_material_id;
                data.material(material_id);
                
                printf("usemtl default\n");
            }
            
            for(int v= 2; v +1 < (int) idp.size(); v++)
            {
                int idv[3]= { 0, v -1, v };
                for(int i= 0; i < 3; i++)
                {
                    int k= idv[i];
                    int p= (idp[k] < 0) ? int(positions.size()) + idp[k] : idp[k] -1;
                    int t= (idt[k] < 0) ? int(texcoords.size()) + idt[k] : idt[k] -1;
                    int n= (idn[k] < 0) ? int(normals.size())   + idn[k] : idn[k] -1;
                    
                    if(p < 0) break; // error
                    if(t >= 0) data.texcoord(texcoords[t]);
                    if(n >= 0) data.normal(normals[n]);
                    data.vertex(positions[p]);
                }
            }
        }
        else if( token[0] == 'm' )
        {
            ss >> token;
            materials= read_materials( std::string(pathname(filename) + token).c_str() );
            data.mesh_materials(materials.data);
        }
        else if( token[0] == 'u' )
        {
            ss >> token;
            material_id= -1;
            for(unsigned int i= 0; i < (unsigned int) materials.names.size(); i++)
                if(materials.names[i] == token)
                    material_id= i;

            if(material_id == -1)
            {
                // force une matiere par defaut, si necessaire
                if(default_material_id == -1)
                    default_material_id= data.mesh_material(Material());
                
                material_id= default_material_id;
            }

            // selectionne une matiere pour le prochain triangle
            data.material(material_id);
        }
    }
    
    fclose(in);
    
    if(error)
        printf("loading mesh '%s'...\n[error]\n%s\n\n", filename, line_buffer);
    
    return data;
}


MeshData read_mesh_data( const char *filename )
{
    FILE *in= fopen(filename, "rt");
    if(in == NULL)
    {
        printf("[error] loading mesh data '%s'...\n", filename);
        return MeshData();
    }
    
    // printf("loading mesh data '%s'...\n", filename);
    
    MeshData data;
    MaterialLib materials;
    int default_material_id= -1;
    int material_id= -1;
    
    std::vector<int> idp;
    std::vector<int> idt;
    std::vector<int> idn;
    
    char tmp[1024];
    char line_buffer[1024];
    bool error= true;
    for(;;)
    {
        // charge une ligne du fichier
        if(fgets(line_buffer, sizeof(line_buffer), in) == NULL)
        {
            error= false;       // fin du fichier, pas d'erreur detectee
            break;
        }
        
        // force la fin de la ligne, au cas ou
        line_buffer[sizeof(line_buffer) -1]= 0;
        
        // saute les espaces en debut de ligne
        char *line= line_buffer;
        while(*line && isspace(*line))
            line++;
        
        if(line[0] == 'v')
        {
            float x, y, z;
            if(line[1] == ' ')          // position x y z
            {
                if(sscanf(line, "v %f %f %f", &x, &y, &z) != 3)
                    break;
                data.positions.push_back( vec3(x, y, z) );
            }
            else if(line[1] == 'n')     // normal x y z
            {
                if(sscanf(line, "vn %f %f %f", &x, &y, &z) != 3)
                    break;
                data.normals.push_back( vec3(x, y, z) );
            }
            else if(line[1] == 't')     // texcoord x y
            {
                if(sscanf(line, "vt %f %f", &x, &y) != 2)
                    break;
                data.texcoords.push_back( vec2(x, y) );
            }
        }
        
        else if(line[0] == 'f')         // triangle a b c, les sommets sont numerotes a partir de 1 ou de la fin du tableau (< 0)
        {
            idp.clear();
            idt.clear();
            idn.clear();
            
            int next= 0;
            for(line= line +1; ; line= line + next)
            {
                idp.push_back(0); 
                idt.push_back(0); 
                idn.push_back(0);         // 0: invalid index
                
                next= 0;
                if(sscanf(line, " %d/%d/%d %n", &idp.back(), &idt.back(), &idn.back(), &next) == 3) 
                    continue;
                else if(sscanf(line, " %d/%d %n", &idp.back(), &idt.back(), &next) == 2)
                    continue;
                else if(sscanf(line, " %d//%d %n", &idp.back(), &idn.back(), &next) == 2)
                    continue;
                else if(sscanf(line, " %d %n", &idp.back(), &next) == 1)
                    continue;
                else if(next == 0)      // fin de ligne
                    break;
            }
            
            // force une matiere par defaut, si necessaire
            if(material_id == -1)
            {
                if(default_material_id == -1)
                {
                    // creer une matiere par defaut
                    default_material_id= int(data.materials.size());
                    data.materials.push_back(Material());
                }
                
                material_id= default_material_id;
                printf("usemtl default\n");
            }
            
            for(int v= 2; v +1 < int(idp.size()); v++)
            {
                int idv[3]= { 0, v -1, v };
                for(int i= 0; i < 3; i++)
                {
                    int k= idv[i];
                    int p= (idp[k] < 0) ? int(data.positions.size()) + idp[k] : idp[k] -1;
                    int t= (idt[k] < 0) ? int(data.texcoords.size()) + idt[k] : idt[k] -1;
                    int n= (idn[k] < 0) ? int(data.normals.size())   + idn[k] : idn[k] -1;
                    
                    if(p < 0) break; // error
                    
                    data.position_indices.push_back(p);
                    data.texcoord_indices.push_back(t);
                    data.normal_indices.push_back(n);
                }
                
                assert(material_id != -1);
                data.material_indices.push_back(material_id);
            }
        }
        
        else if(line[0] == 'm')
        {
           if(sscanf(line, "mtllib %[^\r\n]", tmp) == 1)
           {
               materials= read_materials( std::string(pathname(filename) + tmp).c_str() );
               data.materials= materials.data;
           }
        }
        
        else if(line[0] == 'u')
        {
           if(sscanf(line, "usemtl %[^\r\n]", tmp) == 1)
           {
               material_id= -1;
               for(unsigned int i= 0; i < (unsigned int) materials.names.size(); i++)
                    if(materials.names[i] == tmp)
                        material_id= i;
                
                if(material_id == -1)
                {
                    // force une matiere par defaut, si necessaire
                    if(default_material_id == -1)
                    {
                        // creer une matiere par defaut
                        default_material_id= int(data.materials.size());
                        data.materials.push_back(Material());
                    }
                    
                    material_id= default_material_id;
                }
           }
        }
    }
    
    fclose(in);
    
    if(error)
        printf("loading mesh data '%s'...\n[error]\n%s\n\n", filename, line_buffer);
    
    printf("mesh data '%s':\n\tpositions %d indices %d\n\tnormals %d indices %d\n\ttexcoords %d indices %d\n", filename,
        (int)data.positions.size(), (int)data.position_indices.size(), (int)data.normals.size(), (int)data.normal_indices.size(), (int)data.texcoords.size(), (int)data.texcoord_indices.size() );
    return data;
}


int write_mesh( const Mesh& mesh, const char *filename )
{
    if(mesh == Mesh::error())
        return -1;

    if(mesh.primitives() != GL_TRIANGLES)
        return -1;
    if(mesh.positions().size() == 0)
        return -1;
    if(filename == NULL)
        return -1;
    
    FILE *out= fopen(filename, "wt");
    if(out == NULL)
        return -1;
    
    printf("writing mesh '%s'...\n", filename);
    
    const std::vector<vec3>& positions= mesh.positions();
    for(unsigned int i= 0; i < (unsigned int) positions.size(); i++)
        fprintf(out, "v %f %f %f\n", positions[i].x, positions[i].y, positions[i].z);
    fprintf(out, "\n");
    
    const std::vector<vec2>& texcoords= mesh.texcoords();
    bool has_texcoords= (texcoords.size() == positions.size());
    for(unsigned int i= 0; i < (unsigned int) texcoords.size(); i++)
        fprintf(out, "vt %f %f\n", texcoords[i].x, texcoords[i].y);
    fprintf(out, "\n");
    
    const std::vector<vec3>& normals= mesh.normals();
    bool has_normals= (normals.size() == positions.size());
    for(unsigned int i= 0; i < (unsigned int) normals.size(); i++)
        fprintf(out, "vn %f %f %f\n", normals[i].x, normals[i].y, normals[i].z);
    fprintf(out, "\n");
    
    const std::vector<unsigned int>& indices= mesh.indices();
    bool has_indices= (indices.size() > 0);
    unsigned int n= has_indices ? (unsigned int) indices.size() : (unsigned int) positions.size();
    for(unsigned int i= 0; i +2 < n; i+= 3)
    {
        fprintf(out, "f");
        for(unsigned int k= 0; k < 3; k++)
        {
            unsigned int id= has_indices ? indices[i+k] +1 : i+k +1;
            fprintf(out, " %u", id);
            if(has_texcoords && has_normals)
                fprintf(out, "/%u/%u", id, id);
            else if(has_texcoords)
                fprintf(out, "/%u", id);
            else if(has_normals)
                fprintf(out, "//%u", id);
        }
        fprintf(out, "\n");
    }
    
    fclose(out);
    return 0;
}

int write_mesh_data( const MeshData& data, const char *filename )
{
    if(data.positions.size() == 0)
        return -1;
    if(filename == nullptr)
        return -1;
    
    FILE *out= fopen(filename, "wt");
    if(out == NULL)
        return -1;
    
    printf("writing mesh data '%s'...\n", filename);
    
    for(int i= 0; i < int(data.positions.size()); i++)
        fprintf(out, "v %f %f %f\n", data.positions[i].x, data.positions[i].y, data.positions[i].z);
    fprintf(out, "\n");
    
    bool has_texcoords= (data.texcoords.size() == data.positions.size());
    for(int i= 0; i < int(data.texcoords.size()); i++)
        fprintf(out, "vt %f %f\n", data.texcoords[i].x, data.texcoords[i].y);
    fprintf(out, "\n");
    
    bool has_normals= (data.normals.size() == data.positions.size());
    for(int i= 0; i < int(data.normals.size()); i++)
        fprintf(out, "vn %f %f %f\n", data.normals[i].x, data.normals[i].y, data.normals[i].z);
    fprintf(out, "\n");
    
    for(int i= 0; i +2 < int(data.position_indices.size()); i+= 3)
    {
        fprintf(out, "f");
        for(int k= 0; k < 3; k++)
        {
            int id= i+k;
            fprintf(out, " %u", data.position_indices[id] +1);
            if(has_texcoords && has_normals)
                fprintf(out, "/%u/%u", data.texcoord_indices[id] +1, data.normal_indices[id] +1);
            else if(has_texcoords)
                fprintf(out, "/%u", data.texcoord_indices[id] +1);
            else if(has_normals)
                fprintf(out, "//%u", data.normal_indices[id] +1);
        }
        fprintf(out, "\n");
    }
    
    fclose(out);
    return 0;
}


MaterialLib read_materials( const char *filename )
{
    MaterialLib materials;
    
    FILE *in= fopen(filename, "rt");
    if(in == NULL)
    {
        printf("[error] loading materials '%s'...\n", filename);
        return materials;
    }
    
    // printf("loading materials '%s'...\n", filename);
    
    Material *material= NULL;
    std::string path= normalize_path(pathname(filename));
    
    char tmp[1024];
    char line_buffer[1024];
    bool error= true;
    for(;;)
    {
        // charge une ligne du fichier
        if(fgets(line_buffer, sizeof(line_buffer), in) == NULL)
        {
            error= false;       // fin du fichier, pas d'erreur detectee
            break;
        }
        
        // force la fin de la ligne, au cas ou
        line_buffer[sizeof(line_buffer) -1]= 0;
        
        // saute les espaces en debut de ligne
        char *line= line_buffer;
        while(*line && isspace(*line))
            line++;
        
        if(line[0] == 'n')
        {
            if(sscanf(line, "newmtl %[^\r\n]", tmp) == 1)
            {
                materials.names.push_back( tmp );
                materials.data.push_back( Material(Black()) );
                material= &materials.data.back();
                material->name = materials.names.back();
            }
        }
        
        if(material == NULL)
            continue;
        
        if(line[0] == 'K')
        {
            float r, g, b;
            if(sscanf(line, "Kd %f %f %f", &r, &g, &b) == 3)
                material->diffuse= Color(r, g, b);
            else if(sscanf(line, "Ks %f %f %f", &r, &g, &b) == 3)
                material->specular= Color(r, g, b);
            else if(sscanf(line, "Ke %f %f %f", &r, &g, &b) == 3)
                material->emission= Color(r, g, b);
        }
        
        else if(line[0] == 'N')
        {
            float n;
            if(sscanf(line, "Ns %f", &n) == 1)          // Ns, puissance / concentration du reflet, modele blinn phong
                material->ns= n;
            
            if(sscanf(line, "Ni %f", &n) == 1)          // Ni, indice de refraction
                material->ni= n;
        }

        else if(line[0] == 'i')
        {
            int n;
            if(sscanf(line, "illum %d", &n) == 1)       // modele matiere
                material->illum= n;
        }
        
        else if(line[0] == 'm')
        {
            if(sscanf(line, "map_Kd %[^\r\n]", tmp) == 1)
            {
                // ignore les options de mapping, si elles sont presentes
                int next;
                float x, y, z;
                char option;
                for(line= tmp; ; line= line + next)
                {
                    next= 0;
                    if(sscanf(line, "-%[so] %f %f %f %n", &option, &x, &y, &z, &next) == 4)continue;
                    else if(sscanf(line, "-%[os] %f %f %n", &option, &x, &y, &next) == 3)continue;
                    else if(sscanf(line, "-%[os] %f %n", &option, &x, &next) == 2)continue;
                    else break;
                }
                material->diffuse_filename= normalize_path(line);
                // if(relative_path(material->diffuse_filename))
                if( material->diffuse_filename.find(path) == std::string::npos )
                    material->diffuse_filename= normalize_path(path + line);
            }
            if(sscanf(line, "map_Ks %[^\r\n]", tmp) == 1)
            {
                // ignore les options de mapping, si elles sont presentes
                int next;
                float x, y, z;
                char option;
                for(line= tmp; ; line= line + next)
                {
                    next= 0;
                    if(sscanf(line, "-%[so] %f %f %f %n", &option, &x, &y, &z, &next) == 4) continue;
                    else if(sscanf(line, "-%[os] %f %f %n", &option, &x, &y, &next) == 3) continue;
                    else if(sscanf(line, "-%[os] %f %n", &option, &x, &next) == 2) continue;
                    else break;
                }
                material->specular_filename= normalize_path(line);
                // if(relative_path(material->specular_filename))
                if( material->specular_filename.find(path) == std::string::npos )
                    material->specular_filename= normalize_path(path + line);
            }
            if(sscanf(line, "map_Ke %[^\r\n]", tmp) == 1)
            {
                // ignore les options de mapping, si elles sont presentes
                int next;
                float x, y, z;
                char option;
                for(line= tmp; ; line= line + next)
                {
                    next= 0;
                    if(sscanf(line, "-%[so] %f %f %f %n", &option, &x, &y, &z, &next) == 4) continue;
                    else if(sscanf(line, "-%[os] %f %f %n", &option, &x, &y, &next) == 3) continue;
                    else if(sscanf(line, "-%[os] %f %n", &option, &x, &next) == 2) continue;
                    else break;
                }
                material->emission_filename= normalize_path(line);
                // if(relative_path(material->emission_filename))
                if( material->emission_filename.find(path) == std::string::npos )
                    material->emission_filename= normalize_path(path + line);
            }
            if(sscanf(line, "map_Ns %[^\r\n]", tmp) == 1)
            {
                // ignore les options de mapping, si elles sont presentes
                int next;
                float x, y, z;
                char option;
                for(line= tmp; ; line= line + next)
                {
                    next= 0;
                    if(sscanf(line, "-%[so] %f %f %f %n", &option, &x, &y, &z, &next) == 4) continue;
                    else if(sscanf(line, "-%[os] %f %f %n", &option, &x, &y, &next) == 3) continue;
                    else if(sscanf(line, "-%[os] %f %n", &option, &x, &next) == 2) continue;
                    else break;
                }
                material->ns_filename= normalize_path(line);
                // if(relative_path(material->ns_filename))
                if( material->ns_filename.find(path) == std::string::npos )
                    material->ns_filename= normalize_path(path + line);
            }
            // if(sscanf(line, "map_bump %[^\r\n]", tmp) == 1)
            // {
                // material->bump_filename= normalize_path(tmp);
                // if( material->bump_filename.find(path) == std::string::npos )
                    // material->bump_filename= normalize_path(path + tmp);
            // }
        }
    }
    
    fclose(in);
    if(error)
        printf("[error] parsing line :\n%s\n", line_buffer);
    
    return materials;
}
